#!/bin/sh
#
# Script to copy FlagShip re-distributable files into any directory
# of your choice (see also fsman SYS.1.2).
# Note: Other data from the FlagShip package such as the compiler, static 
# libs etc. are NOT redistributable, see also License Agreement in the
# <FlagShip_dir>/docu directory or in http://www.fship.com/license.html    
#
# Copyright (c) 2005 multisoft Datentechik GmbH, Munich, Germany
# Release $Date: 2005/12/14 09:39:05 $

source=/usr/local/FlagShip6
if [ "$FLAGSHIP_DIR" != "" ] ; then
   source=$FLAGSHIP_DIR
fi
target=$1
if [ "$2" = "-nolib" ] ; then
   makelib="N"
else
   makelib=""
fi

echo
echo "$0 copies FlagShip distributable files"
echo "into a directory of your choice."

if [ "$target" = "" -o "$target" = "-h" -o "$target" = "/?" ] ; then
   echo
   if [ "$target" = "-h" -o "$target" = "/?" ] ; then
      echo "Syntax: $0 /target/dir [-nolib]"
      exit
   fi
fi

while [ 1=1 ] ; do
   loop=0
   if [ "$source" = "" ] ; then 
      loop=1
   elif [ ! -d $source -o ! -f $source/bin/FlagShip_p ] ; then
      loop=1
   fi
   while [ $loop -eq 1 ] ; do
      echo
      if [ "$source" != "" ] ; then
         echo "Sorry, $source is not a FlagShip directory."
      fi
      echo -e "Specify the source directory or RETURN only to exit: \c"
      read source
      if [ "$source" = "" ] ; then
         exit
      fi
      if [ -d $source -a -f $source/bin/FlagShip_p ] ; then
         loop=0
      fi
   done

   if [ "$target" = "" ] ; then
      echo -e "Target directory (er empty to exit): \c"
      read target
      if [ "$target" = "" ] ; then
         exit
      fi
   fi
   if [ "$makelib" = "" ] ; then
      echo -e "Is your executable linked statically (y/n): \c"
      read answ
      if [ "$answ" = "y" ] ; then
         makelib="N"
      else
         makelib="Y"
      fi
   fi
   if [ "$target" = "/" ] ; then
      echo "Cannot copy to root directory"
      exit
   fi
   if [ ! -d $target ] ; then
      echo -e "Directory $target does not exist. Create (y/N) ? \c"
      read answ
      if [ "$answ" != "y" ] ; then
         exit
      fi
      mkdir $target
      if [ ! -d $target ] ; then
         echo "Sorry, could not create $target - permissions?"
         exit
      fi
   fi
   echo
   echo "Source = $source"
   echo "Target = $target"
   if [ "$makelib" = "Y" ] ; then
      echo "Copy also dynamic library libFlagShip6.so"
   else
      echo "Do not copy dynamic library libFlagShip6.so"
   fi    
   echo
   echo -e "Is this correct (y/n) or (q) to quit ? \c"
   read answ
   if [ "$answ" = "q" ] ; then
      exit
   fi
   if [ "$answ" = "y" ] ; then
      break
   fi
   source=""
   target=""
   makelib=""
done

if [ ! -d $target/terminfo ] ; then mkdir $target/terminfo ; fi
if [ ! -d $target/xfonts   ] ; then mkdir $target/xfonts   ; fi
if [ ! -d $target/bin      ] ; then mkdir $target/bin      ; fi

cp -a $source/terminfo/*         $target/terminfo/.
cp -a $source/xfonts/*           $target/xfonts/.
cp -a $source/bin/FSwhichterm    $target/bin/.
cp -a $source/bin/FSXTerm.xrdb*  $target/bin/.
cp -a $source/bin/newfs*         $target/bin/.

# optional (you may un-comment if terminfo modification required):
# cp -a $source/bin/*tic194        $target/bin/.
# cp -a $source/terminfo/*tic55    $target/bin/.

if [ "$makelib" = "Y" ] ; then
   if [ ! -d $target/lib ] ; then
      mkdir $target/lib 
   fi
   cp -a $source/lib/libFlagShip6.so $target/lib/.
fi

echo "Done successfully. Content of $target directory:"
if [ "$makelib" = "Y" ] ; then
   xx="$target/lib"
else
   xx=""
fi
ls -la $target $target/bin $xx
echo -e "any key ...\c" ; read answ
ls -la $target/terminfo $target/xfonts 
echo
echo "... and the size of $target directory:"
du $target
echo
if [ -f $target/lib/libFlagShip6.so ] ; then
   echo "Note: you may remove $target/lib/libFlagShip6.so"
   echo "      if linked statically or use the -nolib option"
   echo
fi
#eof
