// Display database structure
// Release $Date: 2007/11/02 08:49:12 $
//
// compile Windows: FlagShip dbfstru.prg
//     run Windows: dbfstru  [-io=t]  [[drive:][\path\]filename[.dbf]]
// compile Linux  : FlagShip dbfstru.prg -o dbfstru
//     run Linux  : ./dbfstru  [[/path/]filename[.dbf]]
//     run Linux  : newfswin ./dbfstru  [[/path/]filename[.dbf]]

#include "box.fh"
parameters cFileName                   // command-line parameter (optional)

local iPos, aStru, oTbr

set font "courier", 10                 // set fixed font for GUI applic
oApplic:Resize(25,80,,.T.)             // resize GUI applic for this font
oApplic:Move(10,20,.T.)                // move GUI applic window y=10 x=20 pixel

if empty(cFileName)                    // command-line parameter given?
   cFileName := space(100)             // no, init blank input
else
   cFileName += space(70)              // yes, add spaces for editing
endif

do while .T.                           // comment-out for single shot
   cls
   set key K_F2 to listdbf             // redirect F2 key
   do while .T.
      @ 1,0 say "Enter database name (optional with path and extension), or F2 for list, or ESC"

      @ 3,0 say "dbf file" get cFileName PICTURE "@S50"
      read
      if lastkey() == K_ESC            // READ aborted?
         quit
      endif

      /* plausibility and availability check
       */
      if empty(cFileName)
         loop
      endif
      iPos := rat(".",cFileName)
      if iPos == 0 .or. substr(cFileName, iPos +1, 1) $ "\/"
         cFileName := padr(trim(cFileName) + ".dbf", 100)
      endif
      if !file(alltrim(cFileName))     // file available?
         ?? chr(7)                     // no, bell
         @ 5,0
         @ 5,0 say "file '" + alltrim(cFileName) + "' not available"
         loop
      endif
      use (alltrim(cFileName)) shared  // open database
      if used()                        // success?
         exit                          // yes, exit loop
      endif
      ?? chr(7)                        // no, bell
      @ 5,0
      @ 5,0 say "cannot open file '" + alltrim(cFileName) + "'"
   enddo
   set key K_F2 to                     // disable F2 redirection

   aStru := DbStruct()                 // get dbf structure
   use                                 // close database

   /* display structure by Tbrowse
    */
   cls
   @ 0,0 say "File: '" + alltrim(cFileName) + "' - Exit by ESC"
   @ 1,0,maxrow(),51 box B_SINGLE      // for -io=t only
   oTbr := TbrowseArr(2,1,maxrow() -1,50)
   oTbr:ReadOnly  := .T.               // not editable
   oTbr:CanAppend := .F.               // may not append
   oTbr:UserArray := aStru
   oTbr:RowCache  := max(oTbr:RowCount, len(aStru))
   oTbr:AddColumn(TbColumnNew("FieldName ", .T.))
   oTbr:AddColumn(TbColumnNew("Type", .T.))
   oTbr:AddColumn(TbColumnNew("      Size", .T.))
   oTbr:AddColumn(TbColumnNew("      Deci", .T.))
   oTbr:Exec()
enddo                                  // comment-out for single shot
quit


******************************************************
function listdbf()
local aDirs, aFiles, cDir, ii, iPos, iSelect, oGet, iSize := 0
memvar cFileName

cDir := ""
iPos := rat("/", strtran(cFileName, "\","/"))
if iPos > 0
   cDir := left(cFileName, iPos)             // extract path
endif
aDirs := directory(cDir + "*.dbf")           // get databases in curr/given dir
if empty(aDirs)
   Alert("Sorry, no *.dbf files available in;" + ;
         if(empty(cDir),curdir(),cDir) )
   return
endif
aFiles := array(len(aDirs))
for ii := 1 to len(aDirs)
   aFiles[ii] := aDirs[ii,1]
   iSize := max(iSize, len(aFiles[ii]))
next
aFiles := Asort(aFiles)                       // sort file names
@ 5,0 say "Database files in " + if(empty(cDir),curdir(),cDir)
@ 6,2,24,iSize+6 box B_SINGLE                 // here, apply for -io=t only

iSelect := Achoice(7,3, 23,iSize+5, aFiles)        // display files
if iSelect > 0                                     // if one file was selected,
   cFileName := padr(cDir + aFiles[iSelect], 100)  //  feed the name to
   oGet := GetActive()                             //  current READ field
   if valtype(oGet) == "O"
      oGet:UpdateBuffer()
   endif
endif
@ 4,0 clear
return

/* eof */
