* +-------------------------------------------------------------------------
* | Check FlagShip installation, screen & keyboard                          
* |                                                                         
* | Author : J.V.Balek, multisoft GmbH, Munich, 1998                      
* | Release $Date: 2005/11/21 13:39:08 $
* +-------------------------------------------------------------------------
* | UNIX   : copy this source to your actual working or home directory, e.g.
* |          cp <FlagShip_dir>/examples/fscheck.prg ~/.                                  
* | Compile: FlagShip fscheck.prg -ofscheck                                 
* | Run    : fscheck   (or ./fscheck if current dir not in PATH)            
* +-------------------------------------------------------------------------
* | Windows: copy this source to your actual working directory, e.g.                                
* |          copy <FlagShip_dir>\examples\fscheck.prg                                  
* | Compile: FlagShip fscheck.prg                                
* | Run    : fscheck                                                        
* +-------------------------------------------------------------------------

// #define CLEAR_SCREEN_FULL       /* activate if problems with clear screen */

public FlagShip, Clipper
private terminal
private outmap, term_out, inmap, term_key
private screen1, screen2

#ifdef FlagShip5
# define VFS
#else
# ifdef FlagShip6
#  define VFS
# endif
#endif

#ifdef VFS
   if AppIoMode() == "B"      // Basic mode?
      ? "Sorry, this is an interactive application and"
      ? "requires -io=t or -io=g compiler/run-time switch"
      quit
   endif
   if AppIoMode() == "G"      // GUI mode?
      ? "Note: this tool is designed mainly for "
      ?? "Terminal i/o" GUICOLOR "R+"
      ?? " mode."
      ? "Would you like to continue (y/n/esc) ? "
      key = inkey(0)
      if lower(chr(key)) != "y"
         quit
      endif
   endif
   
   set font "courier", 12
   
   SET GUITRANSL ASCII    ON
   SET GUITRANSL LINES    ON
   SET GUITRANSL BOX      ON
   SET GUITRANSL TEXTDRAW ON
#endif
   
t1 = "Check FlagShip (or Clipper) environment."
t2 = "----------------------------------------"
t3 = "Info: you may abort the following tests at any time by pressing the"
t4 = "escape [Esc] key and return to the menu -or- by pressing the CRTL-K key."
sp1= space((max_col() - len(t1)) /2)
sp2= space((max_col() - len(t4)) /2)

scr_ok   = .F.                                             && not ok yet
clear_ok = .F.
protocol = .F.
store "" to prt_file, term_out, term_key, outmap, inmap
if FlagShip
   terminal = FS_SET ("term", @outmap, @term_out, @inmap, @term_key)
else
   terminal = "???"
endif
mapp_out = !empty(term_out)
mapp_in  = !empty(term_key)
temp_terminal = ""

declare msg[20]
afill (msg, "")

msg[ 3]= "--> Check the terminfo setting, especially for parameters:"
msg[ 4]= "or use the extended terminal definitions in /usr/lib/terminfo/FStinfo.src"
msg[ 5]= "    e.g. TERM=FSansi;export TERM                          ... for sh,ksh"
msg[ 6]= "    or   setenv TERM FSansi                               ... for csh"
msg[ 7]= "Refer also to the FlagShip User Manual, chapter 9"
msg[ 8]= " and curses manual pages."
msg[ 9]= "--> Check if the internal terminal setup (if any) corresponds"
msg[10]= "    to your actual terminal TERM=" + terminal
msg[11]= "--> Check the protocol setting using stty"
msg[12]= "    or set it with:  stty tab3 -istrip cs8 "
msg[13]= "--> Check 'mapchan' (if available) and disable it (e.g. 'mapchan -n')"
msg[14]= "--> Check 'mapkey' or 'ttymap' (if available) to confirm the FlagShip"
msg[15]= "    Installation Notes of your soft/hardware portation"
msg[16]= "--> If running X or other window software: "
msg[17]= "    check/set the correct font (e.g. xset, cmdtool -fn , hpterm -fn etc.)"
msg[18]= "--> If running with terminal emulation: check the emulation settings"
msg[19]= "--> If not console: reduce the transmission baud rate"

dispbegin()
dummy = replicate(".", maxcol())
for ii = 1 to maxrow()
   ? dummy  
next
screen1 = savescreen (0,0, maxrow(), maxcol())
dummy = replicate(" ", maxcol())
for ii = 1 to maxrow() +1
   ? dummy  
next
screen2 = savescreen (0,0, maxrow(), maxcol())
dispend()


do clear
? t1
? t2
?
? t3
? t4
?
? "Would you like to protocol the session "

if FlagShip
   prt_file = FS_SET ("printfile")
   ? "into print file " + prt_file
else
   ?? "to the printer"
endif
begin sequence
   if YES_NO ()
      set print on
      protocol = .T.
   endif
end

color = iscolor()
if color
   color1 = "W+/B,N/W,,,B/BG"
   color2 = "GR+/B,N/W,,,B/BG"
else
   color1 = "W+/N,N/W,,,N/W"
   color2 = "W+/N,N/W,,,WU"
endif


declare xmenu[8], xhelp[8]
xmenu[1] = "F: full checking      "
xmenu[2] = "T: terminal info      "
xmenu[3] = "S: screen output      "
xmenu[4] = "C: colors             "
xmenu[5] = "M: screen mapping     "
xmenu[6] = "K: keyboard: FN & curs"
xmenu[7] = "A: keyboard: any key  "
xmenu[8] = "Q: quit               "

xhelp[1] = "Perform all test sequentially"
xhelp[2] = "Test actual terminal, input/output mapping"
xhelp[3] = "Test screen output (terminfo settings)"
xhelp[4] = "Test screen colors and attributes"
xhelp[5] = "Test screen output (mapping) extd.char.set"
xhelp[6] = "Test keyboard, cursor & function keys"
xhelp[7] = "Test keyboard, any key you wish"
xhelp[8] = "Terminate this 'fscheck' program"

choice   = 0
xchoice  = 0
terminal = "?term?"
if FlagShip
   terminal = FS_SET ("terminal")
endif

do while .T.
   begin sequence
      do clear
      if choice # 1
         ? if (scr_ok, sp1+t1, t1)
         ? if (scr_ok, sp1+t2, t2)
         ?
         ? if (scr_ok, sp2+t3, t3)
         ? if (scr_ok, sp2+t4, t4)
         ?
         choice = main_menu (xmenu, xhelp, scr_ok, xchoice)
         xchoice = if (choice = 1, 2, choice)
      endif

      if xchoice > 3 .and. xchoice < len(xmenu) .and. !scr_ok
         ?
         ? "Did you check the SCREEN OUTPUT and is the output o.k."
         if !YES_NO()
            ? "Please, do it first"
            do wait_quit
            break
         endif
         scr_ok = .T.
       clear_ok = .T.
      endif

      do case
      case xchoice = 2
         do GENERAL_SET
         xchoice = if (choice = 1, 3, choice)
      case xchoice = 3
         do CHECK_SCREEN                                   && sets scr_ok
         xchoice = if (choice = 1, 4, choice)
      case xchoice = 4
         do CHECK_COLOR
         xchoice = if (choice = 1, 5, choice)
      case xchoice = 5
         do CHECK_MAPP
         xchoice = if (choice = 1, 6, choice)
      case xchoice = 6
         do KEYB_CHECK
         xchoice = if (choice = 1, 7, choice)
      case xchoice = 7
         do KEYB_ANY_KEY
         xchoice = if (choice = 1, 8, choice)
      otherwise
         if FlagShip .and. protocol
            @ 10+len(xmenu),max_col()
            ?
            ? "The protocol is stored in the " + prt_file + " file"
         endif
         set color to
         quit
      endcase
   end
   if mapp_out .or. !empty(temp_terminal)
      FS_SET ("outmap", "FSchrmap.def", term_out)
   endif
enddo

quit

* +-------------------------------------------------------------------------+
* | GENERAL_SET : check the FlagShip and UNIX settings                      |
* +-------------------------------------------------------------------------+
procedure general_set
private x, blank, err_scr, err_key, err_fi1, err_fi2

err_fi1 = .F.
err_fi2 = .F.
err_scr = .F.
err_key = .F.

if .not. FlagShip
   ?
   ? "Terminal/keyboard settings for Clipper conform to the MS-DOS standard"
   ?
   terminal = ""
   do wait_quit
   return
endif

terminal = FS_SET ("term", @outmap, @term_out, @inmap, @term_key)

do clear
? "General settings info about FlagShip:"
? replicate("-",78)
?
x = "Your actual terminal TERM is       : "
blank = space(len(x))

? x + terminal
?
? "The screen output mapping file "
if empty(outmap)
   ?? '"FSchrmap.def" CANNOT BE FOUND'
   err_scr = .T.
   err_fi1 = .T.
else
   ?? "is  : " + outmap
   x = substr("The actual terminal " + term_out + blank, 1, len(blank))
   if empty(term_out)
      ? x + "is NOT present there, NO MAPPING USED"
      err_scr = .T.
   else
      ? x + "is present there, screen will be mapped."
   endif
endif
?
? "The keyboard input mapping file "
if empty(outmap)
   ?? '"FSkeymap.def" CANNOT BE FOUND'
   err_key = .T.
   err_fi2 = .T.
else
   ?? "is : " + inmap
   x = substr("The actual terminal " + term_key + blank, 1, len(blank))
   if empty(term_key)
      ? x + "is NOT present there, NO MAPPING USED"
      err_key = .T.
   else
      ? x + "is present there, screen will be mapped."
   endif
endif

if err_scr
   ?
   ? "NOTE: your SCREEN OUTPUT will probably not work correctly, especially"
   ? "      line drawing characters, colors and the extended IBM character set."
   margin = space(6)
   if err_fi1
      ?
      ? "      It would be BEST to make the default FSchrmap.def (or your) mapping"
      ? "      file available to FlagShip by inserting the correct path into PATH"
      ? "      or by setting the environment variable SCRMAP"
   end
endif
if err_key
   ?
   ? "NOTE: your KEYBOARD INPUT will probably not work correctly, especially"
   ? "      cursor movement keys and function keys, because no keybord mapping"
   ? "      for your terminal " + terminal + " is defined in FSkeymap.def."
   ? "      You can use CTRL-keys instead, e.g. CTRL-x for cursor down, "
   ? "      see appendix and chapter 9 in the user manual."
   if err_fi1
      ?
      ? "      But, it would be BEST to make the default FSkeymap.def (or your) mapping"
      ? "      file available to FlagShip by inserting the correct path into PATH"
      ? "      or by setting the environment variable SCRMAP"
   end
endif
if err_fi1 .or. err_fi2
   ?? ", e.g."
   ?  "         SCRMAP=/usr/xxx/yyy ; export SCRMAP    (for sh and ksh)"
   ?  "         setenv SCRMAP /usr/xxx/yyy             (for csh)"
   ?  "     See more in chapter 9 of the User Manual."
end
?
do wait_quit

var = 12345.678
?
?
?
? "Testing the numeric output:"
?
? LTRIM(str(var,10,3)) + " <--- do you see 12345.678 (y) or 12345,678 (n)  "
if !YES_NO()
   ?
   ? "--> Before the program execution, disable the national numeric conversion,"
   if .not. empty(gete("LANG"))
      ? "    e.g. by disabling or reseting the environment variable LANG using"
      ? '         LANG="" ; export LANG                       (sh)'
      ? '         unsetenv LANG   -or-  setenv LANG ""        (csh)'
   endif
   ? "    to avoid difficulties within numeric input/output of SAY/GET/READ"
   ? "    See the UNIX manual for international support and 'man locale'"
   ?
   do wait_quit
endif
return

* +-------------------------------------------------------------------------+
* | CHECK_SCREEN: check the terminal, lines, rows                           |
* |                                                                         |
* | out: scr_ok, clear_ok                                                   |
* +-------------------------------------------------------------------------+
procedure check_screen
private x,y,chr,i,j,len,last12,dummy,act_row, a1,a2, mx,my, line0, scr

x = MAX_COL()
y = MAX_ROW()
last12 = ""
line0  = ""
scr_ok = .T.

a1 = "Checking SCREEN output."
a2 = "-----------------------"

******************************************* columns
do clear
for i = 0 to x
   chr = '.'
   if i % 5 = 0
      chr = ','
   end
   if i % 10 == 0
      chr = if (i > 99, chr(48+int((i-100)/10)), chr(48+int(i/10)))
   endif
   ?? chr
   line0 = line0 + chr
   if i > x - 12
      last12 = last12 + chr
   endif
next

?
? a1
? a2
?
? "Your terminal has " + ltrim(str(x +1)) + " columns,"
? "you should see:"
?
? "0....,....1..   up to   " + last12
?
? "on the first line. O.k."

if !YES_NO()
   scr_ok = .F.
   ?
   ? msg[3]
   ? "    cols#80  (or the terminal value)"
   output_fs(.T.)
   ? msg[18]
   ?
   ? msg[7]+ msg[8]
   do wait_quit
endif

******************************************* lines
do clear
for i = 0 to y
   ? strzero (i,3)
   do case
   case i = 5
      ?? "   " + a1
   case i = 6
      ?? "   " + a2
   case i = 8
      ?? "   Your terminal has " + ltrim(str(y +1)) + " rows (lines),"
   case i = 9
      ?? "   you should see 000 to " + strzero(y,3) + " on the left screen side"
   endcase
next
?? " is this o.k."
if !YES_NO()
   scr_ok = .F.
   ?
   ? msg[3]
   ? "    lines#25 (or the terminal value)"
   output_fs(.T.)
   ? msg[18]
   ?
   ? msg[7]+ msg[8]
   do wait_quit
endif


******************************************* scrolling up
do clear
for i = 0 to y
   ? strzero (i,3)
   do case
   case i = y-1
      ?? "   " + a1
   case i = y
      ?? "   " + a2
   endcase
   next
? "NEW"
? "NEW   Testing scrolling: you should see lines 002 ... " + strzero(y,3) + " o.k."
if !YES_NO()
   scr_ok = .F.
   ?
   ? msg[3]
   ? "    ind=<string> "
   output_fs(.T.)
   ? msg[18]
   ?
   ? msg[7]+ msg[8]
   ?
   do wait_quit
endif

if .not. scr_ok
   ?
   ? "--> You should terminate the program now, correct the settings and restart"
   ? "    again thereafter. If you continue, unpredictable results may occur."
   ? "    Abort now"
   if YES_NO()
      quit
   endif
endif

******************************************* x,y positioning
?
do clear

x   = MAX_COL()
y   = MAX_ROW()
my  = int(y/2)
mx  = max(round(x/y,1), 1)
len = max(int((x / y) + 0.9), 1)

@ 0,0 say line0
@ y,0 say line0

draw_X ()                                                  && draw large 'X'

for i = 0 to y                                             && line numbers
   dummy = ltrim(str(i))
   @ i, 0 say dummy
   @ i, x - len(dummy) +1 say dummy
next

dummy = "Do you see a large 'X' diagonal over the WHOLE screen"
@ my, max(int((x - len(dummy) - 15) / 2),1) say ""
?? dummy
if !YES_NO()
   scr_ok = .F.
   do clear
   ? "You have serious screen positioning problems, the application"
   ? "using the input/output system (@,SAY,GET) will not work properly"
   ?
   for i = 9 to 12
      ? msg[i]
   next
   ? msg[18]
   ? msg[19]
   ? msg[3]
   ? "    home=<string>                            and"
   ? "    cup=<string>                             and"
   ? "    cub1=<string>   -or-   cub=<string>      and"
   ? "    cuf1=<string>   -or-   cuf=<string>      and"
   ? "    cud1=<string>   -or-   cud=<string>      and"
   ? "    cuu1=<string>   -or-   cuu=<string>      and"
   ? "    - comment or delete:   ht=<string>       and/or"
   ? "    - comment or delete:   it#<number>       and/or"
   ? "    - comment or delete:   cbt=<string>      and/or"
   ? "    - comment or delete:   sgr=<string>      "
   output_fs(.F.)
   ?
   ? msg[7 ]+ msg[8 ]
   do wait_quit
end

if .not. scr_ok
   do abort_now
endif



******************************************* clear screen, full and partially
?
do clear

x   = MAX_COL()
y   = MAX_ROW()
my  = int(y/2)
mx  = int(x/2)

@ 0,0, y,x box "........."
draw_X ()

center(0, a1)
center(1, a2)
center(2, "Clear screen, fully and partially")
set console off
? "Clear screen, fully and partially"
set console on

save screen to scr
my_wait_center (y)

clear_ok = .T.

@ my+1,0 clear                                          && @ y,x clear
@ my+2,0 say ""
? "Is the lower half of the screen cleared"
if !YES_NO()
   scr_ok   = .F.
   clear_ok = .F.
endif

restore screen from scr
@ 0,0 clear to my,x                                     && @ y,x clear to y,x
@ 0,0 say ""
? "Is the upper half of the screen cleared"
if !YES_NO()
   scr_ok   = .F.
   clear_ok = .F.
endif

restore screen from scr
@ my+1,0 clear to y,mx                                  && @ y,x clear to y,x
set console off
? "Is the lower left part of the screen cleared"
set console on
@ my +3,0 say "Is the lower left part of"
@ my +4,0 say "the screen cleared"

if !YES_NO()
   scr_ok   = .F.
   clear_ok = .F.
endif

restore screen from scr
for i = 0 to my                                         && @ y,x
   @ i,mx+1
next
set console off
? "Is the upper right part of the screen cleared"
set console on
@ my -3,mx+2 say "Is the upper right part of"
@ my -2,mx+2 say "the screen cleared"
if !YES_NO()
   scr_ok   = .F.
   clear_ok = .F.
endif

restore screen from scr
my_wait_center (y)
clear                                                   && clear
?
?? "Is the WHOLE screen cleared"
if !YES_NO()
   scr_ok   = .F.
   clear_ok = .F.
endif

if !clear_ok
   do clear
   ?
   ? msg[3 ]
   ? "    clear=<string>"
   ? "    ed=<string>"
   ? "    home=<string>"
   ? "    cup=<string>"
   ? "    - comment or delete:   ht=<string>       and/or"
   ? "    - comment or delete:   it#<number>       and/or"
   ? "    - comment or delete:   cbt=<string>      and/or"
   ? "    - comment or delete:   sgr=<string>      "
   output_fs(.T.)
   ? msg[18]
   ? msg[19]
   ?
   ? msg[7 ]+ msg[8 ]
   do wait_quit

   do abort_now
endif


******************************************* scrolling +/-
do clear
for i = 0 to y
   ? "FIX" + strzero (i,3)
   do case
   case i = 1
      ?? "   " + a1
   case i = 2
      ?? "   " + a2
   case i = 3
      ?? "   Scrolling up and down within the window lines 9 ... 19"
   case i = 4
      ?? "   (with step interval 1 second)"
   case i = 8 .or. i = 20
      ?? replicate("-",max_col() - 7)
   endcase
next
@ 8, max_col() say ""
for i = 9 to 19
   ? "new" + strzero (i,3)+ " old text at line " + ltrim(str(i))
next
@ 22,9 say "...wait..."
for i = 1 to 7
   inkey (1)
   scroll (9,0, 19,max_col(), 1)
   @ 19,0 say "new" + strzero (i+19,3) + " NEW text at line 19 after scrolling up"
next
@ 22,6 say ""
?? "   Do you see the new016 ... new026 lines in the window"
if !YES_NO()
   @ 23,0 clear
   ? msg[3]
   ? "    ind=<string>    and"
   ? "    ri=<string> "
   output_fs(.T.)
   ? msg[7]+ msg[8]
   do wait_quit
else
   @ 9,0 clear to 19,max_col()
   @ 22,9
   @ 22,9 say "...wait..."
   @ 8, max_col() say ""
   for i = 9 to 19
      ? "new" + strzero (i-9,3)+ " old text at line " + ltrim(str(i))
   next
   for i = 1 to 7
      inkey (1)
      scroll (9,0, 19,max_col(), -1)
      @ 9,0 say "new" + strzero (-i,3) + " NEW text at line 9 after scrolling down"
   next
   @ 22,6 say ""
   ?? "   Do you see the new-07 ... new003 lines in the window"
   if !YES_NO()
      @ 23,0 clear
      ? msg[3]
      ? "    ri=<string> "
      output_fs(.T.)
      ? msg[7]+ msg[8]
      do wait_quit
   endif
endif
return

* +-------------------------------------------------------------------------+
* | CHECK_MAPP:   check the full ASCII char.set & mappings                  |
* |                                                                         |
* | out: scr_ok, clear_ok                                                   |
* +-------------------------------------------------------------------------+
procedure check_mapp
private x,y,chr,i,j,len,last12,dummy,act_row, a1,a2, mx,my, line0, scr

x = MAX_COL()
y = MAX_ROW()
last12 = ""
line0  = ""
scr_ok = .T.

a1 = "Checking SCREEN output"
a2 = "----------------------"

do clear
? a1
? a2
? "Checking the mapping and the output for extended (IBM-PC8) ASCII character set"

temp_terminal = ""

declare aa[2], ab[2]
aa[1] = chr(92)
aa[2] = chr(91)
ab[1] = "a backslash"
ab[2] = "left bracket"
test_out (aa, ab, 1, 10)

declare ba[4], bb[4]
ba[1] = chr(154) + chr(220) + chr(219)    && U-umlaut: ASCII, ISO, HP
ba[2] = chr(132) + chr(228) + chr(204)    && a-umlaut: ASCII, ISO, HP
ba[3] = chr(218)                          && upper left corner
ba[4] = chr(192)                          && lower left corner
bb[1] = "U-umlaut"
bb[2] = "a-umlaut"
bb[3] = "upper left corner or +"
bb[4] = "lower left corner or +"
test_out (ba, bb, 2, 14)

if !empty(temp_terminal)
   @ 20,0 say ""
   ?
   ? "Your terminal mapping was redefined to " + temp_terminal + " during this test"
   ? "Should I continue with the screen test"
   if !YES_NO ()
      ? "--> Insert your actual terminal " + terminal
      ? "    into FSchrmap.def according to the terminal " + temp_terminal
      ?
      ? msg[7]
      ?
      do wait_quit
      return
   else
      ? "Remember: if problems occurs: press ctrl-K, wait 3 seconds"
      ? "          and press ctrl-K again."
      do wait_quit
   end
   a1 = a1 + " with screen mapping " + temp_terminal + " (see FSchrmap.def)"
endif

******************************************* lower ASCII set
do clear with .T.
? a1
? a2
? "Checking the output for the extended (IBM-PC8) ASCII character set"
? "1. characters 32(dec) to 127(dec), standard ASCII :"

act_row = row()
out_char (32, 127)
?
? "You should see:"
? "   32: space,   33: exclamation mark,  34: apostrophe ..."
? "   64: at sign, 65: 'A',  65: 'B' ...  92: backslash ...."
?
if !empty(temp_terminal) .or. mapp_out
   ? "Note: some of the characters may be mapped to graphically similar chara-"
   ? "      cters, if the full character set is not available on your terminal"
   ?
endif
? "Is the output o.k."
if !YES_NO()
   ? "--> check the above mapping again."
   do wait_quit
   return
endif

******************************************* upper ASCII set
do clear with .T.
? a1
? a2
? "2. characters 128(dec) to 255(dec), extended PC-8 ASCII set :"

out_char (128, 255)
?
? "You should see:"
? "   132: a-umlaut, 153: O-umlaut,   225: beta"
? "   196: horizontal line or -       179: vertical line or : or |"
? "   218: upper left corner or +     191: upper right corner or +"
?
if !empty(temp_terminal) .or. mapp_out
   ? "Note: some of the characters may be mapped to graphically similar chara-"
   ? "      cters, if the full character set is not available on your terminal"
   ?
endif
? "Is the output o.k."
if !YES_NO()
   ? "--> check the above mapping again."
   do wait_quit
   return
endif

******************************************* lower ASCII set 1..31
restscreen (0,0, maxrow(), maxcol(), screen1)
do clear with .T.
? a1
? a2
? "3. in the program generally unused characters 1(dec) to 31(dec)"
? "   Note: on some terminals, problems may occur because of conflicts"
? "         with the control chars for the actual stty setting."
? "   Check it"
if YES_NO()
   out_char (1, 31)
   ?
   ? "You may see:"
   ? "   24: up arrow   25: down arrow   26: right arrow   27: left arrow"
   ?
   if !empty(temp_terminal) .or. mapp_out
      ? "Note: some of the characters may be mapped to graphically similar chara-"
      ? "      cters, if the full character set is not available on your terminal"
      ?
   endif
   ? "Is the output o.k."
   if YES_NO()
      ? "--> fine."
   else
      ? "--> You may try to map some of these characters using FSchrmap.def"
      ? "    Check your terminal description, use the alternate or protected"
      ? "    character set"
      output_fs(.F.)
      ?
      ? msg[7]
   endif
   do wait_quit
end

******************************************* drawing ASCII set
restscreen (0,0, maxrow(), maxcol(), screen1)
do clear with .T.
? a1
? a2
? "Checking the output for the extended (IBM-PC8) ASCII character set"
? "4. drawing characters with the extended PC-8 ASCII set :"
@ 6,0 to 10,15
@ 8,2 say "single line"

* @ 6,20,10,35 box chr(218)+chr(196)+chr(191)+chr(179)+ ;
*                  chr(217)+chr(196)+chr(192)+chr(179) + chr(176)
* the above statement may cause "express. too complex" with few
* compilers, and are therefore splitted into two statements:
x_box = chr(218)+chr(196)+chr(191)+chr(179)
x_box = x_box + chr(217)+chr(196)+chr(192)+chr(179) + chr(176)
@ 6,20,10,35 box x_box

@ 8,23 say "filled box"

@ 12,0 to 16,15 double
@ 14,2 say "double line"

@  6,40 to 16,70 double
@ 10,40 say chr(199) + replicate(chr(196),29) + chr(182)
@ 12,40 say chr(204) + replicate(chr(205),29) + chr(185)
@  7,50 to 15,50
@  7,60 to 15,60 double
@  6,50 say chr(209)
@  6,60 say chr(203)
@ 10,50 say chr(197)
@ 10,60 say chr(215)
@ 12,50 say chr(216)
@ 12,60 say chr(206)
@ 16,50 say chr(207)
@ 16,60 say chr(202)
@ 11,53 say "boxes"
@ 18,0 say ""
if !empty(temp_terminal) .or. mapp_out
   ? "Note: some of the line drawing characters may be mapped to - : + characters,"
   ? "      if the drawing character set is not available on your terminal"
   ?
endif
? "Are the boxes o.k."
if !YES_NO()
   ? "--> Check the above extended character set"
   do wait_quit
endif
restscreen (0,0, maxrow(), maxcol(), screen1)
do clear with .T.

return

* +-------------------------------------------------------------------------+
* | service routine for CHECK_SCREEN                                        |
* +-------------------------------------------------------------------------+
function output_fs
parameters newline
if substr(terminal,1,2) != "FS"
   if newline
      ?
   endif
   ? msg[4]
   ? msg[5]
   ? msg[6]
   if newline
      ?
   endif
endif
return .T.
* +-------------------------------------------------------------------------+
* | service routine for CHECK_SCREEN                                        |
* +-------------------------------------------------------------------------+
function out_char
parameters from, to
private x,y,i
y = row() + 2
x = 0
for i = from TO to
   @ y,x say str(i,3) + " " + chr(i)
   x = x + 6
   if x > max_col() - 5
      x = 0
      y = y + 1
   end
next
return .T.

* +-------------------------------------------------------------------------+
* | service routine for CHECK_SCREEN                                        |
* +-------------------------------------------------------------------------+
function draw_x
private i,j,x,y,my,mx,len,dummy

x   = MAX_COL()
y   = MAX_ROW()
my  = int(y/2)
mx  = max(round(x/y,1), 1)
len = max(int((x / y) + 0.9), 1)
dummy = replicate ("*", len)

for i = 0 to y                                           && draw > part of X
   j = round (if (i <= my, i* mx, (y-i)* mx), 0)
   @ i, j say dummy
next
for i = y to 0 step -1                                   && draw < part of X
   j = round (1 + x - (if(i >= my, y-i+1, i+1)) * mx, 0)
   if j + len > x -1
      j = x - len +1
   endif
   @ i,j say dummy
next
return .T.

* +-------------------------------------------------------------------------+
* | service routine for CHECK_SCREEN                                        |
* +-------------------------------------------------------------------------+
function my_wait_center
parameters row
private scr
scr = savescreen (row,0, row, max_col())
center(row, " Press any key ... ")
inkey(0)
restscreen (row,0, row, max_col(), scr)
if lastkey() = 27
   choice = 0
   break
endif
return .T.

* +-------------------------------------------------------------------------+
* | service routine for CHECK_SCREEN                                        |
* +-------------------------------------------------------------------------+
procedure abort_now
   ?
   ? "--> You should terminate the program now, correct the settings and restart"
   ? "    again thereafter. If you continue, unpredictable results may occur."
   ? "    Abort now"
   if YES_NO()
      quit
   endif
return

* +-------------------------------------------------------------------------+
* | service routine for CHECK_SCREEN                                        |
* +-------------------------------------------------------------------------+
function center
parameters row, text
@ row, int((max_col() - len(text)) / 2) say text
return .T.

* +-------------------------------------------------------------------------+
* | service routine for CHECK_MAP                                           |
* +-------------------------------------------------------------------------+
function test_out
parameters arr1, arr2, part, row
private i,ii, y1,y2, try_map, set_map, first, first_msgout, chr, pos,scr

declare fschrmap[10]
#ifdef FGSLINUX
fschrmap[1] = "fsansi"
fschrmap[2] = "fsVT100"
fschrmap[3] = "fslinux"
fschrmap[4] = "fslinxterm"
fschrmap[5] = "fssun"
fschrmap[6] = "fssunxterm"
fschrmap[7] = "fssuncons"
fschrmap[8] = "fswy60-25"
fschrmap[9] = "fs97801-725"
fschrmap[10] = "fs97801-825c"
#else
fschrmap[1] = "FSansi"
fschrmap[2] = "FSvt100"
fschrmap[3] = "FSlinux"
fschrmap[4] = "FSlinxterm"
fschrmap[5] = "FSsun"
fschrmap[6] = "FSsunxterm"
fschrmap[7] = "FSsuncons"
fschrmap[8] = "FSwy60-25"
fschrmap[9] = "FS97801-725"
fschrmap[10] = "FS97801-825c"
#endif

*X  <-- it this character xxxxxxxxxxxxxxxxxxxx                (y|j|o / N) ?
*       Should I try to use a other mapping from FSchrmap.def (y|j|o / N) ?

try_map = .F.
set_map = .F.
first   = .T.
first_msgout = .T.
@ 4,0 clear to 8,max_col()
if !empty(temp_terminal)
   @ 4,0 say "using temporary mapping for TERM=" + temp_terminal
endif

pos = 0
if len(arr1[1]) > 1
   for i = 1 to len(arr1[1])
      chr = substr(arr1[1],i,1)
      y1 = row
      y2 = y1 + 1
      if char_out(y1, chr, arr2[1])
         chr = substr(arr1[2],i,1)
         y1 = row + 2
         y2 = y1 + 1
         if char_out(y1, chr, arr2[2])
            ?
            @ 5,0 say ""
            do case
            case i = 1
               ?? "--> Your terminal seems work correctly"
            case i = 2
               ?? "--> Your terminal uses ISO characters"
            otherwise
               ?? "--> Your terminal uses the HP roman character set"
            end
            pos = i
   exit
         endif
      endif
   next
   if pos > 1
      ?
      @ y2,0
      @ y2,7 say ""
      ?? "Should I try to use the " + fschrmap[pos] + " mapping from FSchrmap.def"
      @ y2,62 say ""
      if YES_NO(.F.)
         if out_warning()
            @ 8,0
            @ 7,max_col() say ""
            if (FS_SET ("outmap", "FSchrmap.def", fschrmap[pos]))
               temp_terminal = fschrmap[pos]
               ? "-->      The mapping for " + fschrmap[pos] + " is active."
               set_map = .T.
            else
               ? "-->      Sorry, the mapping for " + fschrmap[pos] + " is not available."
            endif
         endif
      endif
      @ y2,0
   else
      set_map = .T.
   endif
endif

for loop = 1 to len(fschrmap)
   try_map = .F.
   @ row,0 clear
   for i = 1 to len(arr1)
      chr = substr(arr1[i],1,1)
      y1 = row + (i-1) * 2
      y2 = y1 + 1
      if !char_out(y1, chr, arr2[i])
         if !try_map .and. first_msgout
            save screen to scr
            test_msgout (part)
            rest screen from scr
            first_msgout = .F.
         endif
         @ y2,7 say "Should I try to use"
         if mapp_out
            @ y2, col() say " a other mapping"
         else
            @ y2, col() say " mapping from FSchrmap.def"
         end
         @ y2,62 say ""
         if YES_NO(.F.)
            try_map = .T.
   exit
         endif
      endif
   next
   if set_map .and. !try_map
exit
   endif

   if try_map
      if first
         first = .F.
         @ 8,0
         if !out_warning()
            exit
         endif
      end
      set_map = .F.

      for ii = loop to len(fschrmap)
         @ 8,0
         @ 7,max_col() say ""
         ? "         Setting mapping for " + fschrmap[ii] + ", o.k."
         if YES_NO ()
            loop = ii
            FS_SET ("outmap", "FSchrmap.def", fschrmap[loop])
            temp_terminal = fschrmap[loop]
            set_map = .T.
      exit
         endif
      next
   else
      exit
   endif
next
return .T.

function char_out
parameters row, chr, text
if protocol
   set console off
      ? "Testing output for chr(" + ltrim(str(asc(chr))) + ") = " + text
      ?? if (!empty(temp_terminal), " with " + temp_terminal, "")
      ?? " o.k."
   set console on
end
@ row,0 clear to row+1,max_col()
@ row,0  say chr + "  <-- is this character " + text
@ row,62 say ""
return YES_NO()

function test_msgout
parameter part
private i
?
? "--> Your screen terminal is probably set to the national 7 bit character set."
? "    If possible, use 8 bit settings"
?
if mapp_out
   ? "--> The output (screen) mapping for your terminal TERM=" + terminal
   ? "    is defined in the file " + outmap
   ? "--> Check this mapping"
end
if substr(terminal,1,2) != "FS"
   ?? " or use one of FlagShip predefined terminals"
end
?
for i = 9 to 19
   ? msg[i]
next
?
? msg[7 ]
?
do wait_quit
return .T.

function out_warning
@ 4,0 say "WARNING: wrong terminal mapping may in the EXTREME case disable your "
@ 5,0 say "-------  tty connection or reset the terminal to the default status."
@ 6,0 say "         If such a case happens, press ctrl-K, pause 3 seconds"
@ 7,0 say "         and press ctrl-K again, reset the terminal and start again."
? "         Continue"
if !YES_NO()
   return .F.
end
return .T.

* +-------------------------------------------------------------------------+
* | CHECK_COLOR : check colors                                              |
* +-------------------------------------------------------------------------+
procedure check_color
private i,j,k,col,row,add,dummy


do clear
set color to
if color
   ? "Checking COLOR output."
   ? "----------------------"
else
   ? "Checking screen attributes."
   ? "---------------------------"
endif
?
? "Your terminal has " + if(color,"","NOT ") + "color capability. Correct"
if !YES_NO()
   ?
   ? "Check the terminfo setting, especially for parameters"
   ? "   colors#8,pairs#64"
   output_fs(.F.)
   do wait_quit
   color = if (color, .F., .T.)
endif

SET GUICOLORS ON     // to display colors in GUI mode too

if color
   declare colpair[8]
   colpair[ 1] = "W  white"
   colpair[ 2] = "N  black"
   colpair[ 3] = "B  blue"
   colpair[ 4] = "G  green"
   colpair[ 5] = "BG cyan"
   colpair[ 6] = "R  red"
   colpair[ 7] = "RB magenta"
   colpair[ 8] = "GR brown"

   set color to &color1
   @ 4,0 clear
   ?
   ? "Is the current output white on blue"
   if !YES_NO()
      ? "Is the current output white on red"
      if YES_NO()
         set color to
         @ 8,0 clear
         ?
         ? "Your curses library supports the oposite color mapping (IBM or ISO), than"
         ? "set in the actual FlagShip configuration."
         ?
         ? "-> modify the file /usr/include/FlagShip.h :"
         ? "   cd /usr/include"
         ? "   cp FlagShip.h FlagShip.h.sav"
         ? "   vi FlagShip.h"
         ? "-> look at the definition on line 9: #define IBM_COLORS"
         ? "     if active : mark it as comment  -->  /* #define IBM_COLORS */"
         ? "     if comment: disable the comment -->  #define IBM_COLORS"
         ? "-> and recompile this application fscheck.prg again."
      else
         set color to
         @ 8,0 clear
         ?
         ? "Your curses library does not support the standard IBM or ISO colors,"
         ? "but you can modify the actual FlagShip configuration:"
         ?
         ? "->  cd /usr/include"
         ? "->  cp FlagShip.h FlagShip.h.sav"
         ? "->  vi FlagShip.h"
         ? "->  Look at the definition on line 9: #define IBM_COLORS"
         ? "    if active : mark it as comment  -->  /* #define IBM_COLORS */"
         ? "->  Go to file bottom, change the definition FS_COLOR_RED from 4"
         ? "    to the correct number (0...8), check also the other colors."
         ? "->  Recompile this application fscheck.prg again."
      endif
      do wait_quit
   endif
else
   declare colpair[4]
   colpair[ 1] = "W  white"
   colpair[ 2] = "N  black"
   colpair[ 3] = "U  underline"
   colpair[ 4] = "I  reverse"
endif

set color to
@ 4,0 clear
col = 0
row = 4
for k = 1 to if(color,2,4)
   do case
   case k = 1
      add = ""
   case k = 2
      add = "+"
   case k = 3
      add = "*"
   case k = 4
      add = "+*"
   end
   for i = 1 to len(colpair)
      for j = 1 to len(colpair)
         if i = j
            loop
         endif
         dummy = trim(substr(colpair[j],1,2)) + add + "/" +;
                 trim(substr(colpair[i],1,2))
         setcolor (dummy)
         dummy =  substr("  " + dummy + space(9), 1, 10)
         @ row, col say dummy
         col = col + 11
         if (col > max_col() -10)
            col = 0
            row = row + if (len(colpair) > 4, 1, 2)
         endif
      next
      if col > 0 .and. j > 5
         col = 0
         row = row + 1
      endif
   next
next
set color to

SET GUICOLORS OFF // disable colors in GUI mode

?
if color
   ? "Are the colors o.k."
   if !YES_NO()
      ? "--> Check the available color attributes in your terminal manual."
      ? "    Some terminals cannot display the full color set or"
      ? "    do not support the background colors."
      ? "--> Check the terminfo setting for the entry"
      ? "    colors#8, pairs#64"
      ? "--> Check the correct terminfo setting for"
      ? "    op=<string>, "
      ? "    setf=<string>,          setb=<string>, "
      ? "    setaf=<string>,         setab=<string>,     (SVR4 system only)"
      ? "    bold=<string>,          rev=<string>, "
      ? "    sgr0=<string>, "
      ? "    or mark  sgr=<string>   and/or"
      ? "             smso=<string>  and/or"
      ? "             ht=<string>    as comment "
      output_fs(.F.)
   endif
else
   ? "Is the monochrome screen output o.k."
   if !YES_NO()
      ? "--> Check the available screen attributes in your terminal manual."
      ? "    Some terminals cannot display inverse or blinking characters."
      ? "--> Check the correct terminfo setting for"
      ? "    op=<string>, "
      ? "    blink=<string>, "
      ? "    bold=<string>,          rev=<string>, "
      ? "    smul=<string>,          rmul=<string>, "
      ? "    sgr0=<string>, "
      ? "    or mark  sgr=<string>   and/or"
      ? "             smso=<string>  and/or"
      ? "             ht=<string>    as comment "
      output_fs(.F.)
   endif
endif
do wait_quit


return

* +-------------------------------------------------------------------------+
* | KEYB_CHECK  : check the keys according to FlagShip/Clipper definitions  |
* +-------------------------------------------------------------------------+
procedure keyb_check

clear typeahead

a1 = "Checking KEYBOARD input."
a2 = "------------------------"

do fill_keys

do clear
? a1
? a2
?
? "1. check cursor keys"
? "--------------------"
? "If the key is not available on your keyboard: press space key"
?
do keyb_warning
do keyb_cursor

clear typeahead

do clear
? a1
? a2
?
? "2. check function keys"
? "----------------------"
? "If the key is not available on your keyboard: press space key"
?
do keyb_warning
do keyb_fn_keys

return


* +-------------------------------------------------------------------------+
* | KEYB_CURSOR : check the cursor keys (from KEYB_CHECK)                   |
* +-------------------------------------------------------------------------+
procedure keyb_cursor
private ii, dummy, ok, text

ok = .T.
for ii = 1 to 19
   clear typeahead
   dummy = substr(key_name[ii] + space(20), 1, 20)
   ? "Press the " + dummy + "key now (or 'q') --> "
   pressed = inkey(0)
   if ii = 1 .and. pressed = asc("q")
      return
   endif
   if pressed = asc("q")
      exit
   endif
   if !output_key(ii, pressed)
      ok = .F.
   endif
next

text = if (ok, " and o.k.", "")
if ii # 20
   ? "---> aborted. Do you need help "
else
   ?
   ? "Test of cursor keys complete" + text + ", do you need help "
endif
if YES_NO()
   ?
   if !ok
      ? "Some of the keys pressed were not correct."
      ?
   endif
   ? msg[3 ]
   ? "    kcuu1=<string>,  kcud1=<string>,  kcub1=<string>,  kcuf1=<string>,"
   ? "    kpp=<string>,    knp=<string>,    khome=<string>,  kend=<string>,"
   ? "    kich1=<string>,  kdch1=<string>,  kbs=<string>,"
   ? "    ked=<string>,    kel=<string>,    kri=<string>,    kind=<string>,"
   ? "    khs=<string>,    kctab=<string>"
   output_fs(.F.)
   ? msg[9 ]
   ? msg[10]
   ? msg[11]
   ? "    for conflicts and/or set it using e.g.: 'stty susp ^@' etc."
   for ii = 13 to 18
      ? msg[ii]
   next
   ?
   ? msg[7 ]
   ?
   do wait_quit
endif
return

* +-------------------------------------------------------------------------+
* | KEYB_FN_KEYS: check the function keys (from KEYB_CHECK)                 |
* +-------------------------------------------------------------------------+
procedure keyb_fn_keys
private ii, dummy, ok, text

ok = .T.
for ii = 21 to 68
   clear typeahead
   dummy = substr(key_name[ii] + space(20), 1, 20)
   ? "Press the " + dummy + "key now (or 'q') --> "
   pressed = inkey(0)
   if ii = 21 .and. pressed = asc("q")
      return
   endif
   if pressed = asc("q")
      exit
   endif
   if !output_key(ii, pressed)
      ok = .F.
   endif
next

text = if (ok, " and o.k.", "")
if ii # 69
   ? "---> aborted. Do you need help "
else
   ?
   ? "Test of cursor keys complete" + text + ", do you need help "
endif
if YES_NO()
   ?
   if !ok
      ? "Some of the keys pressed were not correct."
      ?
   endif
   ? msg[3 ]
   ? "    kf1=<string>,  kf2=<string>,   .... etc, up to"
   ? "    kf47=<string>, kf48=<string>"
   output_fs(.F.)
   ? msg[9 ]
   ? msg[10]
   ? msg[11]
   ? "    for conflicts and/or set it using e.g.: 'stty susp ^@' etc."
   for ii = 13 to 18
      ? msg[ii]
   next
   ?
   ? msg[7 ]
   ?
   do wait_quit
endif
return

* +-------------------------------------------------------------------------+
* | KEYB_ANY_KEY: check the escape sequences for any key                    |
* +-------------------------------------------------------------------------+
procedure keyb_any_key
private x,y,outval

a1 = "Checking KEYBOARD input."
a2 = "------------------------"

clear typeahead
do clear
? a1
? a2
? "End test with the 'q' key"
?
do keyb_warning
do fill_keys

do while .T.
   ? "Press any key now (or 'q') --> "
   clear typeahead
   x = inkey(0)
   if upper(chr(x)) == "Q"
      exit
   endif
   y = x
   outval = ""
   do while (.T.)
      if x != 0
        outval = outval + str(x,3,0) + " "
      else
         exit
      endif
      x = inkey()
   enddo
   output_key(0, y, outval)
enddo
return

* +-------------------------------------------------------------------------+
* | OUTPUT_KEY  : output pressed key                                        |
* +-------------------------------------------------------------------------+
function output_key
parameters ii, pressed, string
private jj, ok, ret_cd

if pcount() < 3
   string = ""
endif
if len(string) > 4
   ?? "seqence: " + string
   return .T.
endif

ret_cd = .T.
jj = ii
if ii > 0
   if pressed == key[ii]
      ?? "ok,   "
   else
      ?? chr(7)
      ?? "FALSE "
      ii = 0
      ret_cd = .F.
   endif
endif
if ii = 0
   for jj = 1 to len(key)
      if key[jj] # 0 .and. key[jj] = pressed
         exit
      endif
   next
endif

ok = .F.
if jj > 0 .and. jj <= len(key) .and. pressed < len(key)
   if key[jj] == pressed
      ok = .T.
   endif
endif
if ok
   ?? key_name[jj]
else
   if pressed >= 32 .and. pressed <= 255
      ?? "letter [" + chr(pressed) + "]"
   elseif pressed  < 32
      ?? "ctrl key [^" + chr(pressed+64) + "]"
   else
      ?? "UNKNOWN key pressed"
      ret_cd = .F.
   endif
endif
do while col() < 74
   ?? " "
enddo
?? "= " + ltrim(str(pressed))

return ret_cd

* +-------------------------------------------------------------------------+
* | KEYB_WARNING: service for keyboard check                                |
* +-------------------------------------------------------------------------+
procedure keyb_warning

? "Note: if the stty setting is not checked yet, some of the pressed keys may"
? "      in extreme cases abort your application, because stty has a higher"
? "      preference then your application (refer chapter 9 of User manual)."
? "      If so, check the conflicts with stty setings and the predefined "
? "      FlagShip ctrl-keys (appendix A3 in User Manual) and change the stty"
? "      setting using e.g. 'stty intr ^C' or 'stty quit ^A' or 'stty susp ^@'."
? "      The same conflicts may occur with some terminal emulation programs."
?
return

* +-------------------------------------------------------------------------+
* | MAIN_MENU   : output main menu, return choice                           |
* +-------------------------------------------------------------------------+
function main_menu
parameters array, help, screen_ok, xchoice
private hi,br,left

if !screen_ok
   * first entry: simulate PROMPT/MENU TO until screen is ok
   ?
   set print off
   for i = 1 to len(array)
      ? array[i] + space(10) + help[i]
   next
   ?
   xchoice = ""
   do while type("xchoice") != 'N'
      wait "---------------------- enter choice (F,T,S,C,M,K,A,Q): " to xchoice
      if empty(xchoice) .or. lastkey() = 27
         xchoice = "Q"
      end
      for i = 1 to len(array)
         if (upper(xchoice)) == substr(array[i],1,1)
            xchoice = i
            exit
         endif
      next
   enddo
   if protocol
      set print on
   endif

else
   hi   = len(array)
   br   = len(array[1])
   left = int((max_col() - br) /2)
   xchoice = xchoice +1
   @ 8,left-1 to 8+hi+1, left+br

   set message to 8+hi+2 center
   for i = 1 to len(array)
      @ 8+i,left prompt array[i] message help[i]
   next
   menu to xchoice
endif
return xchoice

* +-------------------------------------------------------------------------+
* | CLEAR       : - safe clear screen, if not ok yet                        |
* |               - ctandard clear, if already checked                      |
* +-------------------------------------------------------------------------+
procedure clear
parameter must
private i, spaces

spaces = " "
if pcount() < 1
   must = .F.
endif

if must .or. !clear_ok
#ifdef CLEAR_SCREEN_FULL
   spaces = space(max_col() + 1)
   set print off
   for i = 1 to max_row() + 1
      ? spaces
   next
   if protocol
      set print on
   endif
#else
   dispbegin()
   restscreen (0,0, maxrow(), maxcol(), screen1)
   restscreen (0,0, maxrow(), maxcol(), screen2)
   dispend()
#endif
endif
clear

return

* +-------------------------------------------------------------------------+
* | WAIT_QUIT   : wait until key pressed, check for Esc key                 |
* +-------------------------------------------------------------------------+
procedure wait_quit
wait
if lastkey() = 27
   choice = 0
   break
endif
return

* +-------------------------------------------------------------------------+
* | yes_no() wait for key, check it for y/n or Esc                          |
* |                                                                         |
* | returns .T. if y,Y,j,J,o,O was entered                                  |
* +-------------------------------------------------------------------------+
function yes_no
parameters p1
if pcount() < 1
   p1 = .T.
endif
if !p1
   set print off
endif
private key
?? " (y|j|o / N) ? "
key = 0
do while (key != 27) .and. .not. upper(chr(key)) $ "YJON"
   key = inkey(0)
enddo
if protocol .and. !p1
   set print on
endif
if key = 27
   choice = 0
   break
endif
if key > 32
   ?? upper(chr(key))
endif
return upper(chr(key)) $ "YJO"


* +-------------------------------------------------------------------------+
* | FILL_KEYS   : assign function keys into array                           |
* +-------------------------------------------------------------------------+
procedure fill_keys
private ii
public key[70],key_name[70]

afill (key, 0)

key [ 1] = 4
key [ 2] = 19
key [ 3] = 5
key [ 4] = 24
key [ 5] = 1
key [ 6] = 6
key [ 7] = 18
key [ 8] = 3
key [ 9] = 22
key [10] = 7
key [11] = 8
key [12] = 9
key [13] = 2
key [14] = 26
key [15] = 29
key [16] = 23
key [17] = 31
key [18] = 30
key [19] = 27
key [21] = 28
for ii = 2 to 10
   key [ii +20] = -ii +1                            && F2..F10
next
for ii = 1 to 10
   key [ii +32] = -ii -9                            && shiftF1..F10
next
for ii = 1 to 10
   key [ii +44] = -ii -19                           && ctrlF1..F10
next
for ii = 1 to 10
   key [ii +56] = -ii -29                           && altF1..F10
next
key [31] = -40                                      && F11
key [32] = -41                                      && F12
key [43] = -42                                      && shiftF11
key [44] = -43                                      && shiftF12
key [55] = -44                                      && ctrlF11
key [56] = -45                                      && ctrlF12
key [67] = -46                                      && altF11
key [68] = -47                                      && altF12

key_name [ 1] ="cursor right"
key_name [ 2] ="cursor left"
key_name [ 3] ="cursor up"
key_name [ 4] ="cursor down"
key_name [ 5] ="home"
key_name [ 6] ="end"
key_name [ 7] ="PgUp"
key_name [ 8] ="PgDn"
key_name [ 9] ="insert"
key_name [10] ="delete"
key_name [11] ="backspace"
key_name [12] ="tab"
key_name [13] ="CTRL cursor right"
key_name [14] ="CTRL cursor left"
key_name [15] ="CTRL home"
key_name [16] ="CTRL end"
key_name [17] ="CTRL PgUp"
key_name [18] ="CTRL PgDn"
key_name [19] ="Esc"
for ii = 1 to 12
   key_name [ii +20] = "F" + ltrim(str(ii))         && 21..32 = F1...F12
next
for ii = 1 to 12
   key_name [ii +32] = "shift-F" + ltrim(str(ii))   && 33..44 = shiftF1...F12
next
for ii = 1 to 12
   key_name [ii +44] = "ctrl-F" + ltrim(str(ii))    && 45..56 = ctrlF1...F12
next
for ii = 1 to 12
   key_name [ii +56] = "alt-F" + ltrim(str(ii))     && 57..68 = altF1...F12
next
return

* +-------------------------------------------------------------------------+
* | FS_SET() dummy function, enable it for Clipper only                     |
* +-------------------------------------------------------------------------+
* function FS_SET
* parameters p1,p2,p3,p4,p5
* return .T.
* function max_row
* return 24
* function max_col
* return 79

