/* For FlagShip 5 and 6
 * Copyright (c) 2003..05 by multisoft GmbH, Munich, Germany 
 *
 * You may freely modify this source according to your needs and/or
 * re-use any parts of it in your source code.
 * Release $Date: 2005/11/23 10:22:26 $
 *--------------------------------------------------------------------------
 * This application demonstrates use of nested PROMPTs with additional
 * clauses like HEIGH, WIDTH, COLOR, GUICOLOR, TOOLTIP, STYLE, SELECT etc.
 * and different look & feel of the display
 *
 * Copy the source to your working directory, and
 * Compile: FlagShip menu_prompt.prg -o menu_prompt
 * Run    : ./menu_prompt               ## in GUI mode
 *     or : newfswin menu_prompt -io=t  ## in terminal i/o mode
 *
 * MS-Windows: FlagShip menu_prompt.prg [-io-a|g|t|b] 
 * Run:        menu_prompt  -or-  menu_prompt -io=t
 */

#include "box.fh"

// set font "courier", 10

// SET GUITRANSL LINES ON
// SET GUITRANSL BOX   ON
SET MESSAGE TO 18 center

@ 2,5 say "Select item by cursor-keys + Enter" COLOR "G+" GUICOLOR "B+"
if AppIoMode() == "G"
   ?? " or by mouse click" GUICOLOR "B+"
endif

StartMenu()    // execute choices

@ 15,0 clear
wait "Done, press any key ..."
quit

//--------------------------------------------------------------------- 
// main menu. Invokes nested sub-menus in PROMPT's via SELECT clause
function StartMenu()
   local iRet
   
   while .T.
      @ 5, 5 PROMPT "&One"    SELECT {|p1,p2,p3| one(p1,p2,p3) }     ;
                              TOOLTIP "first item in main menu"      ;
                              COLOR "W+/B,GR+/R,,R+/B"               ;
                              MESSAGE "with simple submenu"
      @ 5,col()+1 PROMPT "&Two"    ;
                              SELECT {|p1,p2,p3| two(p1,p2,p3) }     ;
                              COLOR "W+/B,GR+/R,,R+/B"               ;
                              TOOLTIP "second item in main menu"     ;
                              MESSAGE "with complex submenu"
      @ 5,col()+1 PROMPT "Th&ree"  ;
                              COLOR "W+/B,GR+/R,,R+/B"               ;
                              GUICOLOR "N/W,W+/B"                    ;
                              STYLE {BOX_RAISED,BOX_SUNKEN}          ;
                              TOOLTIP "third item in main menu"      ;
                              MESSAGE "no submenu"
      @ 5,col()+1 PROMPT "&Four"   ;
                              COLOR "W+/B,GR+/R,,R+/B"               ;
                              SELECT {|p1,p2,p3| four(p1,p2,p3)  }   ;
                              TOOLTIP "fourth item in main menu"     ;
                              MESSAGE "no submenu"
      @ 5,col()+1 PROMPT "&Quit"   ;
                              COLOR "W+/G,GR+/R,,R+/G"               ;
                              SELECT {|p1,p2,p3| myQuit(p1,p2,p3) }  ;
                              TOOLTIP "Exit menu"                    ;
                              MESSAGE "exit, no submenu"
      menu to iRet

      DisplStatus(21, .F., "selected in main menu:"+ltrim(iRet))
      if iRet == 0 .or. iRet == 5
         exit
      endif   
   enddo
   return
   
//--------------------------------------------------------------------- 
// sub-menu 1, selected by SELECT from StartMenu() which is still 
// active. It therefore uses nested (own) PROMPT's.
static function one(iPos, oMenu, oPrompt)
   local _oPrompt       // holds nested PROMPTs
   local iRet, sScr
   local cc := oMenu:column(), r1 := rr := oMenu:Row() +1

   DisplStatus(21, .F., "Main menu="+ltrim(iPos))
   
   sScr := savescreen(r1,cc, r1+4,cc+10)
   @ ++rr, cc PROMPT "one-&one"   WIDTH 11 CENTER MESSAGE "Item #1, centered"
   @ ++rr, cc PROMPT "one-&two"   WIDTH 11 CENTER MESSAGE "Item #2, centered"
   @ ++rr, cc PROMPT "one-t&hree" WIDTH 11 CENTER MESSAGE "Item #3, centered"
   MENU TO iRet
   
   DisplStatus(22, .T., "submenu selection=" + ltrim(iret))
   
   restScreen(r1,cc, r1+4,cc+10, sScr)
   return
   
//--------------------------------------------------------------------- 
// sub-menu 2, selected by SELECT from StartMenu() which is still 
// active. Displays additional selection via Achoice() and sub-submenus 
// via own (nested) PROMPTs. 
static function two(iPos, oMenu, oPrompt)
   local _oPrompt    // holds nested PROMPTs
   local iRet := 1, iRet2 := 0, sScr1, vSave, r1,c1,c2, rr, cc
   local cSetColor := SetColor()
   local lSetConf  := Set(_SET_CONFIRM)
   local cToolTip  := "Sub-submenu"
   local cMessage  := "confirm by Return"
   local aItems := {"two-&one","two-&two","two-&deep","","two-&five","&exit"}
   local aAccess:= {.T., .T., .T., .F., .T., .T.}
   local oColor3, oColor4, aColor3, cMsg3
   
   DisplStatus(21, .F., "Main menu="+ltrim(iPos))

   r1 := oMenu:Row() +2
   c1 := oMenu:Column()
   sScr1 := savescreen(r1-1,c1,r1+8,c1+30)

   set color to "W/N,N/W,W+/N,W+/N,N/N"   // 5th pair = unselectable item 
   
   oColor3 := ColorPair{ Color{250,250,250}, Color{8,93,139} }
   oColor4 := ColorPair{ Color{8,93,139}, Color{250,250,250} }
   aColor3 := { {{255,255,255},{8,93,139}}, ;  // fg,bg standard W+/B
                {{255,255,0},  {255,0,0 }} }   // fg,bg selected GR+/R+
   
   while iRet != 6
      /* vertical sub-menu via Achoice
       */
      iRet := Achoice(r1,c1,,, aItems, aAccess,, iRet,, @vSave)
      DisplStatus(22, .F., "sub-menu 2 selection=" + ltrim(iret))
      if iRet == 0 .or. iRet == 6
         exit
      endif
      
      SET CONFIRM ON                   // <enter> in selection required
      
      rr := r1 + iRet -1               // sub-menu row
      c2 := c1 + 10                    // first column
      @ rr-1, c2 clear to rr +3, maxcol()
      
      /* horizontal sub-submenus via nested PROMPT
       */
      do case
      case iRet == 1
         @ rr, c2+2     PROMPT "sub-&one"                         ;
                        GUICOLOR "N/W,R+/B"                       ;
                        STYLE BOX_NONE                            ;
                        MESSAGE (cMessage) TOOLTIP (cTooltip)
         @ rr, col() +1 PROMPT "sub-&two"                         ;
                        GUICOLOR "N/W,R+/B"                       ;
                        STYLE BOX_NONE                            ;
                        MESSAGE (cMessage) TOOLTIP (cTooltip) 
         @ rr, col() +1 PROMPT "sub-t&hree"                       ;
                        GUICOLOR "N/W,R+/B"                       ;
                        STYLE BOX_NONE                            ;
                        MESSAGE (cMessage) TOOLTIP (cTooltip) 
         cc := col()
         @ rr -1, c2 TO rr +1, cc      // draw frame
      case iRet == 2
         @ rr, c2+2     PROMPT "sub-&one"                         ;
                        COLOR "R+/N,W+/R"  GUICOLOR "R+/W,R+/W"   ;
                        STYLE {BOX_RAISED,BOX_SUNKEN}             ;
                        WIDTH 10 CENTER HEIGHT 2                  ;             
                        MESSAGE (cMessage) TOOLTIP (cTooltip)
         @ rr, col() +1 PROMPT "sub-&two"                         ;
                        COLOR "G+/N,W+/G"  GUICOLOR "R+/W,R+/W+"  ;
                        STYLE {BOX_RAISED,BOX_SUNKEN}             ;
                        WIDTH 10 CENTER HEIGHT 2                  ;             
                        MESSAGE (cMessage) TOOLTIP (cTooltip) 
         @ rr, col() +1 PROMPT "sub-t&hree"                       ;
                        COLOR "B+/N,W+/B"  GUICOLOR "R+/W,R+/W"   ;
                        STYLE {BOX_RAISED,BOX_SUNKEN}             ;
                        WIDTH 10 CENTER HEIGHT 2                  ;             
                        MESSAGE (cMessage) TOOLTIP (cTooltip) 
      case iRet == 3
         cMsg3 := "select by cursor + enter or by hot-key"
         SET CONFIRM OFF         // <enter> in selection not required
         @ rr, c2+2     PROMPT "sub-&one"                         ;
                        COLOR "W/B,N/W" GUICOLOR aColor3          ;
                        STYLE BOX_PLAIN LINEWIDTH 0               ;
                        SELECT {|p1,p2,p3| two_3(p1,p2,p3)  }     ;
                        MESSAGE (cMsg3) TOOLTIP (cTooltip)
         @ rr, col() +1 PROMPT "sub-&two"                         ;
                        STYLE BOX_PLAIN  LINEWIDTH 1              ;
                        COLOR "W/B,N/W" GUICOLOR aColor3          ;
                        SELECT {|p1,p2,p3| two_3(p1,p2,p3)  }     ;
                        MESSAGE (cMsg3) TOOLTIP (cTooltip) 
         @ rr, col() +1 PROMPT "sub-t&hree" ;
                        STYLE BOX_PLAIN LINEWIDTH 2               ;
                        COLOR "W/B,N/W" GUICOLOR aColor3          ;
                        SELECT {|p1,p2,p3| two_3(p1,p2,p3)  }     ;
                        MESSAGE (cMsg3) TOOLTIP (cTooltip) 
      otherwise
         @ rr, c2+2     PROMPT "sub-&one"                         ;
                        STYLE BOX_PLAIN     LINEWIDTH 0           ;
                        WIDTH 10 CENTER HEIGHT 1.5                ;
                        GUICOLOR "GR+/B,W+/R+"                    ;
                        MESSAGE (cMessage) TOOLTIP (cTooltip)
         @ rr, col() +1 PROMPT "sub-&two"                         ;
                        STYLE BOX_PLAIN     LINEWIDTH 1           ;
                        WIDTH 10 CENTER HEIGHT 1.5                ;
                        GUICOLOR "GR+/B,W+/R+"                    ;
                        MESSAGE (cMessage) TOOLTIP (cTooltip) 
         @ rr, col() +1 PROMPT "sub-t&hree"                       ;
                        STYLE BOX_PLAIN     LINEWIDTH 3           ;
                        WIDTH 10 CENTER HEIGHT 1.5                ;
                        GUICOLOR "GR+/B,W+/R+"                    ;
                        MESSAGE (cMessage) TOOLTIP (cTooltip) 
      endcase
      MENU TO iRet2
      
      set(_SET_CONFIRM, lSetConf)   // restore previous status
      
      DisplStatus(22, .F., "selections: submenu=" + ltrim(iret) + ;
                  " this sub-submenu=" + ltrim(iRet2) )
      @ rr-1, c2 clear to rr +3, maxcol()
   enddo
   restscreen(r1-1,c1,r1+8,c1+30, sScr1)  
   SetColor(cSetColor)
   return   

//--------------------------------------------------------------------- 
// sub-menu 3/1, selected by SELECT from two() 
// uses own (nested) PROMPTs. 
static function two_3(iPos, oMenu, oPrompt)
   local _oPrompt    // holds nested PROMPTs
   local iRet, sScr
   local cc := oMenu:column(), r1, rr 

   DisplStatus(22, .F., "sub-submnu="+ltrim(iPos))
   
   while .T.
      r1 := rr := oMenu:Row() +1
      sScr := savescreen(r1,cc, r1+5,cc+30)
      @ ++rr, cc  PROMPT "item &one from prev#" + ltrim(iPos)  ;
                  WIDTH 24 CENTER                              ;
                  COLOR "W/B,N/W"                              ;
                  MESSAGE "Item #1, centered"                  ;
                  SELECT {|p1,p2,p3| alert("selected #"+ltrim(p1)) } 
      @ ++rr, cc  PROMPT "item &two from prev#" + ltrim(iPos)  ;
                  WIDTH 24 CENTER                              ;
                  COLOR "W/B,N/W"                              ;
                  MESSAGE "Item #2, centered"                  ;
                  SELECT {|iSubPos,oSubMenu,oSubPrompt|        ;
                           alert("selected #"+ltrim(iSubPos) + ;
                                 ";" + oSubMenu:Caption) } 
      @ ++rr, cc  PROMPT "item t&hree from prev#" + ltrim(iPos) ;
                  WIDTH 24 CENTER                              ;
                  COLOR "W/B,N/W"                              ;
                  MESSAGE "Item #3, centered"                  ;
                  SELECT {|p1,p2,p3| alert("selected #"+ltrim(p1)) } 
      @ ++rr, cc  PROMPT "&exit this submenu"                  ;
                  WIDTH 24 CENTER                              ;
                  COLOR "W/B,N/W"                              ;
                  MESSAGE "Item #4, centered"                  ;
                  SELECT {|p1,p2,p3| alert("selected #"+ltrim(p1)) } 
      MENU TO iRet

//    DisplStatus(23, .F., "sub-sub-submenu selection=" + ltrim(iret))

      restScreen(r1,cc, r1+5,cc+30, sScr)
      if iRet == 0 .or. iRet == 4
         exit
      endif
   enddo
   return .F.     // don't exit from above menu

   
//--------------------------------------------------------------------- 
// UDF invoked by SELECT from StartMenu() which is still active
static function four(iPos, oMenu, oPrompt)
   return two_3(iPos, oMenu, oPrompt)

//--------------------------------------------------------------------- 
// UDF invoked by SELECT from StartMenu() which is still active
static function myQuit(iPos, oMenu, oPrompt)
   local iRet := Alert("Quit the applic now?",{"No","Yes"})
   if iRet == 2
      return .T.     // exit MENU TO
   endif
   return .F.        // don't exit MENU TO
   
//--------------------------------------------------------------------- 
// helper, display current state of selections
static function DisplStatus(row, wait, cText)
   local rr := row(), cc := col(), key := 0
   @ row,0 clear
   @ row,0 say cText color "R/N" guicolor "R+/W" // display text
   if wait
      inkey(3)
   endif
   setpos(rr,cc)
   return NIL
     

