* For FlagShip 5 and 6
* Copyright (c) 2001..05 by multisoft GmbH, Munich, Germany 
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the 
* modified source code when your application is distributed in source, 
* but may not remove the copyright notice.
* 
* Release $Date: 2005/11/21 13:45:14 $
*
*------------------------------------------------------------------------
* Example of TopBar and PopUp with user defined i/o handler for FoxPro
*
* Compile: FlagShip -fxp -DFOXPRO menufox.prg 
* Run:     ./.a.out
*
* MS-Windows: FlagShip menufox.prg -fxp -DFOXPRO [-io-a|g|t|b]
* Run:        menufox  -or- menufox -io=t
*------------------------------------------------------------------------

#include "set.fh"    // set() constants
#include "inkey.fh"  // inkey() constants, e.g. K_UP. K_LEFT etc.

//-----------------------------------------------------------------------
// User defined input handler for the PopUp class, optional. 
//   Assigned by oPopUp:InputBlock := {|obj,key| myPopHandler(obj,key) }
//   and called from oPopUp:Display() during its execution
// This UDF simulates default behavior of the oPopUp:Display() handler.
// Note: you should NOT change here the selection via oPopUp:Select(pos)
// since the oPopUp:Current is used later to restore the item visibility.
//
// The code block (and hence this UDF) receives 2 parameters:
// - the current PopUp object
// - the Inkey() value pressed
// This handler should return numeric value <ret>
//   ret > 0 : continue with selection/display item# <ret>
//   ret = 0 : abort further execution of this object, exit popup:Display()
//   ret < 0 : execute selection (codeblock or popup) on item# <ret>

FUNCTION myPopHandler(oPopUp,key)
local pos, ii, keyLower as intvar

pos := oPopUp:Current

do case
case key == K_LEFT                           // left: display prev.menu
   keyboard chr(K_LEFT) + chr(K_ENTER)
   pos := 0
case key == K_RIGHT                          // right: display next menu
   keyboard chr(K_RIGHT) + chr(K_ENTER)
   pos := 0
case key == K_UP .or. key == K_SH_TAB        // up, changed
   pos := oPopUp:GetPrev()
   if pos == 0             // no previous items
      pos := if(set(_SET_WRAP), oPopUp:GetLast(), oPopUp:GetFirst() )
   endif
case key == K_DOWN .or. key == K_TAB         // down, changed
   pos := oPopUp:GetNext()
   if pos == 0             // no next items
      pos := if(set(_SET_WRAP), oPopUp:GetFirst(), oPopUp:GetLast() )
   endif
case key == K_PGUP .or. key == K_HOME        // 1st
   pos := oPopUp:GetFirst() 
case key == K_PGDN .or. key == K_END         // last
   pos := oPopUp:GetLast() 
case key == K_ENTER .or. key == K_SPACE      // enter          
   pos := -pos       // forces to process UDF/codeblock or enter submenu
case key == K_ESC .or. key == 0
   pos := 0          // exit the popup, return 0
otherwise
   keyLower := asc(lower(chr(key)))
   /* 
    * Find shortkey (stored in oMenuItem object == oPopUp:Data), 
    * specified by the "&" or "&\" prefix in oMenuItem:Caption
    */
   for ii := 1 to oPopUp:ItemCount
      if keyLower == oPopUp:Data:ShortKey .and. oPopUp:Data:Enabled
         pos := -ii   // behaves like enter key
         exit
      endif
   next
   if pos >= 0
      /* 
       * Key not found yet, look for 1st valid char of caption, but 
       * consider only if shortkey is not specified for the item. 
       */
      for ii := 1 to oPopUp:ItemCount
         if oPopUp:Data:ShortKey == 0 .and. oPopUp:Data:Enabled .and. ;
               keyLower == asc(lower(left(ltrim(oPopUp:Data:Caption),1)))
            pos := -ii   // behaves like enter key
            exit
         endif
      next
   endif
endcase
return pos

* eof *

