#ifdef COPYRIGHT_NOTE
/* For FlagShip 5 and 6. 
 * Copyright (c) 2004..05 by multisoft GmbH, Munich, Germany
 * You may modify this source according to your needs and distribute the
 * object code freely with your executable. You may distribute also the 
 * modified source code when your application is distributed in source, 
 * but may not remove the copyright notice.
 * 
 * Release $Date: 2005/11/21 13:45:45 $
 */
#endif
* ---------------------------------------------------------------------
* This program demonstrates the use of PC-8 character set,
* line drawings and boxes and the use of SET GUITRANSLATE ...
*
* Copy this file to your working directory and compile/run on
* Unix/Linux: FlagShip pc8lines.prg ; ./a.out -or- newfswin a.out -io=t
* MS-Windows: FlagShip pc8lines.prg ; pc8lines -or- pc8lines -io=t
*
* See also fsman section LNG.5.3 and LNG.5.4 for further details,
* or umlauts.prg for more examples. 
* ---------------------------------------------------------------------

#include "font.fh"

// #define NATIONAL_CHARS  /* enable if required */
#define TEST_CHARS         /* disable if test not required */
#define TEST_CHRSET        /* disable if test not required */
#define TEST_LINES         /* disable if test not required */
#define TEST_BOXES         /* disable if test not required */

/* you may disable the SET FONT statement to use
 * defaults or try another one:
 */
SET FONT "courier", 12
// SET FONT "helvetica", 12

#ifdef NATIONAL_CHARS
   #ifdef FS_WIN32
     m->oApplic:Font:CharSet(FONT_ISO8859_2)   // set default font to Latin-2
     FS_SET("guikey","FSguikey.latin2_pl")     // polish 'programmer' AltGr keys
   #else
     ? "using LANG=" + getenv("LANG")          // on Unix/Linux automatically with LANG=pl_PL:ISO-8859-2
   #endif
   if AppIoMode() == "T" .and. !empty(p1)
      ? "inmap=",  fs_set("inmap","FSkeymap.pl")
      ? "outmap=", fs_set("outmap","FSchrmap.pl")
   endif
   wait
#endif

if AppIoMode() == "G"
   /* In GUI mode, move and resize the window
    */
   m->oApplic:Move(50, 50, .T.)
   m->oApplic:Resize(27, 80, , .T.)
endif

#ifdef TEST_CHARS
   /* The current character set is displayed, w/o translation
    */
   @ 0,10 say "Plain chars 0..255 without SET GUITRANSL" COLOR "R+/N" GUICOLOR "R+"
   drawChar()
   wait

   /* Semi graphic PC-8 charset is translated
    */
   cls
   SET GUITRANSL TEXT ON
   @ 0,10 say "Chars 0..255 with SET GUITRANSL TEXT ON" COLOR "R+/N" GUICOLOR "R+"
   drawChar()
   SET GUITRANSL TEXT OFF
   wait
#endif

#ifdef TEST_CHRSET
   /* Display the character set plain, with SET GUITRANSL TEXT & ANCII ON/OFF
    */
   cls
   @ 0,10 say "Chars 0..255 with SET GUITRANSL TEXT & ASCII" COLOR "R+/N" GUICOLOR "R+"
   drawCharSet()
   wait
#endif

#ifdef TEST_LINES
   /* Horizontal and vertical lines are drawn w/o translation
    */
   cls
   @ 0,10 say "Horizontal and vertical lines w/o translation" COLOR "R+/N" GUICOLOR "R+"
   drawLines()
   wait

   /* Horizontal and vertical lines are drawn with translation
    */
   cls
   SET GUITRANSL TEXT ON
   SET GUITRANSL LINES ON
   @ 0,10 say "Horizontal and vertical lines with translation" COLOR "R+/N" GUICOLOR "R+"
   drawLines()
   SET GUITRANSL TEXT OFF
   SET GUITRANSL LINES OFF
   wait
#endif

#ifdef TEST_BOXES
   /* Draw boxes using @..TO.., @..SAY CHR(...) and @..BOX
    * see also examples/boxcommand.prg for more box examples
    */
   cls
   SET GUITRANSL TEXT  ON
   SET GUITRANSL LINES ON
   SET GUITRANSL BOX   ON
   @ 0,10 say "Boxes using SET GUITRANSL translation" COLOR "R+/N" GUICOLOR "R+"
   drawBoxes()
   SET GUITRANSL TEXT  OFF
   SET GUITRANSL LINES OFF
   SET GUITRANSL BOX   OFF
   wait
#endif
quit

/****************************************************************************
 * Draw/display the current character set 0..255
 */
function drawChar()
local ii, row := 2, col := 1

for ii := 0 to 255
   if ii == 128         // new screen
      wait
      @ 1,0 clear 
      row := 2
      col := 1
   endif
   if ii % 16 == 0 .and. ii != 0 .and. ii != 128   // new line
      col := 1
      if AppIoMode() == "T" .and. maxrow() < 26
         row += 2   // in Terminal i/o mode: output would be too large
      else
         row += 3   
      endif
   endif
   @ row, col      say strzero(ii,3)
   @ row +1, col+1 say chr(ii)
   col += 4
next
return NIL

/****************************************************************************
 * Draw/display the current character set 0..255 
 * plain, SET GUITRANSL ASCII and SET GUITRANSL TEXT
 */
function drawCharSet()
local ii, row := 2, iNum := 0

SET GUICOLORS ON
while iNum < 256
   @ row,  0 say strzero(iNum,3) + ".." + strzero(iNum + 31,3) color "R"
   @ row+1,0 say "- ASCII ON"    color "R"
   @ row+2,0 say "- TEXT  ON"    color "R"
   @ row+3,0 say "- TEXT+ASCII"  color "R"
   for ii := 0 to 31
      @ row, 13 + ii * 2 SAY chr(iNum)       // plain

      SET GUITRANSL ASCII ON
      @ row+1, 13 + ii * 2 SAY chr(iNum)     // ASCII ON
      SET GUITRANSL ASCII OFF
      
      SET GUITRANSL TEXT ON
      @ row+2, 13 + ii * 2 SAY chr(iNum)     // TEXT ON
      SET GUITRANSL TEXT OFF
      
      SET GUITRANSL ASCII ON
      SET GUITRANSL TEXT  ON
      @ row+3, 13 + ii * 2 SAY chr(iNum)     // ASCII ON, TEXT ON
      SET GUITRANSL ASCII OFF
      SET GUITRANSL TEXT  OFF
      
      iNum++
   next
   row += 5
   if iNum == 128
      setpos(row -1,0)
      wait
      row := 2
      @ row,0 clear
   endif
enddo
SET GUICOLORS OFF
setpos(row -1,0)
return NIL

/****************************************************************************
 * Draw lines. A part of the screen is pre-set by "X" to see
 * the alignment of semi-graphic PC8 characters.
 */
function drawLines()
local ii, iLastRow := 18

//---- using @..TO.. ----
for ii := 3 to 15
   @ ii,0  say strzero(ii,2) color "B+/N" guicolor "B+"
   @ ii,40 say strzero(ii,2) color "B+/N" guicolor "B+"
next
for ii := 12 to 15
   @ ii,4  say replicate("X",20) color "B+/N" guicolor "B+"
   @ ii,44 say replicate("X",20) color "B+/N" guicolor "B+"
next

@ 2,0 say "using @..TO.."
if AppIoMode() == "G"            // display comments
   if !set(_SET_GUIDRAWLINE)
      @ 17,0 say "Note: Line drawing by @..TO.."        
      @ 18,0 say "is disabled in GUI mode when"    
      @ 19,0 say "SET GUITRANSL LINES is OFF"   
      iLastRow := 19
   else
      @ 17,0 say "SET GUITRANSL LINES is ON"
      @ 18,0 say "Line starts/ends in char center"
   endif
endif
@ 4,4  TO 4,25                   // horizontal single
@ 5,4  TO 5,25 DOUBLE            // horizontal double

@ 6,4 TO 10,4                    // vertical single
@ 6,7 TO 10,7 DOUBLE             // vertical double
@ 6,11 TO 10,11                  // vertical + horizontal
@ 7,9 TO  7,13
@ 8,9 TO  8,13 DOUBLE
@ 9,9 TO  9,13

@ 12,5 to 12,10 DOUBLE           // horizontal over "X"
@ 15,5 to 15,10 DOUBLE           // horizontal over "X"
@ 13,6 to 14,6 DOUBLE            // vertical over "X"
@ 12,8 to 15,8 DOUBLE            // vertical over "X"

@ 12,13 to 15,13 DOUBLE          // --+  vertical
@ 12,12 to 12,13 DOUBLE          //   |  horizontal
@ 15,12 to 15,13 DOUBLE          // --+  horizontal
                                 
//---- using @..DRAW ---
@ 2,16 say "and @..DRAW"
@ 8,16 DRAW TO 6,20 WIDTH 2 COLOR "G+"    // draw a green house (GUI only)
@ DRAW TO  8,24 WIDTH 2 COLOR "G+"
@ DRAW TO  8,16 WIDTH 2 COLOR "G+"
@ DRAW TO 11,16 WIDTH 2 COLOR "G+"
@ DRAW TO 11,21 WIDTH 2 COLOR "G+"
@ DRAW TO 10,21 WIDTH 2 COLOR "G+"
@ DRAW TO 10,22 WIDTH 2 COLOR "G+"
@ DRAW TO 11,22 WIDTH 2 COLOR "G+"
@ DRAW TO 11,24 WIDTH 2 COLOR "G+"
@ DRAW TO  8,24 WIDTH 2 COLOR "G+"

@ 9,17 DRAW TO 9,20 WIDTH 2 COLOR "G+"    // draw the house window
@ DRAW TO 10.5,20 WIDTH 2 COLOR "G+"
@ DRAW TO 10.5,17 WIDTH 2 COLOR "G+"
@ DRAW TO 9, 17   WIDTH 2 COLOR "G+"

//---- using chr(...) -----
@ 2,40 say "using @.. say CHR(...)"
if AppIoMode() == "G"
   if !set(_SET_GUIDRAWTEXT)
      @ 17,40 say "Note: PC-8 translation is"   
      @ 18,40 say "disabled in GUI mode when"   
      @ 19,40 say "SET GUITRANSL TEXT is OFF"   
      iLastRow := 19
   else
      @ 17,40 say "SET GUITRANSL TEXT is ON"
      @ 18,40 say "Line starts/ends at char boundary"
   endif
else
   @ 17,40 say "Line starts/ends at char boundary"
   #ifndef FS_WIN32
      @ 19,0 say "Note: If you don't see semi-graphic characters, but A-umlaut etc,"
      @ 20,0 say "your environment is not set correctly. Best to invoke the application"
      @ 21,0 say 'via "newfswin  ' + execname() + ' -io=t" in GUI environment (KDE, Gnome)'
      @ 22,0 say 'or  "newfscons ' + execname() + '" otherwise; see fsman section REL'
      iLastRow := 21
   #endif
endif
@ 4,44 say replicate(chr(196),15)   // horizontal single
@ 5,44 say replicate(chr(205),15)   // horizontal double
for ii := 6 to 10                   // vertical single
   @ ii,44 say chr(179)
   @ ii,47 say chr(186)
   @ ii,51 say chr(179)
next

@ 7,49 say chr(196) + chr(196) + chr(197) + chr(196) + chr(196)
@ 8,49 say chr(205) + chr(205) + chr(216) + chr(205) + chr(205)
@ 9,49 say replicate(chr(196),5)

@ 12,45 say replicate(chr(205),6)   // horizontal over "X"
@ 15,45 say replicate(chr(205),6)   // horizontal over "X"
@ 13,46 say chr(186)
@ 14,46 say chr(186)
for ii := 12 to 15
   @ ii,48 say chr(186)
next
@ 12,52 say chr(205) + chr(187)     // --+
@ 13,53 say chr(186)                //   |
@ 14,53 say chr(186)                //   |
@ 15,52 say chr(205) + chr(188)     // --+

setpos(iLastRow +1,0)
return NIL

/****************************************************************************
 * Draw boxes using @..TO.., @..SAY CHR(...) and @..BOX
 * See more @..BOX examples in boxcommand.prg
 */
function drawBoxes()

//---- using @..TO.. ----
@ 2,1 say "using @..TO.."
@ 4,1 say chr(218)
@ 4,2 TO 4,5
@ 4,6 say chr(191)
@ 5,1 TO 6,1
@ 5,6 TO 6,6
@ 7,1 say chr(192)
@ 7,2 TO 7,5
@ 7,6 say chr(217)

@ 4,8  TO 4,13
@ 4,8  TO 7,8
@ 4,13 TO 7,13
@ 7,8  TO 7,13


//---- using @..SAY CHR(..) ----
@ 2,21 say "using @..SAY CHR(..)"
@  4,21 say chr(218) + repli(chr(196),2) + chr(194) + ;
            repli(chr(196),2) + chr(210) + repli(chr(196),2) + chr(191)
@  5,21 say chr(179) + space(2) + chr(179) + ;
            space(2) + chr(186) + space(2) + chr(179)
@  6,21 say chr(195) + repli(chr(196),2) + chr(197) + ;
            repli(chr(196),2) + chr(215) + repli(chr(196),2) + chr(180)
@  7,21 say chr(179) + space(2) + chr(179) + ;
            space(2) + chr(186) + space(2) + chr(179)
@  8,21 say chr(198) + repli(chr(205),2) + chr(216) + ;
            repli(chr(205),2) + chr(206) + repli(chr(205),2) + chr(181)
@  9,21 say chr(179) + space(2) + chr(179) + ;
            space(2) + chr(186) + space(2) + chr(179)
@ 10,21 say chr(192) + repli(chr(196),2) + chr(193) + ;
            repli(chr(196),2) + chr(208) + repli(chr(196),2) + chr(217)

@  4,35 say chr(201) + repli(chr(205),2) + chr(209) + ;
            repli(chr(205),2) + chr(203) + repli(chr(205),2) + chr(187)
@  5,35 say chr(186) + space(2) + chr(179) + ;
            space(2) + chr(186) + space(2) + chr(186)
@  6,35 say chr(199) + repli(chr(196),2) + chr(197) + ;
            repli(chr(196),2) + chr(215) + repli(chr(196),2) + chr(182)
@  7,35 say chr(186) + space(2) + chr(179) + ;
            space(2) + chr(186) + space(2) + chr(186)
@  8,35 say chr(204) + repli(chr(205),2) + chr(216) + ;
            repli(chr(205),2) + chr(206) + repli(chr(205),2) + chr(185)
@  9,35 say chr(186) + space(2) + chr(179) + ;
            space(2) + chr(186) + space(2) + chr(186)
@ 10,35 say chr(200) + repli(chr(205),2) + chr(207) + ;
            repli(chr(205),2) + chr(202) + repli(chr(205),2) + chr(188)
 
//---- using @..BOX ----
#include "box.fh"
@ 12,1 say "using @..BOX"
@ 14, 1, 18, 7 BOX B_SINGLE
@ 14,11, 18,17 BOX B_DOUBLE
@ 14,21, 18,27 BOX B_SINGLE_DOUBLE
@ 14,31, 18,37 BOX B_DOUBLE_SINGLE
@ 14,41, 18,47 BOX B_PLAIN

@ 20,0 say "See more @..BOX examples in boxcommand.prg"
         
return NIL

/* eof */

