* tbrowse_db.prg: Tbrowse with database incl. edit and append. 
* See also tbrowse_ar.prg for Tbrowse example with arrays.

* This example demonstrates the features of FS5 and FS6 (VFS) including a
* simple programming (with default input handler) and example of use the 
* Tbrowse class. Some parts here will therefore not run with FlagShip 4 and 
* Clipper - although FS4 and Clipper('87..5.3) applications remains backward 
* compatible to FS5 and FS6.
* Release $Date: 2007/11/02 08:50:14 $
*
* Compile/run: FlagShip [-io=t|g] tbrowse_db.prg ; ./a.out
* MS-Windows : FlagShip [-io=t|g] tbrowse_db.prg ; tbrowse_db.exe

#ifndef FlagShip5
   wait "Sorry, this example uses the extended VFS functionality"
   quit
#endif

#include "fspreset.fh"
#include "inkey.fh"
#include "tbrclass.fh"
#include "basclass.fh"
#include "box.fh"

/* enable additional features or options by commenting out:
 */
// #define WITH_FIXFONT
#define INCREASE_CACHE
// #define DISPLAY_DOTS
// #define WITH_INDEX
// fs_set("devel", .T.)    // enable developer's warning, if any

/* the used database is created automatically if it does not exist
 */
if !file("tbrowse_db.dbf")
   ? "creating database tbrowse_db.dbf"
   dbcreate("tbrowse_db", {{"Key","C",10,0}, {"ID","N",10,0}, ;
       {"Name","C",30,0}, {"Non_Edit","C",15,0}, ;
       {"Address","C",50,0}, {"Zip","N",6,0},{"City","C",34,0}, ;
       {"Long500","C",500,0}, {"Memo","M",10,0}, ;
       {"num73","N",7,3}, {"num15","N",15,0}, {"Logic","L",1,0} } )
   use tbrowse_db excl
   for ii := 1 to 100
      append blank
      repla Key with ltrim(ii), ID with ii, ;
            name with "name of record " + ltrim(ii), ;
            Non_Edit with "fix value", ;
            Address with "abcdefghijklmnopqrstuvw ABCDEFGHIJKLMNOPQRSTUVW", ;
            zip with ii + 1000, city with repli("city ",10)
      if ii == 1 .or. (ii % 25) == 0
         repla Key with trim(key) + repli("x",10)
         repla ID with 1234567890
         repla non_edit with trim(non_edit) + repli("x",15)
         repla name with trim(name) + repli("x",30)
         repla zip with 123456
         repla Long500 with repli("Long text ", 49) + "ending 500"
         repla Memo with "Memo text of record " + ltrim(ii)
         repla num73 with (ii % 1000) + 0.123
         repla num15 with ii
         repla Logic with .T.
      endif
   next       
   use 
   if file("tbrowse_db" + indexext())
      delete file ("tbrowse_db" + indexext())  // reindex required
   endif
endif

#ifdef WITH_INDEX
   if !file("tbrowse_db" + indexext())
      use tbrowse_db excl
      index on upper(key) to tbrowse_db
      use
   endif
   use tbrowse_db SHARED NEW index tbrowse_db
#else
   use tbrowse_db SHARED NEW
#endif

if AppIoMode() == "G"   // meaningful for GUI mode only
#  ifdef WITH_FIXFONT
   m->oApplic:Font:FontName("courier")
   m->oApplic:Font:SizePoint(12)
#  endif
endif

/* Note: Colors are ignored in GUI, unless
 * SET GUICOLOR ON is set
 */
set color to "W/B,gr+/r,gr+/b,gr+/b,w+/b"

clear screen

#ifdef DISPLAY_DOTS
   for ii := 0 to 31
      @ ii,0 say "..." + strzero(ii,2) + "....1....:....2" color "g+/b"
   next
   @ 5,0 say repli("0123456789",7)
#endif

*  m->oApplic:Resize(500,650,.T., 1) // pixel, 1=inner,0=outer

/* Note: The @..BOX is not displayed in GUI, unless
 * SET GUITRANSL BOX ON is set
 */
@ 5,4,21,61 box B_SINGLE color ("gr+/b")

/* Create Tbrowse database object and
 * add columns = all database fields.
 * Note: These few lines is the only work required 
 *       in your application :-)
 */
oBr := TbrowseDb(6,5, 20,60, NIL, NIL, "My Browse") 
  oBr:Trim       := .T.    // to trim displayed data of type 'C'
* oBr:ReadOnly   := .T.    // if editing not desired
* oBr:CanAppend  := .F.    // if append data not desired
  oBr:IncrSearch := .T.    // if increm. key-search is desired (with index)

  oBr:ColAdjust  := 10     // adjust large columns visib. left + 10pix

  oBr:COLSEP    := " " + chr(179) + " "   // "  "
  oBr:HEADSEP   := chr(196,197,196)       // ""
* oBr:FOOTSEP   := chr(196,193,196)       // ""
  oBr:COLORSPEC := "W/B, GR+/B, BG+/W, GR+/B, R+/B, N/W"

for ii := 1 to Fcount()
   oBr:AddColumn( TbColumnNew(FieldName(ii), FieldBlock(FieldName(ii)) ))
   if FieldType(ii) == "M"                               // Memo-Field:
      oBr:GetColumn(ii):MemoPos := Rectangle{9,8,17,57}  // optional
   endif
   if FieldName(ii) == "NON_EDIT"                        // my special field:
      oBr:GetColumn(ii):ReadOnly := .T.                  // optional
   endif
next

#ifdef INCREASE_CACHE   /* if enabled at the top: */
   /* increase the table slice (default = visible rows), i.e.
    * allow movement/browsing by vertical scrollbar (in GUI only)
    * and faster skip by PgDn/PgUp in increments of 50 records 
    */
   oBr:RowCache := min(50, int(reccount() / 2) +1)
#endif

if AppIoMode() == "G"   // meaningful for GUI mode only
#ifndef DISPLAY_DOTS
   cls
   @ 1,5 say "The [^] and [v] buttons on the scrollbar are user-modifiable"
   @ 2,5 say "Currently, Left-mouse=Up/Down, Left+Shift/Ctrl/Alt=PgUp/PgDn"
#endif
endif
/* You may specify the handling of mouse click on [^] and [v] buttons
 * at the vertical scrollbar. The default is defined in initio.prg,
 * where the array value 0 means no action, 1=Up/Down, 2=PgUp/PgDn
 * Note: to achieve behavior of VFS5.x to VFS6.x, set all 2 and .F.
 */
m->_aGlobSetting[GSET_G_A_TBROW_VSBUTT] := ;
   {{1,2,2,2}, ;  // left  mouse click, left+shift,  left+Ctrl,  left+Alt
    {2,2,2,2}, ;  // mid   mouse click, mif+shift,   mid+Ctrl,   mid+Alt
    {2,2,2,2}, ;  // right mouse click, right+shift, right+Ctrl, right+Alt
    .T.}          // .T. = simulate key press, .F. execute directly

DisplayUse()   // just for your convenience

/* Note: instead of invoking the convenient VFS :Exec() method used here, 
 * you also may use the usual keyboard handler, same as in FS4 and Clipper.
 * The here used TbrDbHandler() is available in source code and is
 * replaceable by your own UDF using the :Handler instance, e.g.:
 * oBr:Handler := {|obj| TbrDbHandler(obj) } // this is the default for dbf
 */
oBr:Exec()                                   // calls UDF assigned by :Handler

/* reset cursor position
 */
setpos(21,60)
wait

************************** helper ***************************
FUNCTION DisplayUse()
if AppIoMode() != "G"
   return
endif
@ 7,61 draw to 7,65 color "R+"
@ draw to 8,66 color "R+"
@ draw to 8,64 color "R+"
@ draw to 7,65 color "R+"
@ 8,65 draw to 18,65 color "R+"
@ 19,65 draw to 18,64 color "R+"
@ draw to 18,66 color "R+"
@ draw to 19,65 color "R+"
@ draw to 19,61 color "R+"

@ 11,63 say "Here: " 
@ 11,col() say "page/scroll area =" guicolor "R+"
@ 12,63 say "oBr:RowCache = 50 items" guicolor "R+"
@ 13,63 say "Next page = PgUp/PgDn" 
@ 14,63 say "Use Scrollbar or cursor up/dn" 
@ 15,63 say "or wheel for items within page" 

@ 5.8,61 draw to 5.3,62 width 2 color "G+"
@ draw to 6.3,62 width 2 color "G+"
@ draw to 5.8,61 width 2 color "G+"
@ 5.8,62 draw to 5.8,67 width 2 color "G+"
@ 5.8,68 say "Scroll up button"

@ 19.3,61 draw to 18.8,62 width 2 color "G+"
@ draw to 19.8,62 width 2 color "G+"
@ draw to 19.3,61 width 2 color "G+"
@ 19.3,62 draw to 19.3,67 width 2 color "G+"
@ 19.3,68 say "Scroll down button" 

@ 20.5,63 say "Scroll buttons: up/down = item"
@ 21.5,63 say "Shift+Button = up/down page"
return

** eof

