#ifdef COMMENT_ONLY
* For FlagShip 6 
* Copyright (c) 2005..07 by multisoft GmbH, Munich, Germany 
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the 
* modified source code when your application is distributed in source, 
* but may not remove the copyright notice.
* 
* Release $Date: 2007/03/20 14:44:54 $
*
* Copy the source to your working directory,
* Compile: FlagShip -na -c -w -m [-io=a] [-FSintern] FSerror.prg
* Link   : FlagShip myapp*.prg FSerror.o [-io=a|g|t|b] [-o myapp]
#endif

/****************************************************************************
 * FSerror.prg    Error handler interface for FlagShip.                     *
 *                                                                          *
 * Note: if the environment variable FSERRORLOG is set, all errors are      *
 *       additionally printed to the text file given herewith, e.g.:        *
 *       on Linux,Unix:  export FSERRORLOG=/tmp/myerror.log   #activate     *
 *                       ./a.out  (or newfswin ./a.out)       #execute      *
 *                       unset FSERRORLOG                     #de-activate  *
 *       on MS-Windows:  set FSERRORLOG="c:\tmp\myerror.txt"  #activate     *
 *                       myapp.exe                            #execute      *
 *                       set FSERRORLOG=                      #de-activate  *
 *                                                                          *
 *       FSERRORLOG=fileName as [drive:][path]filename[.ext]                *
 *       FSERRORLOG=1 -> auto, path and name of executable + .err extension *                     
 *       FSERRORLOG=2 -> log output to stderr (may be redirected by 2>file) *                    
 *                                                                          *
 * Warning: modifying the C code requires a good knowledge of C programming *
 *          and the Open C System of FlagShip. Otherwise, do not modify     *
 *          the program part within #Cinline and #endCinline.               *
 ****************************************************************************/

#define INCLUDE_PRINT          /* you may disable it = comment-out */
#define WARN_IGNORE            /* you may disable it = comment-out */

#include "error.fh"
#include "errclass.fh"
#include "set.fh"
#include "box.fh"

#Cinline
static char *fileID="@@# $RCSfile: FSerror.prg,v $ $Revision: 1.24 $ $Date: 2007/03/20 14:44:54 $";
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "FSopenc.h"
#include "FSerrors.h"
#include <malloc.h>
#ifndef FS_WIN32
# include "unistd.h"
#endif

extern void exit();
extern void fgs_quit();
extern int __sig_quit;
extern struct                       /* array of pointers to error messages */
   __error_report __error_report[];

extern int last_errno;      /* last error number                           */
extern int errstat;         /* error status (1= display errors, 0= don't)  */
extern int _GlobAppInit;    /* 1=GUI, 2=text, 4=basic */
extern FSvar *ovStdio;      /* _oStdIo object */
extern unsigned int FsDebugMask, FsDebugMask2, FsDebugMask3;

/* extern int _filbuf(); extern int endwin()   */
/* not required, not available on every system */

static int alreadyin = 0;   /* recursion check flag                        */
static int saveFsDebugMask(int mode);

#define ABORT_COUNT 5
/*----------------------------------------------------------------
 | recursive error check                                          
 | this check avoids recursive calls to the errorsystem,          
 |  (prevent segmentation when error during errorblock-evaluation)
 |
 | flag = 1: increase recursion stack
 |        0: decrease recursion stack
 *----------------------------------------------------------------*/
static void err_check_recursion(flag)
int flag;                          /* do not change this function, else    */
{                                  /* an error occuring in the err handler */
                                   /* can only result in a segmentation    */
   int ii;                         /* violation due to stack overflow      */
   fn_stack * stack;               /* can only result in a segmentation    */
   FSvar *myArgv[1];
   
   if (!flag) {
     alreadyin --;
     if(alreadyin < 0)
        alreadyin = 0; 
     return;
   }
   if(alreadyin > 4) {             /* Error during quit, no way to get out */
      if(ovStdio && IS_VAR_OBJ(ovStdio)) {
         myArgv[0] = ovStdio;
         OBJ_METHEXEC("windowend", -1, 1, myArgv); /* m->_oStdIo:WindowEnd() */
      }
      exit(-1);
   }
   
   alreadyin++ ;
   if(alreadyin > ABORT_COUNT) {        /* Error during error, recursion */

      fprintf(stderr, "\rFATAL ERROR in module %s %d: %s()\r\n",
      old_fn_stack->prg_name, old_fn_stack->lin_no, old_fn_stack->module_name);

      fprintf(stderr, "Called from:\r\n");
      for (stack = old_fn_stack->old_fn_stack; stack && stack->old_fn_stack;
                                               stack = stack->old_fn_stack) {
          if (stack && stack->lin_no >= 0)
              printf("%10s %3d: %10s()\r\n",
                     stack->prg_name, stack->lin_no, stack->module_name);
      }
      if(_GlobAppInit == 1) {             /* GUI */
         if(alreadyin > (ABORT_COUNT +5))
            _exit(798);                   /* exit the hard way */
            
         fprintf(stderr, "Waiting 5 seconds ...\r\n");
         for(ii=0; ii < 5 ; ii++) {
            _fgs_ProcessUserEvents();     /* in cssignal.c */
            sleep(1);
         }
      } else {   
         printf("Press return ...\r\n");
         UDF_EXEC(inkey)(1, &IZERO_VAR);
      }
      fgs_quit();
   }
}

/* -------------------------------------------------------------------------
 * save/clear/restore FsDebugMask ... FsDebugMask3
 * mode = 0: save and clear
 *      = 1: restore
 */
static int saveFsDebugMask(int mode)
{
   static unsigned int save1 = 0, save2 = 0, save3 = 0;
   
   if(mode == 0) {
      save1 = FsDebugMask;
      save2 = FsDebugMask2;
      save3 = FsDebugMask3;
      FsDebugMask  = FsDebugMask2 = FsDebugMask3 = 0;
   } else {
      FsDebugMask  = save1;
      FsDebugMask2 = save2;
      FsDebugMask3 = save3;
   }
   return 0; 
}

/****************************************************************************
 * rt_error(p1,p2) = Display Run time error, C callable fn()    
 *     
 * rt_error() or _rt_error() call sequence in FS5:
 *  run-time lib
 * [-> _rt_error() ]       prg   src/FSerror.prg
 *  -> rt_error()          C     src/FSerror.prg
 *  -> _errcaller()        prg   src/FSerror.prg
 *  -> _DEFERROR()         prg   src/FSerror.prg
 *  -> __erralert()        prg   stdio/bberalrt.c 
 *  -> _gStdIo:ErrAlert()  prg   io_gui/ostdio_g.prg 
 *     -> self:Alert()
 *     [-> gMsgListBox{} ]
 *     --> exit = quit() or continue (ret = -1)
 */
FSvar *rt_error (errno, data)
int errno;                         /* Error code number                    */
char *data;                        /* Error context description data       */
{
   FSvar * argv[3];
   err_check_recursion(1);
   last_errno  = errno;
   if (errstat != 0) {
#ifdef FSintern
      if(FsDebugMask || FsDebugMask2 || FsDebugMask3)
         fprintf(stderr, "\n***RTE %d (%d %s)\n",errno,ERRTYP_RT,data); 
#endif
      saveFsDebugMask(0);      
      argv[0] = SET_VAR_NUM(VAR_NEW, (double) errno);
      argv[1] = SET_VAR_NUM(VAR_NEW, (double) ERRTYP_RT);
      argv[2] = SET_VAR_CHR(VAR_NEW, data);
      SET_VAR_NUM(VAR_NEW, (double) errno);
      UDF_EXEC(_errcaller)(3, argv);
      saveFsDebugMask(1);      
   }
   err_check_recursion(0);
   return(ERROR_VAR);
}

/****************************************************************************
 * Illegal break error, C callable fn()    
 */
void fs_break_error()
{
#ifdef FSintern
   if(FsDebugMask || FsDebugMask2 || FsDebugMask3)
      fprintf(stderr, "\n***RTE: Break without a BEGIN SEQUENCE\n"); 
#endif
   rt_error(RTILLBRK, "Break without a BEGIN SEQUENCE");
   fgs_quit();
}

/****************************************************************************
 * ftl_error(p1,p2) = Display Fatal error, C callable fn() 
 */
FSvar *ftl_error (errno, data)
int errno;                         /* Error code number                    */
char *data;                        /* Error context description data       */
{
   FSvar * argv[3];
   err_check_recursion(1);
   last_errno  = errno;
   if (errstat != 0) {
#ifdef FSintern
      if(FsDebugMask || FsDebugMask2 || FsDebugMask3)
         fprintf(stderr, "\n***RTE fatal %d (%d %s)\n",errno,ERRTYP_FTL,data); 
#endif
      saveFsDebugMask(0);      
      argv[0] = SET_VAR_NUM(VAR_NEW, (double) errno);
      argv[1] = SET_VAR_NUM(VAR_NEW, (double) ERRTYP_FTL);
      argv[2] = SET_VAR_CHR(VAR_NEW, data);
      SET_VAR_NUM(VAR_NEW, (double) errno);
      UDF_EXEC(_errcaller)(3, argv);
      saveFsDebugMask(1);      
   }
   err_check_recursion(0);
   fgs_quit();
   return (FSvar *) 0;
}

/****************************************************************************
 * io_error(p1,p2) = Display i/o error, C callable fn() 
 */
FSvar *io_error (errno, data)
int errno;                           /* Error code number                  */
char *data;                          /* Error context description data     */
{
   FSvar * argv[3];
   err_check_recursion(1);
   last_errno  = errno;
   if (errstat != 0) {
#ifdef FSintern
      if(FsDebugMask || FsDebugMask2 || FsDebugMask3)
         fprintf(stderr, "\n***RTE io %d (%d %s)\n",errno,ERRTYP_IO,data); 
#endif
      saveFsDebugMask(0);      
      argv[0] = SET_VAR_NUM(VAR_NEW, (double) errno);
      argv[1] = SET_VAR_NUM(VAR_NEW, (double) ERRTYP_IO);
      argv[2] = SET_VAR_CHR(VAR_NEW, data);
      SET_VAR_NUM(VAR_NEW, (double) errno);
      UDF_EXEC(_errcaller)(3, argv);
      saveFsDebugMask(1);      
   }
   err_check_recursion(0);
   return(ERROR_VAR);
}


/****************************************************************************
 * int_error(p1,p2,p3,p4) = Display internal error, C callable fn() 
 */
FSvar *int_error (errno, data, file_name, file_line)
int errno;              /* error (message) number, used to print message   */
char *data;             /* context specific message                        */
char *file_name;        /* file name of module reporting error             */
int file_line;          /* line number of file with module reporting error */
{
   FSvar * argv[5];
   err_check_recursion(1);
   last_errno  = errno;
   if (errstat != 0) {
#ifdef FSintern
      if(FsDebugMask || FsDebugMask2 || FsDebugMask3)
         fprintf(stderr, "\n***RTE intern %d (%d %s)\n",errno,ERRTYP_INT,data); 
#endif
      saveFsDebugMask(0);      
      argv[0] = SET_VAR_NUM(VAR_NEW, (double) errno);
      argv[1] = SET_VAR_NUM(VAR_NEW, (double) ERRTYP_INT);
      argv[2] = SET_VAR_CHR(VAR_NEW, data);
      argv[3] = SET_VAR_NUM(VAR_NEW, (double) file_line);
      argv[4] = SET_VAR_CHR(VAR_NEW, file_name);
      UDF_EXEC(_errcaller)(5, argv);
      saveFsDebugMask(1);      
   }
   err_check_recursion(0);
   return(ERROR_VAR);
}

#endCinline
/**************** end of .c, begin of .prg part ***************************/

#ifdef FSintern
static giSTACKBEGIN := 0   /* this prg was compiled w/o procname() stack */ 
#else
static giSTACKBEGIN := 2   /* compilation w/o -FSintern switch */
#endif


/*-------------------------------------------------------------------------- 
 * _Rt_Error(p1,p2) = Display Run time error, .prg callable fn()  
 *                    equivalent to rt_error() for .c sources
 * An alternative is calling directly the _ErrCaller(p1..p3), but it 
 * do not check for recursions as opposite to *rt_error()     
 */
FUNCTION _Rt_Error(iErrno AS IntVar, cData AS Char)

#ifndef FSintern
   giSTACKBEGIN++    // avoid displaying stack of this _Rt_Error()
#endif

#Cinline   
   rt_error(VAR_INTNUM(VAR_NAME_LOCPAR(ierrno)),
            VAR_CHR(VAR_NAME_LOCPAR(cdata)) );
#endCinline

#ifndef FSintern
   giSTACKBEGIN--
#endif
return NIL


PROTO FUNC __SplitTxt(@text AS String, [@lines AS Intvar])    AS Array
PROTO FUNC ErrorNew  ()                                       AS Error
PROTO STATIC FUNCTION _ErrorLog(cMsg AS char, cStack as char, lHead as logic) AS usual

/*----------------------------------------------------------------*
 | call error handler                                             |
 *----------------------------------------------------------------*/
FUNCTION _errcaller(e_code, e_sever, e_decr, e_line, e_file)

local erro := errornew()     AS Error
local descript               AS String
local msg                    AS Usual
local aDescr, aTmp           AS Array
local ii, jj, kk, ret := 0   AS INT
local lowLevel               AS Usual

// The very first thing we do here is to check if _errcaller is already
// on the stack in case the above static C variable alreadyin is > 1.
// If not, we assume the previous errorblock call to be broken by the
// FlagShip BREAK() function.

#Cinline
   ret = alreadyin;
#endCinline

IF ret > 1                         // "recursive error" condition given
   ii := giSTACKBEGIN
   DO WHILE .T.
     msg := TRIM(PROCNAME(ii++))
     IF EMPTY(msg)                 // false alarm, BREAK happened
#Cinline
         alreadyin --;
#endCinline
       EXIT
     ENDIF
      IF (msg) == "_ERRCALLER"     //  truly recursive
         EXIT
      ENDIF
   ENDDO
ENDIF

#Cinline
   /* Get message of errorcode into msg and info if error disabled into ret 
    */
   SET_VAR_CHR(VAR_NAME_LOCAL(msg),
      __error_report[ret = ((int) VAR_INTNUM(VAR_NAME_LOCPAR(e_code)))].err_msg);
   if(__error_report[ret].when < fs_set_flag[0]) 
      ret = 1;
   else 
      ret = 0;
#endCinline

if ret == 1            // Error disabled (depends on current FS_set("devel"))
   return NIL
endif
e_decr := strtran(e_decr, chr(10), ";")
if AppIoMode() == "G"
   e_decr := strtran(e_decr, "\", "/")    // avoid wrong html interpretation
endif
aDescr := __splittxt(e_decr)
descript = ""
for ii := 1 to len(aDescr)
   if len(aDescr[ii]) > 76
     jj := at (",", aDescr[ii])
     if jj > 0 .and. jj < len(aDescr)
         aDescr[ii] = strtran(aDescr[ii], ",", ",;")
         aTmp := __splittxt(aDescr[ii])
       ASIZE (aDescr, len(aDescr) + len(aTmp) -1)
       aDescr[ii] := aTmp[1]
       for jj := 2 to len(aTmp)
         kk := ii + jj - 1
         AINS (aDescr, kk)
         aDescr[kk] := aTmp[jj]
       next
      endif
      if len(aDescr[ii]) > 76
       aDescr[ii] := left(aDescr[ii],56) + " ... " + right(aDescr[ii],15)
     endif
   endif
   descript += aDescr[ii] + if (ii < len(aDescr), ";", "")
next

erro:ARGS          := {}
erro:CANDEFAULT    := .F.
erro:CANRETRY      := .F.
erro:CANSUBSTITUTE := .F.
erro:CARGO         := NIL
erro:DESCRIPTION   := msg + ";" + descript
erro:FILENAME      := ""
erro:GENCODE       := e_code
erro:OPERATION     := ""
erro:OSCODE        := DOSERROR()
erro:SEVERITY      := e_sever
erro:SUBCODE       := IF(e_line == NIL,  0, e_line)
erro:SUBSYSTEM     := IF(e_file == NIL, "BASE", e_file)
erro:TRIES         := 0

msg := ERRORBLOCK()
if type("oApplic") != "O" .or. type("_oStdIO") != "O" .or. ;
   (type("_InitIoDone") == "L" .and. ! m->_InitIoDone) // still in initio.prg
   lowLevel := .T.
endif
IF (VALTYPE(msg) != "B")
   msg := { |x,y,low| _DEFERROR(x, y, low)}
ENDIF
EVAL(msg, erro, NIL, lowLevel)       // Pass error to curr. error handler
return (NIL)


/*----------------------------------------------------------------*
 | default error handler, user modifiable                         |
 *----------------------------------------------------------------*/
FUNCTION _DEFERROR(e_obj AS Error, [start_stack AS IntVar], [_lowLevel])

LOCAL option := {}                                            AS Array
LOCAL optlist:= {"Abort", "Ignore", "Callstack", "Debug", ;
                 "Print", "Errormessage"}                     AS Array
LOCAL replopt := 0, ii, jj, old, choice := 0, aLen := 0       AS Intvar
LOCAL prc, msg_str2, msg_str, cc, cText                       AS String
LOCAL sav_print, sav_cons, is_develop := .F.                  AS Logical
LOCAL old_row, old_col, iRet                                  AS Intvar
LOCAL color, cFrame, cErrFile                                 AS Usual
LOCAL iCursesOn                                               AS INT
LOCAL aProcCall := {}                                         AS Array
LOCAL lowLevel := if(valtype(_lowLevel) == "L",_lowLevel, .F.)

#Cinline
   if(!ovStdio) {
      fprintf(stderr,"No StdIo object, exiting with ret 99 in %s:%d %s\n",
         __FILE__, __LINE__, CprocStack2()); 
      exit(99);
   }
#endCinline

if !lowLevel 
   old_row := row()
   old_col := col()                      
endif

// ??## procstack()
// ??## "   MaxRow()=",ltrim(MaxRow()),"MaxCol()=",ltrim(MaxCol()),"Max_Row()=",ltrim(Max_Row()),"Max_Col()=",ltrim(Max_Col())

*** #define START_STACK  2
if start_stack == NIL .or. start_stack < 0
   start_stack := giSTACKBEGIN  // = 2, with 1, also _ERRCALLER() would be displayed
endif

IF e_obj == NIL
    msg_str := "UNKNOWN ERROR"
ELSE
   IF(e_obj:SEVERITY != ES_FTLERR)
      IF e_obj:SEVERITY == ES_RTERR
         msg_str := "Run-time"
      ELSEIF e_obj:SEVERITY == ES_IOERR
         msg_str := "I/O"
      ELSE
         msg_str := "Internal"
      ENDIF
   ELSE
      msg_str := "Fatal"
   ENDIF
   msg_str += " error "
   if e_obj:GENCODE == EG_DEVWARNING               // Developer Warning ?
      if e_obj:SEVERITY == ES_RTERR .or.;
         e_obj:SEVERITY == ES_IOERR
         if !fs_set("develop")                     // fs_set("devel") set ?
            return NIL                             // no, ignore output
         endif
      endif
      msg_str += "(Developer warning) "
      is_develop := .T.
   else
      msg_str += str(e_obj:GENCODE,3)
   endif
   if !empty(e_obj:FILENAME)
      msg_str += " in " + e_obj:FILENAME
   else
      msg_str += " in " + PROCNAME(giSTACKBEGIN) + ;
                 "(" + LTRIM(STR(PROCLINE(giSTACKBEGIN),5)) + ")"
   endif
   msg_str +=  ";" + e_obj:DESCRIPTION

   IF e_obj:SUBCODE != 0                     // info about subsystem available
      msg_str += ";" + e_obj:SUBSYSTEM
      IF e_obj:SUBCODE != -9999              // display subsystem number?
        msg_str += " (" +  LTRIM(STR(e_obj:SUBCODE,3)) + ")"
      ENDIF
   ENDIF
ENDIF

if !lowLevel 
   DISPEND()
endif

// print to log file specified by FSERRORLOG, if set
cErrFile := ProcFile(giSTACKBEGIN) + ".prg " + ProcStack(giSTACKBEGIN)
_ErrorLog(msg_str, cErrFile, .T.)   
   
DO WHILE .T.
   option := {}
   if is_develop .and. !lowLevel
      AADD (option, optlist[2])                                // Ignore
      AADD (option, optlist[1])                                // Abort
      AADD (option, optlist[3])                                // Callstack
      AADD (option, optlist[4])                                // Debug
#ifdef INCLUDE_PRINT
      if AppIoMode() == "T"
         AADD (option, optlist[5])                             // Print
      endif
#endif
      color := if (iscolor(), "W+/BG,W+/R", NIL)
   elseif lowLevel
      option := {"press Enter to quit"}                        // no Debugger yet
   else
      AADD (option, optlist[1])                                // Abort
      IF e_obj != NIL .and. e_obj:SEVERITY != ES_FTLERR
         AADD (option, optlist[2])                             // Ignore
         AADD (option, optlist[3])                             // Callstack
         AADD (option, optlist[4])                             // Debug
      ELSE
         AADD (option, optlist[3])                             // Callstack
      ENDIF
#ifdef INCLUDE_PRINT
      if AppIoMode() == "T"
         AADD (option, optlist[5])                             // Print
      endif
#endif
   endif

   IF choice == 0
      if lowlevel
         // this is an emergency break if error occurs
         // before application is initialized etc.
         
         _bAlert(msg_str, option, NIL, NIL, NIL, .F.)  // same as  _bStdIo:ErrAlert(...)
         choice := 1   // abort
      else
         // __ERRALERT() displays the err message indirectly:
         // calls the current _oStdIo:ErrAlert(p1..p4) to do
         choice := __ERRALERT(msg_str, option, color, e_obj)  // 4th par 18.09.01
      endif
   ENDIF
   IF choice < 0 .or. (choice > len(option) .and. !lowlevel)
      EXIT
   ENDIF

#ifdef WARN_IGNORE
   if choice == 0    // esc == warning 
      choice := ascan(option, "Ignore")
   endif
#endif

   DO CASE
   CASE choice == 1 .and. lowlevel
      m->oStdIo:WindowEnd()
      #Cinline
      exit(99);
      #endCinline
      quit
      
   CASE choice > 0 .and. option[choice] == "Ignore"  // Ignore
#ifdef WARN_IGNORE
      if set(_SET_WARN_IGNORE) .and. !is_develop
         cText := "Warning: Ignoring an error may cause subsequent;" + ;
                  "follow-up errors or an unpredictable behaviour !; ;" + ;
                  "Are you sure, you want to ignore this error ?" 
         if lowlevel
            iret := _bAlert(cText, {"No,error message","Yes,ignore error"}, NIL, NIL, NIL, .F.) 
         else
            iRet := Alert(cText, {"No,error message","Yes,ignore error"}, ;
                          if (iscolor(), "W+/BG,W+/R", NIL), ;
                          _GlobSetting(GSET_T_C_ERRBOX, B_PLAIN) )
         endif
         if iRet <= 1
            choice  := 0
            loop
         endif
      endif
#endif  
      EXIT

   CASE choice > 0 .and. option[choice] == "Callstack"  // Callstack
      ii := START_STACK
      aProcCall := {} 
      WHILE (LEN(prc := PROCNAME(ii)) != 0)
         aadd(aProcCall, ;
               padr(PADR(lower(PROCFILE(ii) + ".prg"), 15) +  ;
                    STR(PROCLINE(ii), 5) + ": " + prc + "()", 34) )
         ii++
      ENDDO
      // if the callstack is too depth for a single column,
      // display it in two columns
      aLen := len(aProcCall)
      if alen > 3 .and. aLen > (MaxCol(.T.) -4)
         jj := round((aLen / 2) + 0.2, 0)
         for ii := 1 to jj
            aProcCall[ii] := aProcCall[ii] + "  " + ;
               if((ii + jj) > aLen, space(34), aProcCall[ii + jj] )   
         next
         asize(aProcCall, jj)
      endif
      old := choice
      option[old] := optlist[6]                      // Errormessage
      // Display it (again via _oStdIo:ErrAlert(..) )
      choice := __ERRALERT(aProcCall, option, color, e_obj) // 4th par 18.09.01
      IF choice == old
         choice  := 0                               // go back
      ENDIF
      LOOP

   CASE choice > 0 .and. option[choice] == "Debug"  // Debugger
#Cinline
      __sig_quit |= 1;    /* Call debugger as soon as we are out of err.sys */
      _sstep |= 1;
#endCinline
      EXIT

   CASE choice > 0 .and. option[choice] == "Print"  // Print
#ifdef INCLUDE_PRINT
      if !lowLevel
         sav_print := SET (_SET_PRINTER)
         sav_cons  := SET (_SET_CONSOLE)
         SET CONSOLE OFF
         SET PRINTER ON
         ?
         ? "---Error Callstack and Print screen---"
         AEVAL (__splittxt(msg_str), {|x| QOUT(x)} )
         ?
         ii := START_STACK
         WHILE (LEN(prc := PROCNAME(++ii)) != 0)
            ? PADR(PROCFILE(ii), 10) +  STR(PROCLINE(ii), 5) + ;
              ":" + PADR(prc + "()", 12)
         ENDDO
         ?
         err_print()
         SET (_SET_PRINTER, sav_print)
         SET (_SET_CONSOLE, sav_cons )
         choice := 0
         LOOP
      endif
#else
      EXIT
#endif

   OTHERWISE
      if lowLevel
         m->oStdIo:WindowEnd()
         #Cinline
         exit(99);
         #endCinline
      endif
      QUIT

   ENDCASE
ENDDO

if !lowLevel 
   DEVPOS(old_row, old_col)                         // reset old cursor pos.
endif
PUBLIC FS_LAST_ERR := e_obj
RETURN NIL

/*------------------------*
 | optional: print screen |
 *------------------------*/
#ifdef INCLUDE_PRINT
   STATIC FUNCTION err_print ()
   LOCAL_INT ymax := MAXROW()                       // used also in C
   LOCAL     chr_arr[ymax +1]                       // e.g. chr_arr[25]
   LOCAL     ii := 0
   LOCAL     cTxt := ""
   
   if AppIoMode() != "T"
      return NIL
   endif

#Cinline
// #ifndef FS_WIN32
   {
      UDF_PROT(savescreen);                       /* if not used yet */
      UDF_PROT(maxcol);                           /* if not used yet */

      FSvar    *actscreen = VAR_NEW;
      WINDOW   *winPtr;
      int      x, y, xmax;
      char     *oneline, *str;

      actscreen = UDF_EXEC(savescreen)(0,0);               /* SAVE SCREEN   */
      winPtr    = VAR_SCR (actscreen);
      xmax      = (int) VAR_INTNUM (fsFnNAME(maxcol)(0,0) );  /* xmax=MAXCOL() */

      oneline = malloc( (unsigned) ((xmax +1) * sizeof(unsigned char)));
      if (oneline != NULL) {
         for (y=0; y <= ymax; y++) {
            str = oneline;
            for (x=0; x <= xmax; x++)
               *(str++) = (unsigned char)
                          (mvwinch(winPtr,y,x) & A_CHARTEXT);    /* curses */
            SET_VAR_CHRLEN (VAR_ARRELEM (VAR_NAME_LOCAL(chr_arr), y+1),
                            oneline, xmax +1);
         }
         free (oneline);
      }
   }
// #endif
#endCinline
   cTxt := ";Current screen:;"
   aeval(chr_arr, {|x| cTxt += x + ";"})
   _ErrorLog(cTxt, "", .F.)
   
   ii := 0
   ? SPACE(4) + REPLICATE("---------+", 8)
   AEVAL (chr_arr, {|x| QOUT(STR(ii++,2) + ":", x)} )
   ? SPACE(4) + REPLICATE("---------+", 8)
   ?
   
   RETURN NIL
#endif

/*----------------------------------------------------------------
 * print (append) error to log file specified by 
 * environment variable FSERRORLOG, if any. Considered is:
 * FSERRORLOG=fileName or fileName_with_path                      
 * FSERRORLOG=1 -> auto, path and name of executable + .err extension                      
 * FSERRORLOG=2 -> output to stderr                     
 */
STATIC FUNCTION _ErrorLog(cMsg AS char, cStack as char, lHeader as logic) as usual
   local cTxt, cExe, cExt              as char
   local cFile := getenv("FSERRORLOG") as usual
   
   if empty(cFile)               // no FSERRORLOG, no action
      return NIL
   endif
   
   cExe := ExecName(.T.)
   if len(cFile) == 1 
      if val(cFile) == 1         // auto: <exe_path>/<exe_name>.err
         cFile := cExe
         if "." $ cFile
            cExt  := substr(cFile, rat(".", cFile))
            cFile := substr(cFile, 1, rat(".", cFile) -1)
            if empty(cExt) .or. empty(cFile) .or. ;
                  right(cFile,1) $ ".:/\" .or. "/" $ cExt .or. "\" $ cExt
               cFile := cExe
            endif
         endif
         cFile += ".err"
      elseif val(cFile) == 2     // output to stderr (can be redirected)
         cFile := "-stderr-"
         if AppIoMode() == "G"
            OpenConsole()
         endif
      endif   
   endif
         
   if lHeader   
      cTxt := ";****** " + cExe + " on " + dtoc(date()) + " " + time() 
      cTxt += ";" + cMsg 
      cTxt += ";---"
      if !empty(cStack)
         cTxt += ";Source/Callstack  = " + cStack
      endif
      cTxt += ";FlagShip version  = " + Version()
      cTxt += ";Running i/o mode  = " + AppIoMode() + ;
              ", compiled mode = " + AppCompiled()
      if AppIoMode() == "T"
         cTxt += ", TERM=" + getenv("TERM") + ", LANG=" + getenv("LANG")
      endif
      cTxt += ";Current directory = " + CurDir()
      cTxt += ";Current WorkArea  = " + ltrim(select())
      cTxt += ";Current Database  = "
      if !used()
         cTxt += "none"
      else
         cTxt += dbf(,.T.) 
         cTxt += if(IsDbExcl(),  " EXCL", " SHARED")
         cTxt += ";- alias = '" + alias() + "' record = " + ltrim(recno()) + ;
                 " (of " + ltrim(lastrec()) + ")"
         cTxt += if(NetErr(),    " NETERR", "")
         cTxt += if(IsDbFlock(), " FLOCK", "")
         cTxt += if(IsDbRlock(), " RLOCK", "")
         cTxt += if(Eof(),       " EOF", "")
         cTxt += if(Bof(),       " BOF", "")
         cTxt += if(Found(),     " FOUND", "")
         if IndexCount() > 0
            cTxt += ";- index(es) : "
            aeval(IndexNames(), {|cIdx| cTxt += cIdx + ", "})
            if right(cTxt,2) == ", "
               cTxt := left(cTxt, len(cTxt) -2)
            endif
            if IndexCheck() != 0
               cTxt += ";  - check(!): " + ltrim(IndexCheck()) + " see IndexCheck()"
            endif
            cTxt += ";  - current : order=" + ltrim(IndexOrd())
            cTxt += if(IndexOrd() > 0, " key=[" + IndexKey() + "]", "")
         endif
      endif
      cTxt += ";"
   endif
   cTxt := strtran(cTxt, ";" , chr(13) + chr(10))  // CR+LF
#Cinline
   {
      // i/o errors and insuficient access rights are ignored here,
      // i.e. log is not printed on permission failure or disk full
      int fd;
#ifdef FS_WIN32
      int mode = _S_IREAD | _S_IWRITE;
#else
      int mode = S_IREAD | S_IWRITE | S_IRGRP | S_IWGRP | S_IROTH;
#endif
      char  *cOutFile = (char *)VAR_CHR(VAR_NAME_LOCAL(cfile));
      FSvar *vText = VAR_NAME_LOCAL(ctxt);
      
      if(strcmp(cOutFile, "-stderr-") == 0) {
         fprintf(stderr, "%s", VAR_CHR(vText));
         fflush(stderr);
      } else {
         fd = open(cOutFile, O_RDWR | O_CREAT | O_APPEND, mode);
         if(fd > 0) {
            write(fd, VAR_CHR(vText), VAR_ISLEN(vText) -1);
            close(fd);
         }
      }
   }
#endCinline   
   return NIL
   
*** eof ***

