/****************************************************************************
 * rddcb4c.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  
 *               by using CodeBase 5.1 or 6.1, 6.54 multiplatform version 
 *               (not included in the FlagShip package, contact Sequiter Inc.)
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995 multisoft Datentechnik GmbH, Munich. All rights reserved worldw.*
 * Release $Date: 2004/10/27 12:53:13 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains parts of the full RDD driver:                         *
 *    oCb4:deleted           ACCESS deleted() status                        *
 *    oCb4:driver            ACCESS name of the RDD driver                  *
 *    oCb4:fcount            ACCESS no of fields in ea record               *
 *    oCb4:lastrec           ACCESS no of records = reccount                *
 *    oCb4:lupdate           ACCESS date of last update                     *
 *    oCb4:name              ACCESS name of the DBserver                    *
 *    oCb4:readonly          ACCESS read only file status                   *
 *    oCb4:rddname           ACCESS name of the RDD driver                  *
 *    oCb4:reccount          ACCESS no of records = lastrec                 *
 *    oCb4:recno             ACCESS current record no.                      *
 *    oCb4:recno             ASSIGN current record no.                      *
 *    oCb4:shared            ACCESS determines if the dbf was open SHARED   *
 *                                                                          *
 *    oCb4:append()          METHOD append blank                            *
 *    oCb4:asstring()        METHOD == name access                          *
 *    oCb4:commit()          METHOD flush current database file             *
 *    oCb4:delete(...)       METHOD delete, mark record for delete          *
 *    oCb4:deleteall()       METHOD mark all database records as deleted    *
 *    oCb4:fieldpos()        METHOD retrieve the ordin.posit of spec. field *
 *    oCb4:fieldname()       METHOD retrieve the name of the specif.  field *
 *    oCb4:recall(...)       METHOD recall, unmark deleted record           *
 *    oCb4:recallall()       METHOD remove all deleted marks                *
 *                                                                          *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */
#include "set.fh"                                /* std. FS settg _SET_xxx  */

extern FSint CB4FSudf(isScope)();                /* rddcb4e.c               */
extern FSvar *CB4FS_CALL_METH(goto)();           /* in rddcb4b.c            */
extern FSvar *CB4FS_CALL_METH(eval)();           /* in rddcb4e.c            */
extern FSvar *CB4FS_CALL_METH(fieldinfo)();      /* in rddcb4c.c            */
extern FSvar *CB4FS_CALL_ACCESS(filespec)();     /* in rddcb4d.c            */
extern FSvar *CB4FS_CALL_METH(skip)();           /* in rddcb4b.c            */
extern FSvar *CB4FS_CALL_METH(gotop)();          /* in rddcb4b.c            */
extern FSvar *CB4FS_CALL_ACCESS(name)();
extern FSint  CB4FSudf(callError)();             /* in rddcb4b.c            */
extern FSint  CB4FSudf(CSkip)();                 /* in rddcb4b.c            */
extern FSint  CB4FSudf(moveChildRelation)();     /* in rddcb4r.c            */
extern FSvar *fsFnNAME(ctod)();                  /* == std. fn CTOD()       */
extern FSvar *fsFnNAME(set)();                   /* == std. fn SET()        */
extern FSint  CB4FSudf(myAutoRFlock)();          /* in rddcb4b.c            */
extern FSint  CB4FSudf(myAutoUnlock)();          /* in rddcb4b.c            */

static FSint myDeleCheck ();                     /* forward declaration     */
static FSvar *cb4DeleteOne ();                   /* forward declaration     */
static FSvar *cb4RecallOne ();                   /* forward declaration     */

#define CB4_APPREADONLY  "Attempt to APPEND record;on READ ONLY file in alias %s"
#define CB4_DELEDELETE   "Attempt to delete"
#define CB4_DELERECALL   "Attempt to recall"
#define CB4_DELENOOPEN   "%s record in closed database"
#define CB4_DELEONEOF    "%s record %d;being on EOF() in alias %s"
#define CB4_DELEREADONLY "%s recprd %d;on READ ONLY file in alias %s"
#define CB4_DELENORLOCK  "%s record %d;in alias %s, record lock required"
#define CB4_DELENOFLOCK  "%s all records;in alias %s, file lock required"
#define CB4_FIELDPOSPAR  "cb4:FieldPos(cFldName)"
#define CB4_DELETEPAR    "cb4:Delete ([for],[while],[scope])"
#define CB4_RECALLPAR    "cb4:Recall ([for],[while],[scope])"

/*--------------------------------------------------------------------------*
 | x=oCb4:Deleted ACCESS retrieve the DELETED() status                      |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retL    = TRUE if record marked for deletion, FALSE otherw. |
 | Description: retrieves the CodeBase function                             |
 | FS invocat.: retL := oCb4:DELETED                                        |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (deleted)          /* e.g.  _bb_cb4cdx_deleted_ac(ovSelv) */
{
   DATA4    *dataPtr;                            /* CB5 structure           */

   CB4FS_DEBUG_BEGIN ("Deleted",ovSelf);
   if (dataPtr = CB4FS_GETDATA4PTR)
      return CB4FS_DEBUG_END ("Deleted",
            d4deleted(dataPtr) ? TRUE_VAR:FALSE_VAR); /* CB5 */
   return CB4FS_DEBUG_END ("Deleted",FALSE_VAR);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Driver  ACCESS determine the name of the RDD driver               |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retC    = The driver name according to the compilation      |
 |                        of these rddcb4?.* sources, e.g. "CB4CDX"         |
 | FS invocat.: retC := oCb4:DRIVER                                         |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (driver)            /* e.g.  _bb_cb4cdx_driver_ac(ovSelv) */
{
   FSvar *cvRetVar = VAR_NEW;

   CB4FS_DEBUG_BEGIN ("Driver",ovSelf);
   SET_VAR_CHR (cvRetVar, VAR_CHR(VAR_OBJ(ovSelf,CB4OBJ_DRIVERNAME)));
   return CB4FS_DEBUG_END ("Driver",cvRetVar);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Fcount  ACCESS retrieve number of fields in each record           |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retN    = field count or 0 on error                         |
 | Description: retrieves the CodeBase structure                            |
 | FS invocat.: retN := oCb4:FCOUNT                                         |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (fcount)            /* e.g.  _bb_cb4cdx_fcount_ac(ovSelv) */
{
   DATA4    *dataPtr;                                      /* CB5 structure */
   FSvar    *retNum = VAR_NEW;

   CB4FS_DEBUG_BEGIN ("Fcount",ovSelf);
   dataPtr = CB4FS_GETDATA4PTR;
#ifdef CODEBASE6
   if(!dataPtr || !dataPtr->dataFile)
      return CB4FS_DEBUG_END ("Fcount",IZERO_VAR);
   if (dataPtr = CB4FS_GETDATA4PTR)
      SET_VAR_INT (retNum, dataPtr->dataFile->nFields);    /* CB6 structure */
#else
   if (dataPtr = CB4FS_GETDATA4PTR)
      SET_VAR_INT (retNum, dataPtr->n_fields);             /* CB5 structure */
#endif
   else
      SET_VAR_INT (retNum, 0);
   return CB4FS_DEBUG_END ("Fcount",retNum);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:LastRec ACCESS retrieve number of records of the database         |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retN    = last valid record or 0 if dbf is empty            |
 | Description: retrieves the CodeBase function                             |
 | FS invocat.: retN := oCb4:LASTREC                                        |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (lastrec)          /* e.g.  _bb_cb4cdx_lastrec_ac(ovSelv) */
{
   DATA4    *dataPtr;                                      /* CB5 structure */
   FSvar    *nvRetVar = VAR_NEW;
   long     reccount  = 0L;

   CB4FS_DEBUG_BEGIN ("LastRec",ovSelf);
   if (dataPtr = CB4FS_GETDATA4PTR)
      reccount = d4reccount (dataPtr);                      /* CB5 function */
   return CB4FS_DEBUG_END ("LastRec",
            SET_VAR_INT( nvRetVar, reccount));
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Lupdate ACCESS date of last .dbf update                           |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retD    = date of last update                               |
 | Description: retrieves the CodeBase function                             |
 | FS invocat.: retN := oCb4:LUPDATE                                        |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (lupdate)          /* e.g.  _bb_cb4cdx_lupdate_ac(ovSelv) */
{
   DATA4    *dataPtr;                                      /* CB5 structure */
   FSvar    *dvRetVar   = VAR_NEW;
   FSvar    *cvDateForm = VAR_NEW;
   FSvar    *passArg[2];

   CB4FS_DEBUG_BEGIN ("LUpdate",ovSelf);
   dataPtr = CB4FS_GETDATA4PTR;
#ifdef CODEBASE6
   if(!dataPtr || !dataPtr->dataFile) {
#else
   if (! dataPtr) {
#endif
      SET_VAR_DATE( dvRetVar, 0L);
      return CB4FS_DEBUG_END ("LUpdate", dvRetVar );
   }
   VAR_NEW_ARGS (passArg, 2);
   SET_VAR_INT  (passArg[0], _SET_DATEFORMAT);
   SET_VAR_CHR  (passArg[1], "mm/dd/yy");
   SET_VAR_COPY (cvDateForm, fsFnNAME(set)(2,passArg));    /* d=SET(4,"..") */

   sprintf (repdum, "%2d/%2d/%2d",                         /* convert CB5/6 */
#ifdef CODEBASE6
               dataPtr->dataFile->mm,
			   dataPtr->dataFile->dd, dataPtr->dataFile->yy);/* to string   */
#else
               dataPtr->mm, dataPtr->dd, dataPtr->yy);     /* to string     */
#endif
   SET_VAR_CHR  (passArg[0], repdum);
   SET_VAR_COPY (dvRetVar, fsFnNAME(ctod)(1,passArg));     /* CTOD("...")   */

   SET_VAR_INT  (passArg[0], _SET_DATEFORMAT);
   SET_VAR_COPY (passArg[1], dvRetVar);
   SET_VAR_COPY (cvDateForm, fsFnNAME(set)(2,passArg));    /* SET(4,d)      */

   return CB4FS_DEBUG_END ("LUpdate", dvRetVar );
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Name    ACCESS determine the main part of the database            |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retC    = The associated .dbf name (in uppercase), given    |
 |                        on xx := dbserver {"../xyz.dbf"}, e.g. "XYZ"      |
 | FS invocat.: retC := oCb4:NAME                                           |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (name)                /* e.g.  _bb_cb4cdx_name_ac(ovSelv) */
{
   FSvar *cvRetVar = VAR_NEW;
   FSint posEnd, posBeg, len;
   char  *str;

   CB4FS_DEBUG_BEGIN ("Name",ovSelf);
   SET_VAR_CHR (cvRetVar, "");

   str = VAR_CHR( VAR_OBJ(ovSelf, CB4OBJ_FILEINPNAME));
   len = (int) VAR_ISLEN ( VAR_OBJ(ovSelf, CB4OBJ_FILEINPNAME)) - 1;
   if (!str || !len) {
      return CB4FS_DEBUG_END ("Name",cvRetVar);
   }
   for (posEnd = len -1;
        (posEnd >= 0) && (str[posEnd] != '.') &&
        (str[posEnd] != '/') && (str[posEnd] != '\\') ;
        posEnd--);
   if ((posEnd < 0) || (str[posEnd] == '/') || (str[posEnd] != '\\'))
      posEnd = len;
   for (posBeg = posEnd -1;
        ((posBeg >= 0) && (str[posBeg] != '/') && (str[posBeg] != '\\')) ;
        posBeg--);
   if (posBeg < 0)
      posBeg = 0;
   else
      posBeg++;
   if (posEnd > posBeg) {
      SET_VAR_CHRLEN (cvRetVar, str + posBeg, posEnd - posBeg);
      SET_VAR_UPPER  (cvRetVar);
   }
   return CB4FS_DEBUG_END ("Name",cvRetVar);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:RecCount ACCESS retrieve number of records of the database        |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retN    = last valid record or 0 if dbf is empty            |
 | Description: retrieves the CodeBase function                             |
 | FS invocat.: retN := oCb4:RECCOUNT                                       |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (reccount)        /* e.g.  _bb_cb4cdx_reccount_ac(ovSelv) */
{
   DATA4    *dataPtr;                                      /* CB5 structure */
   FSvar    *nvRetVar = VAR_NEW;
   long     reccount  = 0L;

   CB4FS_DEBUG_BEGIN ("RecCount",ovSelf);
   if (dataPtr = CB4FS_GETDATA4PTR)
      reccount = d4reccount (dataPtr);                      /* CB5 function */
   return CB4FS_DEBUG_END ("RecCount",
      SET_VAR_INT( nvRetVar, reccount));
}

/*--------------------------------------------------------------------------*
 | x=oCb4:RecNo    ACCESS retrieve the current record number                |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retN    = current record,                                   |
 |                        0 if dbf is empty, lastrec()+1 when on EOF()      |
 | Description: retrieves the CodeBase function                             |
 | FS invocat.: retN := oCb4:RECNO                                          |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (recno)              /* e.g.  _bb_cb4cdx_recno_ac(ovSelv) */
{
   DATA4    *dataPtr;                                      /* CB5 structure */
   FSvar    *nvRetVar = VAR_NEW;
   long     recNo     = 0L;

   CB4FS_DEBUG_BEGIN ("RecNo",ovSelf);
   if (dataPtr = CB4FS_GETDATA4PTR)
      recNo = d4recno (dataPtr);                            /* CB5 function */
   if (recNo < 0L) recNo = 0L;
   SET_VAR_INT(nvRetVar, recNo);

/*
 *  Clipper set the lenght of the recno to 7, so do we...
 */
   VAR_ISLEN(nvRetVar) = 7;
   return CB4FS_DEBUG_END ("RecNo", nvRetVar);
}

/*--------------------------------------------------------------------------*
 | oCb4:RecNo=x    ASSIGN new record number == goto(x)                      |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retO    = the object self                                   |
 | Description: retrieves the CodeBase goto function                        |
 | FS invocat.: oCb4:RECNO := expN                                          |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ASSIGN (recno)       /* e.g.  _bb_cb4cdx_recno_as(ovSelv,assVar) */
{
   DATA4    *dataPtr;                                      /* CB5 structure */
   FSvar    *param[2];

   CB4FS_DEBUG_BEGIN ("RecNoAss",ovSelf);
   if (dataPtr = CB4FS_GETDATA4PTR) {
      if (! IS_VAR_NUM(assVar))                  /* RTE data type mismatch  */
         CB4FSudf(callError)(ovSelf, 1, DTMMACH, "RECNO",
                             "cb4:RECNO = expN", __FILE__, __LINE__);
      else {
         param[0] = ovSelf;
         param[1] = assVar;
         CB4FS_CALL_METH(goto)(2, param);        /* perform goto(..)        */
      }
   }
   return CB4FS_DEBUG_END ("RecNoAss", assVar);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:ReadOnly ACCESS determine the file read/only status               |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retL    = TRUE if file was open read/only                   |
 | Description: retrieves the CodeBase status                               |
 | FS invocat.: retL := oCb4:READONLY                                       |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (readonly)       /* e.g.  _bb_cb4cdx_readonly_ac(ovSelv) */
{
   DATA4    *dataPtr;                                      /* CB5 structure */

   CB4FS_DEBUG_BEGIN ("ReadOnly",ovSelf);
   dataPtr = CB4FS_GETDATA4PTR;
#ifdef CODEBASE6
   if(dataPtr && dataPtr->dataFile &&
      dataPtr->dataFile->file.isReadOnly)              /* CB6 structure */
      return CB4FS_DEBUG_END ("ReadOnly (T)",TRUE_VAR);
#else
   if(dataPtr && dataPtr->file.is_read_only)                      /* CB5 structure */
      return CB4FS_DEBUG_END ("ReadOnly (T)",TRUE_VAR);
#endif
   return CB4FS_DEBUG_END ("ReadOnly",FALSE_VAR);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:RddName  ACCESS determine the name of the RDD driver              |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retC    = The driver name according to the compilation      |
 | Description: retrieves the CodeBase status                               |
 | FS invocat.: retC := oCb4:RDDNAME                                        |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (rddname)          /* e.g.  _bb_cb4cdx_rddname_ac(ovSelv) */
{
   FSvar *cvRetVar = VAR_NEW;

   CB4FS_DEBUG_BEGIN ("RddName",ovSelf);
   SET_VAR_CHR (cvRetVar, VAR_CHR(VAR_OBJ(ovSelf,CB4OBJ_DRIVERNAME)));
   return CB4FS_DEBUG_END ("RddName",cvRetVar);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Shared   ACCESS determine the file shared status                  |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retL    = TRUE if file was open shared or SET EXCL OFF      |
 | Description: retrieves the CodeBase status                               |
 | FS invocat.: retL := oCb4:SHARED                                         |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (shared)            /* e.g.  _bb_cb4cdx_shared_ac(ovSelv) */
{
   DATA4    *dataPtr;                                      /* CB5 structure */

   CB4FS_DEBUG_BEGIN ("Shared",ovSelf);
   dataPtr = CB4FS_GETDATA4PTR;
#ifdef CODEBASE6
   if(dataPtr && dataPtr->dataFile && 
      (dataPtr->dataFile->file.lowAccessMode != OPEN4DENY_RW)) /*CB6 struct*/
      return CB4FS_DEBUG_END ("Shared (T)",TRUE_VAR);
#else
   if (dataPtr && ! dataPtr->file.is_exclusive)                 /* CB5 structure */
      return CB4FS_DEBUG_END ("Shared (T)",TRUE_VAR);
#endif
   return CB4FS_DEBUG_END ("Shared",FALSE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:Append()   METHOD appends new empty record.                         |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: Adds a new empty record to the end of the currently         |
 |              selected database file. All fields of the new record        |
 |              are EMPTY(). After APPENDing, the new BLANK record becomes  |
 |              the current record. Check also NETERROR() for errors.       |
 | Multiuser  : In multiuser mode, the appended RECORD will be autom.locked.|
 | FS invocat.: retL := oCb4:APPEND()                                       |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (append)          /* e.g.  _bb_cb4cdx_append_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, "cb4:Append()");
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Append",ovSelf);
   dataPtr = CB4FS_GETDATA4PTR;
   
#ifdef CODEBASE6
   if(!dataPtr || !dataPtr->dataFile)
      return CB4FS_DEBUG_END ("Append",FALSE_VAR);
   if(dataPtr->dataFile->file.isReadOnly) {           /* CB6 structure      */
#else
   if(!dataPtr)
      return CB4FS_DEBUG_END ("Append",FALSE_VAR);
   if (dataPtr->file.is_read_only) {                  /* CB5 structure      */
#endif
      sprintf (repdum, CB4_APPREADONLY,
               d4alias(dataPtr));                     /* CB5 alias name     */
      CB4FSudf(callError)(ovSelf, 2, IOWRIT, "APPEND",
                          repdum, __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("Append",FALSE_VAR);    /* RTE replace on r/o */
   }

   if (d4append_blank (dataPtr))                      /* CB5 function       */
      return CB4FS_DEBUG_END ("Append (F)",FALSE_VAR);    /* failed         */
   return CB4FS_DEBUG_END ("Append (T)",TRUE_VAR);        /* success        */
}

/*--------------------------------------------------------------------------*
 | oCb4:AsString() METHOD returns cb4:Name                                  |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : see cb4:Name                                                |
 | FS invocat.: retL := oCb4:ASSTRING()                                     |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (asstring)      /* e.g.  _bb_cb4cdx_asstring_me(argc, argv) */
{
   return CB4FS_CALL_ACCESS(filespec)(argv[0]);
}

/*--------------------------------------------------------------------------*
 | oCb4:Commit()   METHOD flushes the current database to file              |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: On read only file, rereads the record. Otherwise, flushes   |
 |              the current database and indicies to file and rereads rec.  |
 |              The CodeBase optimization is set OFF, if active.            |
 | FS invocat.: retL := oCb4:COMMIT()                                       |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (commit)          /* e.g.  _bb_cb4cdx_commit_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSint  ok = 0, optim_wr;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, "cb4:Commit()");
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Commit",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Commit",FALSE_VAR);

#ifndef S4OPTIMIZE_OFF
   optim_wr = dataPtr->code_base->optimize_write;     /* CB5 structure      */
   dataPtr->code_base->optimize_write = 0;            /* disable write opt. */
#endif

#ifdef CODEBASE6
   if(!dataPtr->dataFile)
      return CB4FS_DEBUG_END ("Commit",FALSE_VAR);
   if (!dataPtr->dataFile->file.isReadOnly)           /* CB6 structure      */
#else
   if (!dataPtr->file.is_read_only)                   /* CB5 structure      */
#endif
      ok = ! d4flush (dataPtr);                       /* CB5 function       */
   ok = ! d4refresh(dataPtr);                         /* CB5 function       */

#ifndef S4OPTIMIZE_OFF
   dataPtr->code_base->optimize_write = optim_wr;     /* restore write opt. */
#endif
   return CB4FS_DEBUG_END ("Commit",ok ? TRUE_VAR : FALSE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:Delete()   METHOD marks current record (or record scope) for delet. |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (opt) cForCond   | cbForCond                      |
 |              argv[2] = (opt) cWhileCond | cbWhileCond                    |
 |              argv[3] = (opt) nScope                                      |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: Only the deletion mark of the record is changed.            |
 |              The record remains in the database, remove it by PACK()     |
 |              Check deleted records by cb4:Deleted or DELETED() funct.    |
 | Multiuser  : In multiuser mode, at least RECORD locking is required,     |
 |              or SET AUTOLOCK ON must be set                              |
 | Compatib.  : unlike CA/VO, a movement to next record is performed        |
 |              only when a condition or scope is given.                    |
 | FS invocat.: retL := oCb4:DELETE()                                       |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (delete)          /* e.g.  _bb_cb4cdx_delete_me(argc, argv) */
{
   FSvar  *ovSelf;
   FSvar  *tmpArgv[6];
   FSint  tmpArgc = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTARGCNT, CB4_DELETEPAR);
      return CB4FS_DEBUG_END ("Delete",FALSE_VAR);
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Delete",ovSelf);

   if (argc == 1 &&
       (! CB4FSudf(isScope)(ovSelf) ))           /* rddcb4d.c: scope set ?  */
      return CB4FS_DEBUG_END ("Delete",          /* no, perform the         */
                cb4DeleteOne(1,argv));           /* current record only     */

   switch(argc) {
   case 4:
      tmpArgc    = 6;
      tmpArgv[5] = IS_VAR_NUM(argv[3]) ? argv[3] : NIL_VAR;     /* all/rest */
      tmpArgv[4] = IS_VAR_NUM(argv[3]) ? argv[3] : NIL_VAR;     /* next     */
   case 3:
      if (argc == 3) tmpArgc = 4;
      tmpArgv[3] = argv[2];                                     /* whib     */
   case 2:
      if (argc == 2) tmpArgc = 3;
      tmpArgv[2] = argv[1];                                     /* forb     */
   case 1:
      if (argc == 1) tmpArgc = 2;
      tmpArgv[1] = VAR_NEW;
      tmpArgv[1] = SET_VAR_BLOCK(VAR_NEW, cb4DeleteOne);        /* blk fn() */
      tmpArgv[0] = argv[0];
      break;
   default:
      CB4FSudf(callError)(ovSelf, 1, RTARGCNT, "DELETE", CB4_DELETEPAR,
                          __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("Delete",FALSE_VAR);
   }
   return CB4FS_DEBUG_END ("Delete",
              CB4FS_CALL_METH(eval)(tmpArgc, tmpArgv) );   /* cb4:eval(...) */
}

/*--------------------------------------------------------------------------*
 | oCb4:DeleteAll()   METHOD marks all records for deletion                 |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |         opt: argv[1] = expL consider RELATION movement, default=.T.   (1)|
 | Note/compat: (1) is not supported by CA/VO.                              |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: The deletion mark of all database records will be set.      |
 |              All records remains in the database, remove them by PACK()  |
 |              Check deleted records by cb4:Deleted or DELETED() funct.    |
 |              After completion, the dbf pointer is on EOF().              |
 |              Current filter IS considered.                               |
 | Multiuser  : In multiuser mode, at least FILE locking is required,       |
 |              or SET AUTOLOCK ON must be set                              |
 | FS invocat.: retL := oCb4:DELETEALL()                                    |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (deleteall)    /* e.g.  _bb_cb4cdx_deleteall_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSint  autoLock, isFilter, ok = 1;

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("DeleteAll",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR)) {
      myDeleCheck (ovSelf, 0, 1, 1, "DELETEALL");     /* RTE error          */
      return CB4FS_DEBUG_END ("DeleteAll",FALSE_VAR);
   }
   if (! (autoLock = myDeleCheck(ovSelf,
           dataPtr, 1, 1, "DELETEALL")))              /* write attempt ok?  */
      return CB4FS_DEBUG_END ("DeleteAll",FALSE_VAR); /* failed             */

/*
 *  If filter condition is not specified, the skip is performed w/o child
 *  repositioning on ea skipped record.
 *  If filter is specified, relations according to SET RELATION are
 *  performed, since the code block may affect also one of the relation.
 *  This may be disabled by the 1st parameter set to .F.
 *  SET FILTER and SET DELETED is always considered.
 */
   isFilter = IS_VAR_BLK(VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK));
   if ((argc > 1) && !IS_VAR_TRUE(argv[1]))
      isFilter = 0;

   CB4FS_CALL_METH(gotop)(1,argv);                    /* cb4:gotop()        */
   while (ok && !d4eof(dataPtr)) {
      d4delete (dataPtr);                             /* CB5 function       */
      ok = CB4FSudf(CSkip)(ovSelf,dataPtr, 1L, isFilter);         /*  skip  */
   }
   if (!isFilter)
      CB4FSudf(moveChildRelation)(ovSelf, dataPtr);   /* move related childs*/

   if (autoLock < 0)                                  /* autolock ?         */
      CB4FSudf(myAutoUnlock)(ovSelf, dataPtr);
   return CB4FS_DEBUG_END ("DeleteAll",TRUE_VAR);     /* success            */
}

/*--------------------------------------------------------------------------*
 | oCb4:FieldPos()  METHOD returns the field ord.position (1..n)            |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (C) field name                                    |
 | Returns    : (N) the ord.position or 0 on error                          |
 | FS invocat.: retN := oCb4:FIELDPOS(cFld)                                 |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (fieldpos)      /* e.g.  _bb_cb4cdx_fieldpos_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSvar  *nvRetVar = VAR_NEW;
   FSint  fieldPos;

   SET_VAR_INT (nvRetVar, 0);

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_FIELDPOSPAR);            /* RTE param          */
      return nvRetVar;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("FieldPos",ovSelf);

   if ((argc < 2) || ! IS_VAR_CHR(argv[1])) {
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "FIELDPOS",
                          CB4_FIELDPOSPAR, __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("FieldPos",nvRetVar);           /* RTE param  */
   }
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("FieldPos",nvRetVar);        /* dbf not open  */

   fieldPos = d4field_number (dataPtr, VAR_CHR(argv[1]));            /* CB5 */
   if (fieldPos <= 0)
      return CB4FS_DEBUG_END ("FieldPos",nvRetVar);   /* field not found   "FieldPos", */

   SET_VAR_INT (nvRetVar, fieldPos);
   return CB4FS_DEBUG_END ("FieldPos",nvRetVar);
}


/*--------------------------------------------------------------------------*
 | oCb4:FieldName() METHOD returns the field name according to position     |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = nFieldNo                                          |
 | Returns    : the requested info or empty string                          |
 | FS invocat.: ret := oCb4:FIELDNAME(nFld)                                 |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (fieldname)    /* e.g.  _bb_cb4cdx_fieldname_me(argc, argv) */
{
   FSvar  *avLocalParams[3];
   FSvar  *uvRetVar;

   switch(argc)
   {
   default:
   case 2: avLocalParams[2] = argv[1];
   case 1: avLocalParams[0] = argv[0];
   }

   avLocalParams[1] = SET_VAR_INT(VAR_NEW, DBS_NAME);
   argc++;
   
   uvRetVar = CB4FS_CALL_METH(fieldinfo)(argc, avLocalParams);

   if (IS_VAR_NIL(uvRetVar))
	   return CB4FS_DEBUG_END ("FieldName", NULLSTR_VAR);
	  
   return CB4FS_DEBUG_END ("FieldName",uvRetVar);
}

/*--------------------------------------------------------------------------*
 | oCb4:Recall()   METHOD remove the deletion mark of current record        |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (opt) cForCond   | cbForCond                      |
 |              argv[2] = (opt) cWhileCond | cbWhileCond                    |
 |              argv[3] = (opt) nScope                                      |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: Only the deletion mark of the record is changed.            |
 |              Check deleted records by cb4:Deleted or DELETED() funct.    |
 | Multiuser  : In multiuser mode, at least RECORD locking is required,     |
 |              or SET AUTOLOCK ON must be set                              |
 | Compatib.  : unlike CA/VO, a movement to next record is performed        |
 |              only when a condition or scope is given.                    |
 | FS invocat.: retL := oCb4:RECALL()                                       |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (recall)          /* e.g.  _bb_cb4cdx_recall_me(argc, argv) */
{
   FSvar  *ovSelf;
   FSvar  *retVar = VAR_NEW;
   FSvar  *tmpArgv[6];
   FSint  tmpArgc = 0;

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Recall",ovSelf);

   if (argc == 1 &&
       (! CB4FSudf(isScope)(ovSelf) ))           /* rddcb4d.c: scope set ?  */
      return CB4FS_DEBUG_END ("Recall",          /* no, perform the         */
                cb4RecallOne(1,argv));           /* current record only     */

   switch(argc) {
   case 4:
      tmpArgc    = 6;
      tmpArgv[5] = IS_VAR_NUM(argv[3]) ? argv[3] : NIL_VAR;     /* all/rest */
      tmpArgv[4] = IS_VAR_NUM(argv[3]) ? argv[3] : NIL_VAR;     /* next     */
   case 3:
      if (argc == 3) tmpArgc = 4;
      tmpArgv[3] = argv[2];                                     /* whib     */
   case 2:
      if (argc == 2) tmpArgc = 3;
      tmpArgv[2] = argv[1];                                     /* forb     */
   case 1:
      if (argc == 1) tmpArgc = 2;
      tmpArgv[1] = VAR_NEW;
      tmpArgv[1] = SET_VAR_BLOCK(VAR_NEW, cb4RecallOne);        /* blk fn() */
      tmpArgv[0] = argv[0];
      break;
   default:
      CB4FSudf(callError)(ovSelf, 1, RTARGCNT, "RECALL", CB4_RECALLPAR,
                          __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("Recall",FALSE_VAR);
   }
   return CB4FS_DEBUG_END ("Recall",
              CB4FS_CALL_METH(eval)(tmpArgc, tmpArgv) );   /* cb4:eval(...) */
}

/*--------------------------------------------------------------------------*
 | oCb4:RecallAll()   METHOD removes all deletion marks of the curr dbf     |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |         opt: argv[1] = expL consider RELATION movement, default=.T.   (1)|
 | Note/compat: (1) is not supported by CA/VO.                              |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: All deletion marks of the database will be removed.         |
 |              After completion, the dbf pointer is on EOF().              |
 |              Current filter IS considered.                               |
 | Multiuser  : In multiuser mode, at least FILE locking is required,       |
 |              or SET AUTOLOCK ON must be set                              |
 | FS invocat.: retL := oCb4:RECALLALL()                                    |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (recallall)    /* e.g.  _bb_cb4cdx_recallall_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSint  autoLock, isFilter, ok = 1;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, "cb4:RecallAll()");
      return (FALSE_VAR);
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("RecallAll",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR)) {
      myDeleCheck (ovSelf, 0, 0, 1, "RECALLALL");     /* RTE error          */
      return CB4FS_DEBUG_END ("RecallAll",FALSE_VAR);
   }
   if (! (autoLock = myDeleCheck(ovSelf,
          dataPtr, 0, 1, "RECALLALL")))               /* write attempt ok?  */
      return CB4FS_DEBUG_END ("RecallAll",FALSE_VAR); /* failed             */

/*
 *  If filter condition is not specified, the skip is performed w/o child
 *  repositioning on ea skipped record.
 *  If filter is specified, relations according to SET RELATION are
 *  performed, since the code block may affect also one of the relation.
 *  This may be disabled by the 1st parameter set to .F.
 *  SET FILTER and SET DELETED is always considered.
 */
   isFilter = IS_VAR_BLK(VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK));
   if ((argc > 1) && !IS_VAR_TRUE(argv[1]))
      isFilter = 0;

   CB4FS_CALL_METH(gotop)(1,argv);                    /* cb4:gotop()        */
   while (ok && !d4eof(dataPtr)) {
      d4recall (dataPtr);                             /* CB5 function       */
      ok = CB4FSudf(CSkip)(ovSelf,dataPtr, 1L, isFilter);         /*  skip  */
   }
   if (!isFilter)
      CB4FSudf(moveChildRelation)(ovSelf, dataPtr);   /* move related childs*/

   if (autoLock < 0)                                  /* autolock ?         */
      CB4FSudf(myAutoUnlock)(ovSelf, dataPtr);
   return CB4FS_DEBUG_END ("RecallAll",TRUE_VAR);     /* success            */
}

/*--------------------------------------------------------------------------*
 | internal function for delete() method above: delete curr.record          |
 *--------------------------------------------------------------------------*/
static FSvar *cb4DeleteOne (argc, argv)
FSint   argc;
FSvar * argv[];
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSint  autoLock;

   ovSelf = argv[0];
   if (! (dataPtr = CB4FS_GETDATA4PTR)) {
      myDeleCheck (ovSelf, 0, 1, 0, "DELETE");        /* RTE error          */
      return FALSE_VAR;
   }
   if (! (autoLock = myDeleCheck(ovSelf,
          dataPtr, 1, 0, "DELETE")))                  /* write attempt ok?  */
      return FALSE_VAR;                               /* failed             */

   d4delete (dataPtr);                                /* CB5: mark delete   */
   if (autoLock < 0)                                  /* autolock ?         */
      CB4FSudf(myAutoUnlock)(ovSelf, dataPtr);
   return TRUE_VAR;                                   /* success            */
}

/*--------------------------------------------------------------------------*
 | internal function for recall() method above: recall curr.record          |
 *--------------------------------------------------------------------------*/
static FSvar *cb4RecallOne (argc, argv)
FSint   argc;
FSvar * argv[];
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSint  autoLock;

   ovSelf = argv[0];
   if (! (dataPtr = CB4FS_GETDATA4PTR)) {
      myDeleCheck (ovSelf, 0, 0, 0, "RECALL");        /* RTE error          */
      return FALSE_VAR;
   }
   if (! (autoLock = myDeleCheck(ovSelf,
           dataPtr, 0, 0, "RECALL")))                 /* write attempt ok?  */
      return FALSE_VAR;                               /* failed             */

   d4recall (dataPtr);                                /* CB5 function       */

   if (autoLock < 0)                                  /* autolock ?         */
      CB4FSudf(myAutoUnlock)(ovSelf, dataPtr);
   return TRUE_VAR;                                   /* success            */
}

/*--------------------------------------------------------------------------*
 | internal function for delete() & recall() methods above                  |
 *--------------------------------------------------------------------------*/
static FSint myDeleCheck (ovSelf, dataPtr, isDelete, isAll, module)
FSvar  *ovSelf;
DATA4  *dataPtr;
FSint  isDelete, isAll;
FSchar *module;
{
   char   *currAlias = 0;
   long   recNo = 0;
   int    ii;

/*
 *  is database open ?
 */
#ifdef CODEBASE6
   if (!dataPtr || !dataPtr->dataFile) {
#else
   if (!dataPtr) {
#endif
      sprintf (repdum, CB4_DELENOOPEN,
          (isDelete ? CB4_DELEDELETE : CB4_DELERECALL));
      CB4FSudf(callError)(ovSelf, 1, RTNODBF, module, repdum,
                          __FILE__, __LINE__);
      return 0;                                  /* RTE: database not open  */
   }
   recNo     = d4recno(dataPtr);                 /* CB5: record number      */
   currAlias = (char *)d4alias(dataPtr);         /* CB5: alias name         */
/*
 *  is file open for read only ?
 */
#ifdef CODEBASE6
   if (dataPtr->dataFile->file.isReadOnly) {     /* CB6 structure           */
#else
   if (dataPtr->file.is_read_only) {             /* CB5 structure           */
#endif
      sprintf (repdum, CB4_DELEREADONLY,
               (isDelete ? CB4_DELEDELETE : CB4_DELERECALL),
               recNo, currAlias);
      CB4FSudf(callError)(ovSelf, 2, IOWRIT, module, repdum,
                          __FILE__, __LINE__);
      return 0;                                  /* RTE: replace on r/o     */
   }
/*
 *  file locked for ALL ?
 */
   if (isAll) {
      if (d4lock_test_file(dataPtr))             /* CB5: file locked ?      */
         return 1;
      ii = CB4FSudf(myAutoRFlock)(ovSelf, 2, dataPtr, recNo);
      if (ii != 0)
         return ii;
      sprintf (repdum, CB4_DELENOFLOCK,
               (isDelete ? CB4_DELEDELETE : CB4_DELERECALL),
               currAlias);
      CB4FSudf(callError)(ovSelf, 1, RTNOLOCK, module, repdum,
                          __FILE__, __LINE__);
      return 0;                                  /* RTE: flock required     */
   }
/*
 *  replace on EOF() ?
 */
   recNo = d4recno (dataPtr);                    /* CB5: record number      */
   if (d4eof(dataPtr) || recNo <= 0L) {
      sprintf (repdum, CB4_DELEONEOF,
               (isDelete ? CB4_DELEDELETE : CB4_DELERECALL),
               recNo, currAlias);
      CB4FSudf(callError)(ovSelf, 1, RTILLEOF, module, repdum,
                          __FILE__, __LINE__);
      return 0;                                  /* RTE: replace on EOF()   */
   }
/*
 *  is file or record locked ?
 */
   if (d4lock_test_append(dataPtr) > 0)          /* CB5: file locked ||     */
      return 1;                                  /* exclus || append lock ? */

   if (D4LOCK_TEST(dataPtr, recNo) > 0)          /* CB5: record locked ?    */
      return 1;

/*
 *  perform AutoRlock or AutoFlock if not disabled
 */
   ii = CB4FSudf(myAutoRFlock)(ovSelf, (isAll ? 2 : 1), dataPtr, recNo);
   if (ii != 0)
      return ii;

   sprintf (repdum, CB4_DELENORLOCK,
            (isDelete ? CB4_DELEDELETE : CB4_DELERECALL),
            recNo, currAlias);
   CB4FSudf(callError)(ovSelf, 1, RTNOLOCK, module, repdum,
                       __FILE__, __LINE__);
   return 0;                                     /* RTE: replace w/o lock   */

}

/**** eof rddcb4c.c ****/

