/****************************************************************************
 * rddcb4d.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2004 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2004/10/27 09:46:07 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains parts of the full RDD driver:                         *
 *    oCb4:filespec            ACCESS determines the full .dbf name         *
 *    oCb4:forblock            ACCESS the object-wide FOR condition         *
 *    oCb4:forblock            ASSIGN the object-wide FOR condition         *
 *    oCb4:found               ACCESS last seek/locate/continue successful  *
 *    oCb4:header              ACCESS the length of dbf header              *
 *    oCb4:recsize             ACCESS retrives the record length in bytes   *
 *    oCb4:rlocklist           ACCESS array of RLOCKed records              *
 *    oCb4:whileblock          ACCESS the object-wide WHILE condition       *
 *    oCb4:whileblock          ASSIGN the object-wide WHILE condition       *
 *    oCb4:scope               ACCESS the object-wide scope condition       *
 *    oCb4:scope               ASSIGN the object-wide scope condition       *
 *                                                                          *
 *    oCb4:clearscope()        METHOD clears WHILE, FOR and SCOPE condit    *
 *    oCb4:flock()             METHOD perform file   locking                *
 *    oCb4:lockcurrentrecord() METHOD same as rlock()                       *
 *    oCb4:noivarget()         METHOD exception handling for x=cb4:name     *
 *    oCb4:noivarput()         METHOD exception handling for cb4:name=x     *
 *    oCb4:refresh()           METHOD reads the curr.record                 *
 *    oCb4:rlock()             METHOD perform record locking                *
 *    oCb4:rlockverify()       METHOD same as rlock()                       *
 *    oCb4:unlock()            METHOD clears rlock() or flock()             *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/
 
#include "rddcb4fs.h"                            /* RDD API for CodeBase    */

#if S4VERSION == 6502
#include "rddcb4cpp.h"
#endif

extern FSvar *CB4FS_CALL_METH(fieldget)();       /* cb4:fieldget() rddcb4b.c*/
extern FSvar *CB4FS_CALL_METH(fieldput)();       /* cb4:fieldput() rddcb4b.c*/
extern FSvar *CB4FS_CALL_METH(commit)();         /* cb4:comit()    rddcb4c.c*/
extern FSint  CB4FSudf(callError)();             /* in rddcb4b.c            */

#define CB4_RLOCKWARN "cb4:Rlock(recNo) this RDD can lock the;current record only, locking request ignored"
#define CB4_UNLCKWARN "cb4:Unlock(recNo) this RDD can unlock the current;record or File only. Unlock() is performed"
#define CB4_UNKNOWGET "DBserver object %s:%s unknown access"
#define CB4_UNKNOWPUT "DBserver object %s:%s unknown assign"

/*--------------------------------------------------------------------------*
 | x=oCb4:FileSpec  ACCESS determines the full .dbf name incl. path         |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retC    = string with the converted name or NIL             |
 | Description: retrieves the stored name from cb4xxxNEW or DBUSEAREA()     |
 | FS invocat.: retA := oCb4:FILESPEC                                       |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (filespec)        /* e.g.  _bb_cb4cdx_filespec_ac(ovSelv) */
{
   FSvar    *retVar   = VAR_NEW;

   CB4FS_DEBUG_BEGIN ("FileSpec",ovSelf);
   if (IS_VAR_CHR( VAR_OBJ(ovSelf, CB4OBJ_FILEFULLNAME)))
      SET_VAR_COPY( retVar, VAR_OBJ(ovSelf, CB4OBJ_FILEFULLNAME));
   return CB4FS_DEBUG_END ("FileSpec",retVar);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Found     ACCESS the status of last seek, locate, continue        |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retL    = status of the object variable                     |
 | FS invocat.: retL := oCb4:FOUND                                          |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (found)              /* e.g.  _bb_cb4cdx_found_ac(ovSelv) */
{
   FSvar    *retVar   = VAR_NEW;

   CB4FS_DEBUG_BEGIN ("Found",ovSelf);
   if (IS_VAR_LOG (VAR_OBJ(ovSelf, CB4OBJ_FOUNDFLAG)))
      SET_VAR_COPY( retVar, VAR_OBJ(ovSelf, CB4OBJ_FOUNDFLAG));
   else
      SET_VAR_LOG( retVar, 0);
   return CB4FS_DEBUG_END ("Found",retVar);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:ForBlock  ACCESS the object-wide FOR block                        |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retB    = code block or NIL                                 |
 | Description: retrieves the previously assigned code block                |
 | FS invocat.: retB := oCb4:FORBLOCK                                       |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (forblock)        /* e.g.  _bb_cb4cdx_forblock_ac(ovSelv) */
{
   FSvar    *retVar   = VAR_NEW;

   CB4FS_DEBUG_BEGIN ("ForBlockAcc",ovSelf);
   SET_VAR_COPY( retVar, VAR_OBJ(ovSelf, CB4OBJ_FOR_COND));
   return CB4FS_DEBUG_END ("ForBlockAcc",retVar);
}

/*--------------------------------------------------------------------------*
 | oCb4:ForBlock=x  ASSIGN the object-wide FOR block                        |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 |              assVar  = (B) or (C) or (NIL) value to be stored            |
 | Returns    : retO    = the object self                                   |
 | Description: Block or NIL is dtored directly to access variable,         |
 |              whilst CharString is compiled to a code block first         |
 | FS invocat.: oCb4:FORBLOCK := <value>                                    |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ASSIGN (forblock) /* e.g.  _bb_cb4cdx_forblock_as(ovSelv,assVar) */
{
   CB4FS_DEBUG_BEGIN ("ForBlockAss",ovSelf);
   switch (VAR_ISTYPE(assVar)) {
   case 'U':
   case 'B':
   case 'b':
      SET_VAR_COPY (VAR_OBJ(ovSelf, CB4OBJ_FOR_COND), assVar);
      break;
   case 'C':
   case 'E':
   case 'M':
      SET_VAR_COPY (VAR_OBJ(ovSelf, CB4OBJ_FOR_COND),
                    COMPILE_BLOCK(assVar) );
      break;
   default:
      CB4FSudf(callError)(ovSelf, 1, DTMMACH, "FORBLOCK",
                "cb4:FORBLOCK := exprB|exprC|NIL",
                __FILE__, __LINE__);                                 /* RTE */
   }
   return CB4FS_DEBUG_END ("ForBlockAss",assVar);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:WhileBlock  ACCESS the object-wide WHILE block                    |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retB    = code block or NIL                                 |
 | Description: retrieves the previously assigned code block                |
 | FS invocat.: retB := oCb4:WHILEBLOCK                                     |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (whileblock)    /* e.g.  _bb_cb4cdx_whileblock_ac(ovSelv) */
{
   FSvar    *retVar   = VAR_NEW;

   CB4FS_DEBUG_BEGIN ("WhileBlockAcc",ovSelf);
   SET_VAR_COPY( retVar, VAR_OBJ(ovSelf, CB4OBJ_WHI_COND));
   return CB4FS_DEBUG_END ("WhileBlockAcc",retVar);
}

/*--------------------------------------------------------------------------*
 | oCb4:WhileBlock=x  ASSIGN the object-wide WHILE block                    |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 |              assVar  = (B) or (C) or (NIL) value to be stored            |
 | Returns    : retO    = the object self                                   |
 | Description: Block or NIL is dtored directly to access variable,         |
 |              whilst CharString is compiled to a code block first         |
 | FS invocat.: oCb4:WHILEBLOCK := <value>                                  |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ASSIGN (whileblock) /* e.g.  _bb_cb4cdx_whileblock_as(ovSelv,assVar) */
{
   CB4FS_DEBUG_BEGIN ("WhileBlockAss",ovSelf);
   switch (VAR_ISTYPE(assVar)) {
   case 'U':
   case 'B':
   case 'b':
      SET_VAR_COPY (VAR_OBJ(ovSelf, CB4OBJ_WHI_COND), assVar);
      break;
   case 'C':
   case 'E':
   case 'M':
      SET_VAR_COPY (VAR_OBJ(ovSelf, CB4OBJ_WHI_COND),
                    COMPILE_BLOCK(assVar) );
      break;
   default:
      CB4FSudf(callError)(ovSelf, 1, DTMMACH, "WHILEBLOCK",
                          "cb4:WHILEBLOCK := exprB|exprC|NIL",
                          __FILE__, __LINE__);                       /* RTE */
   }
   return CB4FS_DEBUG_END ("WhileBlockAss",ovSelf);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Scope     ACCESS the object-wide scope                            |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retN|retL|NIL                                               |
 | Description: retrieves the previously assigned num or logical value (NIL)|
 | FS invocat.: ret := oCb4:SCOPE                                           |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (scope)              /* e.g.  _bb_cb4cdx_scope_ac(ovSelv) */
{
   FSvar    *retVar   = VAR_NEW;

   CB4FS_DEBUG_BEGIN ("ScopeAcc",ovSelf);
   SET_VAR_INT (retVar, VAR_INTNUM(VAR_OBJ(ovSelf, CB4OBJ_SCOPE)) );
   return CB4FS_DEBUG_END ("ScopeAcc",retVar);
}

/*--------------------------------------------------------------------------*
 | oCb4:Scope=x  ASSIGN the object-wide scope                               |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 |              assVar  = (N) next N records                          -or-  |
 |                        (L) TRUE for REST, FALSE for the ALL scope  -or-  |
 |                        (U) NIL to reset                                  |
 | Returns    : retO    = the object self                                   |
 | Description: the object-wide is compatible to CA/VO                      |
 | FS invocat.: oCb4:SCOPE := <value>                                       |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ASSIGN (scope)       /* e.g.  _bb_cb4cdx_scope_as(ovSelv,assVar) */
{
   CB4FS_DEBUG_BEGIN ("ScopeAss",ovSelf);
   switch (VAR_ISTYPE(assVar)) {
   case 'L':
   case 'I':
   case 'N':
   case 'F':
   case 'U':
      SET_VAR_COPY (VAR_OBJ(ovSelf, CB4OBJ_SCOPE), assVar);
      break;
   default:
      CB4FSudf(callError)(ovSelf, 1, DTMMACH, "SCOPE",
                          "cb4:SCOPE := exprN|exprL|NIL",
                          __FILE__, __LINE__);             /* RTE wrong val */
   }
   return CB4FS_DEBUG_END ("ScopeAss",ovSelf);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Header  ACCESS retrives the size of .dbf header                   |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retN    = header size in bytes or 0 on error                |
 | Description: retrieves the CodeBase structure                            |
 | FS invocat.: retN := oCb4:RECSIZE                                        |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (header)            /* e.g.  _bb_cb4cdx_header_ac(ovSelv) */
{
   DATA4    *dataPtr;                                    /* CB5 structure   */
   FSvar    *retNum   = VAR_NEW;

   CB4FS_DEBUG_BEGIN ("Header",ovSelf);
   if (!(dataPtr = CB4FS_GETDATA4PTR))
      SET_VAR_INT (retNum, 0);
   else
#ifdef CODEBASE6
      SET_VAR_INT (retNum, dataPtr->dataFile->headerLen); /* CB6 structure   */
#else
      SET_VAR_INT (retNum, dataPtr->header_len);         /* CB5 structure   */
#endif
   return CB4FS_DEBUG_END ("Header",retNum);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:RecSize ACCESS retrives the record length in bytes                |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retN    = record length or 0 on error                       |
 | Description: retrieves the CodeBase structure                            |
 | FS invocat.: retN := oCb4:RECSIZE                                        |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (recsize)          /* e.g.  _bb_cb4cdx_recsize_ac(ovSelv) */
{
   DATA4    *dataPtr;                                    /* CB5 structure   */
   FSvar    *retNum   = VAR_NEW;

   CB4FS_DEBUG_BEGIN ("RecSize",ovSelf);
   if (!(dataPtr = CB4FS_GETDATA4PTR))
      SET_VAR_INT (retNum, 0);
   else
      SET_VAR_INT (retNum, d4record_width(dataPtr));    /* CB5 function    */
   return CB4FS_DEBUG_END ("RecSize",retNum);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:RlockList ACCESS retrieve list of currently RLOCKed roecords      |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retA    = array of nemeric values or an empty array         |
 | Description: retrieves the CodeBase structure                            |
 | Compatib.  : Only one record may be locked at a time by CB5              |
 | FS invocat.: retA := oCb4:RLOCKLIST                                      |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (rlocklist)      /* e.g.  _bb_cb4cdx_rlocklist_ac(ovSelv) */
{
   DATA4    *dataPtr;                            /* CB5 structure           */
   FSvar    *retArr   = VAR_NEW;
   long     lockedRec;

   CB4FS_DEBUG_BEGIN ("RLoclList",ovSelf);
   if (!(dataPtr = CB4FS_GETDATA4PTR)) {
      VAR_NEW_ARRAY (retArr, 0);                 /* array of len 0          */
      return CB4FS_DEBUG_END ("RLoclList",retArr);
   }
#ifdef CODEBASE6
# if S4VERSION == 6502
   lockedRec = CB4FS_cpp_LockedRec(dataPtr);   /* CB6.54 requires c++ */
# else
   lockedRec = dataPtr->dataFile->lockedRecords.n_link;/* CB6.1 structure     */
# endif
#else
   lockedRec = dataPtr->locked_record;           /* CB5 structure           */
#endif
   if (lockedRec > 0L) {
      VAR_NEW_ARRAY (retArr, 1);                 /* create array[1]         */
      SET_VAR_INT ( VAR_ARRELEM(retArr, 1), lockedRec);
   } else
      VAR_NEW_ARRAY (retArr, 0);                 /* empty array             */

   return CB4FS_DEBUG_END ("RLoclList",retArr);
}

/*--------------------------------------------------------------------------*
 | oCb4:ClearScope()  METHOD clears the FOR, WHILE and SCOPE condit to NIL  |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : always NIL                                                  |
 | Description: The object wide scope for multi-record operation set by     |
 |              cb4:FORBLOCK, cb4:WHILEBLOCK and cb4:SCOPE are cleared      |
 | FS invocat.: retL := oCb4:CLEARSCOPE()                                   |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (clearscope)  /* e.g.  _bb_cb4cdx_clearscope_me(argc, argv) */
{
   FSvar  *ovSelf;

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("ClearScope",ovSelf);
   SET_VAR_NIL (VAR_OBJ(ovSelf, CB4OBJ_FOR_COND));
   SET_VAR_NIL (VAR_OBJ(ovSelf, CB4OBJ_WHI_COND));
   SET_VAR_NIL (VAR_OBJ(ovSelf, CB4OBJ_SCOPE)   );
   return CB4FS_DEBUG_END ("ClearScope",NIL_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:Refresh()  METHOD reread the curr.record from database, discarding  |
 |                        any changes that have been made.                  |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE if reread on CHANGED record successful, FALSE otherw.  |
 | Description: The sucess is possible only if no data movement or not      |
 |              commit() was performed since the changes.                   |
 | FS invocat.: retL := oCb4:REFRESH()                                      |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (refresh)        /* e.g.  _bb_cb4cdx_refresh_me(argc, argv) */
{
   FSvar  *ovSelf;
   FSvar  *vRV;
   DATA4  *dataPtr;                                   /* CB5 structure      */

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Refresh",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Refresh",FALSE_VAR);
   d4changed (dataPtr, 0);                            /* was changed ?      */
                                                      /* CB5 function       */
   vRV = CB4FS_CALL_METH (commit) (1, argv);
   
   return CB4FS_DEBUG_END ("Refresh", vRV);
}

/*--------------------------------------------------------------------------*
 | oCb4:Flock()   METHOD lock the whole file against write access by others |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: The driver internal locking scheme is used.                 |
 |              The preiviously set RLOCK() or FLOCK() is cleared before    |
 |              trying the FLOCK() and the record is flushed.               |
 |              Locking is ignored on EXCLUSIVE or READONLY open database.  |
 | Multiuser  : In multiuser mode, any replaced RECORD has to be locked     |
 |              via FLOCK() or RLOCK() if the database is used in SHARE     |
 |              mode and the SET AUTOLOCK OFF was set while dbf opening     |
 | FS invocat.: retL := oCb4:FLOCK()                                        |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (flock)            /* e.g.  _bb_cb4cdx_flock_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, "cb4:Flock()");
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Flock",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Flock",FALSE_VAR);

#ifdef CODEBASE6
   if ((dataPtr->dataFile->file.isReadOnly) ||        /* EXCL or READ ONLY: */
       (dataPtr->dataFile->file.lowAccessMode == OPEN4DENY_RW))
#else
   if ((dataPtr->file.is_read_only) ||
       (dataPtr->file.is_exclusive))                  /* EXCL or READ ONLY: */
#endif
      return CB4FS_DEBUG_END ("Flock (ex)",TRUE_VAR); /* simulated success  */

   d4flush   (dataPtr);                               /* CB5 update disk    */
   d4refresh (dataPtr);                               /* CB5 reread disk    */

   if (d4lock_file(dataPtr))
      return CB4FS_DEBUG_END ("Flock (F)",FALSE_VAR); /* failed             */
   return CB4FS_DEBUG_END ("Flock (T)",TRUE_VAR);     /* success            */
}

/*--------------------------------------------------------------------------*
 | oCb4:Rlock()   METHOD lock the current or specified record against write |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (opt) record number                               |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: The driver internal locking scheme is used. When the opt.   |
 |              param is not given, or the param is equal to the current    |
 |              record, the curr.record is locked and all previously set    |
 |              RLOCKs or FLOCKs are cleared. Otherwise, the previously     |
 |              set RLOCKs remains active and the new record is locked.     |
 |              Locking is ignored on EXCLUSIVE or READONLY open database.  |
 | Multiuser  : In multiuser mode, any replaced RECORD has to be locked     |
 |              via FLOCK() or RLOCK() if the database is used in SHARE     |
 |              mode and the SET AUTOLOCK OFF was set while dbf opening     |
 | Compatib.  : in CA/VO, multiple RLOCKS() are supported. CodeBase can     |
 |              maintain only a group of records. Therefore, a developer    |
 |              warning occur when attempting to lock OTHER than the        |
 |              current record.                                             |
 | FS invocat.: retL := oCb4:RLOCK( [num] )                                 |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (rlock)            /* e.g.  _bb_cb4cdx_rlock_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   long   recNo, lockRec;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, "cb4:Rlock( [recNo] )");
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Rlock",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Rlock",FALSE_VAR);

#ifdef CODEBASE6
   if ((dataPtr->dataFile->file.isReadOnly) ||        /* EXCL or READ ONLY: */
       (dataPtr->dataFile->file.lowAccessMode == OPEN4DENY_RW))
#else
   if ((dataPtr->file.is_read_only) ||
       (dataPtr->file.is_exclusive))                  /* EXCL or READ ONLY: */
#endif
      return CB4FS_DEBUG_END ("Rlock (ex)",TRUE_VAR); /* simulated success  */

   recNo = lockRec = d4recno (dataPtr);               /* CB5 function       */
   d4flush   (dataPtr);                               /* CB5 update disk    */
   d4refresh (dataPtr);                               /* CB5 reread disk    */

   if (argc > 1) {
      switch (VAR_ISTYPE(argv[1])) {
      case 'I':
      case 'N':
      case 'F':
         lockRec = (long) VAR_INTNUM(argv[1]);
         if (lockRec != recNo) {
            CB4FSudf(callError)(ovSelf, 1, DEVWARNING,
                    "RLOCK", CB4_RLOCKWARN,
                    __FILE__, __LINE__);              /* unsupported        */
            return CB4FS_DEBUG_END ("Rlock",FALSE_VAR);
         }
      case 'U':
         break;
      default:
         CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
                  "RLOCK", "cb4:Rlock( [recNo] )",
                  __FILE__, __LINE__);                /* RTE wrong type     */
         return CB4FS_DEBUG_END ("Rlock",FALSE_VAR);
      }
   }

   if (d4lock(dataPtr, recNo))
      return CB4FS_DEBUG_END ("Rlock (F)",FALSE_VAR); /* failed             */
   return CB4FS_DEBUG_END ("Rlock (T)",TRUE_VAR);     /* success            */
}

/*--------------------------------------------------------------------------*
 | oCb4:RlockVerify()   METHOD determines pending updates of current        |
 |                             record and issue RLOCK() if possible         |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: The driver internal locking scheme is used.                 |
 |              Here, this metod is equivalent to RLOCK().                  |
 |              Locking is ignored on EXCLUSIVE or READONLY open database.  |
 | Multiuser  : In multiuser mode, any replaced RECORD has to be locked     |
 |              via FLOCK() or RLOCK() if the database is used in SHARE     |
 |              mode and the SET AUTOLOCK OFF was set while dbf opening     |
 | FS invocat.: retL := oCb4:RLOCKVERIFY()                                  |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (rlockverify) /* e.g.  _bb_cb4cdx_rlockverify_me(argc,argv) */
{
   return CB4FS_CALL_METH(rlock)((argc <= 1 ? argc : 1), argv);
}

/*--------------------------------------------------------------------------*
 | oCb4:LockCurrentRecord()  METHOD equivalent to RLOCK()                   |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | FS invocat.: retL := oCb4:LOCKCURRENTRECORD()                            |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (lockcurrentrecord) /* e.g.  _bb_cb4cdx_lockcurrentrecord_me(argc,argv) */
{
   return CB4FS_CALL_METH(rlock)((argc <= 1 ? argc : 1), argv);
}

/*--------------------------------------------------------------------------*
 | oCb4:Unlock()  METHOD release specified or the current record or FLOCK() |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (opt) record number, unsupported in this RDD      |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: The driver internal locking scheme is used.                 |
 |              Here, this metod is equivalent to UNLOCK command            |
 |              Locking is ignored on EXCLUSIVE or READONLY open database.  |
 |              The file is flushed (COMMITed) before returning.            |
 | Multiuser  : In multiuser mode, any replaced RECORD has to be locked     |
 |              via FLOCK(), RLOCK() or APPEND(). Subsequent UNLOCK()       |
 |              releases all the preivously lockings.                       |
 | Compatib.  : in CA/VO, multiple RLOCKS() are supported. CodeBase can     |
 |              maintain only a group of records. Therefore, a developer    |
 |              warning occur when attempting to unlock OTHER than the      |
 |              current record.                                             |
 | FS invocat.: retL := oCb4:UNLOCK( [num] )                                |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (unlock)           /* e.g.  _bb_cb4cdx_unlock_me(argc,argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   long   recNo, lockRec;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, "cb4:Unlock( [recNo] )");
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("UnLock",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("UnLock",FALSE_VAR);

   CB4FS_CALL_METH(commit)(1,argv);                   /* cb4:comit()        */

#ifdef CODEBASE6
   if ((dataPtr->dataFile->file.isReadOnly) ||        /* EXCL or READ ONLY: */
       (dataPtr->dataFile->file.lowAccessMode == OPEN4DENY_RW))  
#else
   if ((dataPtr->file.is_read_only) ||
       (dataPtr->file.is_exclusive))                  /* EXCL or READ ONLY: */
#endif
      return CB4FS_DEBUG_END ("UnLock",TRUE_VAR);     /* simulated success  */

   recNo = lockRec = d4recno (dataPtr);               /* CB5 function       */

   if (argc > 1) {
      switch (VAR_ISTYPE(argv[1])) {
      case 'I':
      case 'N':
      case 'F':
         lockRec = (long) VAR_INTNUM(argv[1]);
         if (lockRec != recNo) {
            CB4FSudf(callError)(ovSelf, 1, DEVWARNING,
               "UNLOCK", CB4_UNLCKWARN,
               __FILE__, __LINE__);                   /* unsupported        */
/*          return CB4FS_DEBUG_END ("UnLock",FALSE_VAR); */ /* enable on request  */
         }
      case 'U':
         break;
      default:
         CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
             "UNLOCK", "cb4:Unlock( [recNo] )",
             __FILE__, __LINE__);                     /* data type mismatch */
         return CB4FS_DEBUG_END ("UnLock",FALSE_VAR);
      }
   }
   if (d4unlock(dataPtr))
      return CB4FS_DEBUG_END ("UnLock",FALSE_VAR);    /* failed             */
   return CB4FS_DEBUG_END ("UnLock",TRUE_VAR);        /* success            */
}

/*--------------------------------------------------------------------------*
 | oCb4:NoiVarGet()  METHOD exception handling to determine field name      |
 |                          and return its value.                           |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (C) field name                                    |
 | Returns    : field value or NIL                                          |
 | Description: The exception may occur when the FlagShip compiler          |
 |              cannot determine the field directly (e.g. in macro),        |
 |              or when the fieldname is given as xx := oCB4:anyname        |
 |              and <anyname> is unknown ACCESS method of this object.      |
 |              This method should NOT be called directly.                  |
 | FS invocat.: ret := oCb4:anyname                                         |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (noivarget)     /* e.g.  _bb_cb4cdx_noivarget_me(argc,argv) */
{
   FSvar  *ovSelf;
   FSvar  *newArgv[2];
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSint  fieldNo;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, "cb4:NoiVarGet(cFieldName)");
      return NIL_VAR;
   }
   ovSelf = argv[0];
   if (argc < 2 || ! IS_VAR_CHR(argv[1])) {
      CB4FSudf(callError)(ovSelf, 1, RTMISARG,
               "NOIVARGET", "cb4:NoiVarGet(cFieldName)", __FILE__, __LINE__);
      return NIL_VAR;
   }
   CB4FS_DEBUG_BEGIN ("NoiVarGet",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("NoiVarGet",NIL_VAR);

   fieldNo = d4field_number(dataPtr, VAR_CHR(argv[1]));   /* CB5 get fld no */
   if (fieldNo <= 0) {
      sprintf (repdum, CB4_UNKNOWGET,
                  CB4FSrddName, VAR_CHR(argv[1]));        /* RTE unknown acc*/
      CB4FSudf(callError)(ovSelf, 1, RTNOFLD, 
               "NOIVARGET", repdum, __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("NoiVarGet",NIL_VAR);
   }
   newArgv[0] = argv[0];
   newArgv[1] = SET_VAR_INT (VAR_NEW, fieldNo);           /* faster access  */

   return CB4FS_DEBUG_END ("NoiVarGet", CB4FS_CALL_METH(fieldget)(2, newArgv) );
}

/*--------------------------------------------------------------------------*
 | oCb4:NoiVarPut()  METHOD exception handling to determine field name      |
 |                          and replace it by given value.                  |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (C) field name                                    |
 |              argv[2] =     replaced value                                |
 | Returns    : field value or NIL                                          |
 | Description: The exception may occur when the FlagShip compiler          |
 |              cannot determine the field directly (e.g. in macro),        |
 |              or when the fieldname is given as oCB4:anyname := xxx       |
 |              and <anyname> is unknown ASSIGN method of this object.      |
 |              This method should NOT be called directly.                  |
 | FS invocat.: ret := oCb4:anyname                                         |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (noivarput)     /* e.g.  _bb_cb4cdx_noivarput_me(argc,argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSint  fieldNo;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0]) ||
       ! IS_VAR_CHR(argv[1]) || IS_VAR_NIL(argv[2])) {
      rt_error(RTMISARG, "cb4:NoiVarPut()");
      return NIL_VAR;
   }
   ovSelf = argv[0];
   if (argc < 3 || ! IS_VAR_CHR(argv[1]) || IS_VAR_NIL(argv[2])) {
      CB4FSudf(callError)(ovSelf, 1, RTMISARG, "NOIVARPUT",
               "cb4:NoiVarPut(cFieldName,value)", __FILE__, __LINE__);
      return NIL_VAR;
   }

   CB4FS_DEBUG_BEGIN ("NoiVarPut",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("NoiVarPut",NIL_VAR);

   fieldNo = d4field_number(dataPtr, VAR_CHR(argv[1]));   /* CB5 get fld no */
   if (fieldNo <= 0) {
      sprintf (repdum, CB4_UNKNOWPUT,
                  CB4FSrddName, VAR_CHR(argv[1]));        /* RTE unknown ass*/
      CB4FSudf(callError)(ovSelf, 1, RTNOFLD, 
               "NOIVARPUT", repdum, __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("NoiVarPut",NIL_VAR);
   }
   return CB4FS_DEBUG_END ("NoiVarPut", CB4FS_CALL_METH(fieldput)(argc,argv) );
}


/**** eof rddcb4d.c ****/

