/****************************************************************************
 * rddcb4h.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2007 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2007/01/25 15:23:31 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains parts of the full RDD driver:                         *
 *    oCb4:indexext            ACCESS index extention                       *
 *    oCb4:indexkey()          METHOD returns the key expression            *
 *    oCb4:indexord()          METHOD ord posit of the control index/order  *
 *    oCb4:indexcount()        METHOD info about the order or index         *
 *    oCb4:orderinfo()         METHOD info about the order or index         *
 *    oCb4:setindex()          METHOD add index file to current list        *
 *    oCb4:setorder()          METHOD set controlling index/order/tag       *
 *    oCb4:seek()              METHOD seek the specified key                *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */

extern FSint fgsGetSoftSeek();                              /* SET SOFTS ?  */
extern FSvar *fsFnNAME(dtos)();                             /* DTOS()       */
extern FSint CB4FSudf(checkFile)();                         /* rddcb4a.c    */
extern FSint CB4FSudf(setIndexArr)();                       /* rddcb4g.c    */
extern FSint CB4FSudf(callError)();                         /* rddcb4b.c    */
extern FSvar *CB4FS_CALL_METH(gotop)();                     /* rddcb4b.c    */
extern FSint fgsExtractName();
static FSint getOrderPos();
static FSint setNewTag  ();

#define CB4_ARGUMENTSTR   "%s;argument %d"
#define CB4_SETINDEXPAR   "cb4:SetIndex(cFileName)"
#define CB4_SETORDERPAR   "cb4:SetOrder(cOrderName|nOrderNum,[cFileName])"
#define CB4_INDEXKEYPAR   "cb4:IndexKey([cName|nPos])"
#define CB4_ORDINFPARAM   "cb4:OderInfo(nWhat,[nOrd|cOrd],[cIdxFile],[uNewVar])"
#define CB4_ORDINFPAR1    "cb4:OderInfo(%d) %s"
#define CB4_SEEKPAR       "cb4:Seek(uValue,[lSoft])"
#define CB4_CBFUNCTFAIL   "%s;CodeBase function fails with error %d"
#define CB4_NOTSUPPRTED   "not supported by this RDD"
#define CB4_WRONGREQ      "unknown request"
#define CB4_ACCESSONLY    "%s;access only supported for request #%d"

/*--------------------------------------------------------------------------*
 | x=oCb4:IndexExt  ACCESS determine the default index extension            |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retC    = The driver dependant name of the index file ext.  |
 |                        e.g. ".idx" upper/lower according to fs_set()     |
 | FS invocat.: retC := oCb4:INDEXEXT                                       |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (indexext)        /* e.g.  _bb_cb4cdx_indexext_ac(ovSelv) */
{
   FSvar *cvRetVar = VAR_NEW;

   CB4FS_DEBUG_BEGIN ("IndexExt",ovSelf);
   SET_VAR_CHR (cvRetVar, VAR_CHR(VAR_OBJ(ovSelf,CB4OBJ_IDX_EXT)));
   return CB4FS_DEBUG_END ("IndexExt",cvRetVar);
}

/*--------------------------------------------------------------------------*
 | oCb4:SetIndex()    METHOD add index file (and all orders/tags included   |
 |                    there) to the list of open indexes                    |
 |                    Similar to/called from std.funct. OrdListAdd()        |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (C) name of the index file                        |
 | Returns    : TRUE on success, FALSE otherwise                            |
 | Description: On success, the list of open indicies is appended           |
 |              The active order number remains unchanged, except when      |
 |               the first index is assigned (order = 1).                   |
 |              On multiple-tag indexes, all tags (orders) of the index     |
 |               file are added into the list.                              |
 |              To select the controlling index/order/tag, use SetOrder()   |
 | FS invocat.: retL := oCb4:SETINDEX(cFileName)                            |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (setindex)       /* e.g.  _bb_cb4cdx_setindex_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   INDEX4   *indexPtr;                                /* CB5 structure      */
   TAG4     *tagPtr;                                  /* CB5 structure      */
   TAG4INFO *tagInfoPtr;                              /* CB5 structure      */
   FSint    ii, ok, pos = 0, err = 0;
   FSvar    *aIndexArr;
   char     *fullName;
   extern   FSchar fileNameBuff[];

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_SETINDEXPAR);            /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   if (argc < 2) {
      CB4FSudf(callError)(ovSelf, 1, RTMISARG, "SETINDEX",
               CB4_SETINDEXPAR, __FILE__, __LINE__);
      return FALSE_VAR;
   }

   CB4FS_DEBUG_BEGIN ("SetIndex",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("SetIndex",FALSE_VAR);  /* database not open  */

   if (!IS_VAR_CHR(argv[1]) || IS_VAR_EMPTY(argv[1])) {
      sprintf (repdum, CB4_ARGUMENTSTR, CB4_SETINDEXPAR, 1);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "SETINDEX",
          repdum, __FILE__, __LINE__);               /* RTE: wrong arg type */
      return CB4FS_DEBUG_END ("SetIndex",FALSE_VAR);
   }
/*
 *  Add default extension, transform upper/lower case according to fs_set()
 *  Check if index file exist, RTE if not.
 */
   if (! CB4FSudf(checkFile)
         (VAR_CHR(argv[1]), fileNameBuff, &fullName,
          VAR_CHR(VAR_OBJ(ovSelf,CB4OBJ_IDX_EXT)), CB4_SETINDEXPAR, 1))
      return CB4FS_DEBUG_END ("SetIndex",FALSE_VAR);

   strcpy (fileNameBuff, fullName);
/*
 *  Open the index file,
 *  add the data to the list of open indexes/orders/tags.
 *  if first index, SET ORDER = 1
 */
   indexPtr = i4open (dataPtr, fileNameBuff);         /* CB5: open index    */
   if (! indexPtr) {
      err = dataPtr->code_base->error_code;
      goto error_cb;
   }

   aIndexArr = VAR_OBJ (ovSelf, CB4OBJ_INDEX4ARR);
#ifdef CODEBASE6
   indexPtr->codeBase->error_code = 0;
#else
   indexPtr->code_base->error_code = 0;
#endif

   for( tagPtr = 0, ii = 0 ;; ii++ ) {
      tagPtr = (TAG4 *)l4next( &indexPtr->tags, tagPtr ) ;     /* CB5 funct */
#ifdef CODEBASE6
      if ( indexPtr->codeBase->error_code < 0 ) {              /* CB6 struc */
         err = indexPtr->codeBase->error_code;
#else
      if ( indexPtr->code_base->error_code < 0 ) {             /* CB5 struc */
         err = indexPtr->code_base->error_code;
#endif
         goto error_cb;
      }
      if ( tagPtr == 0 )
         break;
      if (ii == 0)
         pos = CB4FSudf(setIndexArr)(ovSelf, 1, fileNameBuff,
                        t4alias(tagPtr), ii+1, 1, indexPtr, tagPtr);
      else
         CB4FSudf(setIndexArr)(ovSelf, 1, fileNameBuff,
                        t4alias(tagPtr), ii+1, 0, indexPtr, tagPtr);
   }

   if (pos == 1) {
      ok = setNewTag (ovSelf, pos, dataPtr, aIndexArr);
      if (ok) CB4FS_CALL_METH(gotop)(1, argv);
      return CB4FS_DEBUG_END ("SetIndex", (ok ? TRUE_VAR : FALSE_VAR));
   }

   return CB4FS_DEBUG_END ("SetIndex", TRUE_VAR);

 error_cb:
   sprintf (repdum, CB4_CBFUNCTFAIL, CB4_SETINDEXPAR, err);
   CB4FSudf(callError)(ovSelf, 1, DEVWARNING, "SETINDEX",
         repdum, __FILE__, __LINE__);                /* Devel: CB5 failure  */
   return CB4FS_DEBUG_END ("SetIndex",FALSE_VAR);

}

/*--------------------------------------------------------------------------*
 | oCb4:SetOrder()    METHOD set controlling index/order/tag from list      |
 |                           of open indicies (see cb4:SetIndex() )         |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (CN) name or number of the order (tag) 10chr upp  |
 |              argv[2] = (C opt)  name of the index file                   |
 | Returns    : TRUE on success, FALSE otherwise                            |
 | Description: No indicies or orders/tags are opended by this method,      |
 |              use cb4:SetIndex() for this purposes.                       |
 | FS invocat.: retL := oCb4:SETINDEX(cFileName)                            |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (setorder)       /* e.g.  _bb_cb4cdx_setorder_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *aIndexArr;
   FSint    ok, ordPos = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_SETORDERPAR);            /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   if (argc < 2) {
      CB4FSudf(callError)(ovSelf, 1, RTMISARG, "SETORDER",
               CB4_SETORDERPAR, __FILE__, __LINE__);
      return FALSE_VAR;
   }
   CB4FS_DEBUG_BEGIN ("SetOrder",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("SetOrder",FALSE_VAR);  /* database not open  */
   if (IS_VAR_NIL(aIndexArr = CB4FS_GETINDEXARR))
      return CB4FS_DEBUG_END ("SetOrder",FALSE_VAR);  /* indexes  not open  */

	// accomodate set order 0
	if (VAR_ISTYPE (argv[1]) == 'I' || VAR_ISTYPE (argv[1]) == 'N' || VAR_ISTYPE (argv[1]) == 'F')
	{
		if (VAR_INTNUM(argv[1]) == 0)
		{
			d4tag_select (dataPtr, NULL);
			SET_VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM), 0);
			return CB4FS_DEBUG_END ("SetOrder", (ok ? TRUE_VAR : FALSE_VAR));
		}
	}

   ordPos = getOrderPos (ovSelf, argv[1],
                         (argc > 2 ? argv[2] : NIL_VAR),
                         aIndexArr, CB4_SETORDERPAR);
   if (ordPos == 0)
      return CB4FS_DEBUG_END ("SetOrder",FALSE_VAR);  /* indexes  not open  */

   ok = setNewTag (ovSelf, ordPos, dataPtr, aIndexArr);

/* if you wish to execute GO TOP at every SET INDEX, enable:
 */
//  if (ok) CB4FS_CALL_METH(gotop)(1, argv);
   return CB4FS_DEBUG_END ("SetOrder", (ok ? TRUE_VAR : FALSE_VAR));
}

/*--------------------------------------------------------------------------*
 | oCb4:Seek()        METHOD move dbf to the record having the specified    |
 |                    key value according to the current control index/tag  |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (CNIDL) the value to be searched for.             |
 |              argv[2] = (L opt) soft seek on ?                            |
 |                        when not specified or NIL, the current SET SOFTS  |
 |                        is used.                                          |
 | Returns    : TRUE if found, FALSE otherwise.                             |
 |              also, cb4:EOF() and cb4:FOUND() are set                     |
 |              When no index is open, always FALSE is returned.            |
 | FS invocat.: retL := oCb4:SEEK(expr, [lSoft])                            |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (seek)               /* e.g.  _bb_cb4cdx_seek_me(argc,argv) */
{
   FSvar    *ovSelf;
   FSint    found = 0, ok = 0, err = 0;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *aIndexArr;
   FSvar    *cvTemp;
   FSvar    *param[1];
#ifdef TEST_ON
   FSvar    *aSubArr;
   TAG4     *tagPtr;                                  /* CB5 structure      */
#endif

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_SEEKPAR);                /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   if (argc < 2) {
      CB4FSudf(callError)(ovSelf, 1, RTMISARG, "SEEK",
               CB4_SEEKPAR, __FILE__, __LINE__);
      return FALSE_VAR;
   }
   CB4FS_DEBUG_BEGIN ("Seek",ovSelf);

   SET_VAR_LOG (VAR_OBJ(ovSelf,CB4OBJ_FOUNDFLAG), 0); /* not found yet      */

   if (! (dataPtr = CB4FS_GETDATA4PTR)) {
      err = RTNODBF;                                  /* database not open  */
      goto err_ret;
   }
   if (IS_VAR_NIL(aIndexArr = CB4FS_GETINDEXARR)) {
      err = INTNDXNO;
      goto err_ret;
   }

   if (! IS_VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM)) ||
       VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM)) == 0)
      return CB4FS_DEBUG_END ("Seek",FALSE_VAR);      /* SET ORDER TO 0     */

#ifdef TEST_ON
   aSubArr = VAR_ARRELEM(CB4FS_GETINDEXARR, 1);
   tagPtr  = VAR_SPECIAL (VAR_ARRELEM(aSubArr, CB4OBJ_INDEX_T4PTR));
   fprintf(stderr, "\nSeek/%d: tagPtr= %x = %x recno=%d",
        __LINE__, tagPtr, d4tag_selected(dataPtr), (int)d4recno(dataPtr));
#endif

   switch (VAR_ISTYPE(argv[1])) {
   case 'C':
   case 'E':
   case 'M':
      found = d4seek (dataPtr, VAR_CHR(argv[1]));
      break;
   case 'I':
   case 'N':
   case 'F':
      found = d4seek_double (dataPtr, VAR_FPNUM(argv[1]));
      break;
   case 'D':
      cvTemp = VAR_NEW;
      param[0] = argv[1];
      SET_VAR_COPY (cvTemp, fsFnNAME(dtos)(1,param));
      found = d4seek (dataPtr, VAR_CHR(cvTemp));
      break;
   default:
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "SEEK",
              CB4_SEEKPAR, __FILE__, __LINE__);          /* RTE: wrong type */
      return CB4FS_DEBUG_END ("Seek",FALSE_VAR);
      break;
   }

   switch (found) {
   case r4success:
      ok = 1;
      /* consider SET DELETED */
      while ( fgsGetDelete() && d4deleted(dataPtr) ) {
         switch(VAR_ISTYPE(argv[1])) {
         case 'C':
         case 'E':
         case 'M':
            found = d4seekNext (dataPtr, VAR_CHR(argv[1]));
            break;
         case 'I':
         case 'N':
         case 'F':
            found = d4seekNextDouble (dataPtr, VAR_FPNUM(argv[1]));
            break;
         case 'D':
            found = d4seekNext (dataPtr, VAR_CHR(cvTemp));
            break;
         }
         if (found != r4success) {
	         ok = 0;
	         break;
	      }
      }
      if( ok ) {
         SET_VAR_LOG (VAR_OBJ(ovSelf,CB4OBJ_FOUNDFLAG), 1);      /* found flag */
	      break;
      } else if (found != r4after && found != r4entry) {
         d4go_eof(dataPtr);
	      break;
      }
      /* fall through to softseek check */
      /* modified softseek behavior to match SEEK command */
   case r4after:                                              /* softseek ? */
      if ( (argc > 2 && IS_VAR_TRUE(argv[2])) ||
	        (argc <= 2 && fgsGetSoftSeek() ) ) {
         ok = 1;
         SET_VAR_LOG (VAR_OBJ(ovSelf,CB4OBJ_FOUNDFLAG), 1);   /* found flag */
         /* consider SET DELETED */
         if ( fgsGetDelete() && d4deleted(dataPtr) )
            CB4FSudf(CSkip)(ovSelf,dataPtr, 1L, 1);        /*  skip deleted */
         break;
      }
      /* fall through to default (goto eof) if softseek is off */
   default:
      ok = 0;
      d4go_eof(dataPtr);
   }
   return CB4FS_DEBUG_END ("Seek",SET_VAR_LOG(VAR_NEW, ok));

 err_ret:
   CB4FSudf(callError)(ovSelf, -err, DEVWARNING, "SETINDEX",
         CB4_SEEKPAR, __FILE__, __LINE__);             /* Devel: dbf/index n.open  */
   return CB4FS_DEBUG_END ("Seek",FALSE_VAR);

}


/*--------------------------------------------------------------------------*
 | oCb4:IndexKey()    METHOD returns the key expression                     |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (CN) name or ordinal number of the order in list. |
 |                             If NIL or "" or 0 is specified, the curr.    |
 |                             controlling list is returned.                |
 |              argv[2] = (C opt)  name of the index file                   |
 | Returns    : (C) key expression or "" on error                           |
 | Compatib.  : CA/VO requires the paramater, FS does not                   |
 |              CA/VO does not support argv[2]                              |
 | FS invocat.: retC := oCb4:INDEXKEY([c|nPosit]);                          |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (indexkey)       /* e.g.  _bb_cb4cdx_indexkey_me(argc,argv) */
{
   FSvar    *ovSelf, *cvRetVar = VAR_NEW;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   INDEX4   *indexPtr;                                /* CB5 structure      */
   TAG4     *tagPtr;                                  /* CB5 structure      */
   FSint    ii, ordPos = 0;
   FSvar    *aIndexArr, *aSubArr;

   SET_VAR_CHR (cvRetVar, "");

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_INDEXKEYPAR);            /* RTE: missing args  */
      return NIL_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("IndexKey",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("IndexKey",FALSE_VAR);  /* database not open  */
   if (IS_VAR_NIL(aIndexArr = CB4FS_GETINDEXARR))
      return CB4FS_DEBUG_END ("IndexKey",cvRetVar);   /* indexes  not open  */

   if (argc > 1)
      ordPos = getOrderPos (ovSelf, argv[1],
                            (argc > 2 ? argv[2] : NIL_VAR),
                             aIndexArr, CB4_INDEXKEYPAR);

   if (!ordPos && IS_VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM)))
      ordPos = VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM));

   if (!ordPos || ordPos > VAR_ISDIM(aIndexArr))
      return CB4FS_DEBUG_END ("IndexKey",cvRetVar);             /* nothing  */

   aSubArr = VAR_ARRELEM(aIndexArr, ordPos);
   tagPtr  = VAR_SPECIAL( VAR_ARRELEM (aSubArr, CB4OBJ_INDEX_T4PTR));
#ifdef CODEBASE6
   SET_VAR_CHR(cvRetVar, (char *)expr4source(tagPtr->tagFile->expr));   /* CB6 fn   */
#else
   SET_VAR_CHR(cvRetVar, (char *)expr4source(tagPtr->expr));            /* CB5 fn   */
#endif

   return CB4FS_DEBUG_END ("IndexKey",cvRetVar);
}

/*--------------------------------------------------------------------------*
 | oCb4:IndexOrd()    METHOD returns the ordinal position of the            |
 |                    controlling index/order                               |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : (N) posit or 0 on eror                                      |
 | FS invocat.: retN := oCb4:INDEXORD()                                     |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (indexord)       /* e.g.  _bb_cb4cdx_indexord_me(argc,argv) */
{
   FSvar    *ovSelf;
   FSint    ordPos = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, "cb4:IndexOrd()");           /* RTE: missing args  */
      return NIL_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("IndexOrd",ovSelf);

   if (IS_VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM)))
      ordPos = VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM));
   return CB4FS_DEBUG_END ("IndexOrd",SET_VAR_INT (VAR_NEW, ordPos));
}
/*--------------------------------------------------------------------------*
 | oCb4:IndexCount()    METHOD returns number of indices                    |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (indexcount)    /* e.g.  _bb_cb4cdx_indexcount_ac(ovSelv) */
{
   FSvar *myArgv[2];
   FSvar *retVar;
   myArgv[0] = ovSelf;
   myArgv[1] = SET_VAR_INT(VAR_NEW, DBOI_ORDERCOUNT);
   retVar = OBJ_METHEXEC("orderinfo", -1, 2, myArgv);
   if(IS_VAR_NUM(retVar))
      return retVar;
   return IZERO_VAR;
}

/*--------------------------------------------------------------------------*
 | oCb4:OrderInfo()   METHOD returns the info about the specified order/tag |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = requested info according to 'rddsys.fh'           |
 |                 DBOI_CONDITION      order condition string      --> char |
 |                 DBOI_CUSTOM         custom RDD                 <*-> log  |
 |                 DBOI_EXPRESSION     order expression string     --> char |
 |                 DBOI_FILEHANDLE     file handle no              *-> num  |
 |                 DBOI_FULLPATH       full path & name            --> char |
 |                 DBOI_HPLOCKING      high perform CA/VO lock     *-> log  |
 |                 DBOI_INDEXEXT       index extension (".cdx")   <--> char |
 |                 DBOI_INDEXNAME      index file name main part   --> char |
 |                 DBOI_ISCOND         condition flag set ?        --> log  |
 |                 DBOI_ISDESC         descending flag set?        --> log  |
 |                 DBOI_KEYCOUNT       no.of keys in the order     *-> num  |
 |                 DBOI_KEYDEC         no.of decimals in the key   --> num  |
 |                 DBOI_KEYSINCLUDED   no.of keys of cond.index    --> num  |
 |                 DBOI_KEYSIZE        size of the key             --> num  |
 |                 DBOI_KEYTYPE        type of the key             --> char |
 |                 DBOI_KEYVAL         evaluate key to value       *-> any  |
 |                 DBOI_LOCKOFFSET     lock offset NewIndexLock()  *-> num  |
 |                 DBOI_NAME           order name                  --> char |
 |                 DBOI_NUMBER         posit.of the order in list  --> num  |
 |                 DBOI_ORDERCOUNT     no.of orders in the file    --> num  |
 |                 DBOI_POSITION       locical record no.in order  *-> num  |
 |                 DBOI_RECNO          physic. record no.in order  --> num  |
 |                 DBOI_SCOPEBOTTOM    bottom boundary of order    *-> num  |
 |                 DBOI_SCOPETOP       top    boundary of order    *-> num  |
 |                 DBOI_SETCODEBLOCK   key converted to code block --> block|
 |                 DBOI_UNIQUE         unique flag set?            --> log  |
 |                 DBOI_DBFNAME        FS only, dbf name of INDEX  *-> char |
 |                 DBOI_INDEXCHECK     FS only, index ok?          *-> log  |
 |              argv[2] = (opt) nOrderNum|cOrderName within index file      |
 |              argv[3] = (opt) cFileName of index file                     |
 |              argv[4] = (opt) assigned value                              |
 | Returns    : the requested info or NIL on error                          |
 | FS invocat.: ret := oCb4:ORDERINFO(DBOI_xxx,[nOrd|cOrd],[cFile],[uNew])  |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (orderinfo)    /* e.g.  _bb_cb4cdx_orderinfo_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   INDEX4 *indexPtr;                                  /* CB5 structure      */
   TAG4   *tagPtr, *tagTmp;                           /* CB5 structure      */
   FSvar  *uvRetVar = VAR_NEW;
   FSvar  *aIndexArr, *aSubArr;
   FSint  ok, what, ordPos = 0, ii, err = 0;
   FSint  currentOrder = 0;                           /* Order no. (nOrd)   */
   extern FSchar charBuff500[];                       /* global buffer      */

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTARGCNT, CB4_ORDINFPARAM);            /* RTE param          */
      return NIL_VAR;
   }
   ovSelf = argv[0];
   if (argc < 2) {
      CB4FSudf(callError)(ovSelf, 1, RTMISARG, "ORDERINFO",
             CB4_ORDINFPARAM, __FILE__, __LINE__);
      return NIL_VAR;
   }
   CB4FS_DEBUG_BEGIN ("OrderInfo",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("OrderInfo",NIL_VAR);   /* database not open  */
   if (IS_VAR_NIL(aIndexArr = CB4FS_GETINDEXARR))
      return CB4FS_DEBUG_END ("OrderInfo",NIL_VAR);   /* indexes  not open  */

   switch (VAR_ISTYPE(argv[1])) {
   case 'I':
   case 'N':
   case 'F':
      what = VAR_INTNUM (argv[1]);
      break;
   default:
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "ORDERINFO",
           CB4_ORDINFPARAM, __FILE__, __LINE__);      /* RTE: illegal type  */
      return CB4FS_DEBUG_END ("OrderInfo",NIL_VAR);
   }

   if (IS_VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM)))
      currentOrder = VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM));
/*
 *  determine the specified or current order number
 */
   if (argc > 2)
      ordPos = getOrderPos (ovSelf, argv[2],
                            (argc > 3 ? argv[3] : NIL_VAR),
                             aIndexArr, CB4_ORDINFPARAM);
   if (!ordPos && IS_VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM)))
      ordPos = VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM));

   if (!ordPos || ordPos > VAR_ISDIM(aIndexArr)) {
      sprintf (repdum, CB4_ARGUMENTSTR, CB4_ORDINFPARAM, 2);
      CB4FSudf(callError)(ovSelf, 1, DEVWARNING, "ORDERINFO",
           repdum, __FILE__, __LINE__);               /* RTE: illegal val   */
      return CB4FS_DEBUG_END ("OrderInfo",NIL_VAR);   /* nothing to check   */
   }

   ok = 1;
   if (currentOrder != ordPos)
      ok = setNewTag (ovSelf, ordPos, dataPtr, aIndexArr);

   if (!ok)
      return CB4FS_DEBUG_END ("OrderInfo (F)",NIL_VAR);

   aSubArr  = VAR_ARRELEM (aIndexArr, ordPos);
   tagPtr   = VAR_SPECIAL (VAR_ARRELEM (aSubArr, CB4OBJ_INDEX_T4PTR ));
   indexPtr = VAR_SPECIAL (VAR_ARRELEM (aSubArr, CB4OBJ_INDEX_I4PTR ));

/*
 *  Access
 */
   switch (what) {
   case DBOI_CONDITION:
#ifdef CODEBASE6
      SET_VAR_CHR(uvRetVar, (char *)expr4source(tagPtr->tagFile->filter));/* CB6 fn */
#else
      SET_VAR_CHR(uvRetVar, (char *)expr4source(tagPtr->filter));         /* CB5 fn */
#endif
      break;
   case DBOI_EXPRESSION:
#ifdef CODEBASE6
      SET_VAR_CHR(uvRetVar, (char *)expr4source(tagPtr->tagFile->expr)); /* CB6 fn  */
#else
      SET_VAR_CHR(uvRetVar, (char *)expr4source(tagPtr->expr));          /* CB5 fn  */
#endif
      break;
   case DBOI_FULLPATH:
      SET_VAR_COPY(uvRetVar, VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_FNAME));
      break;
   case DBOI_INDEXEXT:
      SET_VAR_COPY(uvRetVar, VAR_OBJ(ovSelf,CB4OBJ_IDX_EXT));
      break;
   case DBOI_INDEXNAME:
      if (fgsExtractName (VAR_CHR(VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_FNAME)),
                          charBuff500, 11) )
         SET_VAR_CHR (uvRetVar, charBuff500);
      break;
   case DBOI_ISCOND:
#ifdef CODEBASE6
      SET_VAR_LOG(uvRetVar, strlen((char *)expr4source(tagPtr->tagFile->filter)) > 0);/* CB6 fn*/
#else
      SET_VAR_LOG(uvRetVar, strlen((char *)expr4source(tagPtr->filter)) > 0);         /* CB5 fn*/
#endif
      break;
   case DBOI_ISDESC:
      SET_VAR_LOG(uvRetVar, t4is_descending(tagPtr));           /* CB5 fn   */
      break;
   case DBOI_KEYDEC:
#ifdef CODEBASE6
      SET_VAR_INT(uvRetVar, tagPtr->tagFile->keyDec);           /* CB5 stru */
#else
      SET_VAR_INT(uvRetVar, tagPtr->key_dec);                   /* CB5 stru */
#endif
      break;
   case DBOI_KEYSIZE:
#ifdef CODEBASE6
      SET_VAR_INT(uvRetVar,
				  (strlen((char *)expr4source(tagPtr->tagFile->filter)))); /* CB6 fn*/
#else
      SET_VAR_INT(uvRetVar, (strlen((char *)expr4source(tagPtr->filter))));/* CB5 fn*/
#endif
      break;
   case DBOI_KEYTYPE:
#ifdef S4UNIX
#ifdef CODEBASE6
      charBuff500[0] = (char) tagPtr->tagFile->keyType;
#else
      charBuff500[0] = (char) tagPtr->key_type;
#endif
      charBuff500[1] = 0;
      SET_VAR_CHR (uvRetVar, charBuff500);
#endif
      break;
   case DBOI_NAME:
      SET_VAR_COPY(uvRetVar, VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_TNAME));
      break;
   case DBOI_NUMBER:
      SET_VAR_INT(uvRetVar, ordPos);
      break;
   case DBOI_ORDERCOUNT:
      for( tagTmp = 0, ii = 0 ;; ii++ ) {
         tagTmp = (TAG4 *)l4next( &indexPtr->tags, tagTmp ) ;   /* CB5 fn   */
         if ( tagTmp == 0 ) break;
      }
      SET_VAR_INT(uvRetVar, ii);
      break;
   case DBOI_RECNO:
      SET_VAR_INT (uvRetVar, d4recno(dataPtr));                 /* CB5 fn   */
      break;
   case DBOI_SETCODEBLOCK:
#ifdef CODEBASE6
      strcpy (charBuff500, (char *)expr4source(tagPtr->tagFile->filter));
#else
      strcpy (charBuff500, (char *)expr4source(tagPtr->filter));
#endif
      if (*charBuff500)
         SET_VAR_COPY (uvRetVar, fgsMakeCB (charBuff500));
      break;
   case DBOI_UNIQUE:
      SET_VAR_LOG(uvRetVar, t4unique(tagPtr));                  /* CB5 fn   */
      break;
   case DBOI_CUSTOM:
   case DBOI_DBFNAME:
   case DBOI_FILEHANDLE:
   case DBOI_HPLOCKING:
   case DBOI_INDEXCHECK:
   case DBOI_KEYCOUNT:
   case DBOI_KEYSINCLUDED:
   case DBOI_KEYVAL:
   case DBOI_LOCKOFFSET:
   case DBOI_POSITION:
   case DBOI_SCOPEBOTTOM:
   case DBOI_SCOPETOP:
      sprintf  (repdum, CB4_ORDINFPAR1, what, CB4_NOTSUPPRTED);
      CB4FSudf(callError)(ovSelf, 1, DEVWARNING, "ORDERINFO",
           repdum, __FILE__, __LINE__);               /* Devel.Warning      */
      err = 1;
      break;
   default:
      sprintf  (repdum, CB4_ORDINFPAR1, what, CB4_WRONGREQ);
      CB4FSudf(callError)(ovSelf, 1, RTARGVAL, "ORDERINFO",
           repdum, __FILE__, __LINE__);               /* RTE: illegal val   */
      err = 1;
   }

/*
 *  Assign
 */
   if (err == 0 && argc > 4) {
      switch (what) {
      case DBOI_INDEXEXT:
         if (IS_VAR_CHR(argv[4]) && VAR_ISLEN(argv[4]) == 5)
            SET_VAR_COPY(VAR_OBJ(ovSelf,CB4OBJ_IDX_EXT), argv[4]);
         break;
      default:
         sprintf  (repdum, CB4_ACCESSONLY,
                      CB4_ORDINFPARAM, what);         /* unsupp ASSIGN      */
         CB4FSudf(callError)(ovSelf, 1, DEVWARNING,
               "ORDERINFO", repdum,
               __FILE__, __LINE__);                   /* Devel.Warning      */
      }
   }

   ok = 1;
   if (currentOrder && currentOrder != ordPos)
      ok = setNewTag (ovSelf, currentOrder, dataPtr, aIndexArr);

   return CB4FS_DEBUG_END ("OrderInfo", (ok ? uvRetVar : NIL_VAR));
}


/*--------------------------------------------------------------------------*
 | internal function, determines the order position                         |
 *--------------------------------------------------------------------------*/
static FSint getOrderPos (ovSelf, var, cvFileName, aIndexArr, cText)
FSvar *ovSelf;
FSvar *var;                         /* input order as N,I,C type            */
FSvar *cvFileName;                  /* opt. file name when type(var) == "C" */
FSvar *aIndexArr;                   /* 2dim array, list with indexes/orders */
FSchar *cText;                      /* param text for error and type "C"    */
{
   FSint    ordPos = 0, ii;
   FSvar    *aSubArr;
   FSchar   *cDummy;
   extern   FSchar  charBuff500[];                    /* global buffer      */
   extern   FSchar  fileNameBuff[];                   /* global buffer      */

   switch (VAR_ISTYPE (var)) {
   case 'I':
   case 'N':
   case 'F':
      ordPos = VAR_INTNUM(var);
      break;
   case 'C':
   case 'E':
   case 'M':
      if (IS_VAR_EMPTY(var))
         break;
      if (! fgsExtractName (VAR_CHR(var), charBuff500, 11)) /* Upper        */
         break;
      if (IS_VAR_CHR(cvFileName) &&
          !IS_VAR_EMPTY(cvFileName)) {                      /* check file ? */
         if (! CB4FSudf (checkFile)(cvFileName, fileNameBuff,
                &cDummy, VAR_CHR(VAR_OBJ (ovSelf,CB4OBJ_IDX_EXT)),
                cText, -2) )                                /* convert only */
            *fileNameBuff = '\0';
      } else
        *fileNameBuff = '\0';
      if (!IS_VAR_ARR(aIndexArr))
         break;
      for (ii=1; ii <= VAR_ISDIM(aIndexArr); ii++) {
         aSubArr = VAR_ARRELEM(aIndexArr, ii);
         if (strcmp (VAR_CHR(VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_TNAME)),
                     charBuff500) == 0) {
            ordPos = ii;
            if (*fileNameBuff) {                            /* check File ? */
               if (strcmp (fileNameBuff,
                      VAR_CHR(VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_FNAME))) == 0)
                  break;
            } else
              break;
         }
      }
      break;
   }
   return ordPos;
}

/*--------------------------------------------------------------------------*
 | internal function, set new controlling tag from list                     |
 *--------------------------------------------------------------------------*/
static FSint setNewTag (ovSelf, ordPos, dataPtr, aIndexArr)
FSvar   *ovSelf;
FSint   ordPos;
DATA4   *dataPtr;                                     /* CB5 structure      */
FSvar   *aIndexArr;
{
   FSvar    *aSubArr;
   TAG4     *tagPtr;                                  /* CB5 structure      */

   if (!IS_VAR_ARR(aIndexArr))
      return 0;
   SET_VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_INDEXORDNUM), ordPos);
   aSubArr = VAR_ARRELEM(aIndexArr, ordPos);
   tagPtr  = VAR_SPECIAL (VAR_ARRELEM(aSubArr, CB4OBJ_INDEX_T4PTR));
   if (!tagPtr) return 0;
   d4tag_select (dataPtr, tagPtr);                    /* CB5 function       */
   if (tagPtr != d4tag_selected(dataPtr))
	  return 0;
   return 1;
}
/**** eof rddcb4h.c ****/

