/****************************************************************************
 * rddcb4k.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2004 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2004/10/27 10:00:04 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains parts of the full RDD driver:                         *
 *                                                                          *
 *    oCb4:average()         METHOD AVERAGE                                 *
 *    oCb4:count()           METHOD COUNT FOR                               *
 *    oCb4:join()            METHOD JOIN                                    *
 *    oCb4:sort()            METHOD SORT ON ...                             *
 *    oCb4:sum()             METHOD SUM                                     *
 *    oCb4:total()           METHOD TOTAL                                   *
 *    oCb4:update()          METHOD UPDATE ON .. FROM..                     *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */

extern FSint  CB4FSudf(callError)();                /* in rddcb4b.c         */
extern FSint  CB4FSudf(getGlobalScope)();           /* in rddcb4e.c         */
extern FSint  CB4FSudf(CSkip)();                    /* in rddcb4b.c         */
extern FSint  CB4FSudf(decodeScope)();              /* in rddcb4e.c         */
extern FSint  CB4FSudf(decodeCond)();               /* in rddcb4e.c         */
extern FSvar *fsFnNAME(__dbtotal)();                /* == TOTAL ... TO..    */
extern FSvar *fsFnNAME(__dbjoin)();                 /* == JOIN WITH..TO..   */
extern FSvar *fsFnNAME(__dbsort)();                 /* == SORT TO ...       */
extern FSvar *fsFnNAME(__dbupdate)();               /* == UPDATE ON...      */
extern FSvar *fsFnNAME(fieldget)();                 /* == FIELDGET()        */

static FSvar *mySum ();                             /* forward declarat.    */

#define CB4_WRONGARG     "%s;wrong type or missing argument # %d"
#define CB4_WRONGELEM1   "%s;field/expr of Array[%d] does not return numeric"
#define CB4_WRONGELEM2   "%s;code block of Array[%d] does not return numeric"
#define CB4_WRONGELEM3   "%s;wrong type (%c) of Array[%d] expression"
#define CB4_AVERAGEPAR   "cb4:Average(cbExpr|acbExpr,[cbFor,cbWhile,nlScop])"
#define CB4_UPDATEPAR    "cb4:Update(cAlias,cbOnKey,[lRandom],cbUpdate)"
#define CB4_SUMPAR       "cb4:Sum(cbExpr|acbExpr,[cbFor,cbWhile,nlScop])"
#define CB4_COUNTPAR     "cb4:Count([cbFor,cbWhile,nlScop])"
#define CB4_JOINPAR      "cb4:Join(cAlias,cName,[aFields,cbFor,cVia])"
#define CB4_TOTALPAR     "cb4:Total(cName,cbKey,[aFields,cbFor,cbWhile,nlScop])"
#define CB4_SORTPAR      "cb4:Sort(cName,[aFields,cbFor,cbWhile,nlScop,cVia])"


/*--------------------------------------------------------------------------*
 | oCb4:average()    METHOD calculates avarage of num expressions for a     |
 |                          range of records, stores result in array        |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C|N|B|aC|aB|aN) expression to be evaluated   (1,2)|
 |         opt. argv[ 2] (C|B) "for" condition of the src file              |
 |         opt. argv[ 3] (C|B) "while" condition of the src file            |
 |         opt. argv[ 4] (N|L) scope (.F. all,def; .T. rest; n records)     |
 | Note/compat: (1) object and symbol values of VO are unsupported          |
 |              (2) num.values = field no. are unsupported by CA/VO         |
 | Returns    : array containing avarages or empty array on errer           |
 | Compatib.  : CA/VO does not support array of num. expressions            |
 | Description: -similar to the the command AVARAGE..TO                     |
 |              -If relations are set, each skip moves also related childs. |
 |               For disabling this behaviour and to increase speed, use:   |
 |               xx := oCb4:Relation ; oCb4:Average() ; oCb4:Relation := xx |
 |              -If only par1 is specified, the default object scope is     |
 |               used. If not set, the default is CURRENT record only.      |
 | FS invocat.: retA := oCb4:Average(expr,....)                             |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (average)         /* e.g.  _bb_cb4cdx_average_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *avRet;
   FSvar    *tmpVar;
   FSvar    *uvExpr, *uvForCond, *uvWhiCond;
   long     iScope = 0L, iCount = 0L;
   FSint    ii, iDimens;
   FSint    err = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_AVERAGEPAR);            /* RTE: missing args   */
      return NIL_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Average",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR)) {
      VAR_NEW_ARRAY (avRet, 0);
      return CB4FS_DEBUG_END ("Average",avRet);       /* database not open  */
   }
/*
 *  Parameter checking
 */
   uvExpr    = VAR_NEW;
   uvForCond = VAR_NEW;
   uvWhiCond = VAR_NEW;

   if (IS_VAR_CHR(argv[1]) || IS_VAR_ARR(argv[1]) || IS_VAR_BLK(argv[1]))
      SET_VAR_COPY (uvExpr, argv[1]);                  /* eval expr         */
   else
      err = 1;

   if (!err)
      err = CB4FSudf(getGlobalScope)
                    (argc, 2, argv, 1, &uvForCond, &uvWhiCond, &iScope);
   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_AVERAGEPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
         "AVERAGE", repdum, __FILE__, __LINE__);     /* RTE: wrong arg type */
      VAR_NEW_ARRAY (avRet, 0);
      return CB4FS_DEBUG_END ("Average",avRet);
   }

   avRet = mySum (ovSelf, uvExpr, uvForCond, uvWhiCond, iScope,
                  CB4_AVERAGEPAR, "AVERAGE", &iCount);   /* totals -> arr   */

   iDimens = VAR_ISDIM(avRet);
   if (iDimens < 1)
      return CB4FS_DEBUG_END ("Average (F)",avRet);

/*
 *  calculate the average
 */
   for (ii=1; ii <= iDimens; ii++) {
      tmpVar = VAR_ARRELEM(avRet,ii);
      if (iCount && IS_VAR_NUM(tmpVar) )
         SET_VAR_NUM (VAR_ARRELEM(avRet,ii),
                      VAR_FPNUM (tmpVar) / (double)(iCount));
   }
   return CB4FS_DEBUG_END ("Average (T)",avRet);
}

/*--------------------------------------------------------------------------*
 | oCb4:Sum()        METHOD calculates sum of num expressions for a         |
 |                          range of records, stores result in array        |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C|N|B|aC|aB|aN) expression to be evaluated   (1,2)|
 |         opt. argv[ 2] (C|B) "for" condition of the src file              |
 |         opt. argv[ 3] (C|B) "while" condition of the src file            |
 |         opt. argv[ 4] (N|L) scope (.F. all,def; .T. rest; n records)     |
 | Note/compat: (1) object and symbol values of VO are unsupported          |
 |              (2) num.values = NIL = field no. are unsupported by CA/VO   |
 | Returns    : TRUE on sucess, FALSE otherwise                             |
 | Description: -similar to the the command SUM .. TO                       |
 |              -If relations are set, each skip moves also related childs. |
 |               For disabling this behaviour and to increase speed, use:   |
 |               xx := oCb4:Relation ; oCb4:Average() ; oCb4:Relation := xx |
 |              -If only par1 is specified, the default object scope is     |
 |               used. If not set, the default is ALL records.              |
 | FS invocat.: retL := oCb4:Sum(....)                                      |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (sum)                 /* e.g.  _bb_cb4cdx_sum_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *avRet;
   FSvar    *uvExpr, *uvForCond, *uvWhiCond;
   long     iScope = 0L, iCount = 0L;
   FSint    err = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_SUMPAR);                /* RTE: missing args   */
      return NIL_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Sum",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR)) {
      VAR_NEW_ARRAY (avRet, 0);
      return CB4FS_DEBUG_END ("Sum",avRet);           /* database not open  */
   }
/*
 *  Parameter checking
 */
   uvExpr    = VAR_NEW;
   uvForCond = VAR_NEW;
   uvWhiCond = VAR_NEW;

   if (IS_VAR_CHR(argv[1]) || IS_VAR_ARR(argv[1]) || IS_VAR_BLK(argv[1]))
      SET_VAR_COPY (uvExpr, argv[1]);                  /* eval expr         */
   else
      err = 1;

   if (!err)
      err = CB4FSudf(getGlobalScope)
                (argc, 2, argv, LONG_MAX, &uvForCond, &uvWhiCond, &iScope);
   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_SUMPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
         "AVERAGE", repdum, __FILE__, __LINE__);     /* RTE: wrong arg type */
      VAR_NEW_ARRAY (avRet, 0);
      return CB4FS_DEBUG_END ("Sum",avRet);
   }

   avRet = mySum (ovSelf, uvExpr, uvForCond, uvWhiCond, iScope,
                  CB4_SUMPAR, "SUM", &iCount);           /* totals -> arr   */

   return CB4FS_DEBUG_END ("Sum",avRet);
}


/*--------------------------------------------------------------------------*
 | internal function for average() and sum()                                |
 | returns array avRet of totals and count of fields in iptrCount           |
 *--------------------------------------------------------------------------*/
static FSvar *mySum (ovSelf,uvExpr,uvForCond,uvWhiCond,iScope,errText,errFun,iptrCount)
FSvar *ovSelf, *uvExpr, *uvForCond, *uvWhiCond;
long  iScope;
char  *errText, *errFun;
long  *iptrCount;
{
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *avRet  = VAR_NEW;
   FSvar    *avExpr;
   FSvar    *tmpVar, *macroVar;
   FSvar    *passArg[2];
   FSint    ii, imark, imark0, isFor, isWhile, iDimens;
   FSint    err = 0, found = 0;
   long     iCount = 0L;

   imark0 = VAR_DEL_MARK();                           /* mark garbage coll. */

   CB4FS_DEBUG_BEGIN ("(mySum)",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR)) {
      VAR_NEW_ARRAY (avRet, 0);
      return CB4FS_DEBUG_END ("(mySum)",avRet);       /* database not open  */
   }
/*
 *  create and init the expr (avExpr) and return (avRet) array
 *  according to (uvExpr) which may be an array of field# or field names.
 *  avExpr[] contains then code-block which receives the requested value
 *           or type 'I'|'N' containing the ordinal field number.
 */
   iDimens = 1;
   avExpr  = VAR_NEW;
   if (IS_VAR_ARR(uvExpr))
      iDimens = VAR_ISDIM(uvExpr);
   VAR_NEW_ARRAY (avRet,  iDimens);                  /* create n-dim array  */
   VAR_NEW_ARRAY (avExpr, iDimens);                  /* create n-dim array  */

   for (ii=1; (!err && (ii <= iDimens)); ii++) {
      SET_VAR_NUM(VAR_ARRELEM(avRet,ii), 0.0);       /* init avRet[] with 0 */
      if (iDimens == 1)
         tmpVar = uvExpr;
      else
         tmpVar = VAR_ARRELEM(uvExpr,ii);

      switch (VAR_ISTYPE(tmpVar)) {
      case 'b':
      case 'B':                                     /* field is passed      */
         SET_VAR_COPY (VAR_ARRELEM(avExpr,ii),      /* as a code block      */
                       tmpVar);
         break;
      case 'M':
      case 'E':
      case 'C':                                     /* field name is passed */
         SET_VAR_COPY (VAR_ARRELEM(avExpr,ii),      /* as a string "MYFIED" */
                       COMPILE_BLOCK (tmpVar));
         break;
      case 'I':
      case 'N':
      case 'F':                                     /* field name is passed */
         SET_VAR_COPY(VAR_ARRELEM(avExpr,ii),       /* as a field no.       */
                      tmpVar);
         break;
      case 'U':                                     /* NIL = element/field  */
         SET_VAR_INT (VAR_ARRELEM(avExpr,ii), ii);
         break;
      default:
         err = VAR_ISTYPE(tmpVar);
      }
   }
   if (err) goto err_return;

/*
 *  loop trough the dbf for the WHILE / FOR / Scope condition
 */
   isFor   = IS_VAR_BLK(uvForCond);              /* FOR   available ?       */
   isWhile = IS_VAR_BLK(uvWhiCond);              /* WHILE available ?       */
   passArg[0] = ovSelf;

   imark = VAR_DEL_MARK();                       /* mark next vars garbage  */
   while (!d4eof(dataPtr)                        /* while !eof()   .and. ;  */
          && (!isWhile ||
              IS_VAR_EVAL_TRUE(uvWhiCond))       /*   eval(bWhile) .and. ;  */
          && (iScope > 0L) )                     /*   scope=ok              */
   {
      if (!isFor || IS_VAR_EVAL_TRUE(uvForCond))
      {                                          /* if eval(bFor)           */
/*
 *       eval the bExpr array, add results to aRet[]
 */
         iCount++;
         for (ii=1; ii <= iDimens; ii++)         /* calculate sum,          */
         {                                       /* add to aRet[ii]         */
            switch (VAR_ISTYPE(VAR_ARRELEM(avExpr,ii))) {
            case 'b':
            case 'B': tmpVar = VAR_EVAL_BLOCK (VAR_ARRELEM(avExpr,ii));
                      break;
            case 'I':
            case 'N':
            case 'F': passArg[1] = VAR_ARRELEM(avExpr,ii);
                      tmpVar = fsFnNAME(fieldget)(2, passArg);
            }
            if (IS_VAR_NUM(tmpVar)) {            /* returns numeric value?  */
               SET_VAR_NUM (VAR_ARRELEM(avRet,ii),
                   VAR_FPNUM (VAR_ARRELEM(avRet,ii)) + VAR_FPNUM(tmpVar));
            } else {
               err = VAR_ISTYPE(tmpVar);
               goto err_return;
            }
         }
      }
      if (--iScope > 0L)                         /* scope --                */
         CB4FSudf(CSkip)(ovSelf,dataPtr, 1L, 1); /*   skip 1 incl. relat.   */
      VAR_DEL_SINCE_MARK(imark);                 /* garbage colection of    */
   }                                             /*   temp vars in block    */

   VAR_DEL_SINCE_MARK(imark0);                   /* perf.full garb.colection*/
   *iptrCount = iCount;
   return CB4FS_DEBUG_END ("(mySum)",avRet);

 err_return:
   switch (err) {
     case  1: sprintf (repdum, CB4_WRONGELEM1, errText, ii); break;
     case  2: sprintf (repdum, CB4_WRONGELEM2, errText, ii); break;
     default: sprintf (repdum, CB4_WRONGELEM3, errText, err, ii);
   }
   CB4FSudf(callError)(ovSelf, 1, RTARGTYP, errFun, repdum,
                       __FILE__, __LINE__);
   SET_VAR_NIL   (avRet);
   VAR_NEW_ARRAY (avRet, 0);
   return CB4FS_DEBUG_END ("(mySum)",avRet);
}


/*--------------------------------------------------------------------------*
 | oCb4:count()      METHOD counts no of records that match spec. scope     |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |         opt. argv[ 1] (C|B) "for" condition of the src file              |
 |         opt. argv[ 2] (C|B) "while" condition of the src file            |
 |         opt. argv[ 3] (N|L) scope (.F. all,.T. rest; n records)          |
 | Returns    : number of records                                           |
 | Description: -similar to the the command COUNT...TO                      |
 |              -If relations are set, each skip moves also related childs. |
 |               For disabling this behaviour and to increase speed, use:   |
 |               xx := oCb4:Relation ; oCb4:Average() ; oCb4:Relation := xx |
 |              -If only par1 is specified, the default object scope is     |
 |               used. If not set, the default is the CURRENT record only.  |
 | FS invocat.: retN := oCb4:Count(....)                                    |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (count)             /* e.g.  _bb_cb4cdx_count_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *uvExpr, *uvForCond, *uvWhiCond;
   FSvar    *nvRet = VAR_NEW;
   long     iScope = 0L, iCount = 0;
   FSint    ii, imark, imark0, isFor, isWhile, iDimens;
   FSint    err = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_COUNTPAR);              /* RTE: missing args   */
      return NIL_VAR;
   }

   SET_VAR_INT (nvRet, 0);
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Count",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Count",nvRet);         /* database not open  */
/*
 *  Parameter checking
 */
   imark0    = VAR_DEL_MARK();                        /* mark garbage coll. */
   uvExpr    = VAR_NEW;
   uvForCond = VAR_NEW;
   uvWhiCond = VAR_NEW;

   err = CB4FSudf(getGlobalScope)
                 (argc, 1, argv, 1, &uvForCond, &uvWhiCond, &iScope);
   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_COUNTPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
         "COUNT", repdum, __FILE__, __LINE__);       /* RTE: wrong arg type */
      return CB4FS_DEBUG_END ("Count",nvRet);
   }

   isFor   = IS_VAR_BLK(uvForCond);              /* FOR   available ?       */
   isWhile = IS_VAR_BLK(uvWhiCond);              /* WHILE available ?       */

/*
 *  loop trough the dbf for the WHILE / FOR / Scope condition
 */
   imark = VAR_DEL_MARK();                       /* mark next vars garbage  */
   while (!d4eof(dataPtr)                        /* while !eof()   .and. ;  */
          && (!isWhile ||
              IS_VAR_EVAL_TRUE(uvWhiCond))       /*   eval(bWhile) .and. ;  */
          && (iScope > 0L) )                     /*   scope=ok              */
   {
      if (!isFor || IS_VAR_EVAL_TRUE(uvForCond))
         iCount++;                               /* count                   */
      if (--iScope > 0L)                         /* scope --                */
         CB4FSudf(CSkip)(ovSelf,dataPtr, 1L, 1); /*   skip 1 incl. relat.   */
      VAR_DEL_SINCE_MARK(imark);                 /* garbage colection of    */
   }                                             /*   temp vars in block    */

   SET_VAR_INT (nvRet, iCount);
   VAR_DEL_SINCE_MARK(imark0);                   /* perf.full garb.colection*/
   return CB4FS_DEBUG_END ("Count",nvRet);
}

/*--------------------------------------------------------------------------*
 | oCb4:Join()       METHOD joins this database with another one and place  |
 |                          results in new dbf                              |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C)   alias name of the "other" database        (1)|
 |              argv[ 2] (C)   name of the new created dbf               (1)|
 |         opt. argv[ 3] (A)   array of fields acc.to FIELDS clause      (1)|
 |         opt. argv[ 4] (C|B) "for" condition of the src file              |
 |         opt. argv[ 5] (C)   "via" clause of the new dbf               (2)|
 | Notes/comp.: (1) object and symbol values of VO are unsupported          |
 |              (2) via is not supported by CA/VO                           |
 | Returns    : TRUE on sucess, FALSE otherwise                             |
 | Description: compatible to JOIN .. TO ... command                        |
 |              This fn uses the default fn __DBJOIN(...), which itself     |
 |              uses the cb4:methods but does not call cb4:join()           |
 | FS invocat.: retL := oCb4:Join(.....)                                    |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (join)               /* e.g.  _bb_cb4cdx_join_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *passArgv[5];
   FSint    err = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_JOINPAR);                /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Join",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Join",FALSE_VAR);      /* database not open  */
/*
 *  Parameter checking
 */
   VAR_NEW_ARGS(passArgv, 5);

   if (argc > 6) argc = 6;
   switch (argc) {
   case 6:                                                /* via clause     */
      SET_VAR_COPY (passArgv[4], argv[5]);
   case 5:                                                /* for cond       */
      if (CB4FSudf(decodeCond)(argv[4], passArgv[3]))
         err = 4;
   case 4:                                                /* array of fields*/
      SET_VAR_COPY (passArgv[2], argv[3]);
      switch (VAR_ISTYPE(argv[3])) {
        case 'A': if (VAR_ISDIM(argv[3]) < 1) err = 3;
        case 'U': break;
        default : err = 3;
      }
   case 3:                                                /* new dbf name   */
      if (!IS_VAR_CHR(argv[2])) err = 2;
      SET_VAR_COPY (passArgv[1], argv[2]);
   case 2:                                                /* alias name     */
      if (!IS_VAR_CHR(argv[1])) err = 1;
      SET_VAR_COPY (passArgv[0], argv[1]);
      if (!err && (argc < 3)) err = argc;
      break;
   default:
      err = 1;
   }
   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_JOINPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "JOIN",
               repdum, __FILE__, __LINE__);                         /* RTE  */
      return CB4FS_DEBUG_END ("Join",FALSE_VAR);
   }

/* function __dbjoin (cAlias, cFile, aFields, bFor, cVia)
 */
   return CB4FS_DEBUG_END ("Join",fsFnNAME(__dbjoin)(5,passArgv));
}

/*--------------------------------------------------------------------------*
 | oCb4:Total()      METHOD calculates totals  of num expressions for a     |
 |                          range of records, stores result in database     |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C)   name of the new dbf with results          (1)|
 |              argv[ 2] (C|B) key field acc.to ON clause of TOTAL       (1)|
 |         opt. argv[ 3] (A)   array of fields acc.to FIELDS clause      (1)|
 |         opt. argv[ 4] (C|B) "for" condition of the src file              |
 |         opt. argv[ 5] (C|B) "while" condition of the src file            |
 |         opt. argv[ 6] (N|L) scope (.F. all,def; .T. rest; n records)     |
 | Note       : (1) object and symbol values of VO are unsupported          |
 | Returns    : TRUE on sucess, FALSE otherwise                             |
 | Description: compatible to TOTAL .. TO ... command                       |
 |              This fn uses the default fn __DBTOTAL(...), which itself    |
 |              uses the cb4:methods but does not call cb4:total()          |
 | FS invocat.: retL := oCb4:Total(....)                                    |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (total)             /* e.g.  _bb_cb4cdx_total_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *passArgv[8];
   FSint    err = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_TOTALPAR);               /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Total",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Total",FALSE_VAR);     /* database not open  */
/*
 *  Parameter checking
 */
   VAR_NEW_ARGS(passArgv, 8);

   if (argc > 7) argc = 7;
   switch (argc) {
   case 7:
      if (CB4FSudf(decodeScope)(argv[6], passArgv, 5, 7))
         err = 6;                                         /* NEXT, REST|ALL */
   case 6:
      if (CB4FSudf(decodeCond)(argv[5], passArgv[4]))     /* while cond     */
         err = 5;
   case 5:
      if (CB4FSudf(decodeCond)(argv[4], passArgv[3]))     /* for   cond     */
         err = 4;
   case 4:                                                /* array of fields*/
      SET_VAR_COPY (passArgv[2], argv[3]);
      switch (VAR_ISTYPE(argv[3])) {
        case 'A': if (VAR_ISDIM(argv[3]) < 1) err = 3;
        case 'U': break;
        default : err = 3;
      }
   case 3:
      if (CB4FSudf(decodeCond)(argv[2], passArgv[1]))     /* On key         */
         err = 2;
   case 2:
      if (argc < 3) err = 2;
      SET_VAR_COPY (passArgv[0], argv[1]);
      if (! IS_VAR_CHR(argv[1]) || IS_VAR_EMPTY(argv[1]))
         err = 1;
      break;
   default:
      err = 1;
   }
   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_TOTALPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "TOTAL",
               repdum, __FILE__, __LINE__);                         /* RTE  */
      return CB4FS_DEBUG_END ("Total",FALSE_VAR);
   }

/* function __dbtotal (cName, bOnKey, aFields, bFor,
                       bWhile, nNext, nRec, lRest)
 */
   return CB4FS_DEBUG_END ("Total",fsFnNAME(__dbtotal)(8,passArgv));

}

/*--------------------------------------------------------------------------*
 | oCb4:Sort()  METHOD copies selected records to other dbf in sorted order |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C)   name of the new dbf                       (1)|
 |         opt. argv[ 2] (A)   array of fields acc.to FIELDS clause      (1)|
 |         opt. argv[ 3] (C|B) "for" condition of the src file              |
 |         opt. argv[ 4] (C|B) "while" condition of the src file            |
 |         opt. argv[ 5] (N|L) scope (.F. all,def; .T. rest; n records)     |
 |         opt. argv[ 6] (C)   "via" clause of the new dbf               (2)|
 | Notes/comp.: (1) object and symbol values of VO are unsupported          |
 |              (2) via is not supported by CA/VO                           |
 | Returns    : TRUE on sucess, FALSE otherwise                             |
 | Description: -compatible to SORT .. TO ... command                       |
 |              -This fn uses the default fn __DBSORT (...), which itself   |
 |               uses the cb4:methods but does not call cb4:sort()          |
 |              -If neither conditions nor scope is specified, the object   |
 |               condition & scope is used. If not set, ALL records perform.|
 | FS invocat.: retL := oCb4:Sort (....)                                    |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (sort)               /* e.g.  _bb_cb4cdx_sort_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *passArgv[8];
   FSint    err = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_SORTPAR);                /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Sort",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Sort",FALSE_VAR);      /* database not open  */
/*
 *  Parameter checking
 */
   VAR_NEW_ARGS(passArgv, 8);

   if (argc > 7) argc = 7;
   switch (argc) {
   case 7:                                                /* via clause     */
      SET_VAR_COPY (passArgv[7], argv[6]);
   case 6:
   case 5:
   case 4:
      err = CB4FSudf(getGlobalScope)
                 (argc, 3, argv, LONG_MAX, &(passArgv[2]),
                 &(passArgv[3]), &(passArgv[4]));
   case 3:                                                /* array of fields*/
      SET_VAR_COPY (passArgv[1], argv[2]);
      switch (VAR_ISTYPE(argv[2])) {
        case 'A': if (VAR_ISDIM(argv[2]) < 1) err = 2;
        case 'U': break;
        default : err = 23;
      }
   case 2:                                                /* name of new dbf*/
      SET_VAR_COPY (passArgv[0], argv[1]);
      if (! IS_VAR_CHR(argv[1]) || IS_VAR_EMPTY(argv[1]))
         err = 1;
      break;
   default:
      err = 1;
   }
   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_SORTPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "SORT",
               repdum, __FILE__, __LINE__);                         /* RTE  */
      return CB4FS_DEBUG_END ("Sort",FALSE_VAR);
   }

/* function __dbsort (cName, aFields, bFor, bWhile,
                      nNext, nRec, lRest, cVia)
 */
   return CB4FS_DEBUG_END ("Sort",fsFnNAME(__dbsort)(8,passArgv));

}

/*--------------------------------------------------------------------------*
 | oCb4:Update()     METHOD update this dbf with data from another dbf      |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C)   alias name of the "FROM" database         (1)|
 |              argv[ 2] (C|B) "on" key                                  (1)|
 |         opt. argv[ 3] (L)   random flag                                  |
 |              argv[ 4] (B)   code block that perform the update operation |
 |                             e.g. {|| _field->fld := var, ...}            |
 | Notes/comp.: (1) object and symbol values of VO are unsupported          |
 | Returns    : TRUE on sucess, FALSE otherwise                             |
 | Description: compatible to UPDATE ON... command                          |
 |              This fn uses the default fn __DBUPDATE(...), which itself   |
 |              uses the cb4:methods but does not call cb4:update()         |
 | FS invocat.: retL := oCb4:Update(...)                                    |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (update)           /* e.g.  _bb_cb4cdx_update_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *passArgv[4];
   FSint    err = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_UPDATEPAR);              /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Update",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Update",FALSE_VAR);    /* database not open  */
/*
 *  Parameter checking
 */
   VAR_NEW_ARGS(passArgv, 4);

   if (argc > 5) argc = 5;
   switch (argc) {
   case 5:                                                /* for cond       */
      if (!IS_VAR_BLK(argv[4])) err = 4;
      SET_VAR_COPY (passArgv[3], argv[4]);
   case 4:                                                /* random         */
      if (IS_VAR_NIL(argv[3]))
         SET_VAR_LOG (passArgv[2], 0);
      else if (IS_VAR_LOG(argv[3]))
         SET_VAR_COPY (passArgv[2], argv[3]);
      else err = 3;
   case 3:                                                /* on key         */
      if (IS_VAR_CHR(argv[2]))
         SET_VAR_COPY (passArgv[1], COMPILE_BLOCK(argv[2]));
      else if (IS_VAR_BLK(argv[2]))
         SET_VAR_COPY (passArgv[1], argv[2]);
      else
         err = 2;
   case 2:                                                /* alias name     */
      if (!IS_VAR_CHR(argv[1])) err = 1;
      SET_VAR_COPY (passArgv[0], argv[1]);
      if (!err && (argc < 5)) err = argc;
      break;
   default:
      err = 1;
   }
   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_UPDATEPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "UPDATE",
               repdum, __FILE__, __LINE__);                         /* RTE  */
      return CB4FS_DEBUG_END ("Update",FALSE_VAR);
   }

/* function __dbupdate (cAlias, bKey, lRandom, bUpdate)
 */
   return CB4FS_DEBUG_END ("Update",fsFnNAME(__dbupdate)(4,passArgv));
}

/**** eof rddcb4k.c ****/

