* For FlagShip 6 
* Copyright (c) 2002..06 by multisoft GmbH, Munich, Germany 
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the 
* modified source code when your application is distributed in source, 
* but may not remove the copyright notice.
* 
* Release $Date: 2006/07/28 08:00:04 $
*
* Default Input handler for CheckBox, called from CheckBox:Show() 
*
* You may assign YOUR own handler UDF via code block :Handler, i.e.
* oBox:Handler := {|obj| MyUdf(obj) }
* which is then called from :Show() automatically instead of this
* default handler.
*
* Compile     : FlagShip -c -na -m -io=a checkboxhand.prg
* Link Unix   : FlagShip -myapp*.prg checkboxhand.o
* Link Windows: FlagShip myapp*.prg checkboxhand.obj

#include "boxclass.fh"
#include "inkey.fh"
#include "box.fh"
#include "button.fh"
#include "set.fh"
#include "FSerrmsg.h"
#include "getexit.fh"

/* --------------------------------------------------------------------------
 * Default CheckBox input handler
 * This function should be used for oCheckBox:Show() only
 *
 * Parameter passed from oCheckBox:Show() via oCheckBox:Handler block,
 * i.e. here oCheckBox:Handler := {|obj| CheckBoxHandler(obj) }
 *    obj = current object
 * Returns current obj:Buffer status or NIL on error
 */
function CheckBoxHandler(obj)
   local key, iHitTest  as intvar
   local cAppType       as char
   local inGet          as logic
   local lRigthButt := .F., lMidButt := .F. as logic

   if valtype(obj) != "O" .or. !IsObjClass(obj, "_CheckBox") 
      _Rt_Error(RTARGTYP, "CheckBoxHandler(): wrong object class - " + ;
         if(valtype(obj) != "O", "not object", IsObjClass(obj)[1]) )
      return NIL 
   endif

// cAppType := AppIoMode()   // i/o mode: G,T,B
// obj:FilterKeys := .F.

   obj:SetFocus()
   obj:Display()

   inGet := valtype(obj:exitState) == "N" .and. obj:exitState >= GE_NOTSET

   while .T.
      if valtype(obj:Keypress) == "N" .and. obj:Keypress != 0
         key := obj:Keypress
         obj:Keypress := NIL
      else
         key := InkeyTrap(0, INKEY_ALL_BUT_MOVE) // same as Inkey(0) but considers SET KEY
      endif
      
      do case
      case  key == K_ESC                  //--> Esc 
         obj:exitState := GE_ESCAPE       //    exit READ
         exit
         
      case  key == K_ENTER       .or. ;   //--> Return 
            key == K_CTRL_ENTER           
         obj:exitState := GE_ENTER        //    next GET
         exit
         
      case  key == K_DOWN         .or. ;  //--> cursor down 
            key == K_RIGHT        .or. ;  //--> cursor right 
            key == K_CTRL_DOWN    .or. ;  //--> Ctrl-Cursor-Down
            key == K_CTRL_RIGHT   .or. ;  //--> Ctrl-Cursor-Right
            key == K_TAB                  //--> Tab 
         obj:exitState := GE_DOWN         //    next GET item 
         exit
         
      case  key == K_UP           .or. ;  //--> cursor up 
            key == K_LEFT         .or. ;  //--> cursor left 
            key == K_CTRL_UP      .or. ;  //--> Ctrl-Cursor-Up
            key == K_CTRL_LEFT    .or. ;  //--> Ctrl-Cursor-Left
            key == K_SH_TAB               //--> Shift-Tab 
         obj:exitState := GE_UP           //    prev GET item 
         exit
         
      case  key == K_HOME         .or. ;  //--> Home 
            key == K_CTRL_HOME            //--> Ctrl-Home 
         obj:exitState := GE_TOP          //    first GET in list
         exit

      case  key == K_END          .or. ;  //--> End 
            key == K_CTRL_END             //--> Ctrl-End 
         obj:exitState := GE_BOTTOM       //    last GET in list 
         exit

      case  key == K_PGUP         .or. ;  //--> PageUp 
            key == K_PGDN                 //--> PageDown 
         obj:exitState := GE_WRITE        //    exit GET
         exit

      /* MidMouseUp and RightMouseUp (if allowed) will toggle 
       * item in non-exclusive RadioGroup
       */
      case key == K_MBUTTONUP .and. ; 
            ascan(m->_aGlobSetting[GSET_G_A_RADGRP_MOUSE], key) > 0
         lMidButt := .T.                  // considered in LeftMouseUp below
//       obj:Keypress := K_LBUTTONUP
         
      case key == K_RBUTTONUP .and. ; 
            ascan(m->_aGlobSetting[GSET_G_A_RADGRP_MOUSE], key) > 0
         lRigthButt := .T.                // considered in LeftMouseUp below
//       obj:Keypress := K_LBUTTONUP
         
      case key >= K_LBUTTONUP .and. ;         
           ascan(m->_aGlobSetting[GSET_G_A_CHBOX_MOUSE], key) > 0
//       iHitTest := obj:HitTest(Mrow(.T.), MCol(.T.), .T.)
         iHitTest := obj:HitTest(MrowClick(.T.), MColClick(.T.), .T.)
         if iHitTest == 0              //  - mouse click outside curr,
            obj:exitState := GE_MOUSE  //    GetReader() will search 
            exit                       //    for the corresp. GET field
         else
            if !(lMidButt .or. lRigthButt) .and. ;
                  _aGlobSetting[GSET_G_L_CHBOX_SINGLE]
               obj:exitState := GE_ENTER     // simulate Enter = next GET
               exit
            endif
         endif
         lMidButt := lRigthButt := .F.    // MMB, RMB: stay in CheckBox
               
	   case  key > 0 .and. key < 255 .and. chr(key) $ " xX"
	      obj:Select()                     // toggle on/off
         exit
	   case  key > 0 .and. key < 255 .and. chr(key) $ "+yYtT"
	      obj:Select(.T.)                  // set ON
         exit
	   case  key > 0 .and. key < 255 .and. chr(key) $ "-nNfF"
	      obj:Select(.F.)                  // set OFF
         exit
         
	   endcase
   enddo

   if !inGet  
      obj:exitState := GE_NOTUSED
   endif
   
   obj:KillFocus()
   return obj:Buffer
   
/* eof */
