#ifdef COMMENT_ONLY
/* For FlagShip 6. Copyright (c) 1994..2004 by multisoft GmbH, Munich, Germany
 * You may modify this source according to your needs and distribute the
 * object code freely with your executable. You may distribute also the
 * modified source code when your application is distributed in source,
 * but may not remove the copyright notice.
 *
 * User modifiable functions used in NT2/CA3 but dependent on hardware
 * or not allowed for a normal User on a Unix or MS-Windows System.
 * These functions cannot be implemented because they require capabilities
 * of an IBM-PC or violate Unix/Windows protection mechanisms.
 *
 * If you need to change this file, copy it into your local directory,
 * make changes and recompile it by using
 *    FlagShip fs2_udfs.prg -c -m -na
 * and then link it with your application
 *    FlagShip myprog*.prg fs2_udfs.o
 *
 * Release $Date: 2005/09/27 16:02:42 $
 */
#endif

#Cinline
#include "FSopenc.h"
#endCinline

#define FGSLINUXELF

#include "set.fh"
#include "fileio.fh"

static MSG_DOS    := "DOS dependant hardware function, returning "
static MSG_MEMORY := "Memory manipulation not possible, returning "
static MSG_PORT   := "Port manipulation not possible, returning "
static MSG_KEYB   := "Keyboard manipulation not possible, returning "
static MSG_HARDW  := "Hardware manipulation not possible, returning "
static MSG_DATE   := "Date/Time can only be set by su/root, returning "
static MSG_PRINT  := "Printer is usually managed by lpr spooler.;" + ;
                     "To set a com/lpt port, use com_*() functions. Returning "
static MSG_NOVEL  := "There is usually no connection to Novell;" + ;
                     "except via NFS (or IPX on some Unix systems),;" + ;
                     "nor are the Novell APIs available here. Returning "

static _setTic := 0.0

*****************************************************************************
* User modifiable Date Functions
*  Note: you may change the array element strings below to fit
*        to your local language
*****************************************************************************
static aDayOfWeek :=  {"Sunday",    "Monday" , "Tuesday", ;
                       "Wednesday", "Thursday", ;
                       "Friday",    "Saturday" } as array

static aMonths := {"January", "February", "March", ;
                   "April",   "May",      "June", ;
                   "July",    "August",   "September", ;
                   "October", "November", "December"} as array

static fs2_ISOweek := .T. as logic

* ---------------------------------------------------------------------------
* CDow2([dDate]) -> cDayOfWeek
*  converts date to char day of week (or ""). Similar to Cdow()
function CDow2(dDate)
   if dDate == NIL
      dDate := date()
   elseif !DateValid(dDate)
      return ""
   endif
   return aDayOfWeek[dow(dDate)]

* ---------------------------------------------------------------------------
* CMonth2([dDate]) -> cMonth
*  converts date to char month (or ""). Similar to Cmonth()
function CMonth2(dDate)
   if dDate == NIL
      dDate := date()
   elseif !DateValid(dDate)
      return ""
   endif
   return aMonths[month(dDate)]

* ---------------------------------------------------------------------------
* CtoDow(cDay) -> nDay
*  Character (also abbreviated) day of week to number 1..7 (or 0)
function CtoDow(cDay)
   local cDayLoc as char

   if empty(cDay) .or. valtype(cDay) != "C"
      return 0
   endif
   cDayLoc := lower(alltrim(cDay))
   return ascan(aDayOfWeek, {|x| lower(left(x, len(cDayLoc))) == cDayLoc})

* ---------------------------------------------------------------------------
* CtoMonth(cMonth) -> nMonth
*  Character (also abbreviated) month to number 1..12 (or 0)
function CtoMonth(cMonth)
   local cMonthLoc as char

   if empty(cMonth) .or. valtype(cMonth) != "C"
      return 0
   endif
   cMonthLoc := lower(alltrim(cMonth))
   return ascan(aMonths, {|x| lower(left(x, len(cMonthLoc))) == cMonthLoc})

* ---------------------------------------------------------------------------
* DMY([date],[lPeriod],[lFsSet]) -> cDate in format "DD[.] Month [YY]YY"
*
function DMY(p1, p2, p3)
   local retVal := "" as char
   local cYear  := "" as char
   local dDate, lPeriod, lFsSet as usual

   if valtype(p1) == "L"   // parameter shift
      dDate   := date()
      lPeriod := p1
      lFsSet  := p2
   else
      dDate   := p1
      lPeriod := p2
      lFsSet  := p3
   endif
   if dDate == NIL
      dDate := date()
   elseif !DateValid(dDate)
      return ""
   endif
   if valtype(lFsSet) != "L"
      lFsSet := .F.
   endif
   retVal := ltrim(day(dDate))
   if valtype(lPeriod) == "L" .and. lPeriod
      retVal += "."
   endif
   cYear := ltrim(year(dDate))
   if !set(_SET_CENTURY)
      cYear := right(cYear,2)
   endif
   retVal += " " + if(lFsSet, cmonth(dDate), Cmonth2(dDate)) + ;
             " " + cYear
   return retVal

* ---------------------------------------------------------------------------
* MDY([date],[lFsSet]) -> cDate in format "Month DD, [YY]YY"
*
function MDY(p1, p2)
   local cYear  := "" as char
   local dDate, lFsSet as usual

   if valtype(p1) == "L"   // parameter shift
      dDate   := date()
      lFsSet  := p1
   else
      dDate   := p1
      lFsSet  := p2
   endif
   if dDate == NIL
      dDate := date()
   elseif !DateValid(dDate)
      return ""
   endif
   if valtype(lFsSet) != "L"
      lFsSet := .F.
   endif
   cYear := ltrim(year(dDate))
   if !set(_SET_CENTURY)
      cYear := right(cYear,2)
   endif
   return if(lFsSet, cmonth(dDate), Cmonth2(dDate)) + " " + ;
          ltrim(day(dDate)) + ", " + cYear

* ---------------------------------------------------------------------------
* NtoCdow([nDay]) --> cDay
*  Number (1..7) to char day of week (or "")
function NtoCdow(nDay)
   if nDay == NIL
      nDay := dow(date())
   elseif valtype(nDay) != "N" .or. nDay < 1 .or. nDay > len(aDayOfWeek)
      return ""
   endif
   return aDayOfWeek[nDay]

* ---------------------------------------------------------------------------
* NtoCmonth([nMonth]) --> cMonth
*  Number (1..12) to char month (or "")
function NtoCmonth(nMonth)
   if nMonth == NIL
      nMonth := month(date())
   elseif valtype(nMonth) != "N" .or. nMonth < 1 .or. nMonth > len(aMonths)
      return ""
   endif
   return aMonths[nMonth]

* ---------------------------------------------------------------------------
* week([date], [ISOconf]) --> nWeek
*   ISOconf = .F. 01/01/year is always week 1 and 12/31/year is 52 or 53
*           = .T. (default) ISO 8601 conformant handling 
* Note: you may set the default by WeekSetIso(ISOconf)
* ISO-8601 says:
*   "A week is a seven day period within a calendar year, starting on a
*   Monday and identified by its ordinal number within the year; the first
*   calendar week of the year is the one that includes the first Thursday
*   of that year. In the Gregorian calendar, this is equivalent to the week
*   which includes 4 January. First week of year should show Week with a
*   minimum of 4 days. Week 1 will then always include at least a Thursday,
*   Friday, Saturday and Sunday." Note that, by ISO standard, if January 1
*   is on a Friday, Saturday or Sunday, this short week is counted as part
*   of the last week of the previous year and is displayed by Week() as week
*   52 or 53. Similarly if January 1 is on a Tuesday, Wednesday or Thursday,
*   the last few days of the previous year are displayed as week 1.
*
function Week(dDate, [ISOconf])
   local iWeek, iDays, iDiff, iRest as intvar
   local dFirst as date
   local cDate := Set(_SET_DATEFORMAT) as char
   local iFix := 1   as intvar
   local lIso := .F. as logic
   local_int iDate1, iDate2, iAdd, iDOW

   if valtype(dDate) != "D"
      dDate := date()
   endif
   if empty(dDate) .or. !DateValid(dDate)
      return 0
   endif
   lIso := if(valtype(ISOconf) == "L", ISOconf, fs2_ISOweek)
   
   set date to american

   if lIso        // ISO conformant
      if dDate <= ctod("01/03/" + ltrim(year(dDate)) )      // date = year begin
         iDOW := dow(ctod("01/01/" + ltrim(year(dDate)) ))  // and is
         if (iDOW == 6 .or. iDOW == 7 .or. iDOW == 1)       // Fri, Sat, Sun:
            iDOW := dow(dDate)                              // week = last week
            if (iDOW == 6 .or. iDOW == 7 .or. iDOW == 1)    // of year-1
               dDate := ctod("12/31/" + ltrim(year(dDate) -1) )
            else
               iFix := 0
            endif
         endif   
      endif

      if dDate >= ctod("12/29/" + ltrim(year(dDate)) )        // date = year-end
         iDOW := dow(ctod("01/01/" + ltrim(year(dDate) +1) )) // and is > Mon and
         if dow(dDate) >= 2 .and. ((iDow - dow(dDate)) > 0) ; // next 01/01/ is
            .and. (iDOW == 3 .or. iDOW == 4 .or. iDOW == 5)   // Tue, Wed, Thu:
            Set(_SET_DATEFORMAT, cDate)                       // curr week = 1
            return 1
         endif
      endif   

      iDOW := dow(ctod("01/01/" + ltrim(year(dDate)) ) )  // current year starts
      if iDOW == 6 .or. iDOW == 7 .or. iDOW == 1          // at Fri, Sat, Sun:
         iFix := 0                                        // week 1 = second week
      endif
   endif
   dFirst := ctod("01/01/" + ltrim(year(dDate)) )         // date 01/01/year
#Cinline
{
//                 su mo tu mi th fr sa
   int iDiff[7] = {6, 0, 1, 2, 3, 4, 5};     /* week starts on Monday */
   int iDays;
   idate1 = (int)VAR_DATE(VAR_NAME_LOCAL(dfirst));
   idate2 = (int)VAR_DATE(VAR_NAME_LOCPAR(ddate));
   iDays  = idate1 % 7;        // 0:sunday, 1:monday... 6:saturday
   iadd   = iDiff[iDays];
}
#endCinline
   iDate1 -= iAdd    // day week 1 starts for 01/01/year

   iWeek := int((iDate2 - iDate1) / 7) + iFix
   Set(_SET_DATEFORMAT, cDate)
   return iWeek
   
* ---------------------------------------------------------------------------
* weekSetIso([ISOconf]) --> lPrevSetting
function weekSetIso([ISOconf])
   local retVal := fs2_ISOweek
   if valtype(ISOconf) == "L"
      fs2_ISOweek := ISOconf
   endif
   return retVal

* ---------------------------------------------------------------------------
* wom([date]) --> nWeek
function WoM(dDate)
   local cDate := Set(_SET_DATEFORMAT) as char
   local iWeek as intvar
   
   if valtype(dDate) != "D"
      dDate := date()
   endif
   if empty(dDate) .or. !DateValid(dDate)
      return 0
   endif
   
   set date american
   iWeek := week(dDate, .F.) - ;
            week(ctod(strzero(month(dDate),2) + "/01/" + ;
                      ltrim(year(dDate)) ), .F.) + 1
   Set(_SET_DATEFORMAT, cDate)
   return iWeek
          

*****************************************************************************
* User-modifiable functions to access video-functions and / or videomodes
*****************************************************************************
* ---------------------------------------------------------------------------
* CGA40() -> .F.
*     Extended Driver
*     Switches to 40-column mode (color or monochrome)
function CGA40
   return .F.

* ---------------------------------------------------------------------------
* CGA80() -> .F.
*     Extended Driver
*     Switches to 80-column mode (color or monochrome)
function CGA80
   return .F.

* ---------------------------------------------------------------------------
* CHARPIX() -> 1
*     Video Functions
*     Returns the number of pixel lines per character
function CHARPIX()
   return 1

* ---------------------------------------------------------------------------
* CHARWIN() -> ""
*     Video Functions
*     Exchanges particular characters in a screen area
function CHARWIN()
   return ""

* ---------------------------------------------------------------------------
* CLEARSLOW() -> ""
*     Video Functions
*     Deletes a screen area from the outside in with a delay
function CLEARSLOW()
   return ""

* ---------------------------------------------------------------------------
* COLORREPL() -> ""
*     Video Functions
*     Exchanges particular screen attributes
function COLORREPL()
   return ""

* ---------------------------------------------------------------------------
* COLORWIN() -> ""
*     Video Functions
*     Exchanges particular attributes in a screen area
function COLORWIN()
   return ""

* ---------------------------------------------------------------------------
* DSETKBIOS() -> .T.
*     Extended Driver
*     Turns the extended keyboard mode on or off through BIOS
*     (additional keys F11,  F12)
function DSETKBIOS()
   return .T.

* ---------------------------------------------------------------------------
* DSETNOLINE() -> .F.
*     Extended Driver
*     Ignores the next line feed sent to the screen
function DSETNOLINE()
   return .F.

* ---------------------------------------------------------------------------
* DSETQFILE() -> .F.
*     Extended Driver
*     Creates a protocol file when the program ends normally
function DSETQFILE()
   return .F.

* ---------------------------------------------------------------------------
* DSETWINDEB() -> NIL
*     Extended Driver
*     No longer supported because by CA-Clipper 5.x
function DSETWINDEB()
   return NIL

* ---------------------------------------------------------------------------
* DSETWINDOW() -> .F.
*     Extended Driver
*     Reroutes external functions and programs to a window
function DSETWINDOW()
   return .F.

* ---------------------------------------------------------------------------
* EGA43() -> .F.
*     Extended Driver
*     Switches to the 43-line EGA mode
function EGA43()
   return .F.

* ---------------------------------------------------------------------------
* EGAPALETTE() -> ""
*     Video Functions
*     Changes EGA palette colors
function EGAPALETTE()
   return ""

* ---------------------------------------------------------------------------
* FIRSTCOL() -> 0
*     Extended Driver
*     Sets the first visible column of a virtual screen
function FIRSTCOL()
   return 0

* ---------------------------------------------------------------------------
* FIRSTROW() -> 0
*     Extended Driver
*     Sets the first visible line of a virtual screen
function FIRSTROW()
   return 0

* ---------------------------------------------------------------------------
* FONTLOAD() -> -2
*     Video Functions
*     Loads EGA/VGA fonts from another file
function FONTLOAD()
   return -2

* ---------------------------------------------------------------------------
* FONTRESET() -> .F.
*     Video Functions
*     Resets all font and palette changes to the ROM defaults
function FONTRESET()
   return .F.

* ---------------------------------------------------------------------------
* FONTROTATE() -> ""
*     Video Functions
*     Rotates and mirrors images within a font string
function FONTROTATE()
   return ""

* ---------------------------------------------------------------------------
* FONTSELECT() -> -2
*     Video Functions
*     Determines font areas for normal- and high-intensity output
function FONTSELECT()
   return -2

* ---------------------------------------------------------------------------
* GETBOXGROW() -> 0
*     Extended Driver
*     Gets the time delay with which boxes are opened
function GETBOXGROW()
   return 0

* ---------------------------------------------------------------------------
* GETCURSOR() -> 1
*     Extended Driver
*     Determines the setting for the cursor form = how many scan lines
*     are used to display the cursor
function GETCURSOR()
   return 1

* ---------------------------------------------------------------------------
* GETFONT() -> ""
*     Video Functions
*     Queries the current font
function GETFONT()
   return ""

* ---------------------------------------------------------------------------
* GETKXLAT() -> 0
*     Extended Driver
*     Determines the current key code table
function GETKXLAT()
   return 0

* ---------------------------------------------------------------------------
* GETKXTAB() -> ""
*     Extended Driver
*     Retrieves the entire key code table
function GETKXTAB()
   return ""

* ---------------------------------------------------------------------------
* GETLINES() -> 0
*     Extended Driver
*     Determines the number of lines after which the screen display pauses
function GETLINES()
   return 0

* ---------------------------------------------------------------------------
* GETMODE() -> "CGA80"
*     Extended Driver
*     Uses the current screen mode as a function name
function GETMODE()
   return "CGA80"

* ---------------------------------------------------------------------------
* GETPAGE() -> 1
*     Extended Driver
*     Determines the current screen page
function GETPAGE()
   return 1

* ---------------------------------------------------------------------------
* GETPBIOS() -> 0
*     Extended Driver
*     Determines if printing is through DOS or the BIOS
function GETPBIOS()
   return 0

* ---------------------------------------------------------------------------
* GETPXLAT() -> ""
*     Extended Driver
*     Retrieves the current printer table
function GETPXLAT()
   return ""

* ---------------------------------------------------------------------------
* GETSCRMODE() -> 3
*     Extended Driver
*     Determines the number of the active video mode
function GETSCRMODE()
   return 3

* ---------------------------------------------------------------------------
* GETSCRSTR() -> ""
*     Video Functions
*     Queries screen output that was redirected by SETSCRSTR()
function GETSCRSTR()
   return ""

* ---------------------------------------------------------------------------
* GETTAB() -> 7
*     Extended Driver
*     Retrieves tab values for screen output
function GETTAB()
   return 7

* ---------------------------------------------------------------------------
* GETVGAPAL() -> -1
*     Video Functions
*     Determines the palette settings on a VGA card
function GETVGAPAL()
   return -1

* ---------------------------------------------------------------------------
* INPUTMODE() -> 0
*     Extended Driver
*     Determines the previously active or the currently active input mode
function INPUTMODE()
   return 0

* ---------------------------------------------------------------------------
* INVERTATTR() -> 0
*     Video Functions
*     Inverts the foreground and background of an attribute
function INVERTATTR()
   return 0

* ---------------------------------------------------------------------------
* INVERTWIN() -> ""
*     Video Functions
*     Inverts all attributes in an area of the screen
function INVERTWIN()
   return ""

* ---------------------------------------------------------------------------
* ISCGA() -> .T.
*     Video Functions
*     Tests for presence of a CGA card or if one can be emulated
function ISCGA()
   return .T.     // Note: return iscolor() may make sense

* ---------------------------------------------------------------------------
* ISEGA() -> .T.
*     Video Functions
*     Determines if an EGA card is present or can be emulated
function ISEGA()
   return .T.     // Note: return iscolor() may make sense

* ---------------------------------------------------------------------------
* ISHERCULES() -> .T.
*     Video Functions
*     Determines if a HERCULES card is present or can be emulated
function ISHERCULES()
   return .T.     // Note: return !iscolor() may make sense

* ---------------------------------------------------------------------------
* ISMCGA() -> .T.
*     Video Functions
*     Determines if an MCGA card is present or can be emulated
function ISMCGA()    // Note: return iscolor() may make sense
   return .T.

* ---------------------------------------------------------------------------
* ISMONO() -> .T.
*     Video Functions
*     Determines if a monochrome card is present or can be emulated
function ISMONO()
   return .T.     // Note: return !iscolor() may make sense

* ---------------------------------------------------------------------------
* ISPGA() -> .T.
*     Video Functions
*     Determines if a PGA card is present or can be emulated
function ISPGA()
   return .T.     // Note: return iscolor() may make sense

* ---------------------------------------------------------------------------
* ISVGA() -> .T.
*     Video Functions
*     Determines if a VGA card is present
function ISVGA()
   return .T.     // Note: return iscolor() may make sense

* ---------------------------------------------------------------------------
*  KEYREAD() -> ""
*     Extended Driver
*     Reads already processed CA-Clipper keyboard buffer input
function  KEYREAD()
   return ""

* ---------------------------------------------------------------------------
* MAXFONT() -> 1
*     Video Functions
*     Determines the number of available fonts
function MAXFONT()
   return 1

* ---------------------------------------------------------------------------
* MAXPAGE() -> 0
*     Extended Driver
*     Determines the number of available screen pages
function MAXPAGE()
   return 0

* ---------------------------------------------------------------------------
* MONISWITCH() -> .F.
*     Video Functions
*     Switches between monochrome and color screen
function MONISWITCH()
   return .F.

* ---------------------------------------------------------------------------
* MONOCHROME() -> .F.
*     Extended Driver
*     Switches to the monochrome mode
function MONOCHROME()
   return .F.

* ---------------------------------------------------------------------------
* PAGECOPY() -> .F.
*     Extended Driver
*     Copies one screen page to another
function PAGECOPY()
   return .F.

* ---------------------------------------------------------------------------
* PRINTERROR() -> 0
*     Extended Driver
*     Returns the error code for the last printer output
function PRINTERROR()
   return 0

* ---------------------------------------------------------------------------
* RESTCURSOR() -> ""
*     Video Functions
*     Restores a saved cursor position and form
function RESTCURSOR()
   return ""

* ---------------------------------------------------------------------------
* SAVECURSOR() -> 0
*     Video Functions
*     Saves current cursor position and form
function SAVECURSOR()
   return 0

* ---------------------------------------------------------------------------
* SAYDOWN() -> ""
*     Video Functions
*     Displays screen output downward and vertically
function SAYDOWN()
   return ""

* ---------------------------------------------------------------------------
* SAYMOVEIN() -> ""
*     Video Functions
*     Displays screen output with a "move in" effect
function SAYMOVEIN()
   return ""

* ---------------------------------------------------------------------------
* SAYSCREEN() -> ""
*     Video Functions
*     Output to the screen without changing the attribute
* Note: simulated by standard output
function SAYSCREEN(cTxt, nRow, nCol)
   nRow := if(valtype(nRow) == "N", nRow, row())
   nCol := if(valtype(nCol) == "N", nCol, col())
   @ nRow, nCol SAY cTxt
   return ""

* ---------------------------------------------------------------------------
* SAYSPREAD() -> ""
*     Video Functions
*     Displays screen output with "spread" effect
* Note: simulated by standard output
function SAYSPREAD(cTxt, nDelay, nRow, nCol)
   nRow := if(valtype(nRow) == "N", nRow, row())
   nCol := if(valtype(nCol) == "N", nCol, col())
   @ nRow, nCol SAY cTxt
   return ""

* ---------------------------------------------------------------------------
* SCREENATTR() -> 0
*     Video Functions
*     Determines the attribute at a particular position
function SCREENATTR()
   return 0

* ---------------------------------------------------------------------------
* SCREENMARK() -> .F.
*     Video Functions
*     Searches for a string on the screen and marks it with an attribute
function SCREENMARK()
   return .F.

* ---------------------------------------------------------------------------
* SCREENMIX() -> ""
*     Video Functions
*     Mixes characters and attributes of a screen
function SCREENMIX()
   return ""

* ---------------------------------------------------------------------------
* SCREENSTR() -> ""
*     Video Functions
*     Reads a string, including attributes, from the screen
function SCREENSTR()
   return ""

* ---------------------------------------------------------------------------
* SETBELL() -> ""
*     Extended Driver
*     Sets the tone frequency and duration for CHR(7) or "beep"
function SETBELL()
   return ""

* ---------------------------------------------------------------------------
* SETBOXGROW() -> ""
*     Extended Driver
*     Opens boxes with a time delay
function SETBOXGROW()
   return ""

* ---------------------------------------------------------------------------
* SETFONT() -> -1
*     Video Functions
*     Loads the font directly out of a string
* Note: Use SET FONT command instead
function SETFONT()
   return -1

* ---------------------------------------------------------------------------
* SETKXLAT() -> .F.
*     Extended Driver
*     Redefines key codes or lock keys
function SETKXLAT()
   return .F.

* ---------------------------------------------------------------------------
* SETKXTAB() -> .F.
*     Extended Driver
*     Installs key tables
function SETKXTAB()
   return .F.

* ---------------------------------------------------------------------------
* SETLINES() -> ""
*     Extended Driver
*     Determines the number of lines after which the screen display pauses
function SETLINES()
   return ""

* ---------------------------------------------------------------------------
* SETMAXCOL() -> .F.
*     Extended Driver
*     Sets the number of columns for a virtual screen
function SETMAXCOL()
   return .F.

* ---------------------------------------------------------------------------
* SETMAXROW() -> .F.
*     Extended Driver
*     Sets the number of lines for a virtual screen
function SETMAXROW()
   return .F.

* ---------------------------------------------------------------------------
* SETPAGE() -> .F.
*     Extended Driver
*     Selects a new screen page
function SETPAGE()
   return .F.

* ---------------------------------------------------------------------------
* SETPBIOS() -> .T.
*     Extended Driver
*     Redirects print output to the BIOS or DOS, and establishes the timeout
function SETPBIOS()
   return .T.

* ---------------------------------------------------------------------------
* SETPXLAT() -> .F.
*     Extended Driver
*     Establishes translation tables for printer output
function SETPXLAT()
   return .F.

* ---------------------------------------------------------------------------
* SETQNAME() -> .F.
*     Extended Driver
*     Changes the file and path name for the QUIT file
function SETQNAME()
   return .F.

* ---------------------------------------------------------------------------
* SETSCRMODE() -> .F.
*     Extended Driver
*     Establishes a new video mode
function SETSCRMODE()
   return .F.

* ---------------------------------------------------------------------------
* SETSCRSTR() -> .F.
*     Video Functions
*     Redirects screen output into a string
function SETSCRSTR()
   return .F.

* ---------------------------------------------------------------------------
* SETTAB() -> .F.
*     Extended Driver
*     Sets the tab widths for screen outputs
function SETTAB()
   return .F.

* ---------------------------------------------------------------------------
* STRSCREEN() -> ""
*     Video Functions
*     Displays a string with characters and attributes on the screen
* Note: simulated by standard output w/o attributes
function STRSCREEN(cTxtAttr, nRow, nCol)
   nRow := if(valtype(nRow) == "N", nRow, row())
   nCol := if(valtype(nCol) == "N", nCol, col())
   @ nRow, nCol SAY CharOdd(cTxtAttr)
   return ""

* ---------------------------------------------------------------------------
* TRAPINPUT() -> .F.
*     Extended Driver
*     Allows supervision of input commands
function TRAPINPUT()
   return .F.

* ---------------------------------------------------------------------------
* TRAPSHIFT() -> .F.
*     Extended Driver
*     Calls a procedure that depends on switching keys
function TRAPSHIFT()
   return .F.

* ---------------------------------------------------------------------------
* UNTEXTWIN() -> ""
*     Video Functions
*     Replaces an area of characters from a region of the screen
function UNTEXTWIN()
   return ""

* ---------------------------------------------------------------------------
* VGA28() -> .F.
*     Extended Driver
*     Switches to 28-line VGA mode
function VGA28()
   return .F.

* ---------------------------------------------------------------------------
* VGA50() -> .F.
*     Extended Driver
*     Switches to 50-line VGA mode
function VGA50()
   return .F.

* ---------------------------------------------------------------------------
* VGAPALETTE() -> .F.
*     Video Functions
*     Changes VGA palette colors
function VGAPALETTE()
   return .F.

* ---------------------------------------------------------------------------
* VIDEOINIT() -> NIL
*     Video Functions
*     Reinitializes a video system after a RUN
function VIDEOINIT()
   REFRESH
   return NIL

* ---------------------------------------------------------------------------
* VIDEOSETUP() -> 0
*     Video Functions
*     Queries video mode at system start
function VIDEOSETUP()
   return 0

* ---------------------------------------------------------------------------
* VIDEOTYPE() -> 2
*     Video Functions
*     Returns bit-coded information about available video modes == CGA
function VIDEOTYPE()
   return 2


**************************************************
* Functions for extended GET
**************************************************
* ---------------------------------------------------------------------------
* GetInput(cDefault,[nRow],[nColumn],[lSAY],[cPrompt],[cPict]) -> cInput
*     a simple GET/READ
function GetInput(cVal,nRow,nCol,lSAY,cPrompt,cPict)
   local GetList := {}
   nRow := if(valtype(nRow)=="N", nRow, row())
   nCol := if(valtype(nCol)=="N", nCol, col())
   if valtype(cPrompt) == "C"
      @ nRow, nCol SAY cPrompt
      nCol := col() +1
   endif
   if !empty(cPict)
      @ nRow, nCol GET cVal PICTURE (cPict)
   else
      @ nRow, nCol GET cVal
   endif
   READ
   if valtype(lSay) == "L" .and. lSay
      @ nRow, nCol SAY cVal
   endif
   return cVal

* ---------------------------------------------------------------------------
*  getsecret
*     input like GET/READ with picture "@P"
function GetSecret(cVal,nRow,nCol,lSAY,cPrompt,cPict)
   return GetInput(cVal,nRow,nCol,lSAY,cPrompt,cPict, "@P")


**********************************************************
* Functions to access memory, I/O Ports, Keyboard and BIOS
**********************************************************

* ---------------------------------------------------------------------------
* biosdate
*     No permission to access hardware directly
function biosadate()
   fs2_devWarn(MSG_DOS + "empty date")
   return ctod(" ")

* ---------------------------------------------------------------------------
* bootcold
* bootwarm
*     CT3: Triggers a cold/warm boot
*     FS2: would be fatal in multi-user and multi-tasking system
function bootcold()
   alert("BootCold();Ingoring cold boot request;but performing QUIT")
   quit
   return NIL
function bootwarm()
   alert("BootWarm();Ingoring warm boot request;but performing QUIT")
   quit
   return NIL

* ---------------------------------------------------------------------------
* cputype
*     CT3: returns CPU types 8088 to 80486 only <g>
function cputype()
   return ""

* ---------------------------------------------------------------------------
* erroract() -> 0
*     CT3: Recommends an action for a DOS error
*     FS2: ignored
function erroract()
   return 0

* ---------------------------------------------------------------------------
* errorbase() -> errno
*     CT3: DOS error codes 1..12
*     FS2: errno system variable, see /usr/include/[asm/]errno.h
function errorbase()
   return ferror()

* ---------------------------------------------------------------------------
* errorcode() -> errno
*     CT3: DOS 3.1+ error codes
*     FS2: errno system variable, see /usr/include/[asm/]errno.h
function errorcode()
   return ferror()

* ---------------------------------------------------------------------------
* errororg() -> 1
*     CT3: DOS 3.1+ error codes 1..5
*     FS2: n/a, returns 1 == unknown
function errororg()
   return 1

* ---------------------------------------------------------------------------
* filesfree([pos]) -> num
*     CT3: number of available handles
*     FS2: system dependant, here for Linux and BSD/SUN systems, and is
*     similar to "cat /proc/sys/fs/file-nr" which display 3 numbers:
*     "allocated \t free \t maximum \n" file handles
* Note: some Unix systems and kernels handle it dynamically, assigning a new
*     pool of file descriptors to "allocated"; the "free" number will be
*     auto-increased in such a case too; so the returned "free files" value
*     is not 100% useful for these kernels...
function filesfree(pos)
   local_int retVal := 0
   local_int iMode := if(valtype(pos) == "N", pos, 2)
#Cinline
   {
      int iAlloc, iFree, iMax;
      FILE *fi;

      fi = fopen("/proc/sys/fs/file-nr", "r");
      if(fi) {
         fscanf(fi, "%d\t%d\t%d\n", &iAlloc, &iFree, &iMax);
         switch(imode) {
            case 1:  retval = iAlloc; break;
            case 3:  retval = iMax;   break;
            default: retval = iFree;  break;
         }
         fclose(fi);
      }
   }
#endCinline
   return retVal

* ---------------------------------------------------------------------------
* filesmax() -> num
*     CT3: number of available handles
*     FS2: system dependant, here for Linux and BSD/SUN systems
*     similar to "cat /proc/sys/fs/file-max" which usually represents
*     system-wide setting.
*  Note: you may increase the system-wide file count by e.g.
*     echo 8000 > /proc/sys/fs/file-max
function filesmax()
   return filesfree(3)

* ---------------------------------------------------------------------------
* getcountry() -> nCountry
*     CT3: Queries country setting for the operating system - DOS 3.1+
*     FS2: Unix use the LANG environment variable in syntax "xx_XX[.nnn]"
*     so there is a small conversion table here (expand for your needs)
function getcountry()
   local cCountry := getEnv("LANG")
   do case
   case cCountry == "C" .or. cCountry == "POSIX"
      return 1
   case left(cCountry,5) == "en_US"
      return 1
   case left(cCountry,5) == "en_GB"
      return 44
   case left(cCountry,2) == "en"
      return 61
   case left(cCountry,2) == "fr"
      return 33
   case left(cCountry,2) == "de"
      return 49
   case left(cCountry,2) == "pt"
      return 351
   endcase
   return 0    // unknown

* ---------------------------------------------------------------------------
* inbyte                         No permission to manipulate I/O Ports
function inbyte(p1)
   fs2_devWarn(MSG_PORT + "-1;Use com_*() functions instead")
   return -1

// Note: the handling of i/o ports in Unix is similar to handling usual
// files. The INBYTE() function can be read:
//
// local fd := Fopen("/dev/cua0", 0), cChar := " "
// if fd == 0
//   return -1
// endif
// Fread(fd, @cChar, 1)
// Fclose(fd)
// return asc(cChar)
//
// where the /dev/cua0 is a device for 1st serial port, /dev/lp0 for
// the parallel port etc. There are three handles always open: 1=stdin,
// 2=stdout and 3=stderr, so you don't need Fopen/Fclose them.
// As you may see, the above algorithm is VERY inefficient, becouse of
// the open/read/close for every byte.  It is better to Fopen() the device
// once, Fread() as long as you want or until Ferror() etc, and then
// close the device by Fclose().

* ---------------------------------------------------------------------------
* inword                         No permission to manipulate I/O Ports
function inword(p1)
   fs2_devWarn(MSG_PORT + "-1;Use com_*() functions instead")
   return -1

* ---------------------------------------------------------------------------
* isansi() -> lAnsi
*     CT3: Tests to see if the DOS ANSI screen driver is installed
*     FS2: returns .T. in text mode, .F. otherwise
function isansi()
   return AppIoMode() == "T"

* ---------------------------------------------------------------------------
* isat() -> .T.
*     CT3: Determines if a program is running on an AT 80386,80486 <g> and up
*     FS2: not applicable, returns .T.
function isat()
   fs2_devWarn(MSG_HARDW + ".T.")
   return .T.

* ---------------------------------------------------------------------------
* ismath() -> .T.
*     CT3: Determines if a math coprocessor is installed <g>
*     FS2: returns .T.
function ismath()
   return .T.

* ---------------------------------------------------------------------------
*  KbdDisable(lEnable)
*     disables any input - ignored
function KbdDisable(lEnable)
   fs2_devWarn(MSG_HARDW + ".F.")
   return .F.

* ---------------------------------------------------------------------------
*  KbdEmulate(cInput) -> 0
*     CT3: Inserts characters into the BIOS keyboard buffer to emulate
*     keyboard input
*     The <cInput> contains one or more sequences of 2-bytes: kbd + scancode.
*     So we put only the odd chars into keyboard buffer here.
function KbdEmulate(cInput)
   if valtype(cInput) == "C" .and. len(cInput) > 0
      KEYBOARD CharOdd(cInput)
   endif
   return 0

* ---------------------------------------------------------------------------
*  KbdSpeed(nDelay, nFreq) -> .F.
function KbdSpeed(nDelay, nFreq)
   fs2_devWarn(MSG_KEYB + ".F.")
   return .F.

* ---------------------------------------------------------------------------
*  kbdstat             It is a terminal, its status is unknown to the host
function kbdstat
   fs2_devWarn(MSG_KEYB + "0")
   return 0

* ---------------------------------------------------------------------------
*  ksetcaps
*     It is a terminal, its status is unknown to the host
*  ksetnum
*     note that functions ksetins and ksetscroll are true functions:
*     ksetscroll() changes behaviour of cursor keys with windows
*     ksetsins()   toggles the insert mode
function ksetcaps(p1)
   fs2_devWarn(MSG_KEYB + ".F.")
   return .F.
function ksetnum(p1)
   fs2_devWarn(MSG_KEYB + ".F.")
   return .F.

* ---------------------------------------------------------------------------
*  kbdtype
*     It is a terminal, its delivered codes depend on the Terminfo
*     description, not on Hardware
function kbdtype
   fs2_devWarn(MSG_KEYB + "0")
   return 0     // == PC-Keyboard

* ---------------------------------------------------------------------------
* memsize([lMode]) -> nBytes
*     CT3: Determines size of conventional or extended memory
*     FS2: not applicable, returns fix 64mb
function memsize()
   return 64 * 1024 * 1024

* ---------------------------------------------------------------------------
* numbuffers() -> 0
*     CT3: Determines the BUFFERS= DOS/CONFIG.SYS setting
*     FS2: not applicable, returns fix 0
function numbuffers()
   return 0

* ---------------------------------------------------------------------------
* numfiles() -> filesmax()
*     CT3: Determines the FILES= DOS/CONFIG.SYS setting
*     FS2: not applicable, returns filesmax()
function numfiles()
   return filesmax()

* ---------------------------------------------------------------------------
*  numFkey() -> 40
*     number of function keys
function numFKEY()
   return 40

* ---------------------------------------------------------------------------
* osver() -> cDosVersion
*     CT3: Determines the DOS version number, e.g. "3.2" or "5.0" etc.
*     FS2: not applicable, returns fix "3.3"
function osver()
   fs2_devWarn(MSG_DOS + "3.3")
// return os()    // alternative
   return "3.3"

* ---------------------------------------------------------------------------
* outbyte                        No permission to manipulate I/O Ports
*   Note: similar to INBYTE(), but two chars are read
function outbyte(p1, p2)
// Note: the handling of i/o ports in Unix is similar to handling usual
// files. The OUTBYTE() function can be read:
//
// local fd := Fopen("/dev/cua0", 1)
// if fd == 0
//   return .F.
// endif
// Fwrite(fd, p2, 1)  // or Fwrite(fd, p2, len(p2)) for a string
// Fclose(fd)
// return .T.
//
// where the /dev/cua0 is a device for 1st serial port, /dev/lp0 for
// the parallel port etc. There are three handles always open: 1=stdin,
// 2=stdout and 3=stderr, so you don't need Fopen/Fclose them.
// As you may see, the above algorithm is VERY inefficient, becouse of
// the open/read/close for every byte/string.  It is better to Fopen() the
// device once, Fwrite() as long as you want or until Ferror() etc, and
// then close the device by Fclose().
   fs2_devWarn(MSG_PORT + ".F.;Use com_*() functions instead")
   return .F.

* ---------------------------------------------------------------------------
* outword                        No permission to manipulate I/O Ports
*     Note: similar to OUTBYTE(), but two chars are written
function outword(p1, p2)
   fs2_devWarn(MSG_PORT + ".F.;Use com_*() functions instead")
   return .F.

* ---------------------------------------------------------------------------
* pctype() -> nType
*    CT3: Returns the type of computer in use (XT, AT, Compaq etc)
*    FS2: not applicable, returns fix 0
function pctype()
   fs2_devWarn(MSG_DOS + "0")
   return 0

* ---------------------------------------------------------------------------
*  peekbyte                       No permission to manipulate Memory
function peekbyte(p1, p2)
   fs2_devWarn(MSG_MEMORY + "-1")
   return -1

* ---------------------------------------------------------------------------
*  peekstr                        No permission to manipulate Memory
function peekstr(p1, p2, p3)
   fs2_devWarn(MSG_MEMORY + "-1")
   return -1

* ---------------------------------------------------------------------------
*  peekword                       No permission to manipulate Memory
function peekword(p1, p2)
   fs2_devWarn(MSG_MEMORY + "-1")
   return -1

* ---------------------------------------------------------------------------
*  pokebyte                       No permission to manipulate Memory
function pokebyte(p1, p2, p3)
   fs2_devWarn(MSG_MEMORY + "-1")
   return -1

* ---------------------------------------------------------------------------
*  pokeword                       No permission to manipulate Memory
function pokeword(p1, p2, p3)
   fs2_devWarn(MSG_MEMORY + "-1")
   return -1

* ---------------------------------------------------------------------------
*  Scankey([lMode]) -> nVal
*     Queries scan code of DOS keyboard input
*     We emulate it here (roughly!) by Inkey()
function Scankey(lMode)
   fs2_devWarn(MSG_KEYB + "inkey()")
   return inkey(0)

* ---------------------------------------------------------------------------
*  Showkey(p1..p7) -> NIL
*     Continuously displays the INSERT and LOCK status
function Showkey()
   fs2_devWarn(MSG_KEYB + "NIL")
   return NIL

* ---------------------------------------------------------------------------
*  Sound(p1..p3) -> NIL
*     Creates tones (melodies) by designating frequency and duration
*     We simulate it by a simple beep
function Sound()
   local lSetCons  := Set(_SET_CONSOLE, .T.)
   local lSetPrint := Set(_SET_PRINTER, .F.)
   qqout(chr(7))
   Set(_SET_PRINTER,lSetPrint)
   Set(_SET_CONSOLE,lSetCons)
   return NIL

* ---------------------------------------------------------------------------
*  speed
*     CT3: A comparison value of the processor speed to 4.77 MHz PC <g>
*     This is not a real time Operating System, so there is no fixed speed
*     index. Maybe, a speed index based on seconds() or secondscpu() may
*     make sense
function speed(p1)
   // local ii, jj := 0, xx := secondscpu()
   // for ii = 1 to 100000
   //   jj = ii + 1
   // next
   // xx -= secondscpu()
   // return 45000 / xx   && the result value is at least compareable

   fs2_devWarn(MSG_HARDW + "10000")    // we at at least 100x faster
   return 10000

* ---------------------------------------------------------------------------
*  settick([lspeed]) -> .T.
*     starts the timer, value received by gettick()
function settick(ldummy)
   _setTic := Seconds()
   return .T.

* ---------------------------------------------------------------------------
*  gettick() -> nMillisec
*     number of Milliseconds since the last SetTick() call
function gettick()
   return (Seconds() - _setTic) * 1000.0

* ---------------------------------------------------------------------------
*  ssetbreak([lNew]) -> lBreak
*     CT3: Sets and checks the DOS BREAK switch
*     FS2: Not applicable, returns .F.
function ssetbreak()
   return .F.

* ---------------------------------------------------------------------------
*  ssetverify([lNew]) -> lBreak
*     CT3: Sets and checks the DOS VERIFY switch
*     FS2: Not applicable, returns .F.
function ssetverify()
   return .F.

* ---------------------------------------------------------------------------
*  stackfree() -> nByte
*     CD3: Determines the remaining stack space
*     Senseless in multi-user and multi-tasking system, we return fix 10mb
function stackfree()
   return 10 * 1024 * 1024

********************************************
* Functions to access printer directly     *
********************************************

* ---------------------------------------------------------------------------
*  printinit
*     No permission to manipulate directly the Printer
function printinit(p1, p2)
   fs2_devWarn(MSG_PRINT + "0")
   return 0            // return no error

* ---------------------------------------------------------------------------
*  printready
*     We normally write to file, the file is always ready
function printready(p1)
   fs2_devWarn(MSG_PRINT + ".T.")
   return .T.          // return printer is ready

* ---------------------------------------------------------------------------
*  printfile
*     not available in CA3. Prints specified file
function printfile(name, clearHigh)
   local buff, handle
   local currCons  := set(_SET_CONSOLE)
   local currPrint := set(_SET_PRINTER)
   clearHigh := if (valtype(clearHigh) == "L", clearHigh, .F.)
   ** if !clearHigh
   **    RUN ("lp " + name)    // upon request
   ** else ...
   handle = fopen(name)
   if handle <= 1
      return .F.
   endif
   set console off
   set printer on
   while ferror() == 0
      buff := freadtxt(handle)
      if clearHigh
#Cinline
      {
         FSuchar *str;
         str = VAR_CHR (VAR_NAME_LOCAL(buff));
         for (; *str; str++)
            (*str) &= 0x7F;
      }
#endCinline
      endif
      ? buff
   enddo
   fclose (handle)
   set(_SET_CONSOLE, currCons)
   set(_SET_PRINTER, currPrint)
   return .T.

* ---------------------------------------------------------------------------
*  printscr([p1],[p2],[p3])   print screen content
*
*  [p1]  = .T. print row number, .F. content only (default)
*  [p2]  = file name (or driver), if not given = std.printer output file
*  [p3]  = .T. line ends with CR/LF, .F. with LF only (default)
*
*  This is a "manual" solution for Terminal i/o mode. A general
*  Print-Screen for GUI and Terminal i/o is available via the
*  Printer object, simply invoke:
*
*     m->_oPrinter:Setup()
*     m->_oPrinter:ExecPrintScreen(.T.)
*
*  or accessible directly via Menu->File->PrintScreen (usually in GUI mode)
*  which then prints the current screen to selected printer.
*
*  Note: the characters mapped by FSchrmap.def on the screen may cause
*        other printer output, since no reverse mapping
*
function printscr(printRow, printFile, DOSline)
   LOCAL_INT ymax := MAXROW2()                      // used also in C
   LOCAL     chr_arr[ymax +1]                       // e.g. chr_arr[25]
   LOCAL     ii := 0, CrLf, dummy
   LOCAL     lSetPrint := SET(_SET_PRINTER)
   LOCAL     cSetPrint := SET(_SET_PRINTFILE)
   LOCAL     lSetCons  := SET(_SET_CONSOLE)
   LOCAL     nRow := ROW(), nCol := COL()

   printRow  := if (valtype(printRow ) == "L", printRow , .F.)
   printFile := if (valtype(printFile) == "C", printFile, "" )
   CrLf      := if (valtype(DOSline  ) == "L" .and. DOSline, ;
                    chr(13) + chr(10), chr(10) )
#Cinline
   {
#include <malloc.h>

#if !defined(A_CHARTEXT) && defined(FGSAIX)
#define A_CHARTEXT 0xffff
#endif

      extern variable *_bb_savescreen();                /* if not used yet */
      extern variable *_bb_maxcol2();                   /* if not used yet */

      variable *actscreen = VAR_NEW;
      WINDOW   *winPtr;
      int      x, y, xmax;
      unsigned char *oneline, *str;

      actscreen = _bb_savescreen(0,0);                    /* SAVE SCREEN   */
      winPtr    = VAR_SCR (actscreen);
      xmax      = VAR_INTNUM (fsFnNAME(maxcol2)(0,0) );   /* xmax=MAXCOL() */

      oneline = (FSuchar *) malloc((unsigned)((xmax + 1)
											  * sizeof(unsigned char)));
      if (oneline != NULL) {
         for (y=0; y <= ymax; y++) {
            str = oneline;
            for (x=0; x <= xmax; x++)
               *(str++) = (unsigned char)
                          (mvwinch(winPtr,y,x) & A_CHARTEXT);    /* curses */
            SET_VAR_CHRLEN (VAR_ARRELEM (VAR_NAME_LOCAL(chr_arr), y+1),
                            oneline, xmax +1);
         }
         free (oneline);
      }
   }
#endCinline

   SET (_SET_CONSOLE, .F.)                       // set console OFF
   if !empty(printFile)
      SET (_SET_PRINTFILE, printfile)            // set printer TO ..
   endif
   SET (_SET_PRINTER, .T.)                       // set printer ON

   ii := 0
   qqout (CrLf)
   if printRow
**        ---- PrintScreen mm/dd/yy hh:mm:ss ----+---------+---------+
**    00: Content of line 0 ...
**    01: Content of line 1 ...
**      : .....
**    24: Content of line 24...
**        ---------+---------+---------+---------+---------+---------+
      dummy := "---- PrintScreen " + dtoc(date()) + " " + time() + " ----+"
      qqout (SPACE(4) + dummy + REPLICATE("---------+", 4) + CrLf)
      AEVAL (chr_arr, {|x| QQOUT(STRZERO(ii++,2) + ":", x + CrLf)} )
      qqout (SPACE(4) + REPLICATE("---------+", 8) + CrLf)
   else
      AEVAL (chr_arr, {|x| QQOUT(x) + CrLf} )
   endif

   SET (_SET_CONSOLE, lSetCons)                  // set console <orig on/off>
   SET (_SET_PRINTER, lSetPrint)                 // set printer <orig on/off>
   if !empty(cSetPrint)
      SET(_SET_PRINTFILE, cSetPrint)             // set printer <orig file>
   endif
   SETPOS (nRow, nCol)                           // reset cursor

   RETURN NIL

* ---------------------------------------------------------------------------
*  printstat
*     printer status, simulated only for Unix
function printstat(nDev)
   local ret := 0
   if !file("/dev/lp" + str(nDev -1,1))
      ret := 8
   endif
   return ret

********************************************
* Functions to Access drives               *
********************************************

* ---------------------------------------------------------------------------
*  diskchange(cDrive) -> .F.
*     Changes the current disk drive
function DiskChange(cDrive)
   fs2_devWarn(MSG_DOS + ".F.")
   return .F.                    // signals error

* ---------------------------------------------------------------------------
*  diskcheck(cDrive,[nSector]) -> .F.
*     Creates a checksum for a disk
function DiskCheck(cDrive)
   fs2_devWarn(MSG_DOS + ".F.")
   return .F.                    // signals error

* ---------------------------------------------------------------------------
*  diskformat(cDrive,[p2..p5]) -> -1
*     Formats a disk
function DiskFormat(cDrive,p2,p3,p4,p5)
   fs2_devWarn(MSG_DOS + "-1")
   return -1                     // signals error

* ---------------------------------------------------------------------------
*  diskname() -> ""
*     Drive designator for current drive
function DiskName()
#ifdef FS_WIN32
   return upper (left( curdir(), 1 ) )
#else
   fs2_devWarn(MSG_DOS + "''")
   return ""                     // signals error
#endif

* ---------------------------------------------------------------------------
*  diskready(cDrive,[lBios]) -> .T.
*     Tests to see if a disk drive is ready
function DiskReady(cDrive,p2)
   fs2_devWarn(MSG_DOS + ".T.")
   return .T.                    // assumes availablity

* ---------------------------------------------------------------------------
*  diskreadyw(cDrive,[lBios]) -> .T.
*     Queries whether can write to a drive
function DiskReadyw(cDrive,p2)
   fs2_devWarn(MSG_DOS + ".T.")
   return .T.                    // assumes availablity

* ---------------------------------------------------------------------------
*  diskspeed(cDrive) -> 100000
*     Comparison value for the drive speed
function diskSpeed(cDrive)
   fs2_devWarn(MSG_DOS + "100000")
   return 100000         // This is 1000times the speed of an PC with 4.77 MHz

* ---------------------------------------------------------------------------
*  diskstat(cDrive) -> 0
*     Determines the status of a drive
function diskStat(cDrive)
   fs2_devWarn(MSG_DOS + "0")
   return 0                      // assumes no error

* ---------------------------------------------------------------------------
*  disktype(cDrive) -> 248
*     Determines the type of a drive
function diskType(cDrive)
   fs2_devWarn(MSG_DOS + "248")
   return 248                    // assumes harddisk

* ---------------------------------------------------------------------------
*  drivetype(cDrive) -> 3
*     Determines the type of a drive
function driveType(cDrive)
   fs2_devWarn(MSG_DOS + "3")
   return 3                      // assumes harddisk

* ---------------------------------------------------------------------------
*  filecclose() -> .F.
*     Closes a file after backup mode
function fileCclose()
   fs2_devWarn(MSG_DOS + ".F.")
   return .F.                    // signals error, backup n/a

* ---------------------------------------------------------------------------
*  fileccont(cDrive) -> 0
*     Copies sections of a file in backup mode
function fileCcont(cDrive)
   fs2_devWarn(MSG_DOS + "0")
   return 0                      // signals error, backup n/a

* ---------------------------------------------------------------------------
*  filecdati(lMode) -> .T.
*     Determines date mode for FILECOPY()
function fileCdati(lMode)
   return .T.                    // file's date and time

* ---------------------------------------------------------------------------
*  filecopen() -> .F.
*     Tests if the file is still open in the backup mode
function fileCopen()
   fs2_devWarn(MSG_DOS + ".F.")
   return .F.                    // signals closed, backup n/a

* ---------------------------------------------------------------------------
*  filevalid(cName) -> lValidChars
*     Tests if the string contains valid file name,
*     i.e. only chars "a..z", "A..Z", "0..9", "_", "."
function fileValid(cName)
   local_int  ii, iChar
   if valtype(cName) != "C" .or. empty(cName) .or. len(cName) > 250
      return .F.
   endif
   for ii := 1 to len(cName)
      iChar := strpeek(cName,ii)
      if iChar == 46 .or. iChar == 95
         loop
      elseif iChar < 48 .or. iChar > 122
         return .F.
      elseif (iChar >= 58 .and. iChar <= 64) .or. ;
             (iChar >= 91 .and. iChar <= 96)
         return .F.
      endif
   next
   return .T.

* ---------------------------------------------------------------------------
*  floppytype(cDrive) -> 0
*     Determines the exact type of floppy drive
function floppytype(cDrive)
   fs2_devWarn(MSG_DOS + "0")
   return 0                      // signals error

* ---------------------------------------------------------------------------
*  getshare() -> 4
*     Determines the file open (share) mode
function getshare()
   return 4                    // Not locked, r/w by other programs is allowed

* ---------------------------------------------------------------------------
*  numdiskf() -> 99
*     number of installed DOS disk drives
function numdiskf()
   return 99                  // assumes all DOS drives are installed

* ---------------------------------------------------------------------------
*  numdiskh() -> 99
*     number of installed DOS hard disk drives
function numdiskh()
   return 99                  // assumes all DOS drives are installed

* ---------------------------------------------------------------------------
*  numdiskl() -> 99
*     number of installed DOS logical disk drives
function numdiskl()
   return 99                  // assumes all DOS drives are installed

* ---------------------------------------------------------------------------
*  setshare(num) -> .T./.F.
*     Set the file open (share) mode
function setshare(num)
   return num == 4           // Not locked, r/w by other programs is allowed

* ---------------------------------------------------------------------------
* volserial([cDrive]) -> 0
*     Determines the DOS disk serial number
function volserial(cDrive)
   return 0                            // signals not available

* ---------------------------------------------------------------------------
* volume([cDrive]) -> .F.
*     Write volume label for a floppy or hard disk
function volume(cDrive)
   return .F.                          // signals not available

********************************************
* Functions to access novell networks      *
********************************************
* ---------------------------------------------------------------------------
*  nnetsdate
*     Our date IS the server's date
function nnetsdate
   return date()

* ---------------------------------------------------------------------------
*  nnetstime
*     Our time IS the servers time
function nnetstime
   return time()

* ---------------------------------------------------------------------------
*  nnetwhoami
*     The login name of this process
function nnetwhoami
   return gete("LOGNAME")

* ---------------------------------------------------------------------------
*  nnetinuse
*     Actually, we have processed, not Stations
function nnetinuse
   return 32

* ---------------------------------------------------------------------------
*  nnetstanum
*     All users are on one host
function nnetstanum
   fs2_devWarn(MSG_NOVEL + "netname() value")
   return netname()

* ---------------------------------------------------------------------------
*  nnetstaid
*     All users are on one host
function nnetstaid
   fs2_devWarn(MSG_NOVEL + "netname() value")
   return netname()

* ---------------------------------------------------------------------------
*  nnetocnums
*     All users are on one host
function nnetocnums
   parameters p1c
   fs2_devWarn(MSG_NOVEL + "netname() value")
   return netname()

* ---------------------------------------------------------------------------
*  nnetwork
*     Is a novell net active?
*     Since used for testing the network (which is always active), ret = .T.
function nnetwork
   fs2_devWarn(MSG_NOVEL + ".T.")
   return .T.

* ---------------------------------------------------------------------------
*  nnetbanner()
*     Sets or queries the current novel banner name
*     Since used for the network (which is always active), ret = .T.
function nnetbanner()
   fs2_devWarn(MSG_NOVEL + ".T.")
   return .T.

* ---------------------------------------------------------------------------
*  nnetcapact()
*     Is a net capture mode active?
*     Since used for the network (which is always active), ret = .T.
function nnetcapact()
   fs2_devWarn(MSG_NOVEL + ".T.")
   return .T.

* ---------------------------------------------------------------------------
*  nnetcapbeg()
*     Activates the capture mode for an network LPT device
*     Since used for the network (which is always active), ret = .T.
function nnetcapbeg()
   fs2_devWarn(MSG_NOVEL + ".T.")
   return .T.

* ---------------------------------------------------------------------------
*  nnetcapcan()
*     Deactivates the capture mode and aborts the network print job
*     Since used for the network (which is always active), ret = .T.
function nnetcapcan()
   fs2_devWarn(MSG_NOVEL + ".T.")
   return .T.

* ---------------------------------------------------------------------------
*  nnetcapflu()
*     Closes the current network print job and dispatches it
*     Since used for the network (which is always active), ret = .T.
function nnetcapflu()
   fs2_devWarn(MSG_NOVEL + ".T.")
   return .T.

* ---------------------------------------------------------------------------
*  nnetcapend()
*     Terminates the capture mode for an network LPT device
*     Since used for the network (which is always active), ret = .T.
function nnetcapend()
   fs2_devWarn(MSG_NOVEL + ".T.")
   return .T.

* ---------------------------------------------------------------------------
*  nnetcapssf()
*     Sets the Novell capture mode flags
*     Since used for the network (which is always active), ret = .T.
function nnetcapssf()
   fs2_devWarn(MSG_NOVEL + ".T.")
   return .T.

* ---------------------------------------------------------------------------
*  network()
*     Is a IBM LAN net active?
*     Since used for testing the network (which is always active), ret = .T.
function network()
   fs2_devWarn(MSG_NOVEL + ".T.")
   return .T.

******************************************************************************
* fs2_devWarn(cText)
*     reports Developer Warning when FS_SET("devel", .T.) is set
#include "error.fh"

FUNCTION fs2_devWarn(cText)
local oDefErr, oErr
local iStack := 0
if fs_set("devel")
   oDefErr := ErrorBlock()                  // default error handler block
   oErr    := ErrorNew()                    // new error object
   if ProcName(0) == "FS2_DEVWAR"
      iStack := 1                           // ignore this UDF
   endif
   oErr:ARGS          := NIL
   oErr:CANDEFAULT    := .F.
   oErr:CANRETRY      := .F.
   oErr:CANSUBSTITUTE := .F.
   oErr:CARGO         := NIL
   oErr:DESCRIPTION   := PROCNAME(iStack) + "(): " + cText
   oErr:FILENAME      := PROCNAME(iStack +1) + "(" + LTRIM(STR(PROCLINE(iStack +1),5)) + ")"
   oErr:GENCODE       := EG_DEVWARNING
   oErr:OPERATION     := ""
   oErr:OSCODE        := 0
   oErr:SEVERITY      := ES_RTERR
   oErr:SUBCODE       := 0
   oErr:SUBSYSTEM     := "BASE"

   EVAL (oDefErr, oErr)
endif
return NIL


* eof
