* For FlagShip 6
* Copyright (c) 2003..08 by multisoft, Munich, Germany
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the
* modified source code when your application is distributed in source,
* but may not remove the copyright notice.
*
* Release $Date: 2008/01/04 17:23:02 $
*

* You may change this file according to your needs.
* Copy this file to your working directory, change and
* Compile:         FlagShip -na -c -w -m -io=a getsys.prg
* Link Unix/Linux: FlagShip myapp*.prg getsys.o   [-io=a|g|t|b] [-o myapp]
* Link MS-Windows: FlagShip myapp*.prg getsys.obj [-io=a|g|t|b] [-o myapp.exe]

/***************************************************************************
 * getsys.prg = FlagShip 5 & 6 GET/READ system
 *
 * Note, this getsys.prg file replaces in FlagShip 5 and 6 the GET/READ system
 *   "getsys.prg" of FlagShip 4.48 and Clipper, the "old" getsys is renamed
 *    to "getsys4.prg":
 *
 * - This new "getsys.prg" support beneath the Get object also CheckBox,
 *   RadioButton, RadioGroup, CheckButton, PushButton, Topbar classes and
 *   is therefore incompatible to FS4.48 and to Clipper. The object of
 *   this source is available in the FlagShip library.
 *
 * - The old "getsys4.prg" is available for backward compatibility purposes
 *   in the .../system directory and will run in VFS6 as well, but does not
 *   support the additional FS6 classes (the getsys4.prg accept Get class
 *   only). It is therefore not included in the FlagShip library. If you want
 *   to use the "old" getsys4.prg, compile it according to the instructions
 *   in header and add the object file during the link stage.
 *
 * - The public functions
 *       GetActive()               RangeCheck()
 *       GetAlign()                ReadExit()
 *       GetApplyKey()             ReadInsert()
 *       GetDoSetKey()             ReadKey()
 *       GetPostValidate()         ReadModal()
 *       GetPreValidate()          __SetFormat()
 *       GetReader()               Updated()
 *       __KillRead()              ReadSelect(), ReadGetPos()
 *   are available in both versions and are call compatible to each other.
 *   If you have modified the sources by your own, these functions need
 *   to be available in the source to avoid linker failure.
 ****************************************************************************/

#define C87_COMPATIBEL             /* set compatibility to Clipper'87       */
#define CLIPPER_SETKEY             /* use Clipper compat. Setkey() behavior */
* #define GERMAN_MSG               /* activate for german scoreboard msg    */
* #define CTRL_END_SPECIAL         /* activate for ^W and ^End = last GET   */
* #define SUBSCRIPT_IN_READVAR     /* deactivate for READVAR() like Clip87  */
#define ACCEPT_WHEEL               /* deactivate to ignore Mouse Wheel      */

#ifdef C87_COMPATIBEL
#  define EMPTY_DATE_VALID         /* C87: empty date is valid, C5: invalid */
#endif

#include "set.fh"
#include "inkey.fh"
#include "getexit.fh"
#include "getclass.fh"
#include "button.fh"
#include "FSerrmsg.h"

#ifdef GERMAN_MSG
#  define GSMSG_INSERT  "Einfg."
#  define GSMSG_RANGE   "Bereich: "
#  define GSMSG_INVDATE "unguelt.Datum"
#else
#  define GSMSG_INSERT  "Insert"
#  define GSMSG_RANGE   "Range is "
#  define GSMSG_INVDATE "Invalid date "
#endif

// display coordinates for SCOREBOARD

#define K_UNDO       K_CTRL_U
#define SCORE_ROW    0
// #define SCORE_COL 60
#define SCORE_COL()  minmax(0, maxcol() -19, 60)  

// STATIC variables for the active READ

static g_Buff             AS Usual
static g_format           AS Usual
static g_updated  := .f.  AS Logic
static g_killread := .f.  AS Logic
static g_bumpTop  := .f.  AS Logic
static g_bumpBot  := .f.  AS Logic
static g_lastExit := 0    AS IntVar
static g_lastKeyb := 0    AS IntVar
static g_prevKeyb := 0    AS IntVar
static g_lastPos  := 0    AS IntVar
static g_activeGet        AS Usual
static g_procName := ""   AS Char
static g_procLine := 0    AS IntVar
static g_saveGets := .F.  AS Logic
static g_acceptedClass := ;
      ",GET,CHECKBOX,RADIOBUTTO,RADIOGROUP,PUSHBUTTON,LISTBOX,COMBOBOX,TBROWSE,MENU," ;
      as character
static g_toggle_multi := .F. as Logic
static g_hold_readvar := .F. as Logic
static g_selectItem   := 0   as Intvar

// Array elements to store actual settings

#define GSV_KILLREAD       1
#define GSV_BUMPTOP        2
#define GSV_BUMPBOT        3
#define GSV_LASTEXIT       4
#define GSV_LASTPOS        5
#define GSV_ACTIVEGET      6
#define GSV_READVAR        7
#define GSV_READPROCNAME   8
#define GSV_READPROCLINE   9
#define GSV_COUNT          9

PROTOTYPE static func ClearGetSysVars() as array
PROTOTYPE static proc DateMsg()
PROTOTYPE static func GetReadVar(get as object) AS Char
PROTOTYPE static proc PostActiveGet(get as object)
PROTOTYPE static proc RestoreGetSysVars(saved AS array)
PROTOTYPE static func Settle(aGetList AS Array, pos AS IntVar, [newPos], [lSkipOver], [lCycle]) AS IntVar
PROTOTYPE static proc ShowScoreboard()
PROTOTYPE function    _isGlyphValid(p1,p2) as usual
PROTOTYPE FUNCTION    ReadGetPos([p1],[p2],[p3],[p4],[p5]) as usual

/*--------------------------------------------------------------------------*
 | READMODAL()  is the "old" READ                                           |
 |   in FS5, this READ implementation support also other objects like       |
 |   TopBar, MenuItem, RadioButton, RadioGroup, CheckBox, ListBox etc.      |
 |                                                                          |
 | in: p1 (Array) with GET objects                                          |
 |     p2 (Logic) optional. If .T.:clears GET widgets on exit               |
 |     p3 (Logic) optional. If .T.:align, .F.:don't align, NIL:SET GUIALIGN |
 |     p4 (Logic) optional. If .T.:(def) skip over unsatisf. VALID fields   |
 |                             .F.: don't skip over unsatisf. VALID fields  |
 |     p5 (Logic) optional. If .T.:check all VALID conditions at exit,      |
 |                             .F.:(default) don't check all VALIDs at exit |
 |     p6 (Logic) optional. If .T.:keep GetList[] on exit = READ SAVE       |
 |     p7 (Logic) optional. If .T.:cycle, don't exit at first/last GET      |
 | fn: updated status (.T./.F.)                                             |
 *--------------------------------------------------------------------------*/
FUNCTION ReadModal ( aGetList AS array, ;
                    [lClear], [lAlign], [lSkipOver], ;
                    [lCheckVal], [lSaveGets], [lCycle] ) AS usual

local oGetItem                         as object
local iPos, ii                         as IntVar
local lOk := .F.                       as logic
local lSaveExit, lAnyEnabled := .F.    as logic
local savedGetSysVars, cOut, obj, readerRet, cClassName as Usual
local iStack := 0    as intvar

   g_killread := .F.                // status of ReadKill()
   if valtype(lSaveGets) == "L"
      g_saveGets := lSaveGets
   else
      g_saveGets := .F.
   endif

   if ( Valtype(g_format) == "B" )                    // SET FORMAT TO ?
      Eval(g_format)                                  // yes, process
   endif

   /* check/set optional parameters
    */
   if AppIoMode() == "G"
      if valtype(lAlign) != "L"
         lAlign := Set(_SET_GUIALIGN)
      endif
   else
      lAlign := .F.
   endif
   lClear    := if(valtype(lClear)    == "L", lClear,    .F.)
   lSkipOver := if(valtype(lSkipOver) == "L", lSkipOver, .T.)
   lCheckVal := if(valtype(lCheckVal) == "L", lCheckVal, .F.)
   lSaveExit := Set(_SET_EXIT)
   if valtype(lCycle) != "L"
      lCycle := .F.
   elseif lCycle
      Set(_SET_EXIT, .F.)
   endif

   /* check _aGlobSetting[GSET_L_READVAR_HOLD]
    * = hold last GET in ReadVar() after READ ends? (default is .F.)
    */
   g_hold_readvar := _GlobSetting(GSET_L_READVAR_HOLD, .F.)  
   
   /* check if all items in aGetList{} array
    * are objects of supported classes
    */
   for ii := 1 to len(aGetList)
      obj := aGetList[ii]
      lOk := .F.
      cClassName := ""
      if valtype(obj) == "O" 
         if IsObjProperty(obj, 4, "ClassName")
            cClassName := obj:ClassName()
         else
            cClassName := if(IsObjClass(obj, "_TBROWSE"), "TBROWSE", "")
         endif
         if !empty(cClassName)
            lOk := ("," + cClassName + + ",") $ g_acceptedClass
         endif
      endif
         
//      lOk := valtype(obj) == "O" .and. ;
//             IsObjProperty(obj, 4, "ClassName") .and. ;
//             ("," + obj:ClassName() + ",") $ g_acceptedClass
      if !lOk
         if !IsObjProperty(obj, 1, "GetEnabled")
            _Rt_Error(RTCLASSTYPE, "Unsupported class " + IsObjClass(obj)[1])
            ReadVar( "" )
            Set(_SET_EXIT, lSaveExit)
            return .F.
         endif
         obj:GetEnabled := .F.
         _Rt_Error(DEVWARNING, "Unsupported class '" + IsObjClass(obj)[1] + "',;GET field ignored/disabled")
         loop
      endif
      if valtype(obj:GetEnabled) != "L"
         obj:GetEnabled := .T.
      endif
      if obj:GetEnabled
         lAnyEnabled := .T.
      endif
   next
   if len(aGetList) > 0 .and. !lAnyEnabled .and. !lCycle
      SetPos( MaxRow()-1, 0 )                         // READ without objects
      _Rt_Error(DEVWARNING, "nothing to do, all GET fields are disabled")
      if !g_hold_readvar 
         ReadVar( "" )  // don't clear when _aGlobSetting[GSET_L_READVAR_HOLD] is set .T.
      endif
      Set(_SET_EXIT, lSaveExit)
      return .F.
   endif
   if Empty(aGetList) .or. !lOk                       // cannot proccess
      SetPos( MaxRow()-1, 0 )                         // READ without objects
      _Rt_Error(DEVWARNING, "nothing to do, no GET fields available")
      if !g_hold_readvar
         ReadVar( "" )  // don't clear when _aGlobSetting[GSET_L_READVAR_HOLD] is set .T.
      endif
      Set(_SET_EXIT, lSaveExit)
      return (.F.)
   endif

   cOut := _LANGNAMES(1)                              // SCOREINS_MSG ("insert")
   if lAlign                                          // align GET columns
      GetAlign(@aGetList)                             // available below
   endif

   savedGetSysVars := ClearGetSysVars()               // save act.status

   iStack := if(ProcName(0) == "READMODAL", 1, 0)     // recompiled getsys.prg?
   g_procName := ProcName(iStack)                     // save calling UDF
   g_procLine := ProcLine(iStack)                     // for SET KEY
   while _IsNameCB(g_procName)                        // ignore codeblocks
      g_procName := ProcName(++iStack)
      g_procLine := ProcLine(iStack)
   enddo
   iPos := Settle( aGetList, 0, , , lCycle )          // first GET item in list

   while iPos > 0
      oGetItem := aGetList[iPos]                      // catch one GET/Gui object
      PostActiveGet( oGetItem )                       // activate it

      readerRet := 0
      if ValType( oGetItem:reader ) == "B"            // user READER block ?
         readerRet := Eval( oGetItem:reader, oGetItem, @aGetList )    // - yes, execute
      else
         readerRet := GetReader(oGetItem, @aGetList ) // - no, use build-in
      endif
      if len(aGetList) < iPos .or. g_killread
         g_killread := .T.                            // GetList was killed
         exit                                         // in external UDF
      endif
      iPos := Settle( aGetList, iPos, readerRet, lSkipOver, lCycle ) // next GET or 0 == exit

      /* READ done.
       * Check all fields for satisfied VALID condition,
       * if requested by READ EXITCHECK and not ESC exit
       */
      if iPos == 0 .and. lCheckVal .and. !g_killread
         for ii := 1 to len(aGetList)
            oGetItem := aGetList[ii]
            if oGetItem:GetEnabled .and.   ;
               GetPreValidate(oGetItem, .T.) .and. ;
               !GetPostValidate(oGetItem, .T.)        // VALID condition ok?
               iPos := ii
               exit  // leave the for..next loop
            endif
         next
         if iPos > 0                                  // VALID not ok,
            loop                                      // stay in READ
         endif
      endif
   enddo

#ifdef C87_COMPATIBEL
   if ( Set(_SET_SCOREBOARD) .and. Set(_SET_INSERT))  // SET SCOREBOARD ON
      SetPos(SCORE_ROW, SCORE_COL())                  // and INSERT active ?
      if AppIoMode() == "G"
         DispOut(SPACE(strlen2space(cOut)))
      else
         DispOut( SPACE( LEN( cOut ) ))               // yes, clear scoreboard
      endif
   end
#endif

   RestoreGetSysVars(savedGetSysVars)                 // restore status
   SetPos( MaxRow()-1, 0 )                            // S87 compat.

   if lClear .and. IsObjProperty(obj,4,"destroy")
      aeval(aGetList, {|obj| obj:Destroy() } )
   endif
   
   GetActive(NIL)
   Set(_SET_EXIT, lSaveExit)
   
return g_updated

/*--------------------------------------------------------------------------
 * ReadSelect(getlist, iPos|cVarName) -> iPos
 * ReadSelect(iPos|cVarName)          -> iPos
 *    select GET item# <iPos> or GET item with variable name <cVarName> 
 *    It may be called during READ from WHEN, VALID or other callback
 *    and considered on moving to next GET item in Settle(),  
 *    or before starting READ, or with READ SELECT <item> clause    
 * iPos = +/- pos = relative to current position
 *      = 1001..n = absolute position in GetList[n -1000]                                   
 */
FUNCTION ReadSelect(par1, par2) as usual
   local iPos, iCurr     as intvar
   local item, aGetlist, oGet  as usual
   memvar getlist
      
   if pcount() == 1
      aGetlist := GetList
      item     := par1
   elseif pcount() >= 2
      aGetlist := par1
      item     := par2
   else
      return 0
   endif
   if valtype(aGetlist) != "A" .or. len(aGetlist) == 0 .or. empty(item)
      return 0
   endif

   if valtype(item) == "N" 
      iCurr := ReadGetPos(aGetList)             // get position of active GET
      if item >= 1000                           // absolute pos in GetList[]
         item := item - 1000
      else                                      // relative +/- pos to current
         if iCurr == 0
            item := minmax(1,item,len(aGetList))
         else                                   
            item := iCurr + item                // now absolute in GetList[]
         endif  
      endif
      if item < 1 
         item := 1
      elseif item > len(aGetlist)
         item := len(aGetlist)
      endif
      g_selectItem := item
      return item
   endif
   if valtype(item) == "C" 
      item := alltrim(upper(item))
      iPos := ReadGetPos(aGetList, item)
      g_selectItem := iPos
      return iPos
   endif
   return 0

/*--------------------------------------------------------------------------
 * ReadGetPos([getlist])            -> element# of current GET in GetList[]
 * ReadGetPos([getlist,] cVar)      -> element# of variable <cVar> in GetList[]
 * ReadGetPos([getlist,] NIL,cCapt) -> element# of caption <cCapt> in GetList[]
 * ReadGetPos([getlist,] NIL,NIL,nRow,nCol) -> element# of nRow,nCol in GetList[]
 */
FUNCTION ReadGetPos(p1, p2,p3,p4,p5) as usual
   local oGet, oGetCurr, ii, iParCount := pcount() as usual
   memvar GetList
   
   if valtype(p1) != "A"
      p5 := p4
      p4 := p3
      p3 := p2
      p2 := p1
      p1 := Getlist
      iParCount := min(iParCount +1, 5)
   endif

   if valtype(p1) != "A" .or. len(p1) == 0
      return 0
   endif 
   if iParCount == 1          
      oGetCurr := GetActive()  
      if empty(oGetCurr) 
         return 0
      endif
   else
      if iParCount >= 4 .and. (valtype(p4) != "N" .or. valtype(p5) != "N")
         iParCount := 3
      endif
      if iParCount == 3 .and. (valtype(p3) != "C" .or. empty(p3))
         iParCount := 2
      endif
      if iParCount == 2 .and. (valtype(p2) != "C" .or. empty(p2))
         return 0
      endif
   endif
         
   for ii := 1 to len(p1)
      oGet := p1[ii]
      if valtype(oGet) != "O"
         return 0
      endif
      do case
      case iParCount == 1                 // current item
         if IsObjEquiv(oGetCurr, oGet)
            return ii
         endif
      case iParCount == 2                 // var name
         if valtype(oGet:Name) == "C" .and. upper(oGet:Name) == upper(p2)
            return ii
         endif
      case iParCount == 3                 // caption
         if valtype(oGet:Caption) == "C" .and. upper(oGet:Caption) == upper(p3)
            return ii
         endif
      case iParCount >= 4                 // row, col
         if valtype(oGet:row) == "N" .and. oGet:row == p4 .and. ;
            valtype(oGet:col) == "N" .and. oGet:col == p5
            return ii
         endif
      endcase
   next
   return 0            
     
/*--------------------------------------------------------------------------*
 | GETREADER()  standard READER to process ONE Get or other GUI object      |
 |              called from READMODAL()                                     |
 |                                                                          |
 |  Note, this function is assigned per default to get:Handler property     |
 |  and can be executed for single get by oGet:Exec()                       |
 |                                                                          |
 | in: GET or GUI object                                                    |
 | fn: 0 on usual GET field exit, <num> if mouse click on field <num>       |
 *--------------------------------------------------------------------------*/
FUNCTION GetReader( oGetItem AS Object, [aGetList], [lCycle] ) AS USUAL

   local lZeroStatus := fs_set ("zero", .F.) AS Logical
   local lExit, lErrOut := .F.               AS Logical
   local lRetHand                            AS Usual
   local key, keymask, iPos, ii, iHitTest, iMouseCol, iMouseRow  as intvar
   local cKeys                   as char
   local iCurr, retVal := 0      as intVar
   local obj, status             as usual
   local lMouseWarn, lFastRead   as logic

   lFastRead := _GlobSetting(GSET_L_READ_FAST, .F.)   // fast read, default is .F.
// lFastRead := .T.
   lCycle := if(valtype(lCycle) == "L", lCycle, .F.)
   
   if !oGetItem:GetEnabled
      return 0
   endif
   lMouseWarn := m->_oMouse:DevWarning
   m->_oMouse:DevWarning := .F.                       // disable mouse warnings
   g_Buff := NIL
   g_LastKeyb := 0
   g_PrevKeyb := 0
   
   if !empty(oGetItem:Name)                           // remove \0 in name if any
      oGetItem:name := Upper(substr(oGetItem:name, 1, LEN(oGetItem:name)))
   endif
   fs_set ("zero", lZeroStatus)                       // restore orig.status

   if GetPreValidate(oGetItem)                        // WHEN option ok?
      oGetItem:SetFocus()                             // activate object

//    if !IsObjClass(oGetItem, "_TBROWSE") .and. oGetItem:ClassName() == "GET"
      if oGetItem:ClassName() == "GET"
         g_Buff := oGetItem:VarGet()
      endif

      while oGetItem:exitState == GE_NOEXIT ;         // == 0, loop until exit
          .and. !g_killRead                           // or until external terminate
//       if (!IsObjClass(oGetItem, "_TBROWSE") .and.  oGetItem:ClassName() == "GET" .and. oGetItem:TypeOut ) // editable position ?
         if ( oGetItem:ClassName() == "GET" .and. oGetItem:TypeOut ) // editable position ?
            oGetItem:exitState := GE_ENTER            // - no, simul.RETURN
         endif

         if !empty(oGetItem:Message) .and. !lErrOut   // MESSAGE set?
            StatusMessage(oGetItem:Message)           // yes, display it in
         endif                                        // StatBar or SET MESSAGE

         retVal := 0
//       if !IsObjClass(oGetItem, "_TBROWSE") .and. oGetItem:ClassName() == "GET"
         if oGetItem:ClassName() == "GET"
            if _GlobSetting(GSET_G_L_GET_MOUSENEW, .T.) ;
                  .and. AppIoMode() == "G" .and. oGetItem:Type == "C" ;
                  .and. ! ("@K" $ upper(oGetItem:Picture))
               /* in GUI mode: set cursor at field begin, even if
                * mouse clicked on other field somewhere within the field
                */
               GetApplyKey( oGetItem, K_HOME, lCycle )        // set cursor to home pos.
            endif
            /* GET object: loop here == call keyboard handler in
             * GetApplyKey() until user exit the GET field or
             * have clicked the mouse button.
             * Note that in GUI i/o mode, ISO char in inkey() is required
             */
            while oGetItem:exitState == GE_NOEXIT .and. !g_killRead
#ifdef CLIPPER_SETKEY
               /* Setkey() evaluation may terminate the READ
                */
               if lFastRead 
                  /* perform fast keyboard read sequence for GET field 
                   */
                  key := 0
                  while key == 0
                     key := Inkey( 0.1, INKEY_KEYBOARD + INKEY_LUP, , .F. )
                  enddo
                  if key >= 32 .and. key <= 255
                     cKeys := chr(key)
                     key := Inkey( 0.1, INKEY_KEYBOARD + INKEY_LUP, , .F. )
                     while key >= 32 .and. key <= 255
                        cKeys += chr(key)
                        key := Inkey( 0.1, INKEY_KEYBOARD + INKEY_LUP, , .F. )
                     enddo
                     lExit := .F.
                     for ii := 1 to len(cKeys)  // process all keys in buffer
                        GetApplyKey( oGetItem, strpeek(cKeys,ii), lCycle )
                        if oGetItem:exitState != GE_NOEXIT .or. g_killRead
                           lExit := .T.
                           exit     // break key: forget rest
                        endif
                     next
                     if lExit
                        loop  // exit while...
                     endif
                     if key == 0
                        key := Inkey( 0, INKEY_KEYBOARD + INKEY_LUP, , .F. ) 
                     endif
                  endif
                  // continues with last key pressed
               else
                  /* default READ behavior
                   */
                  key := Inkey( 0, INKEY_KEYBOARD + INKEY_LUP, , .F. ) // in ISO mode
               endif
#else
               /* process also recursions via Setkey() and don't exit
                */
               key := InkeyTrap( 0, INKEY_KEYBOARD + INKEY_LUP, , , , .F.)
#endif
               if key < K_MOUSEMOVE
                  g_PrevKeyb := g_LastKeyb
                  g_LastKeyb := key
               endif
               GetApplyKey( oGetItem, key, lCycle )

               if _GlobSetting(GSET_G_L_GET_MOUSENEW, .T.) // should mouse click on other field trigger oGet:Home() ? 
                  if AppIoMode() == "G" .and. ;
                     oGetItem:Type == "C" .and. ;
                     key == K_LBUTTONDOWN .and. oGetItem:exitState == GE_MOUSE
                     /*
                      * LeftMouseDown pressed, cache LeftMouseUp
                      */
#ifdef CLIPPER_SETKEY
                     key := Inkey( 0, INKEY_KEYBOARD + INKEY_LUP, , .F. ) // in ISO mode
#else
                     key := InkeyTrap( 0, INKEY_KEYBOARD + INKEY_LUP, , , , .F.)
#endif
                  endif
               endif 
            enddo // end loop for GetApplyKey() handler
         else
            /* GUI object: its keyboard handler
             * is invoked directly via :Exec()
             */
            oGetItem:exitState := GE_NOTSET
            oGetItem:GuiVar2obj()                     // @..GET var -> object
            clear typeahead                           // clear pending keys
            status := oGetItem:Exec()                 // process GUI object and

            oGetItem:GuiObj2var()                     // store object -> @..GET var
            if oGetItem:exitState == GE_NOTSET .or. ;  // :exitState may be set in handler
               (oGetItem:exitState == GE_MOUSE .and. lastkey() == K_LDBLCLK)
               oGetItem:exitState := GE_ENTER
            endif
         endif

         iMouseRow := MRow(.T.)                       // mouse coord in pixel
         iMouseCol := MCol(.T.)

         if ( !GetPostValidate(oGetItem) )            // check VALID block is any
            oGetItem:exitState := GE_NOEXIT           // invalid: stay here
            if !empty(oGetItem:ErrValBlock)           // remember reset error msg
               lErrOut := .T.
            endif
         else
            lErrOut := .F.
         endif

         if oGetItem:exitState == GE_MOUSE            // mouse click outside of GET
            if valtype(aGetList) == "A" .and. len(aGetList) > 0
               /* Search for a GET field, corresponding
                * to the current mouse coordinates.
                * The get:Caption is considered as well
                */
               iCurr  := 0
               retval := 0
               for iPos := 1 to len(aGetList)
                  obj := aGetList[iPos]
                  if IsObjEquiv(oGetItem, obj)
                     iCurr := iPos
                     if retVal > 0
                        exit
                     else
                        loop                          // ignore current object
                     endif
                  endif
                  if retVal == 0 .and. ;
                        valtype(obj) == "O" .and. obj:GetEnabled .and. ;
                        IsObjProperty(obj, 4, "HitTest")
                     iHitTest := obj:HitTest(iMouseRow, iMouseCol, .T.)
                     if iHitTest != HTNOWHERE         // anywhere in object
                        retVal := iPos
                        if iCurr > 0
                           exit
                        endif
                     endif
                  endif
               next
            endif
            if retVal == 0                            // GET item not found or
               oGetItem:exitState := GE_NOEXIT        // GetList n/a, ignore mouse
            else
               oGetItem:exitState := GE_ENTER         // found: simulate Return in current
            endif
         endif
      enddo

      oGetItem:KillFocus()                            // deactivate object

      if !empty(oGetItem:Message) .or. lErrOut
         StatusMessage()                              // clear curr. message
      endif
   endif
   m->_oMouse:DevWarning := lMouseWarn                // restore developer's warning
 
return retVal

/*--------------------------------------------------------------------------*
 | GETAPPLKEY() check input key, take action on object                      |
 |              note: the actual object should have Focus set               |
 |              called from GETREADER()                                     |
 |                                                                          |
 | in: p1: GET object                                                       |
 |         Note: not for GUI objects which are handled by own keyboard      |
 |               handlers via :Exec()                                       |
 |     p2: key (numer.equivalent) from INKEY                                |
 |                                                                          |
 | This is a keyboard handler for Get object.                               |
 | Note: for backward compatibility reason, the usual loop and Inkey*()     |
 |       is not here but in GetReader()                                     |
 *--------------------------------------------------------------------------*/
FUNCTION GetApplyKey (oGetItem AS object, key AS INTVAR, lCycle as logic)

local cKey                 as Char
local bKeyBlock            as Usual
local iHitTest, iPos, iNew as IntVar
local beep1, beep2, ii     as IntVar
local aGlyph1, aGlyph2     as array
local iToggleMulti := 1    as intvar
local lMouseWarn, lSave    as logic
local multiWait            as numeric

   if set(_SET_MULTIBYTE)
      iToggleMulti := _aGlobSetting[GSET_N_MEMO_GET_MULTIB]    // GET/READ,MemoEdit() with SET MULTIBYTE ON: toggle low/high ascii 
      multiWait    := _aGlobSetting[GSET_N_MEMO_GET_MULTIW]    // GET/READ,MemoEdit() with SET MULTIBYTE ON: inkey() wait for 2nd byte
      beep1        := _aGlobSetting[GSET_N_MEMO_GET_BEEP1]     // GET/READ,MemoEdit() with SET MULTIBYTE ON: beep n-times after 1sec 
      beep2        := _aGlobSetting[GSET_N_MEMO_GET_BEEP2]     // GET/READ,MemoEdit() with SET MULTIBYTE ON: beep n-times after time-out expires
      if beep1 > 0 .and. multiWait > 0.0 .and. multiWait <= 1.0
         beep1 := 0        // disable, would not be reached within 1sec
      endif
      if multiWait > 1.0 .and. beep1 > 0
         multiWait -= 1.0  // decrease 1sec for beep1 waiting
      endif
   endif
   
#ifdef CLIPPER_SETKEY
   if ( bKeyBlock := SetKey(key) ) != NIL             // SET KEY active ?
      GetDoSetKey(bKeyBlock, oGetItem)                // yes, process UDF
      return                                          // and go back for
   end                                                // next key
   if ( bKeyBlock := OnKey(key) ) != NIL              // SET KEY active ?
      GetDoSetKey(bKeyBlock, oGetItem)                // yes, process UDF
      return                                          // and go back for
   end                                                // next key
   if key == 28 .or. key < 0
      if len(ckey := Getfunction(key))  != 0          // SET Function active ?
         keyboard ckey                                // yes, Keyboard it
         return                                       // and go back for
      end                                             // next key
   endif
#endif
   lMouseWarn := m->_oMouse:DevWarning
   m->_oMouse:DevWarning := .F.                       // disable mouse warnings

   do case
   case ( key == K_UP )                               // cursor up:
      oGetItem:exitState := GE_UP                     //    previous GET

   case ( key == K_SH_TAB )                           // shift TAB:
      oGetItem:exitState := GE_UP                     //    previous GET

   case ( key == K_DOWN )                             // cursor down:
      oGetItem:exitState := GE_DOWN                   //    next GET

   case ( key == K_TAB )                              // TAB:
      oGetItem:exitState := GE_DOWN                   //    next GET

   case  key == K_ENTER .or. ;
         key == K_CTRL_ENTER                          // RETURN:
      oGetItem:exitState := GE_ENTER                  //    next GET

   case ( key == K_ESC )                              // ESCAPE:
      if ( Set(_SET_ESCAPE) .or. lCycle)              //    SET ESCAPE ON ?
         oGetItem:undo()                              //    yes, undo
         oGetItem:exitState := GE_ESCAPE              //    exit READ
      end

   case ( key == K_PGUP )                             // PageUp:
      if lCycle
         oGetItem:exitState := GE_TOP                 //    CYCLE: first GET
      else
         oGetItem:exitState := GE_WRITE               //    default: update + exit
      endif

   case ( key == K_PGDN )                             // PageDown:
      if lCycle
         oGetItem:exitState := GE_BOTTOM              //    CYCLE: first GET
      else
         oGetItem:exitState := GE_WRITE               //    default: update + exit
      endif
      
   case ( key == K_CTRL_HOME )                        // Ctrl-HOME :
      oGetItem:exitState := GE_TOP                    //    first GET

#ifdef CTRL_END_SPECIAL
   case (key == K_CTRL_W .and. lCycle)                // ^W with CYCLE
      oGetItem:exitState := GE_WRITE                  //    exit

   case (key == K_CTRL_END .or. ;
         key == K_CTRL_W)                             // ^W and ^END :
      oGetItem:exitState := GE_BOTTOM                 //    go last GET
#else
   case (key == K_CTRL_END .or. ;
         key == K_CTRL_W)                             // ^W and ^END :
      oGetItem:exitState := GE_WRITE                  //    update + exit
#endif

   case (key == K_INS)                                // INSERT:
      if !set(_SET_INSERT_INKEY)
         Set( _SET_INSERT, !Set(_SET_INSERT) )        //    set insert mode
      endif
      ShowScoreboard()                                //    show score

   case (key == K_UNDO)                               // ^U or Alt-U:
      oGetItem:Undo()                                 //    Undo

   case (key == K_HOME)                               // HOME:
      if g_Prevkeyb == K_HOME .and. oGetItem:Home2char  // K_HOME + K_HOME :
         lSave := oGetItem:Home2char                    //    move to real buffer begin
         oGetItem:Home2char := .F.
         oGetItem:Home()                               
         oGetItem:Home2char := lSave
         g_Lastkeyb := -999
      else
         oGetItem:Home()                              // go first pos or 1st char
      endif

   case (key == K_END)                                // END:
      if g_Prevkeyb == K_END .and. oGetItem:End2char  // K_END + K_END :
         lSave := oGetItem:End2char                   //    move to real buffer end
         oGetItem:End2char := .F.
         oGetItem:End()                               
         oGetItem:End2char := lSave
         g_Lastkeyb := -999
      else
         oGetItem:End()                               // go last pos/buff or last char
      endif

   case (key == K_RIGHT)                              // Cursor RIGHT
      oGetItem:Right()                                //    move cursor/exit

   case (key == K_LEFT)                               // Cursor LEFT
      oGetItem:Left()                                 //    move cursor/exit

   case (key == K_CTRL_RIGHT)                         // Ctrl-Cursor RIGHT
      oGetItem:WordRight()                            //    go next word

   case (key == K_CTRL_LEFT)                          // Ctrl-Cursor LEFT
      oGetItem:WordLeft()                             //    go prev. word

   case (key == K_BS)                                 // BACKSPACE
      oGetItem:BackSpace()                            //    delete prev.char

   case (key == K_DEL)                                // DELETE
      oGetItem:Delete()                               //    delete act.char

   case (key == K_CTRL_T)                             // Ctrl-T
      oGetItem:DelWordRight()                         //    delete word right

   case (key == K_CTRL_Y)                             // Ctrl-Y
      oGetItem:DelEnd()                               //    delete to buff end

   case (key == K_CTRL_BS)                            // Ctrl-BACKSPACE
      oGetItem:DelWordLeft()                          //    delete word left

#ifdef ACCEPT_WHEEL         
   CASE key = K_MOUSEWHEELUP                          // mouse wheel up
      oGetItem:exitState := GE_UP                     //    previous GET
   CASE key = K_MOUSEWHEELDN                          // mouse wheel down
      oGetItem:exitState := GE_DOWN                   //    next GET
#endif

   case  key == K_LBUTTONDOWN  .or. ;                 // mouse left button
         key == K_LDBLCLK                             // mouse left button double click
      iHitTest := oGetItem:HitTest(Mrow(.T.), MCol(.T.), .T.)
      if iHitTest == 0
         oGetItem:exitState := GE_MOUSE
      endif

   /* Mid-Button or Alt-C (user modifiable via _aGlobSetting[] assignment):
    * copy marked text (done by Left-Mouse-Down)
    * to X11/Windows clipboard buffer (GUI only)
    */
   case  key == _GlobSetting(GSET_G_N_GET_COPY1, K_MBUTTONDOWN) .or. ;
         key == _GlobSetting(GSET_G_N_GET_COPY2, K_ALT_C)
      if IsObjProperty(oGetItem, 4, "Copy")
          oGetItem:Copy()
      endif 

   /* Shift+Mid-Button or Alt-V (user modifiable via _aGlobSetting[] assignment):
    * paste X11/Windows clipboard buffer
    * into current field position (GUI only)
    */
   case  key == _GlobSetting(GSET_G_N_GET_PASTE1, K_SH_MBUTTONDN) .or. ;
         key == _GlobSetting(GSET_G_N_GET_PASTE2, K_ALT_V)
      if IsObjProperty(oGetItem, 4, "Paste")
          oGetItem:Paste()
      endif

   case  key == K_LBUTTONUP  .or. ;                   // mouse left button up
         key == K_RBUTTONUP  .or. ;                   // mouse right button up
         key == K_MBUTTONUP  .or. ;                   // mouse mid button up
         key == K_RBUTTONDOWN                         // mouse right button down
      oGetItem:exitState := GE_NOEXIT
      if AppIoMode() == "G"
         if "@K" $ upper(oGetItem:Picture) .and. g_Lastkeyb == 0  // no key press yet
            lSave := oGetItem:Home2char
            oGetItem:Home2char := .F.
            oGetItem:Home()                           // move to abs. buffer begin
            oGetItem:Home2char := lSave
         else
            if _GlobSetting(GSET_G_L_GET_MOUSEHOME, .F.) // should mouse click on GET always trigger oGet:Home() ?
               oGetItem:Home()  
            elseif oGetItem:type == "C" .and. ;
                   !_GlobSetting(GSET_G_L_GET_MOUSEOUT, .F.) // should mouse click allow positioning after last char? 
               iPos := len(trim(oGetItem:Buffer)) +1
               if oGetItem:Pos > iPos
                  lSave := oGetItem:End2char
                  oGetItem:End2char := .T.
                  oGetItem:End()                      // move to last+1 valid char in buffer
                  oGetItem:End2char := lSave
               endif
            endif
         endif
      endif

// case set(_SET_MULTIBYTE) .and. key == iToggleMulti
//    g_toggle_multi := if(g_toggle_multi, .F., .T.)  // toggle low/high ascii for multibyte 
         
   otherwise
      if ((key >= 32 .and. key <= 255) .or. ;         // char 32..255
          (key == K_CTRL_U .and. key != K_UNDO))      // or paragraph sign
         cKey := Chr(key)                             //    save char

         if (oGetItem:type $ "NFI" .and. (cKey == "." .or. cKey == ","))
            if "@K" $ upper(oGetItem:Picture)         // simulates Clipper behavior
               oGetItem:DelEnd()                      // which deletes the data
            endif
            oGetItem:ToDecPos()
         else
            if set(_SET_MULTIBYTE) ;                  // SET MULTIBYTE ON
                  .and. oGetItem:type $ "CM" ;        // with char input type?
                  .and. len(cKey) < 2 ;
                  .and. multiWait >= 0.0
               /* Multibyte, and 1st byte is > 127: wait for
                * 2nd byte = glyph, beep after 1sec (if specified)
                */
//             if g_toggle_multi .and. key <= 127     // low-ascii input simul?
//                key += if(key <= 95, 160, 127)      // yes, simulate high-ascii
//                cKey := Chr(key)                    
//             endif     
               if _isGlyphValid(key,1)                // 1st glyph byte?                        
                  if beep1 > 0                      
                     iNew := inkey(1.0, INKEY_KEYBOARD)    
                     if iNew == 0
                        m->_ostdio:Beep(beep1,300)    // beep 1x to remember
                        iNew := inkey(multiWait, INKEY_KEYBOARD)    
                     endif
                  else
                     iNew := inkey(multiWait, INKEY_KEYBOARD) 
                  endif   
                  if _isGlyphValid(iNew,2)            // 2nd glyph byte?
                     cKey += chr(iNew)                // yes
//                elseif g_toggle_multi .and. iNew <= 127
//                   iNew += if(iNew <= 95, 160, 127) 
//                   cKey += chr(iNew) 
                  elseif iNew == 0
                     if beep2 > 0
                        m->_ostdio:Beep(beep2,300)    // time-out, beep 2x, delay 0.3s 
                     endif
                  else
                     keyboard chr(iNew)               // not 2nd glyph, push back
                  endif
               endif
            endif
            if ( Set(_SET_INSERT) )                   // INSERT on ?
               oGetItem:Insert(cKey)                  //    yes, insert char
            else
               oGetItem:Overstrike(cKey)              //    no, type over
            end
//            if ( oGetItem:ClassName() == "GET" .and. oGetItem:typeOut ) // out of buffer ?
            if (!IsObjClass(oGetItem, "_TBROWSE") .and.  oGetItem:ClassName() == "GET" .and. oGetItem:typeOut ) // out of buffer ?
               if ( Set(_SET_BELL) )                  //    SET BELL ON ?
                  ?? Chr(7)                           //    yes, bell
               end
               if ( !Set(_SET_CONFIRM) )              //    SET CONFIRM ON ?
                  oGetItem:exitState := GE_ENTER      //    no, exit GET
               end
            end
         end
      elseif key >= K_MOUSEMOVE .and. key <= K_ALT_MBUTTONUP
         // ignore all mouse movements and clicks (the left/right
         // button-up and left-double-click are considered above)
#ifndef CLIPPER_SETKEY
      elseif key != 0
         // probably Fn key, update/re-read get:buffer from variable
         oGetItem:UpdateBuffer()
#endif
      endif
   endcase
   m->_oMouse:DevWarning := lMouseWarn                // restore developer's warning
return

/*--------------------------------------------------------------------------*
 | GETPREVALI() check the WHEN condition on begin of GET processing         |
 |              called from GETREADER()                                     |
 |                                                                          |
 | in: oGetItem = GET or GUI object                                         |
 |     lCheckOnly = optional, if .T. don't change exitState, default = .F.  |
 | fn: .T. WHEN is ok                                                       |
 *--------------------------------------------------------------------------*/
func GetPreValidate(oGetItem AS Object, [lCheckOnly]) AS LOGICAL

local saveUpdated AS Usual
local when := .t. AS Usual                            // logic (or codeblock)

   if IsObjProperty(oGetItem, 1, "preBlock") .and. oGetItem:preBlock <> NIL
      saveUpdated := g_updated                        // yes, save status
      when := Eval(oGetItem:preBlock, oGetItem)       //   eval WHEN block
      while valtype(when) == "B"                      //   also recursively
         when := Eval(when, oGetItem)
      enddo
      oGetItem:Display()                              //   display act. GET
      ShowScoreBoard()                                //   display score
      g_updated := saveUpdated                        //   restore status
   end

   if valtype(lCheckOnly) == "L" .and. lCheckOnly
      if g_killread
         when := .f.
      endif
   elseif g_killread                                  // READ canceled ?
      when := .f.                                     //    no GET edit
      oGetItem:exitState := GE_ESCAPE                 //    exit READMODAL
   elseif ( !when )                                   // WHEN block not ok
      oGetItem:exitState := GE_WHEN                   //    no GET edit
   else
      oGetItem:exitState := GE_NOEXIT                 // edit the actual GET
   end
return (when)

/*--------------------------------------------------------------------------*
 | GETPOSTVAL() check the VALID condition before exiting GET                |
 |              check date validation, no GET exit on invalid date          |
 |              called from GETREADER()                                     |
 |                                                                          |
 | in: oGetItem = GET or GUI object                                         |
 |     lCheckOnly = optional, if .T. don't change exitState, default = .F.  |
 | fn: .T. if VALID is ok, ESCAPE was pressed or READ was killed            |
 *--------------------------------------------------------------------------*/
func GetPostValidate(oGetItem AS Object, [lCheckOnly]) AS LOGICAL

local saveUpdated                    AS Logic
local changed, dateok                AS Logic
local result := .T., lRestore := .F. AS LOGIC
local valid := .T.                   AS Usual         // logic (or codeblock)
local varOrig, varCurr               AS Usual

   lCheckOnly := if(valtype(lCheckOnly) == "L", lCheckOnly, .F.)

   if ( oGetItem:exitState == GE_ESCAPE )             // ESCAPE pressed ?
      return (.t.)                                    // yes, no VALID cond
   end

   if IsObjClass(oGetItem, "_Get")                    // or: oGetItem:ClassName() == "GET"
      dateok := !oGetItem:BadDate()                   // date value ok ?
#ifdef EMPTY_DATE_VALID
      if !dateok .and. oGetItem:type == "D"           // FlagShip:
         dateok := oGetItem:EmptyDate                 // allow empty date
      end                                             // for C87 compatib.
#endif
      if !dateok                                      // Date valid ?
         oGetItem:Home()                              // no, VALID is false
         if valtype(oGetItem:ErrValBlock) == "B"
            eval(oGetItem:ErrValBlock, oGetItem)      //    display user error msg
         elseif valtype(oGetItem:ErrValBlock) == "C"
            StatusMessage(oGetItem:ErrValBlock)
         else
            DateMsg()                                 //    or std. err msg,
         endif
         ShowScoreboard()                             //    show scoreboard,
         return (.f.)                                 //    stay in GET
      end

      if !lCheckOnly
         /* Many applications checks content of the variable
          * or data field in VALID. We therefore store temprary
          * the current edit value (buffer) into variable or field
          * and restore it at failure, except the variable content
          * was changed by user applic in VALID
          */
         if ( oGetItem:changed )                      // changes within GET ?
            lRestore := .T.                           // yes,
            varOrig := oGetItem:VarGet()              // save current var content
            oGetItem:Assign()                         // write buffer
            varCurr := oGetItem:VarGet()
//          g_updated := .t.                          // into var, set update
            if valtype(g_Buff) == valtype(oGetItem:VarGet())
               if !(oGetItem:VarGet() == g_Buff)
                   g_updated := .t.                   // into var, set update
               endif
            else
               g_updated := .t.                       // into var, set update
            endif
         end
         oGetItem:Reset()                             // redisplay buffer
      endif
   endif

   if IsObjProperty(oGetItem, 1, "postBlock") .and. ;
         oGetItem:postBlock != NIL                    // VALID block avail ?
      saveUpdated := g_updated                        // save status
//    if oGetItem:ClassName() == "GET"
      if !IsObjClass(oGetItem, "_TBROWSE") .and. oGetItem:ClassName() == "GET"
         SetPos( oGetItem:row, oGetItem:col + Len(oGetItem:buffer) )    // S87 compat. cursor
      else
         if IsObjProperty(oGetItem, 2, "row")         // n/a for e.g. Listbox
            SetPos( oGetItem:row, oGetItem:col )
         endif
      endif

      valid := Eval(oGetItem:postBlock, oGetItem)     // eval VALID block
      while VALTYPE(valid) == "B"
         valid := Eval(valid, oGetItem)               // eval VALID block
      enddo
      result := valid

      if !lCheckOnly
//         if oGetItem:ClassName() == "GET"             // apply for GET class only
         if !IsObjClass(oGetItem, "_TBROWSE") .and. oGetItem:ClassName() == "GET"             // apply for GET class only
            oGetItem:Home()                           // FS5: reset editing cursor
            SetPos( oGetItem:row, oGetItem:col )      // reset compat. pos
            ShowScoreBoard()                          // scoreboard
            oGetItem:UpdateBuffer()                   // restore buffer
            g_updated := saveUpdated                  // restore status
            if g_killread .or. !oGetItem:HasFocus     // READ killed in UDF ?
               oGetItem:exitState := GE_ESCAPE        // exit ReadModal()
               result := .T.
               g_killread = .T.
            endif
         elseif g_killread                            // apply for GUI classes
            oGetItem:exitState := GE_ESCAPE           // exit ReadModal()
            result := .T.
            g_killread = .T.
         endif
      endif
   endif
   if !result
      if lRestore                                     // var tempor. changed?
         if varCurr == oGetItem:VarGet()              // yes, var changed by user?
            oGetItem:Varput(varOrig)                  // no, restore original value
         endif
      endif
      if valtype(oGetItem:ErrValBlock) == "B"
         eval(oGetItem:ErrValBlock, oGetItem)         // display user error msg
      elseif valtype(oGetItem:ErrValBlock) == "C"
         StatusMessage(oGetItem:ErrValBlock)          // write error msg text
      endif
   endif
return (result)

/*--------------------------------------------------------------------------*
 | GETDOSETKE() call proc according to SET KEY                              |
 |              called from GETAPPLYKEY()                                   |
 |                                                                          |
 | in: SetKey block from SetKey(key)                                        |
 |     GET object                                                           |
 *--------------------------------------------------------------------------*/
proc GetDoSetKey(keyBlock AS Codeblock, oGetItem AS Object)

local saveUpdated    AS Logic
local ii, iPos       AS intvar

//   if oGetItem:ClassName() == "GET"
   if !IsObjClass(oGetItem, "_TBROWSE") .and. oGetItem:ClassName() == "GET"
      iPos := oGetItem:Pos                            // save cursor position
   endif
   if ( oGetItem:changed )                            // if GET changed:
      oGetItem:Assign()                               // write buffer into
      g_updated := .T.                                // variable
   end
   saveUpdated := g_updated                           // save status
   Eval(keyBlock, g_procName, g_procLine, ReadVar())  // invoke SET KEY funct.

   ShowScoreboard()                                   // show scoreboard
   oGetItem:UpdateBuffer()                            // restore buffer
   g_updated := saveUpdated                           // restore status

   if g_killread .or. !oGetItem:HasFocus              // READ killed in UDF ?
      oGetItem:exitState := GE_ESCAPE                 // exit ReadModal()
      g_killread = .T.

   elseif iPos > 0                                    // disable if not required:
      for ii := 1 to iPos                             // restore cursor position
         oGetItem:Right()                             //  since set to Home()
         if oGetItem:Pos >= iPos                      //  in UpdateBuffer()
            exit
         endif
      next

   endif
return

/*--------------------------------------------------------------------------*
 | GETACTIVE () returns the active GET object from READ                     |
 |              or assigns new active get object                            |
 |              standard Clipper/FlagShip function                          |
 *--------------------------------------------------------------------------*/
FUNCTION GetActive(oGetObject) AS USUAL
local oldActive as USUAL
   if ( ValType(g_activeGet) $ 'OU' )
	   oldActive := g_activeGet
   endif
   if ( PCount() > 0 .and. ValType(oGetObject) $ 'OU' )
      g_activeGet := oGetObject
   end
return ( oldActive )

/*--------------------------------------------------------------------------*
 | UPDATED ()   returns the global get:changed status from all prev.GETS    |
 |              standard Clipper/FlagShip function                          |
 *--------------------------------------------------------------------------*/
FUNCTION Updated() AS LOGICAL
return g_updated

/*--------------------------------------------------------------------------*
 | READEXIT()   set/reset an exit of READ at the first/last GET item via    |
 |              cursor Up/Down. Standard Clipper/FlagShip function          |
 *--------------------------------------------------------------------------*/
FUNCTION ReadExit ([lNew AS Logi]) AS LOGICAL
LOCAL lEntryStatus := Set(_SET_EXIT) as Logical
   if pcount() > 0
      Set(_SET_EXIT, lNew)
   endif
return lEntryStatus

/*--------------------------------------------------------------------------*
 | READINSERT() set/reset the actual INSERT status                          |
 |              standard Clipper/FlagShip function                          |
 *--------------------------------------------------------------------------*/
FUNCTION ReadInsert ([lNew AS Logi]) AS LOGICAL
LOCAL lEntryStatus := Set(_SET_INSERT) as Logical
   if pcount() > 0
      Set(_SET_INSERT, lNew)
   endif
return lEntryStatus

/*--------------------------------------------------------------------------*
 | READKEY()    transform LastKey() codes from FlagShip/Clipper             |
 |              INKEY() value into dBASE III compatible value               |
 *--------------------------------------------------------------------------*/
funct ReadKey () AS INTVAR

local lastkey := lastkey()  AS IntVar
local add     := if (g_updated, 256, 0) AS IntVar

do case
   case lastkey == K_RETURN                           // Enter
      return 15 + add
   case lastkey == K_ESC                              // Esc
      return 12 + add
   case lastkey == K_PGUP                             // PgUp
      return  6 + add
   case lastkey == K_PGDN                             // PgDn
      return  7 + add
   case lastkey == K_CTRL_PGUP                        // ^PgUp
      return 34 + add
   case lastkey = K_CTRL_PGDN                         // ^PgDn
      return 35 + add
   case lastkey == K_CTRL_END .or. ;                  // Ctrl-End
        lastkey == K_CTRL_W                           // Ctrl-W
      return 14 + add
   case lastkey >= 32 .or. ;
        (lastkey == K_CTRL_U .and. lastkey != K_UNDO) // type past end
      return 15 + add
   otherwise
      return 0
endcase

/*--------------------------------------------------------------------------*
 | READKILL()   set/return a flag signaling that the READ should be         |
 |              terminated; ReadKill(.T.) is similar to CLEAR GETS but      |
 |              it does not delete GetList[] items                          |
 *--------------------------------------------------------------------------*/
FUNCTION ReadKill( [lKillFlag] ) AS Logical
LOCAL lEntryStatus as logic
   lEntryStatus := g_killread
   if pcount() > 0 .and. valtype(lKillFlag) == "L"
      g_killread := lKillFlag
   endif
return lEntryStatus

/*--------------------------------------------------------------------------*
 | READSAVE()   set/reset an flag signalling whether the GetList[] should   |
 |              be cleared at exit of READ. The default is .F. except       |
 |              when READ SAVE was used which set it to .T.                 |
 *--------------------------------------------------------------------------*/
FUNCTION ReadSave ([lSave AS Logi]) AS LOGICAL
LOCAL lEntryStatus as Logical
   lEntryStatus := g_saveGets
   if pcount() > 0 .and. valtype(lSave) == "L"
      g_saveGets := lSave
   endif
return lEntryStatus

/*--------------------------------------------------------------------------*
 | READUPDATED() equivalent to Updated() except it can change it status     |
 *--------------------------------------------------------------------------*/
FUNCTION ReadUpdated ([lSet AS Logi]) AS LOGICAL
LOCAL lEntryStatus as Logical
   lEntryStatus := g_updated
   if pcount() > 0 .and. valtype(lSet) == "L"
      g_updated := lSet
   endif
return lEntryStatus

/*--------------------------------------------------------------------------*
 | __SETFORMA() set/reset SET FORMAT TO                                     |
 |              used in std.ch and std.fh (SET FORMAT TO)                   |
 *--------------------------------------------------------------------------*/
func __SetFormat(blk)
   g_format := if ( ValType(blk) == "B", blk, NIL )
return (NIL)

/*--------------------------------------------------------------------------*
 | __KILLREAD() kill active READ                                            |
 |              used in std.ch and std.fh (CLEAR GETS)                      |
 *--------------------------------------------------------------------------*/
proc __KillRead()
   ReadKill( .T. )
return

/*--------------------------------------------------------------------------*
 | RANGECHECK() check RANGE from..to and report "out of rande nnn..nnn"     |
 |              used in std.ch and std.fh (RANGE clause)                    |
 |                                                                          |
 | notes: -parameter 2 is unused, for compatibility to Clipper 5.00 only    |
 |        -value will be displayed in the same format, as the PICTURE       |
 |         within the GET (incompatible to Clipper'87)                      |
 *--------------------------------------------------------------------------*/
func RangeCheck(oGetItem AS Object, junk, lo AS NUMER, hi AS NUMER) AS LOGIC

local cMsg        as Char
local nRow, nCol  as Numer
local xValue      as usual

   if ( !oGetItem:changed )                           // don't ckeck,
      return (.t.)                                    // if buffer unchanged
   end

   xValue := oGetItem:VarGet()

   if ( xValue >= lo .and. xValue <= hi )             // range is
      return (.t.)                                    // o.k.
   end

   if ( Set(_SET_SCOREBOARD) )                        // SET SCOREBOARD ON ?

      cMsg := _LANGNAMES (2) + " "                    // get msg from FSsortab.xxx

      cMsg += Ltrim(Transform(lo, oGetItem:picture)) + ;
              " - " + Ltrim(Transform(hi, oGetItem:picture))    // see note

      if ( Len(cMsg) > MaxCol() )
         cMsg := Substr( cMsg, 1, MaxCol() )
      end

      nRow := Row()                                   // display error msg
      nCol := Col()

      SetPos( SCORE_ROW, Min(60, MaxCol() - Len(cMsg)) )
      DispOut(cMsg)
      SetPos(oGetItem:Row, oGetItem:Col)

      while ( NextKey() == 0 )                        // wait for next
      end                                             // key pressed

      SetPos( SCORE_ROW, Min(60, MaxCol() - Len(cMsg)) )
      if AppIoMode() == "G"
         DispOut(SPACE(strlen2space(cMsg)))
      else
         DispOut( Space(Len(cMsg)) )
      endif
      SetPos(nRow, nCol)                              // clear err msg
   end
return (.f.)

/*--------------------------------------------------------------------------*
 | GetAlign(aGetList) align GUI columns                                     |
 |              used in ReadModal() when SET GUIALIGN is ON or from applic. |
 *--------------------------------------------------------------------------*/
FUNCTION GetAlign([aGetList])
   local aCols := {}                as array
   local oGetItem                   as Object
   local ii, jj, iFirst, iLast      as intvar
   local nGetAsCol, nCol, nCaptLng, nMaxLng as numeric
   local cCapt                      as char
   static nSPACE := 1.0             as numeric  // columns between SAY and GET

   if aGetList == NIL
      aGetList := m->GetList
   endif
   if AppIoMode() != "G" .or. empty(aGetList)
      return NIL                                   // apply for GUI only
   endif
   for ii := 1 to len(aGetList)
      oGetItem := aGetList[ii]

      if oGetItem:GetEnabled ;                        // object enabled for GET?
         .and. valtype(oGetItem:GuiCol) == "N" ;      // :GuiCol  set?
         .and. valtype(oGetItem:GuiCaption) == "C" ;  // :GuiCaption set?
         .and. (valtype(oGetItem:GuiAlign) != "L" .or. oGetItem:GuiAlign)

         cCapt := oGetItem:GuiCaption
         if valtype(oGetItem:GuiCapPict) == "C"
            cCapt := transform(cCapt, oGetItem:GuiCapPict)
         endif
         nCaptLng := Strlen2col(cCapt)
         nCol := Pixel2col(oGetItem:GuiCol)        // :CuiCol is in pixel, we need coord
         /*
          * aCols [n,1] = position in GetList
          *       [n,2] = assumed GET column = next column after SAY len(), or -1 if n/a
          *       [n,3] = SAY column, or 0 if n/a
          *       [n,4] = SAY length in pixel, or 0 if n/a
          */
         aadd(aCols, {ii, nCol + len(cCapt) + 1, nCol, nCaptLng})
      else
         aadd(aCols, {ii, -1, 0, 0})           // SAY caption data not available
      endif
   next
   aSort(aCols, NIL, NIL, {|x,y| x[2] <= y[2]})    // sort on assumed GET columns

   /* if at least two GETs have the same assumed column position,
    * align (move) the widgets to the same, rightmost position
    * determined from the largest SAY text (caption) size
    */
   nGetAsCol := nMaxLng := -1
   iFirst := iLast := 0                // first & last GET# with the same col()
   for ii := 1 to len(aCols)
      if aCols[ii,2] < 0
         loop                                      // ignore w/o caption
      endif
      if aCols[ii,2] != nGetAsCol .and. iFirst > 0
         /* column position changed
          * align previous group, if any
          */
         if iLast > iFirst .and. nMaxLng > 0
            for jj := iFirst to iLast
               oGetItem := aGetList[aCols[jj,1]]
               oGetItem:Move(NIL, aCols[jj,3] + nMaxLng + nSPACE, .F.)  // move widget
            next
         endif
         nGetAsCol := nMaxLng := -1
         iFirst := iLast := 0
      endif
      /* new or the same column group:
       * determine first and last GET position,
       * calculate max caption length for this group
       */
      if iFirst == 0
         iFirst := ii
      endif
      iLast := ii
      nGetAsCol := aCols[ii,2]
      nMaxLng   := max(nMaxLng, aCols[ii,4])     // max caption length
   next
   if iFirst > 0 .and. iLast > iFirst .and. nMaxLng > 0
      /* process last group
       */
      for jj := iFirst to iLast
         oGetItem := aGetList[aCols[jj,1]]
         oGetItem:Move(NIL, aCols[jj,3] + nMaxLng + nSPACE, .F.)  // move widget
      next
   endif
return NIL

/******************** internal service routines only ***********************/

****************************************************
* Settle()
*
* returns new element from GETLIST according to:
*  - the actual position
*  - the EXIT status of the actual GET
*  - the ReadSelect() request
*
* aGetList = array of GET or GUI objects
* iPos     = current GetList position (0..len())
* newPos   = optional, next index pos in GetList to be set
* skipOver = optional, allow skip over unsatisfied VALID fields
* lCycle   = optional, don't exit from 1st/last GET but by ESC or ^W only
*
* Note: return 0 signals an exit of the READ
*       or    -1 to hard exit (= ESC)
*       the EXIT status from the previous GET will be transfered forward
*
STATIC FUNCTION Settle(aGetList AS Array, iPos AS IntVar, [newPos], [lSkipOver], [lCycle]) AS IntVar

local iExitState  AS Usual
local lGetEnabled as logic
local obj         as object
local ii          as intvar

   lSkipOver := if(valtype(lSkipOver) == "L", lSkipOver, .T.)
   lCycle    := if(valtype(lCycle) == "L", lCycle, .F.)

   if iPos == 0
      iExitState := GE_DOWN                           // 2: CursDown
   else
      iExitState := aGetList[iPos]:exitState
   end

   do case
   case iExitState == NIL                             // illegal GET
      return -1                                       //    same as ESC
   case iExitState == GE_ESCAPE                       // 7: Return
      return -1                                       //    exit READ
   case iExitState == GE_WRITE                        // 6: ^W
      return 0                                        //    exit READ
   endcase

   if ( iExitState != GE_WHEN )                       // 8: reset status info
      g_lastPos := iPos
      g_bumpTop := .f.
      g_bumpBot := .f.
   else                                               // move the last exit
      iExitState := g_lastExit                        // exit satus
   end

   do case                                            // move GET element
   case g_selectItem > 0
      iPos := g_selectItem                            // --ReadSelect()
      g_selectItem := 0
      
   case valtype(newPos) == "N" .and. newPos > 0       // - to specified elem#
      if lSkipOver .or. iPos < 1 .or. iPos > len(aGetList) .or. newPos <= iPos
         iPos := newPos                               // --w/o VALID check
      else
         iPos := newPos                               // --with VALID check
         for ii := iPos +1 to newPos -1
            obj := aGetList[ii]
            if obj:GetEnabled .and. ;
                  GetPreValidate(obj, .T.) .and. ;
                  !GetPostValidate(obj, .T.)          // VALID condition ok?
               iPos := ii                             // -- no, stay on unsatisfied
               exit
            endif
         next
      endif

   case ( iExitState == GE_UP )                       // - to prev. item
      iPos --

   case ( iExitState == GE_DOWN )                     // - to next item
      iPos ++

   case ( iExitState == GE_TOP )
      iPos := 1
      g_bumpTop := .T.
      iExitState := GE_DOWN

   case ( iExitState == GE_BOTTOM )
      iPos := Len(aGetList)
      g_bumpBot := .T.
      iExitState := GE_UP

   case ( iExitState == GE_ENTER )
      iPos ++

   endcase

   /* check if the item is not generally disabled
    * if so, get next or previous item# if such available
    */
   if iPos > 0 .and. iPos <= len(aGetList) .and. aGetList[iPos]:GetEnabled
      if iExitState == GE_UP
         while iPos >= 1 .and. !aGetList[iPos]:GetEnabled
            iPos --
         enddo
      else
         while iPos <= len(aGetList) .and. !aGetList[iPos]:GetEnabled
            iPos ++
         enddo
      endif
   endif

   /* check if the iPos# is in the GetSys array range
    */
   if ( iPos <= 0 )                                   // bounce top
      if lCycle
         iPos := len(aGetlist)
         while iPos > 1 .and. !aGetlist[iPos]:GetEnabled
            iPos --
         enddo
         iExitState := GE_NOEXIT
      elseif ( !ReadExit() .and. !g_bumpBot )
         g_bumpTop := .T.
         iPos := g_lastPos
         iExitState := GE_DOWN
      endif
   elseif ( iPos > Len(aGetList) )                    // bounce bottom
      if lCycle
         iPos := 1
         while iPos < len(aGetlist) .and. !aGetlist[iPos]:GetEnabled
            iPos --
         enddo
         iExitState := GE_NOEXIT
      elseif ( !ReadExit() .and. iExitState != GE_ENTER .and. !g_bumpTop )
         g_bumpBot := .T.
         iPos := g_lastPos
         iExitState := GE_UP
      else
         iPos := if(iExitState == GE_ESCAPE, -1, 0)
      endif
   endif

   g_lastExit := iExitState                           // hold last exit state

   if ( iPos != 0 )
      aGetList[iPos]:exitState := iExitState
   end

return (iPos)


****************************************************
* PostActiveGet()
* Post active GET for ReadVar(), GetActive().
*
static proc PostActiveGet(oGetItem as Object)

   GetActive( oGetItem )
   ReadVar( GetReadVar(oGetItem) )
   ShowScoreBoard()

return

****************************************************
* ClearGetSysVars()
* save/clear the actual status info from READ into an internal array
* note: UPDATED status will be cleared, but not restored (S87 compat.)
*
static func ClearGetSysVars() as array

local saved[ GSV_COUNT ] AS Array

   saved[ GSV_KILLREAD ] := g_killread
   g_killread := .f.

   saved[ GSV_BUMPTOP ] := g_bumpTop
   g_bumpTop := .f.

   saved[ GSV_BUMPBOT ] := g_bumpBot
   g_bumpBot := .f.

   saved[ GSV_LASTEXIT ] := g_lastExit
   g_lastExit := 0

   saved[ GSV_LASTPOS ] := g_lastPos
   g_lastPos := 0

   saved[ GSV_ACTIVEGET ] := GetActive(NIL)
   saved[ GSV_READVAR ]   := ReadVar()
   ReadVar( "" )
   saved[ GSV_READPROCNAME ] := g_procName
   g_procName := ""

   saved[ GSV_READPROCLINE ] := g_procLine
   g_procLine := 0

   g_updated := .f.

return (saved)

****************************************************
* RestoreGetSysVars()
* restore the actual status info READ from internal array
* note: UPDATED status will be not restored (S87 compat.)
*
static proc RestoreGetSysVars(saved AS array)

// g_killread := saved[ GSV_KILLREAD ]
   g_bumpTop  := saved[ GSV_BUMPTOP ]
   g_bumpBot  := saved[ GSV_BUMPBOT ]
   g_lastExit := saved[ GSV_LASTEXIT ]
   g_lastPos  := saved[ GSV_LASTPOS ]
   g_procName := saved[ GSV_READPROCNAME ]
   g_procLine := saved[ GSV_READPROCLINE ]
   GetActive( saved[ GSV_ACTIVEGET ] )
   if g_hold_readvar    // is _aGlobSetting[GSET_L_READVAR_HOLD] set .T. ?
      if empty(saved[ GSV_READVAR ] ) .and. !empty( ReadVar() )
         return         // hold last GET name in ReadVar()
      endif
   endif
   ReadVar( saved[ GSV_READVAR ] )  // consider nested READs
return

****************************************************
* GetReadVar()
* appends the NAME of the actual array-variable for READVAR()
*   with [elem] according to get:subscript
* Note: incompatible with Clipper'87 (here array name only)
*
static func GetReadVar(oGetItem as Object) AS Char

local name := Upper(oGetItem:name) as Char

#ifndef SUBSCRIPT_IN_READVAR                      // Clipper-87 compatib. ?
local ii AS IntVar
   local stat := fs_set ("zero", .F.) as logic
   name := Upper(substr(oGetItem:name, 1, LEN(oGetItem:name)))
   fs_set ("zero", stat)

   if (!empty(oGetItem:subscript) .and. !empty(name))
      for ii := 1 to len(oGetItem:subscript)
         name += "[" + ltrim(str(oGetItem:subscript[ii])) + "]"
      next
   end
#endif

return (name)


****************************************************
*  ShowScoreboard()  displays the actual INSERT status
*
static proc ShowScoreboard()

local nRow, nCol  AS Numer
local cMsg        AS Char

   if ( Set(_SET_SCOREBOARD) )                    // SET SCOREBOARD ON ?
      cMsg := _LANGNAMES (1)                      // get msg from FSsortab.xxx
      nRow := Row()                               // save actuall cursor
      nCol := Col()
      SetPos(SCORE_ROW, SCORE_COL())              // see define
      DispOut( if(Set(_SET_INSERT), cMsg, Space(Strlen2space(cMsg)) ) )
//    DispOut( if(Set(_SET_INSERT), cMsg, SPACE(LEN(cMsg)) ))
      SetPos(nRow, nCol)                          // restore cursor
   end
return

****************************************************
*  DateMsg()   displays error on "invalid date"
*
static proc DateMsg()

local nRow, nCol  AS Numer
local cMsg        AS Char

   cMsg := _LANGNAMES (3)                         // get msg from FSsortab.xxx

   if ( Set(_SET_SCOREBOARD) )                    // SET SCOREBOARD ON ?
      nRow := Row()                               // save actuall cursor
      nCol := Col()
      SetPos(SCORE_ROW, SCORE_COL())
      DispOut(cMsg)                               // display "invalid.."
      SetPos(nRow, nCol)                          // reset cursor

      while ( Nextkey() == 0 )                    // wait for the next
      end                                         // key pressed

      SetPos(SCORE_ROW, SCORE_COL())              // clear the message
      if AppIoMode() == "G"
         DispOut(SPACE(strlen2space(cMsg)))
      else
         DispOut(SPACE(LEN(cMsg)))
      endif
      SetPos(nRow, nCol)
   end
return

/*** eof ***/

