* For FlagShip 6 
* Copyright (c) 2004..08 by multisoft, Munich, Germany 
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the 
* modified source code when your application is distributed in source, 
* but may not remove the copyright notice.
* 
* Release $Date: 2008/02/03 11:13:17 $
*
* Note: these functions are available in the FlagShip library. They are
* low-level functions called before your application start.  The 
* corresponding UDF to the current i/o mode will initialize (instantiate) 
* the application's i/o. Therefore neither screen i/o, nor database access 
* is allowed yet! The user INIT Procedures or Functions will be invoked 
* after _*InitIo() but before starting the main of your application, so 
* you may put your initialization (if required) by the usual way there.
* See additional details in fsman section SYS.2.7
*
* Warning: do not change this initio.prg except you exactly _know_
* what you are doing. This is a system file and errors here may cause
* your FlagShip application will not work anymore. Also, this file may
* be changed in further FS versions w/o notification.  
* You simply may change these default settings by assigning your preferred 
* value to the _aGlobSetting[...] array element as described in the 
* "Tuning" chapter of corresponding CMD or FUN manual page.  
* If you nevertheless wish to modify this file, copy this source to your 
* working directory, make changes, and
*
* Compile Windows: FlagShip -na -c -w -m -io=g initio.prg -o initio_g.obj
*             and: FlagShip -na -c -w -m -io=t initio.prg -o initio_t.obj
*             and: FlagShip -na -c -w -m -io=b initio.prg -o initio_b.obj
*             and: FlagShip -na -c -w -m -io=a initio.prg -o initio_a.obj
* Link Windows   : FlagShip myapp*.prg initio_*.obj
*
* Compile Linux  : FlagShip -na -c -w -m -io=g initio.prg -o initio_g.o
*             and: FlagShip -na -c -w -m -io=t initio.prg -o initio_t.o
*             and: FlagShip -na -c -w -m -io=b initio.prg -o initio_b.o
*             and: FlagShip -na -c -w -m -io=a initio.prg -o initio_a.o
* Link Linux     : FlagShip myapp*.prg initio_*.o
*
*-------------------------------------------------------------------------

#include "appclass.fh"
#include "applic.fh"
#include "set.fh"
#include "box.fh"
#include "button.fh"
#include "color.fh"
#include "statbar.fh"

/* Arrays and define's used here globaly
 */
STATIC aSetupDescr := {"WINDOW_TOPROW=", "WINDOW_TOPCOL=", ;
                       "WINDOW_HEIGHT=", "WINDOW_WIDTH=" , ;
                       "WINDOW_TITLE=" , "WINDOW_MODE="  , ;
                       "APPLIC_ENVIR=" }                    as array
STATIC aSetupData  := {NIL, NIL, NIL, NIL, NIL, NIL, ""}    as array 
                         
#define IDX_TOPROW  1
#define IDX_TOPCOL  2
#define IDX_HEIGHT  3
#define IDX_WIDTH   4
#define IDX_TITLE   5
#define IDX_MODE    6   
#define IDX_ENVIR   7   
 
/* common defaults used here
 */
#define NEWLINE chr(13)+chr(10)   /* EOL to write the setup file  */
#define FILEPREFIX "."            /* prefix of the setup file     */
   
/*--------------------------------------------------------------------------
 * User modifiable System Function: _*InitIo(aArgs) 
 *   Instantiates the screen or terminal i/o either according to the
 *   compilation switch -io=..., and/or from a setup file, or via 
 *   command-line parameter, or automatically for the curr.used environment
 *
 * Parameters: aArgs is an array of strings corresponding to the 
 *             command-line parameters when invoking the executable.
 *             The 1st element is the name of the executable itself.
 *             So you may control the behaviour also by the passed
 *             parameters. The below listed command-line parameters 
 *             are not passed to the PARAMETERS of the application.
 *             The here supported entries are:
 * -IO=G[UI] or -IO=B[ASIC] or -IO=T[ERM] resp. to override the default 
 *             envirinment when the applic was compiled with the -io=a 
 *             switch or w/o -io= at all. See also note below.
 * -SETUP=.... entry is considered as the fully qualified name of a
 *             file containing set-up data. If not given, the algorithm
 *             described in ReadDefaults() below is used.
 *             If the setup file is found, it will be updated at program 
 *             end by the EXIT function SaveDefaults() - see below, or by 
 *             selecting the "save" option in the GUI main menu by the user. 
 *             To prepare a new setup file, simply create an empty file 
 *             by "touch ~/.a" or similarily.
 * -NOSETUP    don't read and store from/to the setup file
 * -NOSPOOL    don't init default spool file
 * -TESTIO     reports the used i/o mode
 *
 * NOTE: the default _InitIo() function apply when compiled with -io=a 
 *    switch which produce an "intelligent" GUI/terminal/basic mode but 
 *    huge executables, especially for statically linked app's. 
 *    When this program is compiled with the -io=g or -io=b or -io=t 
 *    switch, the _gInitIo(), _bInitIo() or _tInitIo() is produced
 *    instead and is called from a application compiled with the same
 *    -io= switch. The use of the -io=g or -io=b or -io=t switch may 
 *    significantly reduce the application size. 
 *--------------------------------------------------------------------------*/
#ifdef FSIO_ALL
FUNCTION _InitIo(aArgs)    // apply when compiled by -io=a or w/o -io switch
   #define USE_MENUBAR
   #define USE_READDEFAULTS
   #define USE_EXITCALLBACK

   PROTO FUNCTION InitIoMenu(p1 as object)   as usual
   PROTO FUNCTION gInitIoMenu(p1 as object)  as usual
   PROTO FUNCTION tInitIoMenu(p1 as object)  as usual
#endif

#ifdef FSIO_GUI
FUNCTION _gInitIo(aArgs)   // apply when compiled by -io=g switch
   #define USE_MENUBAR
   #define USE_READDEFAULTS
   #define USE_EXITCALLBACK

   PROTO FUNCTION InitIoMenu(p1 as object)   as usual
   PROTO FUNCTION gInitIoMenu(p1 as object)  as usual
#endif

#ifdef FSIO_TERM
FUNCTION _tInitIo(aArgs)   // apply when compiled by -io=t switch
// #define USE_MENUBAR
// #define USE_READDEFAULTS
// #define USE_EXITCALLBACK
   PROTO FUNCTION tInitIoMenu(p1 as object)  as usual
#endif

#ifdef FSIO_BASIC
FUNCTION _bInitIo(aArgs)   // apply when compiled by -io=b switch
// #define USE_MENUBAR
// #define USE_READDEFAULTS
// #define USE_EXITCALLBACK
#endif

#ifdef USE_READDEFAULTS
   PROTO STATIC FUNCTION ReadDefaults()      as usual
#endif

PROTO FUNCTION _MainCompiledBy()          as char
PROTO FUNCTION _SetSortab()               as usual
PROTO STATIC FUNCTION GlobSetting()       as usual
PROTO STATIC FUNCTION ErrSpoolFile(oErr, bDefErr)  as usual
   
PUBLIC   gcSetupFile            // used in ReadDefaults() and SaveDefaults()
PUBLIC   _InitIoDone := .F.     // used in error handler
LOCAL    ii, pos                    AS IntVar
LOCAL    cIoType := "?"             AS char
LOCAL    cIoDefault,  cHowDetected  AS char
LOCAL    iClockWidth, iStatBarLng   AS IntVar
LOCAL    oBoxClock, oBoxMsg         AS Object
LOCAL    lOk                        AS Logic
MEMVAR   gcSetupFile            

   /* check for command-line param -SETUP=<setup file name>, 
    * e.g. SETUP=/my/data/mySetup.data
    */
   ii := ascan(aArgs, {|x| upper(left(x,7)) == "-SETUP=" })
   if ii > 0 .and. len(trim(aArgs[ii])) > 7
      gcSetupFile := alltrim(substr(aArgs[ii],8))
   endif

   GlobSetting()     // Prepare global settings (see bellow)

// You may also manually define/set a fix setup-file name here (which stores 
// e.g. the last used applic coordinates for GUI mode), especially when the 
// default search algorithm used here (see ReadDefaults() below) is not 
// comfortable or applicable for your application, e.g. 
*  gcSetupFile := getenv("HOME") + ".myApp"
   ii := ascan(aArgs, {|x| upper(trim(x)) == "-NOSETUP" })
   if ii > 0
      gcSetupFile := NIL
#ifdef USE_READDEFAULTS
   else
      ReadDefaults()                     // get the window setup data
#endif
   endif

/**************************************************************************
 * this initio.prg was compiled with -io=g switch
 */
#ifdef FSIO_GUI
   CONSTANT oApplic := _gAppWindow{NIL, ;
                        aSetupData[IDX_TOPROW], aSetupData[IDX_TOPCOL], ;
                        aSetupData[IDX_HEIGHT], aSetupData[IDX_WIDTH],  ;
                        aSetupData[IDX_TITLE] , aSetupData[IDX_MODE]  }
   CONSTANT _oMouse := _gMouse{}           // for Mouse* functions
   CONSTANT _oStatusBar := m->oApplic:StatusBar
   CONSTANT _oPrinter := _gPrinter{}
   
   extern_g()
#endif

/**************************************************************************
 * this initio.prg was compiled with -io=t switch
 */
#ifdef FSIO_TERM
   CONSTANT oApplic := _tAppWindow{}      // no window defaults apply here
   CONSTANT _oMouse := _tMouse{}          // for Mouse* functions
   CONSTANT _oStatusBar := m->oApplic:StatusBar
   CONSTANT _oPrinter := _tPrinter{}
   
   extern_t()
#endif

/**************************************************************************
 * this initio.prg was compiled with -io=b switch
 */
#ifdef FSIO_BASIC
   CONSTANT oApplic := _bAppWindow{}      // no window defaults apply here
   CONSTANT _oMouse := _bMouse{}          // for Mouse* functions
   CONSTANT _oStatusBar := m->oApplic:StatusBar
   CONSTANT _oPrinter := _bPrinter{}
   
   extern_b()
#endif

cIoType := AppIoMode()           // = G/T/B, set in ?AppWindow{}
cHowDetected := "corresponding to -io=" + lower(cIoType) + " compilation"
  
/**************************************************************************
 * The initio.prg was compiled with -io=a switch.
 * If the main program of the application was compiled for a
 * specific i/o mode, i.e. with the "io=g|t|b" switch, this
 * mode is also used for the rest of the application. Otherwise,
 * the default i/o mode can also be stored in the setup file or 
 * may be overriden by the command-line IO=g/t/b parameter.
 */
#ifdef FSIO_ALL

   extern_g()
   extern_t()
   extern_b()
   
   cIoType := aSetupData[IDX_ENVIR]      // last saved: G/T/B or ""
   cHowDetected := "read from the setup file"
 
   // the "io=g|t|b" or "-io=g|t|b" command line parameter
   // overrides the setting read from the setup file
   ii := ascan(aArgs, {|x| upper(left(x,4)) == "-IO=" })
   if ii > 0
     pos := if (ii > 0, 5, 0)
   else
     ii := ascan(aArgs, {|x| upper(left(x,3)) == "IO=" })
     pos := if (ii > 0, 4, 0)
   endif
   if ii > 0
      cIoType := upper(left(ltrim(substr(aArgs[ii],pos)),1))
      cHowDetected := "according to command-line parameter " + aArgs[ii] 
   endif
   
   if _MainCompiledBy() $ "GTB"    // Highest preference, overrides other 
      cIoType = _MainCompiledBy()  // settings: the main program was 
      cHowDetected := "corresponding to main prg compilation"
   endif                       // compiled with the "io=g|t|b" switch

   if !(upper(cIoType) $ "GTB")
      cHowDetected := "detected from the current environment"
      /*
       * The applic environment was not given by IO=... nor in setup
       * and the main app program was compiled with -io=a[uto] switch.
       * Will determine the current environment automatically:
       * - using GUI mode if the X/server or MS-Window32 is running,
       * - otherwise using Terminal mode for Unix and Basic for MS-Win
       */
      cIoDefault := "T"               // Unix w/o GUI = terminal mode
#     ifdef FS_WIN32
      cIoDefault := "B"               // MS-Win has GUI and Basic modes only
#     endif
      cIoType := if(IsGuiMode(), "G", cIoDefault)   
   endif 
     
   do case
   case cIoType == "G"
      CONSTANT oApplic := _gAppWindow{NIL, ;
                            aSetupData[IDX_TOPROW], aSetupData[IDX_TOPCOL], ;
                            aSetupData[IDX_HEIGHT], aSetupData[IDX_WIDTH],  ;
                            aSetupData[IDX_TITLE] , aSetupData[IDX_MODE]  }
   case cIoType == "T"
      CONSTANT oApplic := _tAppWindow{}    
   otherwise
      cIoType := "B"
      CONSTANT oApplic := _bAppWindow{}
   endcase    
   
   do case
   case cIoType == "G"
       CONSTANT _oMouse     := _gMouse{}      // for Mouse* functions
       CONSTANT _oPrinter   := _gPrinter{}
   case cIoType == "T"
       CONSTANT _oMouse     := _tMouse{}      
       CONSTANT _oPrinter   := _tPrinter{}
   otherwise
       CONSTANT _oMouse     := _bMouse{}      
       CONSTANT _oPrinter   := _bPrinter{}
   endcase    

   CONSTANT _oStatusBar := m->oApplic:StatusBar
#endif      /* FSIO_ALL */

/* display the used i/o type, when -TESTIO commandline param given
 */   
ii := ascan(aArgs, {|x| upper(x) == "-TESTIO" })
if ii > 0
   OutErr(chr(10) + "Note: i/o mode used = " + AppIoMode() + ;
                    ", " + cHowdetected + chr(10)) 
   sleep(3)      
endif

/* Set some public constants and protected variables
 */
// CONSTANT    FlagShip   := .T.      // already set internally
PROTECT PUBLIC Clipper    := .F.      // for Clipper compatibility check
PROTECT PUBLIC GetList    := {}       // used for @..GET..
PROTECT PUBLIC _oPrompt   := NIL      // used for @..PROMPT, MENU TO when not LOCAL
PROTECT PUBLIC _oAchoice  := NIL      // used for read-only Achoice() 

CONSTANT oAppWindow := m->oApplic     // alternative var name

/* You may add additional settings here when required, 
 * e.g. set your own window size (default is 80x25).
 * The window is still invisible.  
 * Note: better is to do this at the start of your applic or in a INIT procedure.
 */
* m->oApplic:Resize(300, 400, .T.)                      // 300x400 pixels
* m->oApplic:Move((oApplic:DesktopHeight - 300)/2, ;
*                 (oApplic:DesktopWidth  - 400)/2, .T.) // center

/* enable keyboard input 
 */
SET(_SET_INPUT, .T.)  // not mandatory, this is the default

/* If you do NOT want to create a default printer spooler file in the
 * current or the $FSOUTPUT directory, disable (comment out) the 
 * CALL fgsSetSpoolOn statement below - or start with -noSpool flag. 
 * If disabled, you will need to invoke SET PRINTER TO ... to be able  
 * to pass an output to a printer file e.g. via SET PRINT ON or 
 * SET DEVICE TO PRINT 
 * For a backward compatibility to FS 4.4*, you may alternatively invoke
 * CursesInit() which usually contains this CALL fgsSetSpoolOn
 *
 * Note: we will temprarily redirect the std. error handler here
 * to display a clean error message and advanced user help.
 */
ii := ascan(aArgs, {|x| upper(left(x,8)) == "-NOSPOOL" })
if ii == 0
   static bStdErr as usual
   bStdErr := ErrorBlock({|oErr| ErrSpoolFile(oErr, bStdErr) }) // disable std. error handler
   CALL fgsSetSpoolOn     // or: CursesInit() 
   ErrorBlock(bStdErr)    // enable std. error handler again
endif

#ifdef USE_MENUBAR
   /* 
    * Create the default menu bar
    * The source is available in initiomenu.prg
    */
   #ifdef FSIO_ALL
      if AppIoMode() == "G"
         InitIoMenu(m->oApplic)      // general, compiled with -io=a
      endif
   #endif
   #ifdef FSIO_GUI
      gInitIoMenu(m->oApplic)     // GUI based applic
   #endif
// #ifdef FSIO_TERM
//    tInitIoMenu(m->oApplic)     // terminal based applic
// #endif
#endif

#ifdef USE_EXITCALLBACK
   /* Assign the User-Exit-Request application callback to the InitIoQuit() 
    * function available in the user modifiable initiomenu.prg - or assign
    * your own UDF here (you may redefine it also later via codeblock).
    * The event is emitted by user click on the close [X] window button
    * or the ^K key, i.e. it behaves same as menu click File->Exit (see 
    * initiomenu.prg)
    */
   PROTO FUNCTION InitIoQuit(p1, [p2])                    // in initiomenu.prg
   m->oApplic:NotifyClose := {|event| InitIoQuit(event) } // event = APP_EV_CLOSE or APP_CTRL_K, see applic.fh
#endif

m->oApplic:Display()    // Show the window now

/* Create StatusBar Clock (optional) right aligned
 * and Default Message (mandatory)   left  aligned
 */
if AppIoMode() == "G"
   iClockWidth := m->oApplic:ApplicFont:WidthChar("99:99:99") +3  // StrLen2Pix() not avail yet at this stage
   iStatBarLng := m->oApplic:CurrSize(APP_COLS_SIZE,.T.) - ;
                  _aGlobSetting[GSET_G_N_STATBARRIGHT]            // 28, consider edge
else
   iClockWidth := 8
   iStatBarLng := MaxCol(.T.) +1 
endif     
#ifdef FSIO_GUI
   oBoxClock := _gStatBarItem{SBITEM_CLOCK,      "", "ClockItem",   iStatBarLng - iClockWidth, iClockWidth, .T.}
   oBoxMsg   := _gStatBarItem{SBITEM_DEFMESSAGE, "", "DefMsgItem", 0, oBoxClock:Col(NIL,.T.), .T.}
#endif
#ifdef FSIO_TERM
   oBoxClock := _tStatBarItem{SBITEM_CLOCK,      "", "ClockItem", iStatBarLng - iClockWidth, iClockWidth, .T.}
   oBoxMsg   := _tStatBarItem{SBITEM_DEFMESSAGE, "", "DefMsgItem", 0, oBoxClock:Col(NIL,.T.) -1, .T.}
#endif
#ifdef FSIO_BASIC
   oBoxClock := _bStatBarItem{SBITEM_CLOCK,      "", "ClockItem", iStatBarLng - iClockWidth, iClockWidth, .T.}
   oBoxMsg   := _bStatBarItem{SBITEM_DEFMESSAGE, "", "DefMsgItem", 0, oBoxClock:Col(NIL,.T.) -1, .T.}
#endif
#ifdef FSIO_ALL
   oBoxClock := StatBarItem{SBITEM_CLOCK,      "", "ClockItem", iStatBarLng - iClockWidth, iClockWidth, .T.}
   oBoxMsg   := StatBarItem{SBITEM_DEFMESSAGE, "", "DefMsgItem", 0, oBoxClock:Col(NIL,.T.) -1, .T.}
#endif
oBoxClock:Style := SBITEM_FLAT   // altern: SBITEM_SUNKEN or SBITEM_RAISED
oBoxClock:ClockInterval := 1     // tick every 1 second
m->_oStatusBar:AddItem(oBoxClock)

oBoxMsg:Style := SBITEM_FLAT     // altern: SBITEM_SUNKEN or SBITEM_RAISED
m->_oStatusBar:AddItem(oBoxMsg)

/* Display status bar in GUI mode. 
 * If this is desired also in Terminal i/o, activate the status bar in 
 *    your .prg via m->_oStatusBar:Show()
 * Note: you may re-define the status bar in your application, clear
 *    the already set items via m->_oStatusBar:Clear()
 */
if AppIoMode() == "G"
   m->_oStatusBar:Show()
endif

/* Load the default sorting table
 * The FSsortab.def is usually located in /usr/lib/terminfo directory
 * and is used mainly for lower/upper conversion of chars > 127 
 */
_SetSortab()   // available and redefinable in setsortab.prg

m->_InitIoDone := .T.   // use default error handler now
return NIL

/*--------------------------------------------------------------------------
 * Global settings and adjustments: 
 * you may redefine these settings at any tame later in your application 
 * (except the by (*) marked items which are available later via oApplic 
 * object) by a simple assignment, e.g.
 *   _aGlobSetting[GSET_T_C_COLDBOX] := B_PLAIN  // = "+-+|+-+|"
 * The array-element constants GSET_* are specified in set.fh
 *
 * Note: additional global settings are available via the SET command 
 * as well as Set() or fs_set() functions.
 */
STATIC FUNCTION GlobSetting()
PROTECT PUBLIC _aGlobSetting := array(GSET_ARRAYSIZE)
#ifdef FS_WIN32
_aGlobSetting[GSET_N_WAITCLOSEWIND ] := 5          // wait seconds before exit/close window
#else
_aGlobSetting[GSET_N_WAITCLOSEWIND ] := 0          // wait seconds before exit/close window
#endif
_aGlobSetting[GSET_C_WAITPROMPT    ] := "Press any key to continue ... " // WAIT prompt
_aGlobSetting[GSET_C_AUTOLOCK_PROT ] := ""         // Autolock: name of protocol file or empty 
_aGlobSetting[GSET_N_AUTOLOCK_INFO ] := 5          // Autolock: display retry info every n Seconds, 0 = no msg 
_aGlobSetting[GSET_N_AUTOLOCK_INFOWT]:= 2          // Autolock: Delay for display info message in Seconds 
_aGlobSetting[GSET_N_AUTOLOCK_SLEEP] := 150        // Autolock: Wait for retry in Millisec 
_aGlobSetting[GSET_C_TBRMEMOEMPTY  ] := "<Memo>"   // Tbrowse, empty Memo field
_aGlobSetting[GSET_C_TBRMEMODATA   ] := "<MEMO>"   // Tbrowse, Memo field with data
_aGlobSetting[GSET_L_ROUNDROWCOL   ] := .F.        // round row(), col() to int
_aGlobSetting[GSET_L_ROUNDMAXROWCOL] := .T.        // round maxrow(), maxcol() to int
_aGlobSetting[GSET_L_ACHOICECONFIRM] := .T.        // .T. considers SET CONFIRM ON/OFF in Achoice() 
_aGlobSetting[GSET_L_ACHOICEABORTLR] := .F.        // .T. aborts Achoice() via Cursor Ri/Le
_aGlobSetting[GSET_L_ACHOICESPACESEL]:= .T.        // .T. handles space in Achoice() same as Enter
_aGlobSetting[GSET_L_ACHOICEIGNEMPTY]:= .F.        // .T. ignores empty Achoice arrays, same as Clipper
_aGlobSetting[GSET_L_ACHOICEHOTKEY ] := .T.        // .F. ignores &x hotkeys in Achoice, same as Clipper
_aGlobSetting[GSET_L_ACHOICEUDFSEARCH] := .F.      // search for hotkey/1st letter in Achoice with UDF
_aGlobSetting[GSET_L_ACHOICESINGLE ] := .T.        // Achoice single mouse click = Enter
_aGlobSetting[GSET_L_PROMPTSPACESEL] := .T.        // handle Space in MENU TO same as Enter
_aGlobSetting[GSET_L_PROMPTSINGLE  ] := .T.        // single mouse click in MENU TO = Enter
_aGlobSetting[GSET_L_ADAPT_PICT    ] := .T.        // allow adapt picture if number too large
_aGlobSetting[GSET_L_DBCREATE_CLIP ] := .F.        // .T. = ignores 4th element in DbCreate() array for "C" type
_aGlobSetting[GSET_L_FOUND_SETORDER] := .F.        // hold Found() status during SET ORDER 
_aGlobSetting[GSET_C_SETKEYFILTER  ] := "|_TMEMOEDIT|_GMEMOEDIT|"  // filter of proc names for SET KEY procedure 
// _aGlobSetting[GSET_C_SETKEYFILTER]:= "|_TMEMOEDIT|_GMEMOEDIT|GETDOSETKE|GETAPPLYKE|GETREADER|READMODAL|"  // with re-compiled getsys.prg 
_aGlobSetting[GSET_L_SOFTCR_WITH_LF] := .T.        // soft-cr must include LF for HardCr()
_aGlobSetting[GSET_L_REMCR_VCFIELD ] := .T.        // remove hard and soft-CR at storing VC* field 
_aGlobSetting[GSET_L_BINARY0_VFIELD] := .T.        // allow \0 in V*Z field
_aGlobSetting[GSET_L_READVAR_HOLD  ] := .F.        // hold last GET name for ReadVar() outside READ 
_aGlobSetting[GSET_L_READ_FAST     ] := .F.        // use alternative READ algorithm in getsys.prg 
_aGlobSetting[GSET_N_FOPEN_CREATE  ] := 0          // fopen() par3 0:don't create new file, 1:create, 2:create in write mode 

_aGlobSetting[GSET_N_DBCOMMIT      ] := 4          // 1:asynchronous (def), 2:synchronous, 4:flush (def Win), 6:synchr+flush 
_aGlobSetting[GSET_N_DBCOMMITALL   ] := 4          // 1:asynchronous, 2:synchronous (def), 4:flush (def Win), 6:synchr+flush 
_aGlobSetting[GSET_N_DBCOMMIT_FLUSH] := 3.0        // dbcommit() must perform full flush latest after n seconds, 0=always 
_aGlobSetting[GSET_L_DBCOMMIT_EXCL ] := .F.        // commit and dbcommit() also on exclusive database
_aGlobSetting[GSET_N_UNLOCK_COMMIT ] := 0          // 0:don't commit (def), 1:DbCommit(), 2:DbCommitAll() 
_aGlobSetting[GSET_N_SKIP_COMMIT   ] := 1          // 0:don't commit, 1:DbCommit()=def, 2:DbCommitAll() - yet unused
_aGlobSetting[GSET_L_DBCOPY_LOCK   ] := .F.        // .T. = Flock() source database on COPY TO with SET AUTOLOCK ON
_aGlobSetting[GSET_L_SECURE_COPY   ] := .F.        // .T. = use secure COPY TO and PACK with multiple memo fields
_aGlobSetting[GSET_L_NETERR_USED   ] := .T.        // should neterr() be set if !used() ? 
_aGlobSetting[GSET_N_LOCK_TIMEOUT  ] := 4900       // timeout retry lock in millisec. Increase if RTE347 occurs.
_aGlobSetting[GSET_L_RESTSCR_CLS   ] := .T.        // clear screen region before RestScreen() 

_aGlobSetting[GSET_C_NEWLINE_BASIC ] := chr(13,10) // new line qout() on screen with -io=b
#ifdef FS_WIN32
_aGlobSetting[GSET_C_NEWLINE_PRINT ] := chr(13,10) // new line qout() on printer (Windows)
#else
_aGlobSetting[GSET_C_NEWLINE_PRINT ] := chr(10)    // new line qout() on printer (Linux)
#endif
_aGlobSetting[GSET_L_BOXES_POPUP   ] := .F.        // use popup for InfoBox() etc, enable moving by spec.keys
_aGlobSetting[GSET_N_BOXES_ROWS    ] := 2          // default rows moving InfoBox() etc
_aGlobSetting[GSET_N_BOXES_COLS    ] := 5          // default cols moving InfoBox() etc
_aGlobSetting[GSET_A_BOXES_KEYS    ] := {{K_ALT_LEFT,"L"},{K_CTRL_LEFT,"L"},{K_ALT_RIGHT,"R"},{K_CTRL_RIGHT,"R"},{K_ALT_UP,"U"},{K_CTRL_PGUP,"U"},{K_ALT_DOWN,"D"},{K_CTRL_PGDN,"D"},{K_ALT_HOME,"T"},{K_CTRL_HOME,"T"},{K_ALT_END,"B"},{K_CTRL_END,"B"},{K_PLUS,"T"},{K_MINUS,"B"}}

_aGlobSetting[GSET_T_C_DEFCOLOR    ] := "W/N,N/W,W/N,W/N,N/W" // (*) default color
_aGlobSetting[GSET_T_L_BOX_SUBWIN  ] := .F.        // use sub-windows for InfoBox(), TextBox() etc. also in Terminal mode
_aGlobSetting[GSET_T_C_ERRBOX      ] := B_PLAIN    // RTE and errormsg frame  
_aGlobSetting[GSET_T_C_ALERTBOX    ] := B_PLAIN    // Alert(), MessageBox{}, InfoBox{} frame
_aGlobSetting[GSET_T_C_INFOBOX     ] := B_SINGLE   // InfoBox() frame
_aGlobSetting[GSET_T_C_TEXTBOX     ] := B_PLAIN    // TextBox() frame
_aGlobSetting[GSET_T_C_COLDBOX     ] := B_SINGLE   // Listbox, RadioGroup w/o focus 
_aGlobSetting[GSET_T_C_HOTBOX      ] := B_DOUBLE   // Listbox, RadioGroup w/  focus 
_aGlobSetting[GSET_T_C_COMBOMARK   ] := chr(31)    // Combo dropdown mark  
_aGlobSetting[GSET_T_C_ERRBOXCOLOR ] := "GR+/R,R/W"    // ErrorBox{} color
_aGlobSetting[GSET_T_C_INFBOXCOLOR ] := "W+/BG,BG+/GR" // InfoBox{} color
_aGlobSetting[GSET_T_C_MSGBOXCOLOR ] := "W+/B,B/W"     // MessageBox{} color
_aGlobSetting[GSET_T_C_TEXTBOXCOLOR] := "W+/B,B/W"     // TextBox{} color
_aGlobSetting[GSET_T_C_WARBOXCOLOR ] := "GR+/B,B/W"    // WarningBox{} color
_aGlobSetting[GSET_T_C_AT_TO_SINGLE] := B_SINGLE   // @..TO.. SINGLE 
_aGlobSetting[GSET_T_C_AT_TO_DOUBLE] := B_DOUBLE   // @..TO.. DOUBLE 
_aGlobSetting[GSET_T_C_MENUBORDER  ] := B_SINGLE   // PopUp Menu border 
_aGlobSetting[GSET_T_C_MENUSEPAR   ] := SEPARATOR_SINGLE  // PopUp Menu separator, see button.fh
_aGlobSetting[GSET_T_C_PUSHB_STYLE ] := "<>"       // PushButton default style
_aGlobSetting[GSET_T_C_CHBOX_STYLE ] := "[ X]?"    // CheckBox default style
_aGlobSetting[GSET_T_C_RADBUT_STYLE] := "( *)"     // RadioButton default style
_aGlobSetting[GSET_T_C_SBAR_SEPAR  ] := chr(186)   // StatusBar separator
_aGlobSetting[GSET_T_C_SBAR_SEPCOLR] := ""         // StatusBar separator color = unchanged
_aGlobSetting[GSET_T_L_ACHOICE_LBOX] := .T.        // use extended Achoice() via ListBox
_aGlobSetting[GSET_T_A_MEMOE_SHOW  ] := {}         // MemoEdit() in Terminal i/o: don't show wrap
// _aGlobSetting[GSET_T_A_MEMOE_SHOW]:= {1," ","+","<","G+"} // MemoEdit() in Terminal i/o: show wrap: col rel rght, std, soft, hard, color
_aGlobSetting[GSET_T_L_MEMOE_SOFTCR] := .T.        // MemoEdit() end in Terminal i/o: T=leave chr(141)+chr(10), F=remove
_aGlobSetting[GSET_G_L_MEMOE_SOFTCR] := .T.        // MemoEdit() end in GUI i/o: T=leave chr(141)+chr(10), F=remove
_aGlobSetting[GSET_T_N_MEMOE_TUNE  ] := 127        // MemoEdit() Tuning in Terminal i/o = 1+2+4+8+16+32+64

_aGlobSetting[GSET_A_MEMO_GET_GLYPH1]:= {128,255}          // GET/READ,MemoEdit() with SET MULTIBYTE ON: glyph1 range  
_aGlobSetting[GSET_A_MEMO_GET_GLYPH2]:= {{1,31},{33,255}}  // GET/READ,MemoEdit() with SET MULTIBYTE ON: glyph2 range 
_aGlobSetting[GSET_N_MEMO_GET_MULTIB]:= K_CTRL_A   // GET/READ,MemoEdit() with SET MULTIBYTE ON: toggle low/high ascii 
_aGlobSetting[GSET_N_MEMO_GET_MULTIW]:= 10.0       // GET/READ,MemoEdit() with SET MULTIBYTE ON: inkey() wait for 2nd byte 
_aGlobSetting[GSET_N_MEMO_GET_BEEP1] := 1          // GET/READ,MemoEdit() with SET MULTIBYTE ON: beep n-times after 1sec 
_aGlobSetting[GSET_N_MEMO_GET_BEEP2] := 2          // GET/READ,MemoEdit() with SET MULTIBYTE ON: beep n-times after time-out expires
_aGlobSetting[GSET_L_MEMOE_SOFTHARD] := .F.        // MemoEdit() start: .T.=Soft-CR->Hard-CR, .F.=remove chr(141)+chr(10)
_aGlobSetting[GSET_L_MEMOE_HARDCRLF] := .T.        // MemoEdit() end: .T.=Hard-CR->CR+LF, .F.=Hard-CR->LF
_aGlobSetting[GSET_L_MEMOE_FAST    ] := NIL        // MemoEdit() .T. use fast Inkey() algorithm, NIL = GSET_L_READ_FAST */
_aGlobSetting[GSET_A_COPY_TO_MEMO  ] := {" ",chr(20)}  // COPY TO DELIM/SDF replacement for soft-CR, hard-CR in Memo-Field
#ifdef FS_WIN32
_aGlobSetting[GSET_C_COPY_TO_NEWLINE]:= chr(13,10) // COPY TO DELIM/SDF replacement for record separator = CR+LF
#else
_aGlobSetting[GSET_C_COPY_TO_NEWLINE]:= chr(10)    // COPY TO DELIM/SDF replacement for record separator = LF
#endif

_aGlobSetting[GSET_G_N_WINSIZE_COL ] := 80         // (*) start window size columns 
_aGlobSetting[GSET_G_N_WINSIZE_ROW ] := 25         // (*) start window size rows
_aGlobSetting[GSET_G_N_WINBORDER_X ] := 25         // Window manager horiz border pixel
_aGlobSetting[GSET_G_N_WINBORDER_Y ] := 55         // Window manager vert  border pixel
_aGlobSetting[GSET_G_N_WINFRAME_X  ] := 7          // Window manager left + rigt frame pixel
_aGlobSetting[GSET_G_N_WINFRAME_Y  ] := 27         // Window manager top + bottom frame pixel
_aGlobSetting[GSET_G_N_MINSIZE_COL ] := 100        // (*) minimal x-window size pixel  
_aGlobSetting[GSET_G_N_MINSIZE_ROW ] := 100        // (*) minimal y-window size pixel  
_aGlobSetting[GSET_G_N_MAXSIZE_COL ] := 2000       // maximal x-window size pixel (100..32100) 
_aGlobSetting[GSET_G_N_MAXSIZE_ROW ] := 2000       // maximal y-window size pixel (100..32100) 
_aGlobSetting[GSET_G_C_COL_MAXCHAR ] := "358AMX"   // column width = largest char of ...  
_aGlobSetting[GSET_G_N_ROW_SPACING ] := 2          // row = font:LineHeight() + this pixel value 
_aGlobSetting[GSET_G_N_DRAWLINE    ] := 1          // default line width in pixel for @..DRAW..
_aGlobSetting[GSET_G_N_DRAWBOLD    ] := 3          // default bold line width in pixel
_aGlobSetting[GSET_G_L_ADD_SP_HTML ] := .F.        // if .T.: add space to <html> tag 
#ifdef FS_WIN32
_aGlobSetting[GSET_G_N_REFRESHDRAW ] := 300        // refresh graphics after delay of Millisec (0: now, -1: don't refresh)
#else
_aGlobSetting[GSET_G_N_REFRESHDRAW ] := -1         // auto-refresh not required with X11
#endif
_aGlobSetting[GSET_G_N_KEYEVENT    ] := 2          // react on 1:key press event, 2:key release event
_aGlobSetting[GSET_G_N_TEXTSHAPE   ] := -21        // text cursor shape for SET GUICURSOR ON = CURSOR_UNDERSCORE
_aGlobSetting[GSET_G_N_WAITSHAPE   ] := -6         // text cursor shape for WAIT in GUI mode = CURSOR_HAND
_aGlobSetting[GSET_G_L_ROW_VISIBLE ] := .T.        // ensure Qqout/@..SAY output is visible (scroll vert)
_aGlobSetting[GSET_G_L_COL_VISIBLE ] := .F.        // ensure Qqout/@..SAY output is visible (scroll horiz)
_aGlobSetting[GSET_G_C_ERRBOXTITLE ] := "Error ..." // Title of error box 
// _aGlobSetting[GSET_G_AC_DEFCOLOR] := "N/W,N/W+,N/W+,N/W+,N/W,N/W+" // (*) default GUI color 
#ifdef FS_WIN32
   _aGlobSetting[GSET_G_AC_DEFCOLOR] := "N/#ECE9D8,N/W+,N/W+,N/W+,N/W,N/W+" // (*) default GUI color Windows 16mio
#else
// _aGlobSetting[GSET_G_AC_DEFCOLOR] := "N/#DEDEDE,N/W+,N/W+,N/W+,N/W,N/W+" // (*) default GUI color Linux
   _aGlobSetting[GSET_G_AC_DEFCOLOR] := "N/#EEEEE6,N/W+,N/W+,N/W+,N/W,N/W+" // (*) default GUI color Linux
#endif
* _aGlobSetting[GSET_G_AC_DEFCOLOR ] := {{{RGBCOLOR_BLACK},{RGBCOLOR_GRAY_LIGHT}}, ;
*                                        {{RGBCOLOR_BLACK},{RGBCOLOR_WHITE}} ,     ; 
*                                        {{RGBCOLOR_BLACK},{RGBCOLOR_WHITE}} ,     ; 
*                                        {{0, 0, 0       },{255, 255, 255 }} ,     ; 
*                                        {{RGBCOLOR_BLACK},{RGBCOLOR_GRAY_LIGHT}}, ;
*                                        {{RGBCOLOR_BLACK},{RGBCOLOR_WHITE}} }
_aGlobSetting[GSET_G_N_STATBARRIGHT] := 28         // status bar end pixels from right corner

_aGlobSetting[GSET_G_N_GET_COPY1   ] := K_MBUTTONDOWN  // copy marked text to clipboard buffer
_aGlobSetting[GSET_G_N_GET_COPY2   ] := K_ALT_C        // copy - alternative key
_aGlobSetting[GSET_G_N_GET_PASTE1  ] := K_SH_MBUTTONDN // paste clipboard buffer to get field
_aGlobSetting[GSET_G_N_GET_PASTE2  ] := K_ALT_V        // paste - alternative key

_aGlobSetting[GSET_G_N_MEMO_COPY1  ] := NIL        // copy : use [GSET_G_N_GET_COPY1 ]
_aGlobSetting[GSET_G_N_MEMO_COPY2  ] := NIL        // copy : use [GSET_G_N_GET_COPY2 ]
_aGlobSetting[GSET_G_N_MEMO_PASTE1 ] := NIL        // paste: use [GSET_G_N_GET_PASTE1]
_aGlobSetting[GSET_G_N_MEMO_PASTE2 ] := NIL        // paste: use [GSET_G_N_GET_PASTE2]
_aGlobSetting[GSET_G_C_MEMO_COLOR  ] := ""         // MemoEdit() color in GUI: empty or e.g."#RRGGBB/#RRGGBB[,,#RRGGBB]" 
_aGlobSetting[GSET_G_L_MEMO_COLORON] := .F.        // MemoEdit() consider SET GUICOLOR ON 

_aGlobSetting[GSET_G_L_LISTBOX_ADJ ] := .T.        // use ListBox,Achoice() row/col adjustment?
_aGlobSetting[GSET_G_L_LISTBOX_BOX ] := .T.        // .T. ListBox coordinates are frame, .F. = inner box
_aGlobSetting[GSET_G_L_MEMOEDIT_ADJ] := .T.        // use MemoEdit() row/col adjustment?
_aGlobSetting[GSET_G_N_LISTBOX_TOP ] := -2         // adjust pix ListBox,Achoice(),MemoEdit() top 
_aGlobSetting[GSET_G_N_LISTBOX_BOT ] := 2          // adjust pix ListBox,Achoice(),MemoEdit() bottom
_aGlobSetting[GSET_G_N_LISTBOX_LEFT] := -7         // adjust pix ListBox,Achoice(),MemoEdit() left
_aGlobSetting[GSET_G_N_LISTBOX_RIGH] := 6          // adjust pix ListBox,Achoice(),MemoEdit() right
_aGlobSetting[GSET_G_A_LISTBOX_MOUSE ] := {K_LBUTTONUP,K_MBUTTONUP,K_RBUTTONUP} // select Listbox by mouse click  
_aGlobSetting[GSET_G_L_LISTBOX_SINGLE] := .T.      // select Listbox by sigle mouse click 
_aGlobSetting[GSET_G_N_COMBO_HEIGHT] := 4          // add pix to closed combo line height

_aGlobSetting[GSET_G_L_CHBOX_ADJ   ] := .T.        // use CheckBox row/col adjustment?  
_aGlobSetting[GSET_G_N_CHBOX_COL   ] := 2          // adjust pix left column of CheckBox 
_aGlobSetting[GSET_G_N_CHBOX_ROW   ] := -3         // adjust pix upper row of CheckBox 
_aGlobSetting[GSET_G_N_CHBOX_WIDTH ] := 20         // add pix CheckBox width
_aGlobSetting[GSET_G_N_CHBOX_HEIGHT] := 0          // adjust pix height of RadioButton 
_aGlobSetting[GSET_G_N_CHBOX_CAPCOL] := 10         // CapCol caption pix dist. if auto
_aGlobSetting[GSET_G_L_CHBOX_STYLE ] := .F.        // consider :Style = add 1 column if :Style is set
_aGlobSetting[GSET_G_A_CHBOX_MOUSE ] := {K_LBUTTONUP,K_MBUTTONUP,K_RBUTTONUP} // select CheckBox by mouse click 
_aGlobSetting[GSET_G_L_CHBOX_SINGLE] := .T.        // select CheckBox by sigle mouse click 
 
_aGlobSetting[GSET_G_L_RADBUT_ADJ  ] := .T.        // use RadioButton row/col adjustment?  
_aGlobSetting[GSET_G_N_RADBUT_COL  ] := 2          // adjust pix left column of RadioButton
_aGlobSetting[GSET_G_N_RADBUT_ROW  ] := -3         // adjust pix upper row of RadioButton 
_aGlobSetting[GSET_G_N_RADBUT_WIDTH] := 20         // add pix RadioButton width 
_aGlobSetting[GSET_G_N_RADBUT_HEIGHT]:= 0          // adjust pix RadioButton height 
_aGlobSetting[GSET_G_N_RADBUT_CAPCOL]:= 10         // CapCol caption pix dist. if auto
_aGlobSetting[GSET_G_L_RADBUT_STYLE ]:= .F.        // consider :Style = add 1 column if :Style is set
_aGlobSetting[GSET_G_A_RADBUT_MOUSE ]:= {K_LBUTTONUP,K_MBUTTONUP,K_RBUTTONUP} // select RadioButton by mouse click
_aGlobSetting[GSET_G_L_RADBUT_SINGLE]:= .T.        // select RadioButton by sigle mouse click 

_aGlobSetting[GSET_G_L_RADGRP_ADJ  ] := .T.        // use RadioGroup row/col adjustment?  
_aGlobSetting[GSET_G_N_RADGRP_TOP  ] := 0          // adjust pix RadioGroup top 
_aGlobSetting[GSET_G_N_RADGRP_BOT  ] := 0          // adjust pix RadioGroup bottom 
_aGlobSetting[GSET_G_N_RADGRP_LEFT ] := 0          // adjust pix RadioGroup left 
_aGlobSetting[GSET_G_N_RADGRP_RIGH ] := 0          // adjust pix RadioGroup right 
_aGlobSetting[GSET_G_N_RADGRP_ADCOL] := 0          // adjust pix RadioGroup frame left/right if auto
_aGlobSetting[GSET_G_N_RADGRP_ADROW] := 0          // adjust pix RadioGroup frame top/bottom if auto
_aGlobSetting[GSET_G_N_RADGRP_BUTCOL]:= 0          // adjust pix RadioButton columns
_aGlobSetting[GSET_G_N_RADGRP_BUTROW]:= 7          // adjust pix RadioButton rows
_aGlobSetting[GSET_G_A_RADGRP_MOUSE ]:= {K_LBUTTONUP,K_MBUTTONUP,K_RBUTTONUP} // select RadioGroup by mouse click
_aGlobSetting[GSET_G_L_RADGRP_SINGLE]:= .T.        // select RadioGroup by sigle mouse click 

_aGlobSetting[GSET_G_L_MENUTO_LMB  ] := .T.        // MenuTo: consider Left  Mouse Button
_aGlobSetting[GSET_G_L_MENUTO_MMB  ] := .T.        // MenuTo: consider Mid   Mouse Button
_aGlobSetting[GSET_G_L_MENUTO_RMB  ] := .T.        // MenuTo: consider Right Mouse Button
_aGlobSetting[GSET_G_L_MENUTO_DLMB ] := .T.        // MenuTo: consider double Left  Mouse Button
_aGlobSetting[GSET_G_L_MENUTO_DMMB ] := .F.        // MenuTo: consider double Mid   Mouse Button
_aGlobSetting[GSET_G_L_MENUTO_DRMB ] := .F.        // MenuTo: consider double Right Mouse Button
_aGlobSetting[GSET_G_L_MENUTO_WHEEL] := .T.        // MenuTo: consider Mouse Wheel

_aGlobSetting[GSET_G_L_TBROW_ADJ   ] := .T.        // use Tbrowse row/col adjustment?  
_aGlobSetting[GSET_G_N_TBROW_TOP   ] := -3         // adjust pix Tbrowse top row
_aGlobSetting[GSET_G_N_TBROW_BOT   ] := 0          // adjust pix Tbrowse bottom row
_aGlobSetting[GSET_G_N_TBROW_LEFT  ] := 0          // adjust pix Tbrowse left col
_aGlobSetting[GSET_G_N_TBROW_RIGH  ] := 0          // adjust pix Tbrowse right col
_aGlobSetting[GSET_G_N_TBROW_RHEIGHT]:= 0          // adjust pix Tbrowse row height 
_aGlobSetting[GSET_G_N_TBROW_CWIDTH ]:= 4          // adjust pix TbColumn width 
_aGlobSetting[GSET_G_N_TBROW_COLADJ] := 0          // 0=center large field, >0 adjust pixel left
_aGlobSetting[GSET_G_N_TBROW_WAIT  ] := -9         // cursor shape during stabilize, 0=no change
_aGlobSetting[GSET_G_A_TBROW_VSBUTT] := {{1,2,2,2},{2,2,2,2},{2,2,2,2},.T.} // Tbrowse vert. Scrollbar button handling left/mid/right mouse

_aGlobSetting[GSET_G_N_ALERT_ALIGN ] := 0          // Alert, Infobox etc alignment (0=def, auto)
_aGlobSetting[GSET_G_N_ALERT_WIDTH ] := 0          // Alert, Infobox etc max width pixel (0=def, auto)
_aGlobSetting[GSET_G_N_ALERT_HEIGHT] := 0          // Alert, Infobox etc max height pixel (0=def, auto)
_aGlobSetting[GSET_G_O_ALERT_FONT  ] := NIL        // Alert, Infobox etc font (NIL=def, oApplic:FontWindow)

_aGlobSetting[GSET_L_GET_HOME2CHAR ] := .T.        // oGet:Home2char property default setting 
_aGlobSetting[GSET_L_GET_END2CHAR  ] := .T.        // oGet:End2char  property default setting 
_aGlobSetting[GSET_G_L_GET_ADJ     ] := .T.        // use Get row/col adjustment? 
_aGlobSetting[GSET_G_N_GET_ROW     ] := 0          // adjust pix Get row 
_aGlobSetting[GSET_G_N_GET_COL     ] := -2         // adjust pix Get column 
_aGlobSetting[GSET_G_N_GET_WIDTH   ] := 8          // adjust pix active Get width (< 8 will scroll)
_aGlobSetting[GSET_G_N_GET_HEIGHT  ] := 0          // adjust pix Get height 
_aGlobSetting[GSET_G_L_GET_MOUSENEW] := .T.        // should mouse click on other field select it and trigger oGet:Home() ? (getsys.prg)
_aGlobSetting[GSET_G_L_GET_MOUSEHOME]:= .F.        // should mouse click on GET always trigger oGet:Home() ? (getsys.prg)
_aGlobSetting[GSET_G_L_GET_MOUSEOUT] := .F.        // should mouse click allow positioning behind last char in GET ? (getsys.prg)

_aGlobSetting[GSET_G_L_PRINT_EJECT ] := .F.        // perform eject before _oPrinter:Exec() 
#ifdef FS_WIN32
_aGlobSetting[GSET_G_L_PRINTF_EJECT] := .T.        // Window: perform eject before _oPrinter:ExecFormatted() 
_aGlobSetting[GSET_G_N_PRINTF_FIXPG] := 23         // Window: fix alignment by 23pixels for _oPrinter:ExecFormatted() */
#else
_aGlobSetting[GSET_G_L_PRINTF_EJECT] := .F.        // Linux: don't perform eject before _oPrinter:ExecFormatted() 
_aGlobSetting[GSET_G_N_PRINTF_FIXPG] := 0          // Linux: don't change alignment for _oPrinter:ExecFormatted() */
#endif
_aGlobSetting[GSET_L_ACCEPT_WHEEL  ] := .T.        // accept mouse wheel in GUI mode
_aGlobSetting[GSET_N_WHEEL_STEPS   ] := 4          // wheel driver reports <n> (n >= 1) steps

#ifdef FS_WIN32
_aGlobSetting[GSET_N_CLOSEOPTIMIZE ] := 0          // for future use
_aGlobSetting[GSET_N_COMMITOPTIMIZE] := 0          // for future use
#else
_aGlobSetting[GSET_N_CLOSEOPTIMIZE ] := 1          // for future use
_aGlobSetting[GSET_N_COMMITOPTIMIZE] := 1          // for future use
#endif
_aGlobSetting[GSET_L_WINCMDDETECT  ] := .T.        // detect internal CMD commands for RUN automat.
_aGlobSetting[GSET_L_RUNDISPLAY    ] := .F.        // display RUN command on console
_aGlobSetting[GSET_L_RUNRTERROR    ] := .T.        // display RUN error as RTE

_aGlobSetting[GSET_A_TBROWSE_DISPLMSG]:= {.F.,.F.} // display {messages,deleted} in Tbrowse,DbEdit() 
_aGlobSetting[GSET_A_TBROWSE_MSG   ]  := ;         // messages (and color) for Tbrowse,DbEdit() in GUI
  /* 1*/ {{"Read-only table, browse by cursor or PgUp/PgDn, exit by Enter or Esc", NIL} ,; // text,color
  /* 2*/  {"Browse by cursor or PgUp/PgDn, <b>edit</b> by Enter, <b>exit</b> by Esc", NIL}  ,;
  /* 3*/  {"Browse by cursor or PgUp/PgDn, ", NIL}  ,;               // +[4,1] +[4,2] +[4,3] +[4,4]
  /* 4*/  {"<b>Edit</b>=enter, ", "<b>Add</b>=Ctrl-PgDn+CursDown, ",; 
           "<b>Delete</b>=DEL, ", "<b>Exit</b>=ESC"} ,;              // add to [3,1] if enabled
  /* 5*/  {"<b>Read-only data</b>, not modifiable", "R+/#FFFAB7"} ,;
  /* 6*/  {"<b>Read-only data</b>, browse by cursor or PgUp/PgDn, exit with Esc", "R+/#FFFAB7"} ,;
  /* 7*/  {"Edit Memo data, press Ctrl-W or Esc to finish", "N/#FFF398"} ,;
  /* 8*/  {"Enter or modify data, press Enter to save or Esc to discard changes","N/#FFF398"} ,;
  /* 9*/  {"<b>Deleted record.</b> Browse by cursor or PgUp/PgDn", NIL} ,;
  /*10*/  {"<b>Deleted record</b>, un-delete by DEL key. Browse by cursor or PgUp/PgDn", NIL} ,;
  /*11*/  {"Are you sure to delete record# ", {"No","Yes"}} ,;  // empty text = no prompt
  /*12*/  {"Undelete record# ", {"No","Yes"}} ,;                // empty text = no prompt
  /*13*/  {"", NIL} ,;                                // unused yet
  /*14*/  {"", NIL} ,;                                // clear status message {text, gui-color}
  /*15*/  {"", "N/W+,N/#FFF398"} ,;                   // GUI color for field editing
  /*16*/  {"", "N/#FFF398"} ,;                        // GUI color for Memo editing
  /*17*/  {"", NIL} ,;                                // unused yet
  /*18*/  {"", NIL} }                                 // unused yet
  
return NIL

#command ?# [ <list,...> ]  => outerr (chr(13)+chr(10)) ; outerr( <list> )
/*--------------------------------------------------------------------------
 * If i/o error occurs during setting the spooler file, 
 * print a help message to stderr and exit
 */
STATIC FUNCTION ErrSpoolFile(oErr, bStdErr)
   if valtype(oErr) == "O" .and. ;
		   (oErr:GenCode == 101 .or. oErr:GenCode == 105) 
      ?# "RTE i/o error", ltrim(oErr:GenCode),"occured at "
      ?# ProcStack(1),"in",lower(procfile())+".prg"
      ?# oErr:Description
      ?#
      ?# "At this point, it signals that you have insufficient write permission"
      ?# "to the current directory, where the temporarily spool file should be"
      ?# "created. Either set other directory (e.g. /tmp) for the spool file via"
      ?# "  FSOUTPUT=/tmp ; export FSOUTPUT"
#ifdef FS_WIN32
      ?# "or change user's access rights (read + write) for the current directory"
      ?# curdir()
#else
      ?# "or change the directory permissions via"
      ?# "  su ; chmod a+rwx " + curdir() + " ; exit"
#endif
      ?# "or start the application via '" + execname(.T.) + " -noSpool' and"
      ?# "assign SET PRINTER TO <filename> before you redirect output to printer." 
      ?#
      sleep(3)
      quit_hard()
      return
   endif
   if valtype(bStdErr) != "B" .or. valtype(oErr) != "O"
      ?# "Unexpected error"
      ?#
      quit_hard()
   endif
   eval(bStdErr, oErr)   // display all other errors by std.error handler
   return NIL
   
      
/*--------------------------------------------------------------------------
 * Reads the last window position or predefined defaults
 *
 * When the setup file name is not specified in "gcSetupFile" variable,
 * the default file name is "." + <ApplicName w/o extension>, located 
 *   a) in directory specified by envir.var FSAPPSETUP
 *   b) in the local directory
 *   c) in the directory where the executable resides (if path is available)
 *   d) in the user home directory
 * The name and location is then stored in the "gcSetupFile" public var,
 * if the file is available.
 *
 * The content of the ASCII setup file is here in free format: any/all of
 *   WINDOW_TOPROW=nnnn   (nnnn = pixels value)
 *   WINDOW_TOPCOL=nnnn   (nnnn = pixels value)
 *   WINDOW_HEIGHT=nnnn   (nnnn = pixels value)
 *   WINDOW_WIDTH=nnnn    (nnnn = pixels value)
 *   WINDOW_TITLE=cccc  -or-  WINDOW_TITLE"ccc ccc ccc ccc"
 *   WINDOW_MODE=MDI|SDI
 *   APPLIC_ENVIR=GUI|BASIC|TERM
 * The setup file may contain any other data as well, but only the above
 * tokens (case insensitive) will be considered by default here.
 *
 * You may, of course, freely modify the behaviour to fit your needs! 
 * Instead of MemoRead(), you may alternatively use FOPEN() and read the 
 * data into an array similarily to the SaveDefaults() code.
 * But do NOT use screen i/o here, since the API is not initialized yet.
 *--------------------------------------------------------------------------*/
STATIC FUNCTION ReadDefaults()

#ifdef USE_READDEFAULTS
MEMVAR gcSetupFile                              // from _*initio()
LOCAL cAppName, cAppPath, cFileName, cTmp, cData AS CHAR
LOCAL ii, idx       AS INTVAR
LOCAL lFound := .F. AS LOGICAL

if !empty(gcSetupFile)
    lFound := file(gcSetupFile)
else
    // the file name & location algorithm described above
    cAppName := ExecName(.F.)
    while !empty(cAppName) .and. left(cAppName,1) == "."
       cAppName := substr(cAppName,2)           // remove leading "."
    enddo
    ii := at(".", cAppName)                     // remove extension
    cAppName := if(ii > 1, left(cAppName, ii-1), cAppName)

    cAppPath := ExecName(.T.)
    ii := rat("/", cAppPath)                    // path where exec resides
    cAppPath := if(ii > 1, left(cAppPath,ii-1), "")

    if !empty(cAppName)
        cAppName := FILEPREFIX + cAppName       // e.g. ".myapp" or ".a"
        cTmp := GetEnv("FSAPPSETUP")
        if !empty(cTmp)                         // environment variable?
            cFileName := cTmp + "/" + cAppName 
            lFound := file(cFileName)
        endif
        if !lFound                              // local directory?
           cFileName := "./" + cAppName  
           lFound := file(cFileName)
        endif
        if !lFound .and. !empty(cAppPath)       // application path?
           cFileName := cAppPath + "/" + cAppName  
           lFound := file(cFileName)
        endif
        if !lFound                              // home directory?
           cFileName := "~/" + cAppName  
           lFound := file(cFileName)
        endif
        if lFound                               // something of above found,
           gcSetupFile := cFileName             // store to public var
        endif                                   // for use in SaveDefaults()
    endif
endif
if !lFound 
    return NIL                                  // file not found
endif  
/*
 * The here supported ASCII setup file is in free format,
 * but best is to set each <identifier>=<value> at separate line
 * since this is required for SaveDefaults(), see details there
 */
cData := MemoRead(gcSetupFile)
/*
 * Window coordinates topY, topX, height, width -- all data in pixel
 *   WINDOW_TOPROW=nnnn  WINDOW_TOPCOL=nnnn
 *   WINDOW_HEIGHT=nnnn  WINDOW_WIDTH=nnnn
 * if not given, a centered window for 80x25 (col/row) is calculated 
 */ 
for idx := 1 to 4                                // top...width
    ii := at(aSetupDescr[idx], cData)
    if ii > 0
        aSetupData[idx] := val(substr(cData, ii + len(aSetupDescr[idx])))
    endif
next
/*
 * Windows Title:
 *   WINDOW_TITLE=cccc -or- WINDOW_TITLE="ccc ccc ccc"
 * if not given, the name of the executabke is used otherwise
 */
ii := at(aSetupDescr[IDX_TITLE], cData)         // title
if ii > 0
    ii += len(SetupDescr[IDX_TITLE])
    cTmp := substr(cData, ii, 200)              // here upto 200 chars
    if substr(cTmp, 1, 1) == '"'                // look for 
        ii := at('"', cTmp, 2)                  //   2nd separator
        if ii > 0
           cTmp := substr(cTmp,2,ii-1)          // remove " marks
        endif   
    endif
    if chr(13) $ cTmp                           // abbreviate upto CR
        cTmp := left(cTmp, at(chr(13),cTmp) -1)    
    endif
    if chr(10) $ cTmp                           // abbreviate upto LF
        cTmp := left(cTmp, at(chr(10),cTmp) -1)    
    endif
    if !empty(cTmp)
        aSetupData[IDX_TITLE] := cTmp           // pass back
    endif
endif 
/*
 * GUI window mode: MDI or SDI
 *   WINDOW_MODE=MDI 
 * is considered here, SDI is the default otherwise
 */
ii := at(aSetupDescr[IDX_MODE], cData)          // MDI mode? (SDI is default)
if ii > 0
    ii += len(aSetupDescr[IDX_MODE])
    cTmp := upper(ltrim(substr(cData, ii, 10)))
    if left(cTmp,3) == "MDI"
       aSetupData[IDX_MODE] := APP_MDI + APP_ALLBARS + APP_SBAR_AUTO  // applic.fh
    endif 
endif
/*
 * Application environment: gui/terminal/basic
 *   APPLIC_ENVIR=G...|T.../B... is considered,
 * command-line parameter IO=G/T/B overrides it. 
 * Otherwise, the current environment is determined in _*initio() 
 */
ii := at(aSetupDescr[IDX_ENVIR], cData)         // used environment
if ii > 0
    ii += len(aSetupDescr[IDX_ENVIR])
    cTmp := upper(ltrim(substr(cData, ii, 10)))
    if left(cTmp,1) $ "GTB"
       aSetupData[IDX_ENVIR] := left(cTmp,1)
    endif 
endif

#endif   /* USE_READDEFAULTS */
return NIL

#ifdef FSIO_ALL
/*--------------------------------------------------------------------------
 * Save the current window position in the setup file at program exit
 *
 * The name and location of the setup file is stored in the
 * "gcSetupFile" public var either in ReadDefaults() at app start, 
 * or manually somewhere in the application.
 *
 * Only the winows coordinates will be replaced by current values, all
 * other data remains unchanged, when the old (and new) format is
 * <identifier>=<value>
 * i.e. each one per line. See the aSetupDescr[] array at the begin of 
 * this file for default identifiers.
 *
 * Note: as opposite to ReadDefaults(), you may freely use screen i/o here.
 *     
 * You nay, of course, freely modify the behaviour to fit your needs!
 *--------------------------------------------------------------------------*/
#ifdef USE_READDEFAULTS
EXIT FUNCTION SaveDefaults()

MEMVAR gcSetupFile                  // from _*initio(), ReadDefaults() and app
LOCAL ii, fh, idx       AS INTVAR
LOCAL cData             AS CHAR
LOCAL aSaveSetup := {}  AS ARRAY
LOCAL lError := .F.     AS LOGIC
LOCAL oApp              as Applic

STATIC aCurSizeIdx := {APP_ROWY_TOP,  APP_COLX_TOP, ;
                       APP_ROWS_SIZE, APP_COLS_SIZE}  as array // see applic.fh

if type("gcSetupFile") != "C" .or. empty(gcSetupFile) .or. AppObject() == NIL 
    return NIL
endif
oApp := AppObject()

/* This is an "emergency break" when the user (or programmer) passed
 * a "dangerous" name of existing file - avoids destroying databases 
 * or sources. You may comment this out if not required :-)
 */
ii := rat(".", gcSetupFile)
if ii > 0 .and. file(gcSetupFile)
    if "|" + lower(substr(gcSetupFile,ii+1)) + "|" $ ;
       "|dbf|dbt|dbv|idx|cdx|fpt|mem|lbl|frm|prn|prg|c|h|fh|out|exe|dll|so|a|o|lib|"
        alert("Cannot save the current setup to file;named " + gcSetupFile + ;
              ";because of the too dangerous extension")
        return NIL
    endif
endif

/* Read the current set-up file into an array
 */
fh := Fopen(gcSetupFile, 0)
if fh > 0
    while Ferror() == 0               
        aadd(aSaveSetup, FreadTxt(fh))
    enddo  
    Fclose(fh)
endif

/* create header for a new setup file
 */
if len(aSaveSetup) == 0  
   aadd(aSaveSetup, "# Setup data for " + ExecName(.T.))
   aadd(aSaveSetup, "# created on " + dtoc(date()) + " at " + time())
   aadd(aSaveSetup, "")
endif
    
/* Replace (or fill a new) cordinates of the current window.
 * Other lines remains untouched. The line order is not relevant. 
 */
for idx := 1 to 4
    cData := aSetupDescr[idx]
    ii := ascan(aSaveSetup, {|x| left(ltrim(x),len(cData)) == cData} )
    if ii == 0
        aadd(aSaveSetup, NIL)
        ii := len(aSaveSetup)
    endif
    aSaveSetup[ii] := aSetupDescr[idx] + ;
                      ltrim(oApp:CurrSize(aCurSizeIdx[idx], .T.))
next

/* write the original + changed data back to file
 */
fh := Fcreate(gcSetupFile, 664)  // rw for owner & group, ro for others
if fh > 0
    for ii := 1 to len(aSaveSetup)
        Fwrite(fh, aSaveSetup[ii] + NEWLINE )
        if Ferror() > 0
            lError := .T.
            exit
        endif
    next
    Fclose(fh)
endif
if fh == 0 .or. lError
   alert("Cannot write setup file;named '" + gcSetupFile + ;
         "';check name and permissions")
endif
 
return NIL
#endif   /* USE_READDEFAULTS */
#endif  /* FSIO_ALL */

/* EOF initio.prg */

