* For FlagShip 6
* Copyright (c) 2002..2007 by multisoft GmbH, Munich, Germany
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the
* modified source code when your application is distributed in source,
* but may not remove the copyright notice.
*
* Release $Date: 2007/11/30 16:53:04 $
*
* Note: these functions are available in the FlagShip library. They are
* low-level functions called before your application start.
*
* Warning: do not change this initiomenu.prg except you exactly _know_
* what you are doing. This is a system file and errors here may cause
* your FlagShip application will not work anymore. Also, this file may
* be changed in further FS versions.  You preferably may re-define
* the menu in your application as demonstrated in the <FlagShip_dir>/
* examples/topmenu.prg example. If you nevertheless wish to modify this
* file, copy the source to your working directory, make changes, and
*
* Compile Windows: FlagShip -na -c -w -m -io=g initiomenu.prg -o initiomenu_g.obj
*             and: FlagShip -na -c -w -m -io=t initiomenu.prg -o initiomenu_t.obj
*             and: FlagShip -na -c -w -m -io=a initiomenu.prg -o initiomenu_a.obj
* Link Windows   : FlagShip myapp*.prg initiomenu_*.obj
*
* Compile Linux  : FlagShip -na -c -w -m -io=g initiomenu.prg -o initiomenu_g.o
*             and: FlagShip -na -c -w -m -io=t initiomenu.prg -o initiomenu_t.o
*             and: FlagShip -na -c -w -m -io=a initiomenu.prg -o initiomenu_a.o
* Link Linux     : FlagShip myapp*.prg initiomenu_*.o
*
*-------------------------------------------------------------------------

#include "stdclass.fh"
#include "set.fh"
#include "button.fh"
#include "menuclass.fh"
#include "dialog.fh"

PROTO        FUNCTION InitIoQuit(obj, menuID)      // used also externally
PROTO STATIC FUNCTION LocInitIoQuit(obj, menuID)
PROTO STATIC FUNCTION AboutFlagShip()
PROTO STATIC FUNCTION InitIoAbout(obj, menuID)
PROTO STATIC FUNCTION InitIoFS(obj, menuID)
PROTO STATIC FUNCTION InitIoFont(mode, obj, menuID)
PROTO STATIC FUNCTION InitIoPrSet(obj, menuID)
PROTO STATIC FUNCTION InitIoPrint(obj, menuID)
PROTO STATIC FUNCTION InitIoSave(obj, menuID)
PROTO STATIC FUNCTION InitIoPrScr(obj, menuID)
PROTO STATIC FUNCTION InitIoSubwind(obj, menuID)

/*--------------------------------------------------------------------------
 * User modifiable System Function: InitIoMenu(oApp) -> NIL
 *   Creates the default menu for GUI system.
 *   Invoked from the user modifiable function InitIo(), which is available
 *   in commented source code in the /usr/FSsrc/system/initio.prg file
 *
 * Parameters: oApp is the already instantiated application object. This
 *             parameter is equivalent to the global oApplic constant
 *
 * If redefine the FUNCTION InitIoMenu() by your own same named UDF,
 * you also will need to add all other public functions from this prg
 * file in your source, i.e. at least the FUNCTION InitIoQuit().
 *
 *--------------------------------------------------------------------------*/
#ifdef FSIO_ALL
FUNCTION InitIoMenu(oApp as object)
#endif
#ifdef FSIO_GUI
FUNCTION gInitIoMenu(oApp as object)
#endif
#ifdef FSIO_TERM
FUNCTION tInitIoMenu(oApp as object)
#endif

   local  oPopUp, oPop2, oPopUpFont, oItem, oSubItem as object
   memvar oTopBar

   /* Callback codeblocks, redefine to your needs. There are two different
    * API's which allows you to re-define the default action to your own UDF():
    *
    * 1.Declare somewhere in tha application a PUBLIC _menu_* variable
    *   where _memu_* is either _menu_new, _menu_open, _menu_close etc,
    *   (see the corresponding function header below) and assign a codeblock
    *   performing the required action. For example, if you wish re-declare
    *   the File->Print action to your function MyPrint(), use in your applic:
    *     PUBLIC _menu_print := {|obj, menuID| MyPrint(obj, menuID) }
    *
    * 2.Search the m->oTopBar object for the text or ID-Number and re-assign
    *   the corresponding object property, e.g.:
    *     oMenu := oTopBar:FindMenuProperty( {|obj| obj:Caption == "&About"} )
    *     if valtype(oMenu) == "O" .and. IsObjClass(oMenu,"_MenuItem")
    *       oMenu:Caption := "&About " + ExecName()
    *       oMenu:Data    := {|obj,id| MyUdf(obj,id) }
    *     endif
    */
   local cb_File_New  := {|obj, menuID| alert("File/New (ID="+ltrim(menuID)+")") }     as codeblock
   local cb_File_Open := {|obj, menuID| alert("File/Open (ID="+ltrim(menuID)+")") }    as codeblock
   local cb_File_Close:= {|obj, menuID| alert("File/Close (ID="+ltrim(menuID)+")") }   as codeblock
   local cb_File_SetPr:= {|obj, menuID| InitIoPrSet(obj, menuID) }      as codeblock
   local cb_File_Print:= {|obj, menuID| InitIoPrint(obj, menuID) }      as codeblock
   local cb_File_Save := {|obj, menuID| InitIoSave(obj, menuID) }       as codeblock
   local cb_File_Quit := {|obj, menuID| LocInitIoQuit(obj, menuID) }    as codeblock

   local cb_Opt_defFont := {|obj, menuID| InitIoFont(1, obj, menuID) }  as codeblock
   local cb_Opt_ioFont  := {|obj, menuID| InitIoFont(2, obj, menuID) }  as codeblock
   local cb_Opt_Color   := {|obj, menuID| alert("Opt/Color (ID="+ltrim(menuID)+")") }  as codeblock

   local cb_Wind_Split  := {|obj, menuID| alert("Win/Split (ID="+ltrim(menuID)+")") }  as codeblock
   local cb_wind_Subwin := {|obj, menuID| InitIoSubwind(obj, menuID) }  as codeblock
   local cb_Wind_PrtScr := {|obj, menuID| InitIoPrScr(obj, menuID) }    as codeblock

   local cb_help_applic := {|obj, menuID| InitIoAbout(obj, menuID) }    as codeblock
   local cb_help_fs     := {|obj, menuID| AboutFlagShip() }             as codeblock

// TopBar() crea_MENU_SIZE  tes the menu at the very top of the screen
   CONSTANT oTopBar := TopBar( 0, 0, 78)

   /* user interface, is set .T. when user changes font via Menu->Options
    */
   PUBLIC _gMenuDefFontChanged := .F.
   PUBLIC _gMenuAppFontChanged := .F.

*  File
*   |
*   +-- (New)      -> cb_File_New     API: _menu_new
*   +-- (Open)     -> cb_File_Open    API: _menu_open
*   +-- (Close)    -> cb_File_Close   API: _menu_close
*   +-- (PrintSet) -> cb_File_SetPr   API: _menu_setpr
*   +-- (Print)    -> cb_File_Print   API: _menu_print
*   +++++++++++
*   +-- Save       -> cb_File_Save    API: _menu_save
*   +-- Quit       -> cb_File_Quit    API: _menu_quit

   oPopUp := PopUp()
   oPopUp:ColorSpec:= "b/w,gr+/rb,r/w,g/rb,n+/w,w+/b,w/b" // ignored by GUI
   oTopBar:AddItem( MenuItem ( "&File",oPopUp) )
      oItem := MenuItem( "&New", cb_File_New, K_ALT_N,"Create a new file", 1001)
      oItem:enabled := .F.
      oPopUp:AddItem( oItem )     // "New", disabled, callback = cb_File_New

      oItem := MenuItem( "&Open...",cb_File_Open, K_ALT_O,"Open a file", 1002)
      oItem:enabled := .F.
      oPopUp:AddItem( oItem )     // "Open", disabled, callback = cb_File_Open

      oItem := MenuItem( "&Close...",cb_File_Close, K_ALT_C,"Close a file", 1003)
      oItem:enabled := .F.
      oPopUp:AddItem( oItem )     // "Close", disabled, callback = cb_File_Close

      oPopUp:AddItem( MenuItem( MENU_SEPARATOR ) ) // separator

      // Note: Don't change the ID#50001, otherwise consider
      // to use your ID# in re-defined UDF, see InitIoPrSet() below
      oItem := MenuItem( "P&rinter setup",cb_File_SetPr, K_ALT_R,"Printer setup", 50001)
      oItem:enabled := .T.
      oPopUp:AddItem( oItem )     // "Printer Setup", callback = cb_File_SetPrup

      // Note: Don't change the ID#50002,
      // it is used in SET PRINTER OFF to enable the item.
      // Otherwise you need to change _EnablePrinter() at the bottom of this file
      oItem := MenuItem( "&Print",cb_File_Print, K_ALT_P,"Print spooled output", 50002)
      oItem:enabled := .F.        // enabled after SET PRINT ON + OFF
      oPopUp:AddItem( oItem )     // "Print", disabled, callback = cb_File_Print

      oPopUp:AddItem( MenuItem( MENU_SEPARATOR ) ) // separator

      oItem := MenuItem( "&Save",cb_File_Save, K_ALT_S,"Save defaults", 1006)
      oPopUp:AddItem( oItem )     // "Save", enabled, callback = cb_File_Save

      oPopUp:AddItem( MenuItem( "E&xit", cb_File_Quit, K_ALT_F4,"Quit this application") )

*  Options
*   |
*   +-- Font -+- Window font        -> cb_Opt_defFont  API: _menu_deffont
*   |         +- Default i/o font   -> cb_Opt_ioFont   API: _menu_iofont
*   +-- Color                       -> cb_Opt_Color    API: _menu_color

   oPopUp := PopUp()
   oPopUp :ColorSpec:= "b/w,gr+/rb,r/w,g/rb,n+/w,w+/b,w/b" // ignored by GUI
   oTopBar:AddItem( MenuItem ( "&Options", oPopUp) )
      oPopUpFont := PopUp()
      oPopUp:AddItem( MenuItem( "&Font", oPopUpFont) )     // "Font ->"
         oPopUpFont:AddItem( MenuItem("&Default font", cb_Opt_defFont, K_ALT_D, "Set default application font"))
         oPopUpFont:AddItem( MenuItem("&Input/Output", cb_Opt_ioFont,  K_ALT_O, "Set default i/o font"))

      oPopUp:AddItem( MenuItem( "&Color", cb_Opt_Color, K_ALT_I, "Set default i/o font" ) )


*  Windows
*   |
*   +-- Split        -> cb_Wind_split
*   +-- Sub-Windows  -> cb_Wind_Subwin    (disabled)    ID=9500
*   |                +> subwindow #0      (main screen) ID=9501
*   |                +> subwindow #1..n   (created automatically) ID=9502..9799
*   +-- PrintScreen  -> cb_Wind_PrtScr

   oPopUp := PopUp()
   oPopUp :ColorSpec:= "b/w,gr+/rb,r/w,g/rb,n+/w,w+/b,w/b" // ignored by GUI
   oTopBar:AddItem( MenuItem ( "&Windows", oPopUp) )
      oItem := MenuItem( "&Split", cb_wind_split, NIL, "Split window")
      oItem:enabled := oApp:isMDI()
      oPopUp:AddItem( oItem )

      /* Note: This menu item (with sub-menus) will be automatically
       * enabled when using sub-windows wia MDIopen() or when using
       * Wopen() from the FS2 Toolbox.
       * The menu-id 9500..9799 are RESERVED for these subwindows,
       * (and at least #9501 must be available for the main screen,
       * it message and code block is then used for other sub-windows too),
       * while the menu text is of your choice and will be replaced by
       * the sub-window's header, except for menu-id 9501 = main screen.
       */
      oPop2 := PopUp()
      oItem := MenuItem( "Sub-&Windows", oPop2, NIL, NIL, 9500 )
      oItem:enabled := .F.
         oSubItem := MenuItem("Main window", cb_wind_subwin, NIL, "Select sub-window", 9501)
         oSubItem:enabled := .F.
         oPop2:AddItem( oSubItem )
      oPopUp:AddItem( oItem)

      oPopUp:AddItem( MenuItem( MENU_SEPARATOR ) ) // separator
      oPopUp:AddItem( MenuItem( "&Print screen", cb_Wind_PrtScr, K_CTRL_PRTSCR, "Print current screen" ) )


*  Help
*   |
*   +-- Context Help  F1 -> help(...)
*   +-- About            -> cb_Help_applic          API: _menu_about
*   +-- About FlagShip   -> cb_Help_FS (optional)

   oPopUp := PopUp()
   oPopUp :ColorSpec:= "b/w,gr+/rb,r/w,g/rb,n+/w,w+/b,w/b" // ignored by GUI
   oTopBar:AddItem( MenuItem ( "&Help", oPopUp) )
      oPopUp:AddItem( MenuItem( "Context H&elp", ;
         {|obj, menuID| help(procname(1),procline(1),readvar(), obj)}, ;
         NIL, "Context help"), 9001 )
      oPopUp:AddItem( MenuItem( "&About", cb_help_applic, NIL, "About " + execname(), 9002) )

// Optional: FlagShip info
      oPopUp:AddItem( MenuItem( MENU_SEPARATOR ) ) // separator
      oPopUp:AddItem( MenuItem( "About &FlagShip", cb_help_fs, NIL, "About FlagShip", 9003) )

   oTopBar:Display()    // display MenuBar

//----test: you can print the menu structure by simple:
* outerr(chr(10)+"Structure of 'oTopBar' object in "+lower(procfile())+".prg"+chr(10))
* aeval(oTopBar:MenuStruct(), {|x| outerr(space(x[1] *3), x[4], ;
*       if(valtype(x[3]) == "O" .and. x[3]:IsSeparator(),"(separator)",""), chr(10) )} )
//----or with much more details:
* local _ii, _txt, _obj, _arr
* _arr := oTopBar:MenuStruct()
* outerr(chr(10) + "Structure of 'oTopBar' object of " + ;
*        isObjClass(oTopBar)[1] + " class:")
* for _ii := 1 to len(_arr)
*   _txt := space(_arr[_ii,1] *2) + ltrim(_arr[_ii,2]) + ": " + _arr[_ii,4]
*   _obj := _arr[_ii,3]
*   _txt := padr(_txt,40)
*   if valtype(_obj) == "O" .and. "MENUITEM" $ isObjClass(_obj)[1]
*     _txt += if(_obj:IsSeparator(), " (separator)", "")
*     _txt += if(_obj:IsPopUp(),     " (has PopUp)", "")
*   * _txt += if(_obj:IsCodeblock(), " (has code-block)", "")
*   endif
*   _txt := padr(_txt,53)
*   _txt += "= (" + valtype(_obj) + ") "
*   if valtype(_obj) == "O"
*     _txt += "of "+ isObjClass(_arr[_ii,3])[1]+" class, its :Data = (" + ;
*             valtype(_arr[_ii,3]:Data) + ") "
*     if valtype(_arr[_ii,3]:Data) == "O"
*       _obj := _arr[_ii,3]:Data
*       _txt += "of "+isObjClass(_obj)[1] + " class"
*       if "POPUP" $ isObjClass(_obj)[1]
*          _txt += ", :Top/:Left="+if(_obj:Top == NIL,"NIL",ltrim(_obj:Top))+;
*                  "/"+if(_obj:Left == NIL,"NIL",ltrim(_obj:Left))
*       endif
*     endif
*   endif
*   outerr(chr(10), _txt)
* next
* outerr(chr(10))
//----test end

return NIL

#ifdef FSIO_ALL
/***************************************************************************
 * InitIoQuit() is used here and is also invoked when clicking the
 * close (X) button via oApplic:NotifyClose := {|event| InitIoQuit(event) }
 * in the user modifiable initio.prg source
 *
 * p1: when invoked from menu, p1 == menu object
 *     when invoked via oApplic:NotifyClose or :NotifyCtrlK codeblock,
 *          p1 = APP_EV_CLOSE   == 19  (see applic.fh)
 *       or p1 = APP_EV_CTRL_K  == -11
 * p2: when invoked from menu, menu Id number
 *
 * User's API: assign your UDF via codeblock to PUBLIC _menu_quit, e.g.
 *  PUBLIC _menu_quit := {|obj, menuID| MyUdf(obj, menuID) }
 */
FUNCTION InitIoQuit(p1, p2)
   local iAnsw as intVar
   if type("_menu_quit") == "B"
      return eval(m->_menu_quit, p1, p2)
   endif

   iAnsw := alert("Do you really want to exit the;" + execname() + " application now ?", ;
                  {"No","Yes"},,MBOX_INFO,"Quit ...")
   if iAnsw == 2
      quit
   endif
   return .F.
#endif

STATIC FUNCTION LocInitIoQuit(p1, p2)
   return InitIoQuit(p1, p2)

/***************************************************************************
 * local functions invoked by CodeBlocks here
 *
 * User's API: assign your UDF via codeblock to PUBLIC _menu_save, e.g.
 *  PUBLIC _menu_save := {|obj, menuID| MyUdf(obj, menuID) }
 */
STATIC FUNCTION InitIoSave(obj, menuID)
   if type("_menu_save") == "B"
      return eval(m->_menu_save, obj, menuID)
   endif
   alert("saving screen defaults")
   return NIL

/***************************************************************************
 * local functions invoked by CodeBlocks here
 *
 * User's API: assign your UDF via codeblock to PUBLIC _menu_*, e.g.
 *  PUBLIC _menu_deffont := {|obj, menuID| MyUdf(obj, menuID) } = preset default & exit
 *  PUBLIC _menu_iofont  := {|obj, menuID| MyUdf(obj, menuID) } = preset io & exit
 *  PUBLIC _menu_fd_post := {|obj, menuID, changed| MyUdf(obj, menuID, changed) } = call after dialog, default
 *  PUBLIC _menu_fi_post := {|obj, menuID, changed| MyUdf(obj, menuID, changed) } = call after dialog, io font
 */
STATIC FUNCTION InitIoFont(mode, obj, menuID)
   local oFont    as object
   local lChanged as logic

   if mode == 1   // set default window font
      if type("_menu_deffont") == "B"
         return eval(m->_menu_deffont, obj, menuID)   // only own handling
      endif

      oFont := m->oApplic:FontWindow
      lChanged := oFont:Dialog()          // display font dialog
      if lChanged
         m->oApplic:FontWindow := oFont
         m->_gMenuAppFontChanged := .T.   // reset in applic when considered
      endif
      if type("_menu_fd_post") == "B"     // post-hanling, notification
         return eval(m->_menu_fd_post, obj, menuID, lChanged)
      endif

   elseif mode == 2   // set default i/o font
      if type("_menu_iofont") == "B"
         return eval(m->_menu_iofont, obj, menuID)    // only own handling
      endif

      oFont := m->oApplic:Font
      lChanged := oFont:Dialog()          // display font dialog
      if lChanged
         m->oApplic:Font := oFont
         m->_gMenuDefFontChanged := .T.   // reset in applic when considered
      endif
      if type("_menu_fi_post") == "B"     // post-hanling, notification
         return eval(m->_menu_fi_post, obj, menuID, lChanged)
      endif
   endif
   return NIL

/***************************************************************************
 * Print the current spooler output file via Printer:Exec(). The output
 * file is assigned either automatically by SET PRINT [TO <file>] or
 * manually via m->_oPrinter:OutputFileName
 * Before executing Printer:Exec(), you may set other printer properties,
 * e.g. _oPrinter:Font and _oPrinter:ExecBlock, and/or invoke printer
 * selection via _oPrinter:SetUp()
 *
 * To redefine the functionality to your own UDF named e.g. <myFunct>>, use:
 *    oItem := m->oTopBar:FindMenuProperty({|obj| obj:Id == 50002})
 *    if !empty(oItem)
 *       oItem:Data := {|obj, menuID| myFunct(obj, menuID) }
 *       oItem:Enabled := .T.
 *    endif
 * see also _EnablePrinter() in initio.prg
 *
 * User's API: assign your UDF via codeblock to PUBLIC _menu_print, e.g.
 *  PUBLIC _menu_print := {|obj, menuID| MyUdf(obj, menuID) }
 */
STATIC FUNCTION InitIoPrint(obj, menuID)
   if type("_menu_print") == "B"
      return eval(m->_menu_print, obj, menuID)
   endif
   m->_oPrinter:Exec(3)
   return NIL

/***************************************************************************
 * Printer setup. Calls the Printer:Setup() method. The setup data are
 * used by Printer:Exec() or the above InitIoPrint()
 *
 * To redefine the functionality to your own UDF named e.g. <myFunct>>, use:
 *    oItem := m->oTopBar:FindMenuProperty({|obj| obj:Id == 50001})
 *    if !empty(oItem)
 *       oItem:Data := {|obj, menuID| myFunct(obj, menuID) }
 *    endif
 *
 * User's API: assign your UDF via codeblock to PUBLIC _menu_setpr, e.g.
 *  PUBLIC _menu_setpr := {|obj, menuID| MyUdf(obj, menuID) }
 */
STATIC FUNCTION InitIoPrSet(obj, menuID)
   if type("_menu_setpr") == "B"
      return eval(m->_menu_setpr, obj, menuID)
   endif
   m->_oPrinter:SetUp()
   return NIL

/***************************************************************************
 * Print current screen
 */
STATIC FUNCTION  InitIoPrScr(obj, menuID)
   memvar _oPrinter           // instantiated in initio.prg
   m->_oPrinter:Setup()
   m->_oPrinter:ExecPrintScreen(.T.)
   return NIL

/***************************************************************************
 * Sub-Window selection
 * Note: The pop-up ID# is fix 9500, subwindow items have ID# 9501..9699,
 * by using the m->oTopBar object. You may re-define this handler to
 * your own UDF by assign in the cb_wind_Subwin codeblock at the begin
 * of this source, or by modifying of the corresponding menuItem:Data().
 */
STATIC FUNCTION InitIoSubwind(obj, menuID)
   local_int iWinId := 0
   if valtype(obj) != "O" .or. !IsObjClass(obj,"_gMenuItem")
      return NIL
   endif
   if menuID < 9500 .or. menuID > 9699 .or. ;   // reserved menu ID numbers
      valtype(obj:Cargo) != "N"
      return NIL
   endif

   /* Note: you may use Wselect(obj:Cargo) containing the
    * window ID# (same as return from Wopen) to pass
    * an input focus to the selected sub-window.
    * But be careful, your application must consider this.
    */
    * Wselect(obj:Cargo)  -or- MDIselect(obj:Cargo)
   /*
    * We therefore only pop-up (display) the selected sub-window
    * without changing the current input focus...
    */
   MDIselect(obj:Cargo, .F.)   // or Wselect(obj:Cargo, .F.)
   return NIL

/***************************************************************************
 * Display "about" info about your executable
 *
 * To redefine the functionality to your own UDF named e.g. <myFunct>>, use:
 *    oItem := m->oTopBar:FindMenuProperty({|obj| obj:Id == 9002})
 *    oItem:Data := {|obj, menuID| myFunct(obj, menuID) }
 *
 * User's API: assign your UDF via codeblock to PUBLIC _menu_about, e.g.
 *  PUBLIC _menu_about := {|obj, menuID| MyUdf(obj, menuID) }
 */
STATIC FUNCTION InitIoAbout(obj, menuID)
   local oInfo as object
   if type("_menu_about") == "B"
      return eval(m->_menu_about, obj, menuID)
   endif
   oInfo := InfoBox {NIL, "About ...", ;
      "This " + execname() + " is;... put here your description ...; ;" + ;
      "Release ....;Author ...;Copyright ..."}
   oInfo:show()
   return NIL

/***************************************************************************
 * Display optional About FlagShip info
 *
 * To redefine the functionality to your own UDF named e.g. <myFunct>>, use:
 *    oItem := m->oTopBar:FindMenuProperty({|obj| obj:Id == 9003})
 *    oItem:Data := {|obj, menuID| myFunct(obj, menuID) }
 */
STATIC FUNCTION AboutFlagShip()
   memvar FlagShip_Version, FlagShip_Release, FlagShip_Serno
   memvar FlagShip_Betauntil, FlagShip_Evaluntil

   InfoBox {NIL, "About FlagShip ...", ;
      "This application was developed by;" + ;
      FlagShip_Version + ", Release " + FlagShip_Release + ;
         ", s/n " + FlagShip_Serno + ";" + ;
      if(empty(FlagShip_Betauntil), "", "Beta valid until " + ;
         FlagShip_Betauntil + ";") + ;
      if(empty(FlagShip_Evaluntil), "", "Eval valid until " + ;
         FlagShip_Evaluntil + ";") + ;
      "This version supports up to " + ltrim(usersmax()) + ;
         " users simultaneously."}:show()
   return NIL

/***************************************************************************
 * Enable the menu item File->Print spooled output
 * May be used in the application, is used by SET PRINT OFF, SET PRINT TO
 */
#ifdef FSIO_ALL
FUNCTION _EnablePrinter()
#endif
#ifdef FSIO_GUI
FUNCTION _gEnablePrinter()
#endif
#ifdef FSIO_TERM
FUNCTION _tEnablePrinter()
#endif
local oItem as usual
if type("oTopBar") != "O"
   return .F.
endif
oItem := m->oTopBar:FindMenuProperty({|obj| obj:Id == 50002})
if !empty(oItem)
   oItem:Enabled := .T.
   return .T.
endif
return .F.


/* EOF */

