/**************************************************************************
 * FlagShip 6.x : the basic Printer class
 * Prototypes   : printerclass.fh
 *
 * Compile: FlagShip -na -c -io=a 
 * Link:    FlagShip myapp*.prg o_printer.o|.obj ...your switches...
 *
 * For FlagShip 6. Copyright (c) 2005 by multisoft GmbH, Munich, Germany
 * You may modify this source according to your needs and distribute the
 * object code freely with your executable. You may distribute also the 
 * modified source code when your application is distributed in source, 
 * but may not remove the copyright notice.
 *
 * Note: this source is intended mainly for your information. It may be 
 * changed without further notification in later FS release.
 * 
 * Release $Date: 2005/09/27 20:00:55 $
 */
 
#Cinline
static char *fileID="@@#STDIO# _printer{} $Revision: 1.6 $";
# include "FSopenc.h"
#endCinline

/**************************** _printer **************************************/
#include "printerclass.fh"
#include "error.fh"

#ifdef FS_WIN32
PROTO STATIC FUNCTION winSpool(cFileName, cDriver, nNumCopies, nMarginLeft, nMarginTop)   
#endif

CLASS _Printer ALIAS Printer

/* --- instances are in printerclass.fh
 */

/*---------------------------------------------------------------------------
 * :Init() -> self
 *    creator, instantiated in initio.prg -> _oPrinter
 */
METHOD Init()                             CLASS _Printer
   return self

/*---------------------------------------------------------------------------
 *  :Axit() -> NIL
 *    destructor
 */
METHOD Axit()                             CLASS _Printer
  return NIL

/*---------------------------------------------------------------------------
 * optional property, not used directly by :Exec() but my be used 
 *    and considered in user-defined :ExecBlock
 * The default is .T. or an user-set value from :SetUp()
 */
ACCESS Color                              CLASS _Printer
   return self:lColor
ASSIGN Color(p1)                          CLASS _Printer
   if valtype(p1) == "L"
      self:lColor := p1
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * Return/Set the default driver name (system dependant)
 * The default driver is user-selected in :SetUp()
 * and is e.g. "Printer1" on Unix or driver name in MS-Windows
 * The driver name is used in :Exec() 
 */
ACCESS Driver                             CLASS _Printer
   return self:cPrintDriver
ASSIGN Driver(p1)                         CLASS _Printer
   if valtype(p1) == "C"
      self:cPrintDriver := p1
   endif
   return NIL

/*---------------------------------------------------------------------------
 * :Display([top],[left],[bott],[right],[inPixel],[lineSize],[cUdf]) -> lOk
 *    Display the printer spooler file
 */
METHOD Display([p1],[p2],[p3],[p4],[p5],[p6line],[p7udf])   CLASS _Printer    
   local cFileName, cBuffer as char
   local r1,r2, c1, c2 as intvar             // in pixel
   cFileName := self:InputFileName           // using ACCESS instd.of instance
   if empty(cFileName) .or. !file(cFileName) .or. Directory(cFileName)[1,2] == 0
      return .F.
   endif
   cBuffer := MemoRead(cFileName)
   r1 := IOsetRowPix(p1, 0,        p5)
   r2 := IOsetRowPix(p3, maxrow(), p5)
   c1 := IOsetColPix(p2, 0,        p5)
   c2 := IOsetColPix(p4, maxcol(), p5)
      
   MemoEdit(cBuffer, r1,c1, r2,c2, .F., p7Udf, p6line, NIL, NIL, ;
            NIL, NIL, NIL, .T., self:Font, .T.) 
   return .T.
   
/*---------------------------------------------------------------------------
 * :Driver?margin -> iPixel
 *   Get the horizontal/vertical margin of the printer driver
 -> reimplemented in GUI
 */
ACCESS DriverHmargin             CLASS _Printer
   return self:iDriverHmargin
ACCESS DriverVmargin             CLASS _Printer
   return self:iDriverVmargin

/*---------------------------------------------------------------------------
 * :Driver?ldpi -> iLogPixel_per_inch
 *   Get the horizontal/vertical "logical" driver resolution in DPI (dots
 *   per inch). It is not the real printer resolution, but the unit
 *   used for margins. For laser printers,the returned value is usually 
 *   72 x 72 dpi = 2.83 x 2.83 dots per mm.
 -> reimplemented in GUI
 */
ACCESS DriverHldpi               CLASS _Printer
   return self:iDriverHldpi
ACCESS DriverVldpi               CLASS _Printer
   return self:iDriverVldpi
   
/*---------------------------------------------------------------------------
 * :Exec(par1],[par2]) -> lOk
 * par1: wait seconds after printing (to display messages), default = 5sec
 * par2: 0, NIL = Linux: invoke lpr or lp, in GUI check for Setup()
 *                Windows: invoke native driver, in GUI check for Setup()
 *     : 1      = Linux: cp <file> /dev/lp0
 *                Windows: copy <file> PRN:
 *     : string, e.g. output to "LPT3:" on Windows, or "/dev/lp2" in Unix
 *
 * Before calling _oPrinter:Exec(), either
 * a) execute _oPrinter:Setup() - or will be called here autom. in GUI mode
 * b) or assign _oPrinter:Driver := "My Printer Name"
 * c) or use _oPrinter:Exec(,1)
 * d) or use _oPrinter:Exec(,"device name")
 * otherwise simple copy <printerfile> to standard printer device is used,
 * same as with _oPrinter:Exec(,1). You also may set _oPrinter:NumCopies, 
 * _oPrinter:MarginTop and _oPrinter:MarginLeft beforehand.
 * 
 * Note: The _oprinter:Exec() may also be called from main menu File->Print, 
 *    see initiomenu.prg - STATIC FUNCTION InitIoPrint(obj, menuID) 
 */
METHOD Exec([par1],[par2])                      CLASS _Printer 
   local cFileName, cExeName, cDef, cDevice  as char
   local nSeconds, nWaitSec                  as numeric
   local retVal := .T.                       as logic
   local ii, iCopy, iMode := 0               as intvar
   
   if valtype(self:bExecBlock) == "B"        // user defined exec block?
      return eval(self:bExecBlock, self)     // yes, eval and exit
   endif   

#ifdef FS_WIN32
   cDevice := "PRN:"
#else
   cDevice := "/dev/lp0"
#endif
   
   if valtype(par2) == "N"
      iMode := par2
   elseif valtype(par2) == "C" .and. !empty(par2)
      iMode := 1
      cDevice := par2
   endif
   cDef := "Using copy to " + cDevice
          
   if iMode == 0 .and. empty(self:Driver) 
      /* neither :SetUp() executed yet, nor :Driver assigned
       * invoke :SetUp() in GUI mode now, or use simple copy otherwise
       */   
      if AppIoMode() == "G"
         self:Setup()      // user dialog selecting printer
         if self:lAborted
            return .F.     // abort - or alternativelly:
//          StatusMessage(cDef)
//          _rt_error(EG_DEVWARNING, "The _oPrinter:Setup() was aborted;" + cDef)
//          iMode := 1
         endif
      else
         iMode := 1
      endif
   endif         

   cFileName := self:InputFileName          // using ACCESS instd.of instance
   if empty(cFileName) .or. !file(cFileName) .or. Directory(cFileName)[1,2] == 0
      /* either SET PRINT TO was not executed yet  
       * and/or the input file is not specified by _oPrinter:InputFileName
       * or the file is not available 
       * or the file is empty
       */
      StatusMessage("Nothing to print")
      _rt_error(EG_DEVWARNING, "Nothing to print;check _oPrinter:InputFileName")
      return .F.
   endif
   if iMode == 0 .and. empty(self:Driver)
      StatusMessage(cDef)
      _rt_error(EG_DEVWARNING, "No _oPrinter:Driver specified;" + cDef)
      iMode := 1
   endif
   
   nWaitSec := if (valtype(par1) == "N", par1, 5)
   nSeconds := Seconds()
   
   if iMode == 0   
#ifdef FS_WIN32
      /* --- Windows: use native Windows GDI API, see below 
       */  
      StatusMessage("Printing output file '" + cFileName + "'" + ;
                    " via " + self:Driver)
      winSpool(cFileName, self:Driver, self:NumCopies, ;
               self:MarginLeft, self:MarginTop)          // see below
#else   
      /* --- Linux/Unix: use self:ExecName or self:PrintExecutable or lpr spooler
       */  
      if valtype(self:ExecName) == "C" .and. !empty(self:ExecName)
         cExeName := self:ExecName
      else
         cExeName := self:PrintExecutable    // check for lpr or lp
      endif
      if empty(cExeName)
         StatusMessage("Cannot print via driver, printer exacutable n/a, using copy")
         _rt_error(EG_DEVWARNING, "No _oPrinter:ExecName specified;" + cDef)
         iMode := 1
      else
         /* driver available, use lpr or other spooler utility
          */
         if !empty(self:Driver)
            cExeName += " -P" + self:Driver
         endif
         if self:NumCopies > 1
            cExeName += " " + "-#" + ltrim(self:NumCopies)
         endif
         if self:Columns > 0
            cExeName += " " + "-w" + ltrim(self:Columns)
         endif
         StatusMessage("Printing output file '" + cFileName + "'" + ;
               if(empty(self:Driver),""," via " + self:Driver))
         RUN (cExeName + " " + cFileName)           // execute
         if DosError() > 0
            Tone()
            StatusMessage(cExeName + " reported error " + ltrim(DosError()))
            retVal := .F.
         endif 
      endif
#endif   /* Windows or Unix */
   endif
   if iMode == 1    
      /* use simple copy to default device 
       */  
      iCopy := max(1, self:NumCopies)
      for ii := 1 to iCopy
         COPY FILE (cFileName) TO (cDevice) // execute: copy to printer device
         if DosError() > 0
            Tone()         
            StatusMessage("Error " + ltrim(DosError()) + ;
                          ": '" + DosError2str() + "' reported")
            retVal := .F.
            exit
         endif 
      next
   endif
   
   if nWaitSec > 0
      nWaitSec := nWaitSec + nSeconds - Seconds()
      if nWaitSec > 0
         sleep(nWaitSec)            // wait rest of the time
      endif
   endif
   StatusMessage()                  // clear status message
   return retVal
   
*---------------------------------------------------------------------------
 * User defined code block or NIL, evaluated instead of the default Exec()
 */
ACCESS ExecBlock                          CLASS _Printer
   return self:bExecBlock
ASSIGN ExecBlock(p1)                      CLASS _Printer
   if valtype(p1) $ "BU"
      self:bExecBlock := p1
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * :ExecFormatted([fileName], [oFont]) -> lOk
 *    reads the <fileName>, reformats according to font, invokes the 
 *    printer-spooler according to :SetUp()
 -> Default for Basic/Terminal i/o uses self:Exec() w/o considering fonts
 -> Reimplemented for GUI to consider font, headers, margins
 *  In GUI, following font attributes within the text are supported:
 *    <b>...</b> - Bold font style. 
 *    <i>...</i> - Italic font style. 
 *    <u>...</u> - Underlined font style.         
 */
METHOD ExecFormatted([p1],[p2])             CLASS _Printer
   local cOldFile as char
   if valtype(p1) == "C" .and. !empty(p1) .and. file(p1)
      cOldFile := self:InputFileName
      self:InputFileName := p1
      self:Exec()
      self:InputFileName := cOldFile
      return .T.
   elseif !empty(self:InputFileName)
     self:Exec()
      return .T.
   endif
   return .F.

/*---------------------------------------------------------------------------
 * :ExecPrintScreen([lAdapt]) -> lOk
 *    prints the whole user-screen to default spooler according to :SetUp()
 * The _oprinter:PrintScreen() is also executed from main menu, see 
 *    initiomenu.prg either via STATIC FUNCTION InitIoScrPrint(obj, menuID) 
 *    or user redefined
 -> Reimplemented in Terminal and GUI i/o
 */
METHOD ExecPrintScreen([p1])              CLASS _Printer
   return .F.

/*---------------------------------------------------------------------------
 * optional font object, used by :ExecFormatted() and can be used 
 *    and considered in user-defined :ExecBlock
 * The default is NIL, assign e.g. _oPrinter:Font := Font{"Courier",10}
 */
ACCESS Font                               CLASS _Printer
   return self:oFont
ASSIGN Font(p1)                           CLASS _Printer
   if valtype(p1) $ "OU"
      self:oFont := p1
   endif
   return NIL

/*---------------------------------------------------------------------------
 * Header Left for ExecFormatted()
 *    Can be any string including macros <#>,<date>,<time><file>  
 *    Default is "Page <#>"
 */
ACCESS HeaderLeft                         CLASS _Printer
   return self:cHeadLeft
ASSIGN HeaderLeft(p1)                     CLASS _Printer
   if valtype(p1) == "C"
      self:cHeadLeft := p1
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * Header Mid for ExecFormatted()
 *    Can be any string including macros <#>,<date>,<time><file>  
 *    Default is "<date> <time>"
 */
ACCESS HeaderMid                          CLASS _Printer
   return self:cHeadMid
ASSIGN HeaderMid(p1)                      CLASS _Printer
   if valtype(p1) == "C"
      self:cHeadMid := p1
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * Header Right for ExecFormatted()
 *    Can be any string including macros <#>,<date>,<time><file>  
 *    Default is "<file>"
 */
ACCESS HeaderRight                        CLASS _Printer
   return self:cHeadRight
ASSIGN HeaderRight(p1)                    CLASS _Printer
   if valtype(p1) == "C"
      self:cHeadRight := p1
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * ACC returns either the file lastly set by SET PRINTER TO <file> if the
 *    file is available and not empty, or the default spooler file
 *    FS_SET("printfile") otherwise. It therefore work also after 
 *    SET PRINTER TO. You may reset the PRINTER TO <file> preference 
 *    by _oPrinter:InputUser := "" and/or re-assign the default name 
 *    by _oPrinter:InputFileName := "your_file_name_to_print"
 *    If the spooler file is not available or is empty, "" is returned
 * The :cInputFile is assigned automatically from SET PRINTER ON,
 *    the :cInputUser instance by SET PRINTER TO <file>. 
 */
ACCESS InputFileName                      CLASS _Printer
   local retVal as usual

   if !empty(self:cInputUser) .and. file(self:cInputUser) .and. ;
         Directory(self:cInputUser)[1,2] > 0
      retVal := self:cInputUser
   else
      if empty(self:cInputFile)
         self:cInputFile := FS_SET("printfile")
      endif
      retVal := self:cInputFile
   endif
   if empty(retVal) .or. !file(retVal) .or. Directory(retVal)[1,2] == 0
      retVal := ""
   endif
   return retVal
   
ASSIGN InputFileName(p1)                  CLASS _Printer
   if valtype(p1) == "C" 
      self:cInputFile := alltrim(p1)
   endif
   return NIL

/*---------------------------------------------------------------------------
 * assigned automatically from SET PRINTER TO <file>. 
 */
ASSIGN InputUser(p1)                      CLASS _Printer
   if valtype(p1) == "C" 
      self:cInputUser := alltrim(p1)
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * Access/Assign left/top/right/bottom printer margin in LPI (printer)
 *    Considered in :ExecFormatted() and :ExecPrintScreen() in GUI
 *    Defaults are defined in printerclass.fh, but best set to
 *       MarginLeft & MarginRight = _oPrinter:DriverHmargin and
 *       MarginTop & MarginBottom = _oPrinter:DriverVmargin
 *    i.e. corresponding to the currently used printer driver.
 */
ACCESS MarginBottom              CLASS _Printer
   return self:iPageBotMarg
ASSIGN MarginBottom(p1)          CLASS _Printer
   if valtype(p1) == "N" 
      self:iPageBotMarg := p1
   endif
   return NIL

ACCESS MarginLeft                CLASS _Printer
   return self:iPageLeftMarg
ASSIGN MarginLeft(p1)            CLASS _Printer
   if valtype(p1) == "N" 
      self:iPageLeftMarg := p1
   endif
   return NIL
   
ACCESS MarginRight               CLASS _Printer
   return self:iPageRightMarg
ASSIGN MarginRight(p1)           CLASS _Printer
   if valtype(p1) == "N" 
      self:iPageRightMarg := p1
   endif
   return NIL

ACCESS MarginTop                 CLASS _Printer
   return self:iPageTopMarg
ASSIGN MarginTop(p1)             CLASS _Printer
   if valtype(p1) == "N" 
      self:iPageTopMarg := p1
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * The default is 1 or an user-set value from :SetUp()
 * If set, the number of copies is considered :Exec() and :ExecFormatted()
 */
ACCESS NumCopies                          CLASS _Printer
   return self:iNumCopies
ASSIGN NumCopies(p1)                      CLASS _Printer
   if valtype(p1) == "N" .and. p1 > 0
      self:iNumCopies := p1
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * optional property, not used directly by :Exec() but my be used 
 *    and considered in user-defined :ExecBlock
 * The default is "Portrait" or an user-selected value from :SetUp()
 * Assign accept "P" or "F" as the 1st char
 */
ACCESS Orientation                        CLASS _Printer
   return if(self:lLandscape,"Landscape","Portrait")
ASSIGN Orientation(p1)                    CLASS _Printer
   if valtype(p1) == "C" .and. upper(left(p1,1)) $ "LP"
      self:lLandscape := upper(left(p1,1)) == "L"
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * optional property, not used directly by :Exec() but my be used 
 *    and considered in user-defined :ExecBlock
 * The default is "" or an user-set value from :SetUp()
 */
ACCESS OutputFileName                     CLASS _Printer
   return self:cOutputFile
ASSIGN OutputFileName(p1)                 CLASS _Printer
   if valtype(p1) == "C"
      self:cOutputFile := p1
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * optional property, not used directly by :Exec() but my be used 
 *    and considered in user-defined :ExecBlock
 * The default is "First" or an user-set value from :SetUp()
 * Assign accept "F" (first) or "L" (last) as the 1st char
 */
ACCESS PageOrder                          CLASS _Printer
   return if(self:lLastFirst,"Last","First")
ASSIGN PageOrder(p1)                      CLASS _Printer
   if valtype(p1) == "C" .and. upper(left(p1,1)) $ "FL"
      self:lLastFirst := upper(left(p1,1)) == "L"
   endif
   return NIL

/*---------------------------------------------------------------------------
 * optional property, not used directly by :Exec() but my be used 
 *    and considered in user-defined :ExecBlock
 * The default ret val is .T. when :PageFrom and/or :PageTo is 0, otherwise .F.
 * ASSIGN accept .T. only and reset both :PageFrom and :PageTo to 0
 */
ACCESS PageAll                            CLASS _Printer
   return self:iPageFirst <= 0 .or. self:iPageLast <= 0
ASSIGN PageAll(p1)                        CLASS _Printer
   if valtype(p1) == "L" .and. p1
      self:iPageFirst := 0 
      self:iPageLast  := 0
   endif
   return NIL
      
/*---------------------------------------------------------------------------
 * optional property, not used directly by :Exec() but my be used 
 *    and considered in user-defined :ExecBlock
 * The default is 0 or an user-set value from :SetUp()
 */
ACCESS PageFrom                           CLASS _Printer
   return self:iPageFirst   
ASSIGN PageFrom(p1)                       CLASS _Printer
   if valtype(p1) == "N" .and. p1 >= 0
      self:iPageFirst := p1
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * optional property, not used directly by :Exec() but my be used 
 *    and considered in user-defined :ExecBlock
 * The default is 0 or an user-set value from :SetUp()
 */
ACCESS PageTo                             CLASS _Printer
   return self:iPageLast    
ASSIGN PageTo(p1)                         CLASS _Printer
   if valtype(p1) == "N" .and. p1 >= 0
      self:iPageLast := p1
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * optional property, not used directly by :Exec() but my be used 
 *    and considered in user-defined :ExecBlock
 * The default is "A4" or an user-selected value from :SetUp()
 * ASSIGN accept any char value w/o check.
 */
ACCESS PageSize                           CLASS _Printer
   return self:cPageSize
ASSIGN PageSize(p1)                       CLASS _Printer
   if valtype(p1) == "C"
      self:cPageSize := p1
   endif
   return NIL

/*---------------------------------------------------------------------------
 * Return/Set the default spooler executable (system dependant),
 * user-selected in :SetUp()
 * If not set manually, "lpr" or "lp" is returned in Unix/Linux, this
 * spooler executable is used in :Exec() except for Windows
 */
ACCESS PrintExecutable                     CLASS _Printer
   local retVal := "" as char
   
   if !empty(self:cPrintExeName)
      return self:cPrintExeName
   endif
#ifndef FS_WIN32
   retVal := FindExeFile("lpr")
   if empty(retVal)
      retVal := FindExeFile("lp")
   endif
#endif
   return retVal      
   
ASSIGN PrintExecutable(p1)                 CLASS _Printer
   if valtype(p1) == "C"
      self:cPrintExeName := p1
   endif
   return NIL
   
/*---------------------------------------------------------------------------
 * Dialog to select priner properties.
 * The _oprinter:Setup() is also executed from main menu, see initiomenu.prg
 *    either via STATIC FUNCTION InitIoPrSet(obj, menuID) or user redefined
 -> Reimplemented in Terminal and GUI i/o
 */
METHOD Setup()                            CLASS _Printer    // called also from menu
   return .F.
   
/*---------------------------------------------------------------------------
 * Was the SetUp() aborted by user? 
 */
ACCESS SetupAborted                       CLASS _Printer
   return self:lAborted

/*---------------------------------------------------------------------------
 * equivalent to Display(), for compatibility purposes.
 */
METHOD Show()                             CLASS _Printer    
   return self:Display()
   
/*---------------------------------------------------------------------------
 * set/get tabulator position (in ldpi pixels) for GUI ExecFormatted()
 *    corresponding to \t = chr(9) character of the text line.
 *    Every Tab position occupy one numeric element in one-dimensional 
 *    array, also unsorted, e.g. _oPrinter:TabStop := {10,20,50,120} in 
 *    ldpi units, see ::DriverHldpi. All tab values must be > 0, otherwise 
 *    no tabs will be set. Reset by an empty array or NIL assignment.
 */
ACCESS TabStop                             CLASS _Printer
   return self:aTabs
ASSIGN TabStop(p1)                         CLASS _Printer
   local ok := .T. as logic
   if p1 == NIL .or. empty(p1)
      self:aTabs := {}
   elseif valtype(p1) == "A" .and. AelemType(p1,"N")
      aeval(p1, {|x| if(x <= 0, ok := .F., NIL) } )
      if ok
         self:aTabs := aclone(p1)
         asort(self:aTabs)
      endif
   endif
   return NIL

#ifdef FS_WIN32
/*---------------------------------------------------------------------------
 * Native printing for Windows using GDI API, used in :Exec() mode 0
 * Note that only some _oPrinter data are considered here: file name, 
 * driver name, number of copies, margin left and margin top. You of 
 * course may adapt it to your own needs.
 */
STATIC FUNCTION winSpool(cFileName, cDriver, nNumCopies, nMarginLeft, nMarginTop)   
   local_int iError := 0, iNumCopy := 1, iMargLeft := 10, iMargTop := 0
   local fdFile, iCopies as intvar
   local cOutName as char
   
   if valtype(cFileName) != "C" .or. empty(cFileName)
      return .F.
   endif   
   if valtype(cDriver) != "C" .or. empty(cDriver)
      return .F.
   endif   
   if valtype(nNumCopies) == "N" .and. nNumCopies > 0
      iNumCopy := nNumCopies
   endif
   if valtype(nMarginLeft) == "N" 
      iMargLeft := nMarginLeft
   endif
   if valtype(nMarginTop) == "N" 
      iMargTop := nMarginTop
   endif
   
   for iCopies := 1 to iNumCopy 
      fdFile := Fopen(cFileName, 0)
      if fdFile <= 0
         StatusMessage("Cannot open file '" + cFileName + "'")
         _rt_error(EG_DEVWARNING, "Cannot open file '" + cFileName + "';check access rights")
         return .F.
      endif
      cOutName := ExecName()
      if "\" $ cOutName
         cOutName := substr(cOutName, rat("\", cOutName) +1)
      endif
      if iNumCopy > 1
         cOutName := "Copy# " + ltrim(iCopies) + " from " + cOutName
      else
         cOutName := "Output from " + cOutName
      endif
      
#Cinline
{
#include "FSopenc.h"
#include "FSerrors.h"
#include <windows.h>
#include <wingdi.h>
#include <stdio.h>

   HDC         hDC;
   DOCINFO     di;
   TEXTMETRIC  tmTextInfo;
   int         iRet, iLine, iPage, iPageHeight, iLineHeight;
   char       *cText, *cDriver, *ptrFF;
   FSvar      *vText = VAR_NEW;
   FSvar      *myArgv[1];
   
   UDF_PROT(freadtxt);
   UDF_PROT(ferror);
      
   cDriver = VAR_CHR(VAR_NAME_LOCPAR(cdriver));
   
   hDC = CreateDC("WINSPOOL", cDriver, NULL, NULL);
   if(!hDC) {
      rt_error(DEVWARNING, "cannot CreateDC() for WinSpool");
      ierror = 1;    // cannot create DC
      goto done;
   } 
   memset( &di, 0, sizeof(DOCINFO) );
   di.cbSize = sizeof(DOCINFO);
   di.lpszDocName = VAR_CHR(VAR_NAME_LOCAL(coutname));

   if(!GetTextMetrics(hDC,&tmTextInfo)) {
      rt_error(DEVWARNING, "cannot get text metrics for WinSpool");
      ierror = 2;
      goto done;
   } 

   if(StartDoc(hDC, &di) == SP_ERROR) {                  // Windows API
      rt_error(DEVWARNING, "cannot StartDoc() in WinSpool");
      ierror = 3;
      goto done;
   } 
   
   iLine = 0;
   iPage = 1;
   iLineHeight = tmTextInfo.tmHeight + 2;
   iPageHeight = GetDeviceCaps(hDC, VERTRES); 
   if(iPageHeight == 0)  
      iPageHeight = GetDeviceCaps(hDC, PHYSICALHEIGHT);   
   if(iPageHeight > iLineHeight)  
      iPageHeight -= iLineHeight;      // need to print last line in fully
   if(iPageHeight < (5 * iLineHeight))
      iPageHeight = 0;                 // seems driver data are wrong, ignore
         
// fprintf(stderr,"page height=%d line height=%d\n", iPageHeight, iLineHeight);
// fprintf(stderr,"PHYSICALHEIGHT =%d\n", GetDeviceCaps(hDC, PHYSICALHEIGHT));
// fprintf(stderr,"PHYSICALOFFSETY=%d\n", GetDeviceCaps(hDC, PHYSICALOFFSETY));
// fprintf(stderr,"VERTRES        =%d\n", GetDeviceCaps(hDC, VERTRES));
// fprintf(stderr,"SCALINGFACTORY =%d\n", GetDeviceCaps(hDC, SCALINGFACTORY));

   myArgv[0] = VAR_NAME_LOCAL(fdfile);                   // file handle
   SET_VAR_COPY(vText, UDF_EXEC(freadtxt)(1,myArgv) );   // read one line
   iRet = VAR_INTNUM( UDF_EXEC(ferror)(0,0) );           // check for error

   StartPage(hDC);                                       // Windows API
   
   while(iRet == 0) {  
      cText = VAR_CHR(vText);       // current line as string
       
// fprintf(stderr,"Line:%d Page:%d text='%s' pos=%d\n", iLine +1, iPage, cText, imargtop + iLine * iLineHeight);

      ptrFF = strchr(cText, 12);    // check for user FormFeed
      if(ptrFF) {     
         if(ptrFF > cText) {        // print leading chars upto FF
            *ptrFF = 0;             // replace FF by \0 (in vText!)
            TextOut(hDC, imargleft, imargtop + iLine * iLineHeight, cText, strlen(cText));
         }
         EndPage(hDC);              // printer driver form feed

         StartPage(hDC);
         iLine = 0;
         iPage ++;
         
         ptrFF ++;                  // now ptr to \0 or to rest of line
         if(*ptrFF) {               // print line rest (if any) on next page
            TextOut(hDC, imargleft, imargtop + iLine * iLineHeight, ptrFF, strlen(ptrFF));
            iLine++;
         }
      } else {                      // print whole line
         TextOut(hDC, imargleft, imargtop + iLine * iLineHeight, cText, strlen(cText));
         iLine++;

         if(iPageHeight && (imargtop + iLine * iLineHeight > iPageHeight)) {  
            EndPage(hDC);           // printer driver form feed
            StartPage(hDC);
            iLine = 0;
            iPage ++;
         }
      }
      SET_VAR_COPY(vText, UDF_EXEC(freadtxt)(1,myArgv) );   // read next line
      iRet = VAR_INTNUM( UDF_EXEC(ferror)(0,0) );           // check for error
   }

   EndPage(hDC);
   EndDoc(hDC);
   
done:
   if(hDC)
      DeleteDC(hDC);
}
#endCinline
      Fclose(fdFile) 
      if iError > 0
         exit
      endif
   next  // iCopies
   return iError == 0
#endif  /* FS_WIN32 */
  
/* eof */
