* For FlagShip 5 and 6 
* Copyright (c) 2002..06 by multisoft GmbH, Munich, Germany 
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the 
* modified source code when your application is distributed in source, 
* but may not remove the copyright notice.
* 
* Release $Date: 2006/07/28 08:01:51 $
*
* Default Input handler for PushButton, called from PushButton:Exec() 
*
* You may assign YOUR own handler UDF via code block :Handler, i.e.
* oBox:Handler := {|obj| MyUdf(obj) }
* which is then called from :Exec() automatically instead of this
* default handler.
*
* Compile: FlagShip -c -na -m -io=a pushbutthand.prg
* Link   : FlagShip [-io=t|g|b|a] myapp*.prg pushbutthand.o

#include "boxclass.fh"
#include "inkey.fh"
#include "box.fh"
#include "button.fh"
#include "set.fh"
#include "FSerrmsg.h"
#include "getexit.fh"

/* --------------------------------------------------------------------------
 * Default PushButton input handler
 *
 * Parameter passed from oPushButton:Show() via oPushButton:Handler block,
 * i.e. here oPushButton:Handler := {|obj| PushButtHandler(obj) }
 *    obj = current object
 * Returns current obj:Buffer status or NIL on error
 */
function PushButtHandler(obj)
   local key, iHitTest  as intvar
   local cAppType       as char
   local inGet, hasFSblock := .F. as logic

   if valtype(obj) != "O" .or. !IsObjClass(obj, "_PushButto") 
      _Rt_Error(RTARGTYP, "PushButtonHandler(): wrong object class - " + ;
         if(valtype(obj) != "O", "not object", IsObjClass(obj)[1]) )
      return NIL 
   endif

// cAppType := AppIoMode()   // i/o mode: G,T,B
// obj:FilterKeys := .F.

   obj:SetFocus()
   obj:Display()

   inGet := valtype(obj:exitState) == "N" .and. obj:exitState >= GE_NOTSET
   if inGet
      hasFSblock := valtype(obj:FBlock) == "B" .or. ;
                    valtype(obj:SBlock) == "B"
   endif
   
   while .T.
      key := InkeyTrap(0)     // same as Inkey(0) but considers SET KEY

*     For debug to stderr:
*     ??## "-> CheckBoxHandler() key=",ltrim(key),"=",inkey2str(key),procstack()

      do case
      case  key == K_ESC                  //--> Esc 
         obj:exitState := GE_ESCAPE       //    exit READ
         exit
         
      case  key == K_ENTER       .or. ;   //--> Return 
            key == K_CTRL_ENTER           
	      obj:Select()                     //    simulate push
         if inGet .and. !hasFSblock       //    in @..GET/READ w/o Fblock/Sblock: 
            obj:exitState := GE_ENTER     //    simulate Enter key
         endif
         exit
         
      case  key == K_DOWN         .or. ;  //--> cursor down 
            key == K_RIGHT        .or. ;  //--> cursor right 
            key == K_CTRL_DOWN    .or. ;  //--> Ctrl-Cursor-Down
            key == K_CTRL_RIGHT   .or. ;  //--> Ctrl-Cursor-Right
            key == K_TAB                  //--> Tab 
         obj:exitState := GE_DOWN         //    next GET item 
         exit
         
      case  key == K_UP           .or. ;  //--> cursor up 
            key == K_LEFT         .or. ;  //--> cursor left 
            key == K_CTRL_UP      .or. ;  //--> Ctrl-Cursor-Up
            key == K_CTRL_LEFT    .or. ;  //--> Ctrl-Cursor-Left
            key == K_SH_TAB               //--> Shift-Tab 
         obj:exitState := GE_UP           //    prev GET item 
         exit
         
      case  key == K_HOME         .or. ;  //--> Home 
            key == K_CTRL_HOME            //--> Ctrl-Home 
         obj:exitState := GE_TOP          //    first GET in list
         exit

      case  key == K_END          .or. ;  //--> End 
            key == K_CTRL_END             //--> Ctrl-End 
         obj:exitState := GE_BOTTOM       //    last GET in list 
         exit

      case  key == K_PGUP         .or. ;  //--> PageUp 
            key == K_PGDN                 //--> PageDown 
         obj:exitState := GE_WRITE        //    exit GET
         exit

      case  key == K_LBUTTONUP  .or. ;    //--> mouse left button
            key == K_RBUTTONUP  .or. ;    //--> mouse right button 
            key == K_LDBLCLK              //--> mouse left button double click
         iHitTest := obj:HitTest(Mrow(.T.), MCol(.T.), .T.)
         if inGet
            if iHitTest == 0              //  - mouse click outside curr,
               obj:exitState := GE_MOUSE  //    GetReader() will search 
               exit                       //    for the corresp. GET field
            elseif iHitTest < 0 .and. !hasFSblock
	            obj:Select()               // simulate push
               obj:exitState := GE_ENTER  //   simulate Enter key
               exit                       // 
            endif                         //  - otherwise stay in GET field
         else
            if iHitTest < 0
	            obj:Select()               // simulate push
            endif
            exit                          //  - no Get, exit
         endif
               
	   case  key > 0 .and. key < 255 .and. chr(key) $ " xXyYtT+pP"
	      obj:Select()                     // simulate push
         if inGet .and. !hasFSblock       // - in @..GET/READ w/o Fblock/Sblock: 
            obj:exitState := GE_ENTER     //   simulate Enter key
            exit                          // 
         endif
	   endcase
   enddo

   if !inGet  
      obj:exitState := GE_NOTUSED
   endif

   obj:KillFocus()
   return obj:Buffer
   
/* eof */
