* For FlagShip 6 
* Copyright (c) 2002..06 by multisoft GmbH, Munich, Germany 
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the 
* modified source code when your application is distributed in source, 
* but may not remove the copyright notice.
* 
* Release $Date: 2006/07/28 08:02:21 $
*
* Default Input handler for one RadioButton, called from RadioButton:Show() 
*
* You may assign YOUR own handler UDF via code block :Handler, i.e.
* oRadButt:Handler := {|obj| MyUdf(obj) }
* which is then called from oRadButt:Show() automatically instead 
* of this default handler.
*
* Compile     : FlagShip -c -na -m -io=a radbutthand.prg
* Link Linux  : FlagShip myapp*.prg radbutthand.o
* Link Windows: FlagShip myapp*.prg radbutthand.obj

#include "boxclass.fh"
#include "inkey.fh"
#include "box.fh"
#include "button.fh"
#include "set.fh"
#include "FSerrmsg.h"
#include "getexit.fh"

/* --------------------------------------------------------------------------
 * Default RadioButton input handler for ONE Radio Button
 * This function should be used for oRadButt:Show() only
 *
 * Parameter passed from oRadButt:Show() via oRadButt:Handler block,
 * i.e. here oRadButt:Handler := {|obj| RadButtHandler(obj) }
 *    obj = current object
 * Returns current obj:Buffer status or NIL on error
 */
function RadButtHandler(obj)
   local key, iHitTest  as intvar
   local cAppType       as char
   local inGet          as logic

   if valtype(obj) != "O" .or. !IsObjClass(obj, "_BoxButt") 
      _Rt_Error(RTARGTYP, "RadButtHandler(): wrong object class - " + ;
         if(valtype(obj) != "O", "not object", IsObjClass(obj)[1]) )
      return NIL 
   endif

// cAppType := AppIoMode()   // i/o mode: G,T,B
// obj:FilterKeys := .F.

   obj:SetFocus()
   obj:Display()

   inGet := valtype(obj:exitState) == "N" .and. obj:exitState >= GE_NOTSET

   while .T.
      key := InkeyTrap(0)     // same as Inkey(0) but considers SET KEY

*     For debug to stderr:
*     ??## "-> RadButtHandler() key=",ltrim(key),"=",inkey2str(key)

      do case
      case  key == K_ESC                  //--> Esc 
         obj:exitState := GE_ESCAPE       //    exit READ
         exit
         
      case  key == K_ENTER       .or. ;   //--> Return 
            key == K_CTRL_ENTER           
         obj:exitState := GE_ENTER        //    next GET
         exit
         
      case  key == K_DOWN         .or. ;  //--> cursor down 
            key == K_RIGHT        .or. ;  //--> cursor right 
            key == K_CTRL_DOWN    .or. ;  //--> Ctrl-Cursor-Down
            key == K_CTRL_RIGHT   .or. ;  //--> Ctrl-Cursor-Right
            key == K_TAB                  //--> Tab 
         obj:exitState := GE_DOWN         //    next GET item 
         exit
         
      case  key == K_UP           .or. ;  //--> cursor up 
            key == K_LEFT         .or. ;  //--> cursor left 
            key == K_CTRL_UP      .or. ;  //--> Ctrl-Cursor-Up
            key == K_CTRL_LEFT    .or. ;  //--> Ctrl-Cursor-Left
            key == K_SH_TAB               //--> Shift-Tab 
         obj:exitState := GE_UP           //    prev GET item 
         exit
         
      case  key == K_HOME         .or. ;  //--> Home 
            key == K_CTRL_HOME            //--> Ctrl-Home 
         obj:exitState := GE_TOP          //    first GET in list
         exit

      case  key == K_END          .or. ;  //--> End 
            key == K_CTRL_END             //--> Ctrl-End 
         obj:exitState := GE_BOTTOM       //    last GET in list 
         exit

      case  key == K_PGUP         .or. ;  //--> PageUp 
            key == K_PGDN                 //--> PageDown 
         obj:exitState := GE_WRITE        //    exit GET
         exit

      case key >= K_LBUTTONUP .and. ;         
           ascan(m->_aGlobSetting[GSET_G_A_RADBUT_MOUSE], key) > 0
         if inGet
            iHitTest := obj:HitTest(Mrow(.T.), MCol(.T.), .T.)
            if iHitTest == 0              //  - mouse click outside curr,
               obj:exitState := GE_MOUSE  //    GetReader() will search 
               exit                       //    for the corresp. GET field
            endif                         //  - otherwise stay in GET field
         else
            exit                          //  - no Get, exit
         endif
               
	   case  key > 0 .and. key < 255 .and. chr(key) $ " xX"
	      obj:Select()                     // toggle on/off
         exit
	   case  key > 0 .and. key < 255 .and. chr(key) $ "+yYtT"
	      obj:Select(.T.)                  // set ON
         exit
	   case  key > 0 .and. key < 255 .and. chr(key) $ "-nNfF"
	      obj:Select(.F.)                  // set OFF
         exit
         
	   endcase
   enddo

   if !inGet  
      obj:exitState := GE_NOTUSED
   endif

   obj:KillFocus()
   return obj:Buffer
   
/* eof */
