#ifdef COMMENT_ONLY
* For FlagShip 6 
* Copyright (c) 2002..07 by multisoft, Munich, Germany 
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the 
* modified source code when your application is distributed in source, 
* but may not remove the copyright notice.
* 
* Release $Date: 2007/09/28 08:45:54 $
*
* Default Input handler for TbrowseDB() called via 
* oTbrowse:Exec(). You may replace this handler by your own 
* via oTbrowse:Handler ACC/ASS
*
* Compile     : FlagShip -c -na -m -io=a tbrowsedbhand.prg
* Link Linux  : FlagShip myapp*.prg tbrowsedbhand.o
* Link Windows: FlagShip myapp*.prg tbrowsedbhand.obj
#endif

#include "set.fh"
#include "inkey.fh"
#include "basclass.fh"
#include "tbrclass.fh"
#include "button.fh"

#define ACCEPT_WHEEL    /* disable if not required */

PROTO STATIC FUNCTION myEdit(oTbr as Tbrowse, [lAfterAppend as logic])
PROTO STATIC FUNCTION myAppend(oTbr,lEdit)

/* --------------------------------------------------------------------------
 * Default TbrowseDB() input handler 
 * This function (or any other UDF assined to oTbrowse:Handler block) 
 * is invoked automatically via oTbr:Exec() or via manual call of 
 * TbrDbHandler(oTbr). It is similar to TbrHandler(oTbr,arr) which does 
 * not support dbf handling and require the array name as 2nd param
 * 
 * Notes: 
 * 1) if the Tbrowse object is instantiated via TbrowseDB(), it set
 *    oTbr:Handler := {|obj| TbrDbHandler(obj)}
 * 2) if the Tbrowse object is instantiated via Tbrowse{} or TbrowseNew(),
 *    it set oTbr:Handler := NIL 
 *    so you need to use e.g. oTbr:Handler := {|obj| TbrHandler(obj,myArr)} 
 *    or oTbr:Handler := {|obj| TbrDbHandler(obj)} to use this handler
 * 3) oTbr:ReadOnly := .F. and oTbr:CanAppend := .T. is default
 * 4) oTbr:IncrSearch := .F. (disabled) per default
 * 5) see examples of tbrowse_db.prg and tbrowse_ar.prg
 *
 * Parameters:
 *    oTbr : current Tbrowse object
 */
FUNCTION TbrDbHandler(oTbr as Tbrowse) as usual
   local key            as intvar
   local lEdit, isUpper as logic
   local lAppend        as logic
   local nCursOn        as numer
   local iCurrRec       as intvar         
   local cSearch := ""  as char                 // or static if required
   
   lEdit   := !oTbr:ReadOnly    
   lAppend := oTbr:CanAppend  
   isUpper := IndexOrd() > 0 .and. ;
              upper(left(strtran(indexkey()," ",""),6)) == "UPPER(" 
   nCursOn := SetCursor()
   SetCursor(0)                                 // disable cursor

   WHILE .T.
      WHILE (!oTbr:STABLE)                      // (re)build screen,
         oTbr:STABILIZE()                       // wait for stabilizing
         IF NEXTKEY() != 0                      // optional:
      EXIT                                      // manage async.input
         ENDIF
      ENDDO

#ifdef FlagShip5
      key := InkeyTrap(0, INKEY_ALL_BUT_MOVE )  // get key pressed, process SET KEY
#else
      key := Inkey(0)                           // get key pressed
#endif

      DO CASE
      CASE key = K_LEFT .or. key = K_CTRL_S     // cursor left, prev.column
         oTbr:LEFT()                         
      CASE key = K_RIGHT .or. key= K_CTRL_D     // cursor right, next column
         oTbr:RIGHT()                        
      CASE key = K_UP   .or. key = K_CTRL_E     // cursor up, prev.record
         oTbr:UP()                           
      CASE key = K_DOWN .or. key = K_CTRL_X     // cursor down, next record/append
        oTBr:DOWN() 
        oTBr:ForceStable()
         if lAppend .and. oTbr:HitBottom
            myAppend(oTBr, lEdit)               // add new record and edit
         endif  
         cSearch := ""
      CASE key = K_PGUP .or. key = K_CTRL_R     // page-up, prev.record
         oTbr:PAGEUP()                       
         cSearch := ""
      CASE key = K_PGDN .or. key = K_CTRL_C     // page-down
         oTbr:PAGEDOWN()                     
         cSearch := ""
      CASE key = K_CTRL_PGUP                    // Ctrl page-up, first record
         oTbr:GOTOP()                        
         cSearch := ""
      CASE key = K_CTRL_PGDN                    // Ctrl page-down, last record
         oTbr:GOBOTTOM()                     
         cSearch := ""
      CASE key = K_HOME .or. key = K_CTRL_A     // home, first column
         oTbr:PANHOME()                         
      CASE key = K_END  .or. key = K_CTRL_F     // end, last column
         oTbr:PANEND()                         
      CASE key = K_CTRL_HOME                    // First column
         oTbr:ColPos := 1
      CASE key = K_CTRL_END                     // Last column
         oTbr:ColPos := oTbr:ColCount
      CASE key = K_ESC                          // escape, terminates browsing
         EXIT                                  
      CASE key == K_RETURN .or. ;               // return or mouse double-click: edit curr.cell
          (key == K_LDBLCLK .and. oTbr:HitTest(Mrow(),Mcol()) == HTCELL) 
         if lEdit
            myEdit(oTbr)  
         endif 
         cSearch := ""
      CASE key = K_TAB                          // scroll --> (except in READ)
         if valtype(oTbr:GetEnabled) == "L" .and. oTbr:GetEnabled
            KEYBOARD chr(K_TAB)                 // in READ: next GET field
            EXIT                                  
         else
            oTbr:ScrollRight()
         endif
      CASE key = K_SH_TAB                       // scroll <-- (except in READ)
         if valtype(oTbr:GetEnabled) == "L" .and. oTbr:GetEnabled
            KEYBOARD chr(K_SH_TAB)              // in READ: previous GET field
            EXIT                                  
         else
            oTbr:ScrollLeft()
         endif
#ifdef ACCEPT_WHEEL         
      CASE key = K_MOUSEWHEELUP                 // mouse wheel up, prev.record
         oTbr:UP()                           
      CASE key = K_MOUSEWHEELDN                 // mouse wheel down, next record
        oTBr:DOWN() 
      CASE key == K_SH_MOUSEWHEELUP .or. ;
           key == K_CTRL_MOUSEWHEELUP .or. ;
           key == K_ALT_MOUSEWHEELUP           // mouse wheel up + modifier, prev.page
         oTbr:PAGEUP()                       
         cSearch := ""
      CASE key == K_SH_MOUSEWHEELDN .or. ;
           key == K_CTRL_MOUSEWHEELDN .or. ;
           key == K_ALT_MOUSEWHEELDN           // mouse wheel down + modifier, next page
         oTbr:PAGEDOWN()                     
         cSearch := ""
#endif
      CASE key >= K_MOUSEMOVE .and. key <= K_MAX_KEYNUM  
         // ignore other mouse buttons                

      case oTbr:IncrSearch .and. IndexOrd() > 0 .and. key >= 32 .and. key < 255
         // incremental search (activate by oTbr:IncrSearch := .T.)
         // is reset here by cursor movement, Enter or when data not found
         cSearch += if(isUpper, upper(chr(key)), chr(key))
         iCurrRec := recno()
         seek (cSearch)
         if !found()
            ?? chr(7)
            goto (iCurrRec)
            cSearch := ""
         endif
         oTbr:RefreshAll()

      OTHERWISE
         ?? CHR(7)                             // invalid key
      ENDCASE
   ENDDO                                       // system is unstable now

   SetCursor(nCursOn)                          // restore cursor
   RETURN NIL

/* --------------------------------------------------------------------
 * Edit the current cell
 *    see also FS manual OBJ.3.6
 *    Note: all calculations are in coord
 */
STATIC FUNCTION myEdit(oTbr as Tbrowse, [lAfterAppend as logic])
   local data           as usual
   local nRow, nCol     as numeric   
   local r1, c1, r2, c2 as numeric
   local oTbCol         as TbColumn
   local block, sScreen as usual
   local oRect          as Rectangle
   local getList := {}  as array
   local lReadOnly, lUpdated as logic
   local nCursorOn      as numeric
   local cFieldType     as char
   local cPict          as char
   local nWidth         as numeric
   local iColNum        as intvar
   local lSavePixel := Set(_SET_PIXEL) as logic // current SET PIXEL on/off
   
   Set(_SET_PIXEL, .F.) // all calculations are in coord here

   nCursorOn := Set(_SET_CURSOR, 1)  // set cursor visible
   
   nRow := ROW() 
   nCol := COL() // in coord
   
   IF .not. oTbr:STABLE
      oTbr:ForceStable()
   ENDIF

   iColNum := oTbr:COLPOS
   oTbCol  := oTbr:GetColumn(iColNum)
   block   := oTbCol:Block
   data    := Eval(block)       // Retrieve data using the column block

   lReadOnly := oTbr:ReadOnly .or. oTbCol:ReadOnly
   lUpdated := .F.
   
   if lReadOnly
      @ 0,50 say "Read-Only"
   endif
      
   cFieldType := valtype(Eval(block))
   if cFieldType == "M"                   // Memo field
      if valtype(oTbCol:MemoPos) == "O"
         oRect := oTbCol:MemoPos          // coordinates specified
         r1 := oRect:Top   
         c1 := oRect:Left   
         r2 := oRect:Bottom   
         c2 := oRect:Right
      else                                // calculate coordinates
         r1 := oTbr:Ntop    + 2
         c1 := oTbr:Nleft   + 2
         r2 := oTbr:Nbottom - 2
         c2 := oTbr:Nright  - 2
         if (r2 - r1) <= 3
            r1 := 1
            r2 := maxrow() -1
         endif
         if (c2 - c1) <= 3
            c1 := 1
            c2 := maxcol() -1
         endif
      endif 
      if AppIoMode() == "T"               // in Terminal i/o, draw box
         sScreen := SaveScreen(r1,c1,r2,c2)
         @ r1++,c1++, r2--,c2-- BOX m->_aGlobSetting[GSET_T_C_COLDBOX] 
      endif
         
      data := MemoEdit(data, r1, c1,r2, c2, ;
                        !lReadOnly, NIL,NIL,NIL,NIL, ;
                        NIL,NIL,NIL,.F.,NIL,.T.)      // edit or view only
        
      if AppIoMode() == "T"
         RestScreen(--r1,--c1,++r2,++c2, sScreen)
      endif
      
   elseif !lReadOnly
      r1 := nRow
      c1 := nCol
      nWidth := oTbr:ColWidth(iColNum)
      if AppIoMode() == "G" .and. nWidth > oTbr:ColVisibleWidth() 
         c1 := max(oTbr:Nleft, c1 - nWidth + oTbr:ColVisibleWidth())
      endif 
      if !empty(oTbCol:Picture)
         @ r1,c1 GET data PICT (oTbCol:Picture) 
      else
         nWidth := oTbr:Nright - oTbr:Nleft
         if cFieldType == "C" .and. oTbr:ColWidth(iColNum) > nWidth
            cPict := "@S" + ltrim(int(nWidth))
            c1 := oTbr:Nleft
            @ r1,c1 GET data PICT (cPict) 
         else   
            @ r1,c1 GET data 
         endif 
      endif
      READ clear                    // remove GET widget at finishing
   endif

   if LastKey() != K_ESC .and. !lReadOnly
      Eval(block, data)             // REPLACE if the block is read/write,
      oTbr:RefreshAll()             // or do nothing if block is read only
      oTbr:Updated := .T.
      lUpdated     := .T.
*     oTbr:RefreshCurrent()         // or (faster) if w/o index
   endif

   if lReadOnly
      if cFieldType != "M"
         inkey(2)                   // allow display "Read-Only"
      endif                   
      @ 0,50 say space(StrLen2Space("Read-Only"))   // clear message
   endif
   
   if lUpdated .and. valtype(oTbr:OnUpdate) == "B"
      eval(oTbr:OnUpdate, oTbr, if(lAfterAppend,3,2))   // edited
   endif

   Set(_SET_CURSOR, nCursorOn)      // restore cursor on/off
   SetPos (nRow, nCol)              // reposit cursor
   Set(_SET_PIXEL, lSavePixel)      // restore SET PIXEL on/off
   return NIL

/* --------------------------------------------------------------------
 * Append new record
 */
STATIC FUNCTION myAppend(oTbr, lEdit)
   APPEND BLANK                     // add new, empty record
   skip 0
   unlock
   
   oTbr:Updated := .T.
*  oTbr:GoBottom()
*  oTbr:RefreshCurrent()            // ok (faster) if w/o index
   oTbr:RefreshAll()                // required if with index
   oTbr:ForceStable()

   if oTbr:Updated .and. valtype(oTbr:OnUpdate) == "B"
      eval(oTbr:OnUpdate, oTbr, 1)
   endif
   if lEdit             
      myEdit(oTbr, .T.)                  // edit data
   endif
   return NIL

* eof *
