#ifdef COMMENT_ONLY
* For FlagShip 6 
* Copyright (c) 2004..07 by multisoft, Munich, Germany 
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the 
* modified source code when your application is distributed in source, 
* but may not remove the copyright notice.
* 
* Release $Date: 2007/09/28 08:46:31 $
*
* Default Input handler for TbrowseArr() called via 
* oTbrowse:Exec(). You may replace this handler by your own 
* via oTbrowse:Handler ACC/ASS
*
* Compile     : FlagShip -c -na -m -io=a tbrowsehand.prg
* Link Linux  : FlagShip myapp*.prg tbrowsehand.o
* Link Windows: FlagShip myapp*.prg tbrowsehand.obj
#endif

#include "set.fh"
#include "inkey.fh"
#include "basclass.fh"
#include "tbrclass.fh"
#include "FSerrmsg.h"

#define ACCEPT_WHEEL    /* disable if not required */

PROTO STATIC FUNCTION myEdit(oTbr as Tbrowse, [lAfterAppend as logic])
PROTO STATIC FUNCTION myAppend(oTbr,lEdit)

/* --------------------------------------------------------------------------
 * Default TbrowseArr() input handler 
 * This function (or any other UDF assined to oTbrowse:Handler block) 
 * is invoked automatically via oTbr:Exec() or via manual call of 
 * TbrHandler(oTbr.array). It is similar to TbrDbHandler(oTbr) which  
 * support dbf handling instead
 * 
 * Notes: 
 * 1) if the Tbrowse object is instantiated via TbrowseArr(), it set
 *    oTbr:Handler       := { |obj|     TbrHandler(obj) }
 *    oTbr:gobottomblock := { |obj|     TbrArrBottom(obj) }   
 *    oTbr:gotopblock    := { |obj|     TbrArrTop(obj) }      
 *    oTbr:skipblock     := { |xx,obj|  TbrArrSkip(xx, obj) } 
 * 2) if the Tbrowse object is instantiated via Tbrowse{} or TbrowseNew(),
 *    it set oTbr:Handler := NIL 
 *    so you need to use e.g. oTbr:Handler := {|obj| TbrHandler(obj,myArr)} 
 *    or oTbr:Handler := {|obj| TbrDbHandler(obj)} to use this handler
 *    and assign the corresponding skip blocks
 * 3) oTbr:CanEdit := .T. and oTbr:CanAppend := .T. per default
 * 4) see examples of tbrowse_ar.prg and tbrowse_db.prg
 *
 * Parameters:
 *    oTbr : current Tbrowse object
 */
FUNCTION TbrHandler(oTbr as Tbrowse) as usual
   local key      as intvar
   local lEdit    as logic
   local lAppend  as logic
   local nCursOn  as numer
   
   lEdit   := !oTbr:ReadOnly    
   lAppend := oTbr:CanAppend  
   nCursOn := SetCursor()
   SetCursor(0)                                 // disable cursor
   
   WHILE .T.
      WHILE (!oTbr:STABLE)                      // (re)build screen,
         oTbr:STABILIZE()                       // wait for stabilizing
         IF NEXTKEY() != 0                      // optional:
      EXIT                                      // manage async.input
         ENDIF
      ENDDO

#ifdef FlagShip5
      key := InkeyTrap(0)                       // get key pressed, process SET KEY
#else
      key := Inkey(0)                           // get key pressed
#endif

      DO CASE
      CASE key = K_LEFT .or. key = K_CTRL_S     // cursor left, prev.column
         oTbr:LEFT()                         
      CASE key = K_RIGHT .or. key= K_CTRL_D     // cursor right, next column
         oTbr:RIGHT()                        
      CASE key = K_UP   .or. key = K_CTRL_E     // cursor up, prev.record
         oTbr:UP()                           
      CASE key = K_DOWN .or. key = K_CTRL_X     // cursor down, next record/append
        oTBr:DOWN() 
        oTBr:ForceStable()
         if lAppend .and. oTbr:HitBottom
            myAppend(oTBr, lEdit)               // add new record and edit
         endif  
      CASE key = K_PGUP .or. key = K_CTRL_R     // page-up, prev.record
         oTbr:PAGEUP()                       
      CASE key = K_PGDN .or. key = K_CTRL_C     // page-down
         oTbr:PAGEDOWN()                     
      CASE key = K_CTRL_PGUP                    // Ctrl page-up, first record
         oTbr:GOTOP()                        
      CASE key = K_CTRL_PGDN                    // Ctrl page-down, last record
         oTbr:GOBOTTOM()                     
      CASE key = K_HOME .or. key = K_CTRL_A     // home, first column
         oTbr:PANHOME()                         
      CASE key = K_END  .or. key = K_CTRL_F     // end, last column
         oTbr:PANEND()                         
      CASE key = K_CTRL_HOME                    // First column
         oTbr:ColPos := 1
      CASE key = K_CTRL_END                     // Last column
         oTbr:ColPos := oTbr:ColCount
      CASE key = K_ESC                          // escape, terminates browsing
         EXIT                                  
      CASE key = K_RETURN                       // return, edit curr.cell
         if lEdit
            myEdit(oTbr)  
         endif                     
      CASE key = K_TAB                          // scroll --> (except in READ)
         if valtype(oTbr:GetEnabled) == "L" .and. oTbr:GetEnabled
            KEYBOARD chr(K_TAB)                 // in READ: next GET field
            EXIT                                  
         else
            oTbr:ScrollRight()
         endif
      CASE key = K_SH_TAB                       // scroll <-- (except in READ)
         if valtype(oTbr:GetEnabled) == "L" .and. oTbr:GetEnabled
            KEYBOARD chr(K_SH_TAB)              // in READ: previous GET field
            EXIT                                  
         else
            oTbr:ScrollLeft()
         endif
#ifdef ACCEPT_WHEEL         
      CASE key = K_MOUSEWHEELUP                 // mouse wheel up, prev.record
         oTbr:UP()                           
      CASE key = K_MOUSEWHEELDN                 // mouse wheel down, next record
        oTBr:DOWN() 
      CASE key == K_SH_MOUSEWHEELUP .or. ;
           key == K_CTRL_MOUSEWHEELUP .or. ;
           key == K_ALT_MOUSEWHEELUP           // mouse wheel up + modifier, prev.page
         oTbr:PAGEUP()                       
      CASE key == K_SH_MOUSEWHEELDN .or. ;
           key == K_CTRL_MOUSEWHEELDN .or. ;
           key == K_ALT_MOUSEWHEELDN           // mouse wheel down + modifier, next page
         oTbr:PAGEDOWN()                     
#endif
      CASE key >= K_MOUSEMOVE .and. key <= K_MAX_KEYNUM
         // ignore other mouse buttons
         loop
      OTHERWISE
         ?? CHR(7)                             // invalid key
      ENDCASE
   ENDDO                                       // system is unstable now
   
   SetCursor(nCursOn)                          // restore cursor
   RETURN NIL

/* --------------------------------------------------------------------
 * Default array skipper, assigned to :skipblock
 *                                                 
 * In : p1 (N) Number of Records to skip   
 *      p2 (O) the current Tbrowse object                                  
 * Out: fn (N) Number of Records really skipped                              
 */
FUNCTION TbrArrSkip(toskip as intvar, obj as Tbrowse)
   local iOld as IntVar

   if valtype(obj) != "O" .or. !IsObjClass(obj,"_Tbrowse") .or. ;
         valtype(obj:UserArray) != "A" .or. len(obj:UserArray) == 0
      _Rt_Error(RTOBJNOPROP, "TbrArrSkip() requires 2-dim array,;assign via oTbr:UserArray")
      return NIL
   endif

   if obj:UserArrayPos < 1
      obj:UserArrayPos := 1
   endif
   
   iOld := obj:UserArrayPos

   obj:UserArrayPos += toskip
   
   do case
   case toskip > 0                     // skip forward
      if obj:UserArrayPos > len(obj:UserArray)
         obj:UserArrayPos := len(obj:UserArray)
      endif
   case toskip < 0                     // skip backward
      if obj:UserArrayPos < 1
         obj:UserArrayPos := 1
      endif
   endcase
   return obj:UserArrayPos - iOld      
         
/* --------------------------------------------------------------------
 * Default array skipper, assigned to :gotopBlock
 *                                                 
 * In : p1 (O) the current Tbrowse object                                  
 * Out: 1                              
 */
FUNCTION TbrArrTop(obj as Tbrowse)
   if valtype(obj) != "O" .or. !IsObjClass(obj,"_Tbrowse") .or. ;
         valtype(obj:UserArray) != "A" .or. len(obj:UserArray) == 0
      _Rt_Error(RTOBJNOPROP, "TbrArrTop() requires 2-dim array,;assign via oTbr:UserArray")
      return 0
   endif

   obj:UserArrayPos := 1
   return 1     

/* --------------------------------------------------------------------
 * Default array skipper, assigned to :gobottomBlock
 *                                                 
 * In : p1 (O) the current Tbrowse object                                  
 * Out: last array element                              
 */
FUNCTION TbrArrBottom(obj as Tbrowse)
   if valtype(obj) != "O" .or. !IsObjClass(obj,"_Tbrowse") .or. ;
         valtype(obj:UserArray) != "A" .or. len(obj:UserArray) == 0
      _Rt_Error(RTOBJNOPROP, "TbrArrBottom() requires 2-dim array,;assign via oTbr:UserArray")
      return 0
   endif

   obj:UserArrayPos := len(obj:UserArray)
   return obj:UserArrayPos     

/* --------------------------------------------------------------------
 * Edit the current cell
 *    see also FS manual OBJ.3.6
 *    Note: all calculations are in coord
 */
STATIC FUNCTION myEdit(oTbr as Tbrowse, [lAfterAppend as logic])
   local data                 as usual
   local nRow, nCol           as numeric   
   local r1, c1, r2, c2       as numeric
   local oTbCol               as TbColumn
   local block, sScreen       as usual
   local oRect                as Rectangle
   local getList := {}        as array
   local lReadOnly, lUpdated  as logic
   local nCursorOn            as numeric
   local cFieldType           as char
   local cPict                as char
   local nWidth               as numeric
   local iColNum              as intvar
   local lSavePixel := Set(_SET_PIXEL) as logic // current SET PIXEL on/off
   
   Set(_SET_PIXEL, .F.) // all calculations are in coord here

   nCursorOn := Set(_SET_CURSOR, 1)  // set cursor visible
      
   nRow := ROW() 
   nCol := COL() // in coord
   
   IF .not. oTbr:STABLE
      oTbr:ForceStable()
   ENDIF

   iColNum := oTbr:COLPOS
   oTbCol  := oTbr:GetColumn(iColNum)
   block   := oTbCol:Block
   data    := Eval(block)       // Retrieve data using the column block
   lUpdated := .F.

   lReadOnly := oTbr:ReadOnly .or. oTbCol:ReadOnly

   if lReadOnly
      @ 0,50 say "Read-Only"
   endif
      
   cFieldType := valtype(Eval(block))
   if !lReadOnly
      r1 := nRow
      c1 := nCol
      nWidth := oTbr:ColWidth(iColNum)
      if AppIoMode() == "G" .and. nWidth > oTbr:ColVisibleWidth() 
         c1 := max(oTbr:Nleft, c1 - nWidth + oTbr:ColVisibleWidth())
      endif 
      if !empty(oTbCol:Picture)
         @ r1,c1 GET data PICT (oTbCol:Picture) 
      else
         nWidth := oTbr:Nright - oTbr:Nleft
         if cFieldType == "C" .and. oTbr:ColWidth(iColNum) > nWidth
            cPict := "@S" + ltrim(int(nWidth))
            c1 := oTbr:Nleft
            @ r1,c1 GET data PICT (cPict) 
         else   
            @ r1,c1 GET data 
         endif 
      endif
      READ clear                    // remove GET widget at finishing
   endif

   if LastKey() != 27 .and. !lReadOnly
      Eval(block, data)             // REPLACE if the block is read/write,
      oTbr:RefreshCurrent()         // or do nothing if block is read only
      oTbr:Updated := .T.
      lUpdated     := .T.
   endif

   if lReadOnly
      if cFieldType != "M"
         inkey(2)                   // allow display "Read-Only"
      endif                   
      @ 0,50 say space(StrLen2Space("Read-Only"))   // clear message
   endif
   
   if lUpdated .and. valtype(oTbr:OnUpdate) == "B"
      eval(oTbr:OnUpdate, oTbr, if(lAfterAppend,3,2))   // edited
   endif
   
   Set(_SET_CURSOR, nCursorOn)      // restore cursor on/off
   SetPos (nRow, nCol)              // reposit cursor
   Set(_SET_PIXEL, lSavePixel)      // restore SET PIXEL on/off
   return NIL

/* --------------------------------------------------------------------
 * Append new record
 */
STATIC FUNCTION myAppend(oTbr, lEdit)
   local arr, ii as usual
   
   if valtype(oTbr:UserArray) != "A"
      return NIL
   endif
   arr := aclone(oTbr:UserArray[1])
   if valtype(arr) == "A"
      for ii := 1 to len(arr)
         do case
         case valtype(arr[ii]) == "C"
            arr[ii] := space(len(arr[ii]))
         case valtype(arr[ii]) == "N"
            arr[ii] := 0
         case valtype(arr[ii]) == "L"
            arr[ii] := .F.
         case valtype(arr[ii]) == "D"
            arr[ii] := ctod("")
         endcase
      next
   endif
   aadd(oTbr:UserArray, arr)
        
   oTbr:Updated := .T.
   oTbr:GoBottom()
   oTbr:RefreshAll()  
   oTbr:ForceStable()
   
   if oTbr:Updated .and. valtype(oTbr:OnUpdate) == "B"
      eval(oTbr:OnUpdate, oTbr, 1)
   endif
   if lEdit             
      myEdit(oTbr, .T.)                  // edit data
   endif
   return NIL

* eof *
