*****************************************************************************
* Start-up for Unix, with proper terminal setting, called from dbu.prg
* Compile with -na switch, e.g. FlagShip dbu*.prg -na
* Not required and applicable for Clipper.
* (c) 1995-2003 by multisoft GmbH (JVB), public domain
*
* Release $Date: 2005/01/04 09:35:26 $

#ifndef FlagShip
   function dbuFSinit()                        // dummy fn for Clipper
   return NIL
#else

** #define LINUX_CAN_CHANGE_COLOR              /* see following comments */

/* Currently, the Linux curses cannot re-assign color terminals properly
   from running process. This may be changed with the next release. You may
   then use the fs_set(...) function also for Linux by simply uncommenting
   (activating) the #define LINUX_CAN_CHANGE_COLOR statement above.
*/

#Cinline
#include "FSopenc.h"
 extern char __argv0[];                        /* name of the executable */
#endCinline

****************** dbuFSinit()
*
function dbuFSinit (par1ref, par2ref, par3ref, par4ref, par5ref)
   LOCAL ii, key, currTerm := getenv("TERM")
   LOCAL isUpper, isLower, isPathL, isPathU
   LOCAL termDir := getenv("TERMINFO")
   LOCAL dummy, locPar[5], ClipParam := {"","",""}

/* Note: You may extend BOTH the arrays oldTerm[] and newTerm[] for
   other terminals, when required.
*/
   LOCAL oldTerm := {"console", "ansi",  "ansi-43",  "vt100",  "vt200"  }, ;
         newTerm := {"fslinux", "FSansi","FSansi-43","FSVT100","FSVT200"}

   isUpper := fs_set ("upper", .F.)    // the automatical upper & lower
   isLower := fs_set ("lower", .F.)    // translations must be tempo- 
   isPathL := fs_set ("pathL", .F.)    // rarily disabled when searching
   isPathU := fs_set ("pathU", .F.)    // for eg. /usr/.../F/FSansi

   par1ref := locPar[1] := if (valtype(par1ref) == "C", par1ref, "")
   par2ref := locPar[2] := if (valtype(par2ref) == "C", par2ref, "")
   par3ref := locPar[3] := if (valtype(par3ref) == "C", par3ref, "")
   par4ref := locPar[4] := if (valtype(par4ref) == "C", par4ref, "")
   par5ref := locPar[5] := if (valtype(par5ref) == "C", par5ref, "")

   for ii := 1 to len(oldTerm)
      if !(currTerm == oldTerm[ii])
         loop
      endif
      termDir := if (empty(termDir), "/usr/lib/terminfo", termDir)
      if file(termDir + "/" + substr(newTerm[ii],1,1) + "/" + newTerm[ii])
         ?
         ? "TERM is currently set to '" + oldTerm[ii] + "'. Would you like to"
         ? "     temporarily change it to '" + newTerm[ii] + ;
           "' for color display (y/[n]) ? "
         key := upper(chr(inkey(0)))
         ?? key
         if key == "Y"
            fs_set ("upper", isUpper)
            fs_set ("lower", isLower)
            fs_set ("pathL", isPathL)
            fs_set ("pathU", isPathU)

#ifdef LINUX_CHANGE_COLOR
            fs_set ("setenv", "TERM", newTerm[ii], .T.)
#else
            if ii == 1 .and. getenv("OSTYPE") == "Linux"
               fs_set ("setenv", "TERM", newTerm[ii], .F.)
               initLinux(par1ref, par2ref, par3ref, par4ref, par5ref)
            else
               fs_set ("setenv", "TERM", newTerm[ii], .T.)
            endif
#endif

         endif
         exit
      endif
   next
   fs_set ("upper", isUpper)      // restore the current defaults
   fs_set ("lower", isLower)
   fs_set ("pathL", isPathL)
   fs_set ("pathU", isPathU)
 
// parse the extended parameters, store Param1..3 for Clipper
 
   for ii := 1 to len(locPar)
      dummy := upper(locPar[ii])
      if dummy == "-CASE"
         CaseSensit := 0
         fs_set ("lower", .F.)
         fs_set ("upper", .F.)
      elseif dummy == "-UPPER"
         CaseSensit := 2
         fs_set ("lower", .F.)
         fs_set ("upper", .T.)
      elseif dummy == "-LOWER"
         CaseSensit := 1
         fs_set ("lower", .T.)
         fs_set ("upper", .F.)
      elseif left(dummy,4) == "-FUL"
         // already recognized in dbu()
      elseif dummy == "-MAX"
         // already recognized in dbu()
      elseif dummy == "-E" .or. dummy == "/E"
         ClipParam[3] := "/E"
      elseif dummy == "-C" .or. dummy == "/C"
         ClipParam[2] := "/C"
      elseif dummy == "-M" .or. dummy == "/M"
         ClipParam[2] := "/M"
      elseif substr(dummy,1,5) == "-RDD="
         DefaultRDD := alltrim(substr(locPar[ii],6))
      elseif left(dummy,7) == "-CACHE="   
         m->iCacheRecords := val(substr(dummy, 8))  // used in dbuedit.prg
      elseif dummy == "-NOTRIM" 
         m->lTrimBrowse := .F.                      // used in dbuedit.prg
      elseif !empty(dummy)
         ClipParam[1] := locPar[ii]
      endif
   next
   par1ref := par2ref := par3ref := ""
   if !empty(ClipParam[1])
      par1ref := ClipParam[1]
      par2ref := ClipParam[2]
      par3ref := ClipParam[3]
   else
      par1ref := ClipParam[2]
      par2ref := ClipParam[3]
   endif

return NIL

****************** StartDir()
*
* The generic function StartDir() determines the current directory of the 
* executable.  If no parameters or .F. is specified, it return the path
* of the executable, with .T. the name of executable itself. The path
* will be returned as e.g. "../xx/yy/" when the executable 'a.out' 
* was invoked by "../xx/yy/a.out".  When the executable was found in the 
* Unix PATH instead, it returns "" only.
* It demonstrates the usage of inline-C and Open C API of FlagShip.
* Otherwise, you may simply use the ExecName(.T.) function!
*
FUNCTION StartDir (getExeOnly)
   LOCAL cRet := "", cExe := ""
   local cSlash := "/"
#ifdef FS_WIN32
   cSlash := "\"
#endif
   getExeOnly := if (valtype(getExeOnly) == "L", getExeOnly, .F.)

#Cinline
   SET_VAR_CHR (VAR_NAME_LOCAL(cexe), __argv0);  /* see sect EXT in manual */ 
#endCinline
   IF !getExeOnly                                   // determine path:
      if cSlash $ cExe                              // if path was given,  
         cRet := substr(cExe, 1, rat(cSlash, cExe)) // remove name of executable.
      endif
   ELSE                                             // determine executable:
      if cSlash $ cExe                              // if path was given,  
         cRet := substr(cExe, rat(cSlash, cExe) +1) // remove it 
      else
         cRet := cExe
      endif
   ENDIF  
RETURN cRet 


/* --------------------------------------------- see note above */
#ifndef LINUX_CAN_CHANGE_COLOR

STATIC function initLinux(par1, par2, par3, par4, par5 )
   LOCAL dbuCall := ""                          // name of the executable
   LOCAL key

#Cinline
   SET_VAR_CHR (VAR_NAME_LOCAL(dbucall), __argv0);
#endCinline
   ?
   ? "Note: Linux currently cannot change terminals from application, but"
   ? "you may start it by using 'newfscons " + dbuCall + " [params]'," + ;
     " see man flagship."
   ?
   ? "Should I start " + dbuCall + " as a subprocess now ([y]/n) ? "
   key := upper(chr(inkey(0)))
   key := if (empty(key), "Y", if (key != "N", "Y", key))
   ?? key
   ?
   if key == "Y"
      ?
      RUN (dbuCall + " " + par1 + " " + par2 + " " + par3 + ;
                     " " + par4 + " " + par5)
   endif
   quit
return NIL

#endif
/* enf of FlagShip */   
#endif

#ifndef FlagShip5
function rowvisible()
   return maxrow()

function appiomode()
   return "T"

function inkeytrap(n)
   if pcount() == 0
      return inkey()
   endif
   return inkey(n)
#endif
** eof **

