#ifdef INFO_NOTES
// For FlagShip 4.48+ and 5.x or 6.x 
// Copyright (c) 2001..2004 by multisoft GmbH, Munich, Germany 
// 
// This is a free add-on to FlagShip.  You may modify this source according 
// to your needs and distribute the original or modified source freely with 
// your application, but may not remove the copyright notice. Note that also
// the compiled executable requires installed FlagShip of the same release.
// Author: Jan V.Balek, multisoft GmbH. Report bugs to <bugs@fship.com>
// Release $Date: 2005/05/21 08:11:20 $
#endif
/*-------------------------------------------------------------------------
 * Compile    : make (or nmake with MS-VC6++)
 * or with FS2: make fsifs2 (or "nmake fsifs2" with MS-VC6++)
 * or manually: FlagShip -na -w -m [-DFS2LIB] -Mfsi fsi.prg -o fsi
 * Execute    : fsi  -or- ./fsi  -or-  newfswin fsi  -or-  fsi -io=t
 * Use        : extensive on-line help is available, type HELP when running
 *
 * Hint: Best to create a symbolic link to /usr/bin or /usr/local/bin, e.g.
 *        "ln -s `pwd`/fsi  /usr/bin/fsi" 
 *       on Unix/Linux for an easy invocation from any working directory
 *       or "copy fsi.exe <dir_in_path>" in MS-Windows 
 *-------------------------------------------------------------------------
 * This is a simple interpreter executing FlagShip (xBase) based statements.
 * It accepts standard commands and functions available in the FS library,
 * as well as assign (via the := operator) and access of given variables. 
 *
 * This fsi is not intended to replace the FlagShip compiler. Instead,
 * it allows you a quick on-line checking the intended syntax in your
 * application, maintain databases etc. in a similar way as for example
 * the dBase, FoxPro or similar interpreters do.
 *
 * The given syntax will be tested by the FlagShip Preprocessor and
 * translated according to the <FlagShip_dir>/include/std.fh file. The 
 * FS Compiler will (and cannot) be invoked here, so the given statements 
 * will be macro-evaluated (some says macro-compiled) by the FS run time 
 * system instead. This means: not all library functions are linked-in, 
 * except you declare them by the EXTERN command (see note 1 & fsman EXTERN
 * and function _addExternals() at the bottom of this file).
 *
 * Notes:
 * 1) If you want to use also your own functions, add the objects or the
 * library in the compile line (see details in FlagShip manual sect. FSC). 
 * If you use statically linked library, you will need to add the 
 * corresponding EXTERNAL commands in this source.
 *
 * 2) The same is valid for FS2 library functions. If you own it,
 * compile with -DFS2LIB or "[n]make fsifs2". This will enable also
 * execution of interactive functions like @..GET/READ, DbEval() 
 * etc in second window.
 *
 * 3) All variable names here are prefaced by underscore to avoid 
 * misinterpretation, interference and/or hiding of user auto-private's
 *
 *-------------------------------------------------------------------------
 */

#Cinline
#ifndef FS_WIN32        /* for Unix/Linux */
# include "FSopenc.h"
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <unistd.h>
/*
 * create named pipe for data exchange with FlagShip preprocessor
 */
UDF_DECL(makepipe)
{
   char *cName;
   int  err;
   struct stat fstat;
   
   if(argc < 1 || !IS_VAR_CHR(argv[0]) || IS_VAR_EMPTY(argv[0]) )
      return FALSE_VAR;
   cName = VAR_CHR(argv[0]);
   
   err = mkfifo(cName, 0666);  /* 666 & ~u_mask = 666 & ~022 = 644 = rw-r--r-- */
   if(err != 0) {          
      err = stat(cName, &fstat);       /* error from mkfifo():          */
      if(err != 0)                     /* is the file available?        */
         return FALSE_VAR;             /* - no, probably access denied  */
      if(! S_ISFIFO(fstat.st_mode))    /* is this a pipe/fifo?          */ 
         return FALSE_VAR;             /* - no, other file type         */
   }
   return TRUE_VAR;
}
#endif
#endCinline

/* ------------------ .prg part starts here -------------------------------
 */
#include "fileio.fh"
#include "inkey.fh"
#include "set.fh"

#ifdef FS_WIN32
# define EXE_EXT ".exe"
#else
# define HAS_EGREP
# define EXE_EXT ""
#endif

#define NEWLINE chr(10)
* #define FSIO_BASIC  /* (for FS4.48:) use ACCEPT instead of GET/READ */

/* add your EXTERN statements here, if required
 */
 
PROTO FUNCTION _testForCmd(@_cmd as char, @_transl as logic) as logical
PROTO FUNCTION _testIfCmd (@_cmd as char, @_transl as logic) as logical

static _fsiErr := .F., _fsiTra := .T., _fsiWrp := .T.    as logical
static _fsiLow := .F., _fsiUpp := .F., _fsiCmd := .T.    as logical
static _useAccept := .F., _fsiCmdOpen := .F.             as logical
static _fsiFox := ""                                     as char
static _fsiDefErr                                        as codeblock
static _fsiNoErr := {|oErr| "ERROR"}                     as codeblock
static _aHistory := {}                                   as array
static _FS_WIN32 := .F.                                  as logic

***************************************************************************
* main program
*
FUNCTION fsi(_p1)
   local _dfPipe, _ii   as intVar
   local _cmd, _cmd2    as char
   local _retval, _prep as usual
   local _tmp1, _tmp2, _tmpfile, _tmpPath := "" as char
   local _errorRecurs   as logical
   local _aHistNum      as array
   
   public __tmpEval, _cBinPath := "/usr/bin/", _cInclPath := "-I/usr/include"
   public _winOut := -1
   public _color1 := "W+/B", _color2 := "W+/R"
   memvar __tmpEval, _cBinPath, _cInclPath

   _addExternals()    // link externals, see below
   
#ifdef FS_WIN32      /* MS-Windows part */
   _FS_WIN32 := .T.
   
   _tmpPath := getenv("TEMP")
   if empty(_tmpPath)
      _tmpPath := getenv("TMP")
   endif
   if empty(_tmpPath)
      _tmpPath := curdir()
   endif     
   
   /* check temp path and access rights
    */
   if empty(_p1)
      _p1 := _tmpPath + "\fsi_intepreter"
   endif
   _dfPipe := FOpen(_p1)
   if _dfPipe <= 0
      _dfPipe := FCreate(_p1)
      if _dfPipe > 0
         FClose(_dfPipe)
      endif
      _dfPipe := FOpen(_p1)
   endif   
   if _dfPipe <= 0
      alert("sorry, could not open file;" + _p1 + " for read" + ;
            ";invoke with '" + execname() + " <name>'")
      return
   endif
   FClose(_dfPipe)
#else                /* Unix/Linux part */
   if empty(_p1)
      _p1 := "/tmp/fsi_intepreter"
   endif
   if !makepipe(_p1)
      alert("sorry, could not create named pipe in " + _p1 + ;
            ";invoke with '" + execname() + " <name>'")
      return
   endif
   _dfPipe := FOpen(_p1)
   if _dfPipe <= 0
      alert("sorry, could not open named pipe " + _p1 + " for read")
      return
   endif
#endif

   /* determine/check FlagShip paths
    */
   _cBinPath  := FlagShip_dir() + PATH_SLASH + "bin" + PATH_SLASH
   if !file(_cBinPath + "FlagShip_p" + EXE_EXT, .F.)   
      _tmp1 := alltrim(getenv("FSDEBUG_COMPILER"))
      if !empty(_tmp1)
         if _FS_WIN32
            _cBinPath := strtran(_tmp1, "/", "\")
         else
            _cBinPath := strtran(_tmp1, "\", "/")
         endif
         if right(_cBinPath,1) != PATH_SLASH
            _cBinPath += PATH_SLASH
         endif
      endif      
   endif
   if !file(_cBinPath + "FlagShip_p" + EXE_EXT, .F.)
      alert( "ERROR: cannot locate FlagShip compiler;in " + _cBinPath + ;
             ";Set path via envir. variable FSDEBUG_COMPILER")
      return
   endif

   _cInclPath := FlagShip_dir() + PATH_SLASH + "include"
   if !file(_cInclPath + PATH_SLASH + "std.fh", .F.)   
      _tmp1 := alltrim(getenv("FSDEBUG_INCLUDE"))
      if !empty(_tmp1)
         if _FS_WIN32
            _cInclPath := strtran(_tmp1, "/", "\")
         else
            _cInclPath := strtran(_tmp1, "\", "/")
         endif
      endif      
   endif
   if !file(_cInclPath + PATH_SLASH + "std.fh", .F.)
      alert("ERROR: cannot locate std.fh;in " + _cInclPath + ;
             ";Set path via envir. variable FSDEBUG_INCLUDE")
      return
   endif

   if _FS_WIN32
      if " " $ _cInclPath
         _cInclPath := '-I"' + _cInclPath + '"'
      else
         _cInclPath := '-I' + _cInclPath
      endif
   else
      _cInclPath := '-I' + FlagShip_dir() + '/include'
   endif

   /* temporary file for results
    */
   _tmpFile := TempFileName("/tmp", "fsi_")
   if empty(_tmpFile)
      alert("sorry, could not open temp file in /tmp for write")
      return
   endif

#ifdef FlagShip5
	SET FONT "Courier",10
# ifdef FSIO_BASIC                        /* FS5: -io=b switch used, basic mode */
   _useAccept := .T.
# else
#  ifdef FSIO_ALL                         /* dynamic GUI mode in FS5: */
   _useAccept := AppIoMode() == "B"       /* all but basic use GET    */
#  endif
# endif
#endif

   /* in GUI mode, use screen height -90 pixel
    */
   if AppIoMode() == "G"
      SET GUITRANSL TEXT  ON 
      SET GUITRANSL LINES ON 
      
      local _nScrRows, _nVisRows
      _nScrRows := int(Pixel2row(m->oApplic:DesktopHeight - 90))
      _nVisRows := int(Pixel2row(m->oApplic:DesktopSizeAvail(.T.)[1] - 75))
      _nScrRows := min(_nScrRows, _nVisRows)      
      m->oApplic:Resize(_nScrRows, 82)    // resize screen <maxrows> x 82cols
      m->oApplic:Move(0, 50, .T.)         // move 50 pixel right
   endif
   
   /* start execution
    */
#ifdef FS2LIB 
   _winOut := wopen(22, 75, 47, 155)
   wselect(0)           // back to main window
   set key K_F2 to _winSwitch
   StatusMessage("F2 to switch to execution window", "R+/N", _color1)
#endif

   _cmd := ""
   set typeahead to 4000
   set key 28 to
   set function 1 to "HELP "
   _fsiDefErr := ErrorBlock(_fsiNoErr)
   cls
   
   ?? repli("-", 76)
   ? "Welcome to 'fsi', a nearly full sized interpreter based on"
   ? "  FlagShip (c) " + ltrim(year(date())) + " Copyright by multisoft GmbH, Munich, Germany."
   ? "It executes most of the standard statements, commands and functions, but" 
   ? "is mainly intended for database mainenance in a manner of other interpreted"
   ? "xBase systems like dBase, FoxBase, FoxPro etc. It is an alternative for the"
   ? "stand alone database maintenance program named 'dbu'."
   ? "Type 'help' for additional explanation. Type 'quit' (or 'q') to terminate." 
   if maxcol() < 77
      ? "Note: the help system is best visible with screen width of 77 chars or more,"
      ? "your screen seems to be set to " + ltrim(maxcol()) + " chars."
   endif
   ? repli("-", 76)
   
   while .T.
      /*
       * enter command (or use history file),
       * translate shortcuts
       */
      if len(_aHistNum) == 0
         _cmd := _getCommand("Command:")     // read command
         _cmd := alltrim(_cmd)
         if _cmd == "!?"                
            _cmd := "HELP HIST -10 * 10"     // "!?" equivalence
         elseif left(_cmd,2) == "!?"                
            _cmd := "HELP HIST " + substr(_cmd,3)
         elseif _cmd == "!!"                
            _cmd := "!-1"                    // "!!" equivalence
         endif
         aadd(_aHistory, alltrim(_cmd))
      else
         _cmd := _aHistory[_aHistNum[1]]    // get command from history
         ? padl("(" + ltrim(_aHistNum[1]) +")",7) + "! " + _cmd
         adel(_aHistNum,1) 
         asize(_aHistNum, len(_aHistNum) -1)
      endif
      
      /* remove comments,
       * check for QUIT, HELP, FSISET
       */
      _ii := at("//", _cmd)                  // remove inline comments
      if _ii > 0
         _cmd := if(_ii == 1, "", alltrim(left(_cmd, _ii -1)))
      elseif left(ltrim(_cmd),1) == "*"   
         _cmd := ""                          // remove "*..." comment lines
      endif
      if empty(_cmd)
         loop
      endif
      _cmd := alltrim(_cmd)
      if upper(_cmd) == "Q" .or. upper(_cmd) == "QUIT" .or. upper(_cmd) == "EXIT"
         exit
      endif
      if upper(_cmd) == "HELP" .or. upper(left(_cmd,5)) == "HELP "
         _help(alltrim(substr(_cmd + " ",5)))
         loop
      endif
      if upper(left(_cmd + " ",7)) == "FSISET "
         _setInternal(upper(alltrim(substr(_cmd,7))))
         loop
      endif
 
      if left(_cmd,1) == "!" .and. len(_aHistNum) == 0 
         /*
          * execute command history
          */
         _aHistNum := _ExecuteHist(substr(_cmd,2), len(_aHistory))
         loop
      endif
      
      /* continued statement
       */
      while right(trim(_cmd),1) == ";"       
         if len(_aHistNum) == 0
            _cmd2 := _getCommand("   cont+") // read command
            aadd(_aHistory, alltrim(_cmd2))
         else                                // get cont. command from history
            _cmd2 := _aHistory[_aHistNum[1]] 
            ? padl("(" + ltrim(_aHistNum[1]) +")",7) + "+ " + _cmd
            adel(_aHistNum,1) 
            asize(_aHistNum, len(_aHistNum) -1)
         endif
         _ii := at("//", _cmd2)              // remove inline comments
         if _ii > 0
            _cmd2 := if(_ii == 1, "", alltrim(left(_cmd2, _ii -1)))
         endif
         _cmd := trim(_cmd)
         _cmd := left(_cmd, len(_cmd) -1)    // remove previous line separ.
         _cmd += alltrim(_cmd2)
      enddo
      _cmd := strtran(@_cmd, ";;", ";")
      
      /* process statement, return result
       */
      _retval := _check(_p1, _dfPipe, _tmpFile, _cmd + " ", @_prep)
      if valtype(_prep) == "C" .and. _fsiTra
         _tmp1 := strtran(_cmd,  " ", "")
         _tmp2 := strtran(_prep, " ", "")
         if !(_tmp1 == _tmp2)
            ? " transl:"; _formout(_prep)    // display preprocessor transl.
         endif                               // if differs from user entry
      endif
      ? "       >", _retval                  // display final result
   enddo

   if _dfPipe > 0
      Fclose (_dfPipe)
   endif
   delete file (_p1)
   delete file (_tmpFile)
   ?
quit

/************************************************************************** 
 * get user command
 * uses @..GET/READ except for Basic i/o in FS5, where ACCEPT is used
 */
#define GETSIZE 4000
FUNCTION _getCommand(_cPrompt) 
   local _cRet          as char
   local getlist := {}  as array
   local _iRow,_currNum as intVar
   local _lScore, _lGuiColor as logical
   local _xprin, _xalt  as logical 
	local _cColor  		as char

   set extra off
   if _useAccept
      accept _cPrompt + " " to _cRet
   else
      _lScore  := set(_SET_SCOREBOARD, .F.)
      _currNum := len(_aHistory) +1
      _cRet := space(GETSIZE)
      ? ""                                   // handles scroll
      _iRow := row()
		_cColor := if(AppIoMode() == "T", "W/N,W+/N", "N/W,N/W")
		_lGuiColor := Set(_SET_GUICOLORS, .T.)
      @ _iRow, 0 SAY _cPrompt GET _cRet ;
                              PICT ("@S" + ltrim(min(70, maxcol() - len(_cPrompt) -2))) ;
                              VALID _validUpDn(@_currNum, @_cRet) ;
                              COLOR (_cColor)
//		if AppIoMode() == "G"
//      	READ clear
//		else
      	READ 
//		endif
		Set(_SET_GUICOLORS, lGuiColor)
      if lastkey() == K_ESC
         _cRet := ""
      else
         _cRet := trim(_cRet)
      endif
      setpos(_iRow, 0)
      set(_SET_SCOREBOARD, _lScore) 
   endif
   if _fsiCmd
      set console off
      set extra on
*     _xprin := set(_SET_PRINTER,   .F.)
*     _xalt  := set(_SET_ALTERNATE, .F.)
      ? _cPrompt, _cRet
      set console on
*     set(_SET_PRINTER,   _xprin)
*     set(_SET_ALTERNATE, _xalt)
   endif
   return _cRet

FUNCTION _validUpDn(_currNum, value)
   local _key := lastkey() as intvar
   if _key == K_UP .or. _key == K_PGUP                // or ^E, ^R
      if _currNum > 1
         _currNum --
         value := padr(_aHistory[_currNum], GETSIZE)
			GetActive():varput(value)
      else
         ?? chr(7)
      endif
      return .F.
   elseif _key == K_DOWN .or. _key == K_PGDN          // or ^X, ^C
      if _currNum < len(_aHistory)
         _currNum ++
         value := padr(_aHistory[_currNum], GETSIZE)
      else
         ?? chr(7)
         value := space(GETSIZE)
      endif
      return .F.
   endif
   return .T.
  
/************************************************************************** 
 * Check and evaluate the given command
 */
FUNCTION _check(_pipeName, _dfPipe, _tmpFile, _cmdIn, _prep)
   local   _retVal, _cmd, _cmdLine  as usual
   local   _tmp1, _tmp2             as char
   local   _transl                  as logic
   private _cEval 
   memvar  _cEval, __tmpEval
   
   /* pass the input the FlagShip preprocessor   
    */
   memowrit(_tmpFile, _cmdIn + NEWLINE)

#ifdef HAS_EGREP
   _cmdLine := _cBinPath + 'FlagShip_p -s -q ' + _cInclPath + ;
       ' -DNOSTDIOTRANS  -DNOSTDFUNCT -DNOSTDCLASS -DUSE_SIMPLE_GET' + ;
       ' -DFlagShip5 -DFlagShip6 ' + ;
       _fsiFox + _tmpFile + ;
       ' | egrep -v "#line|^\*|^$|^_PROTO |^_EXPO |^_PROTCT " > ' + _pipeName
#else
   _cmdLine := _cBinPath + 'FlagShip_p -s -q ' + _cInclPath + ;
       ' -DNOSTDIOTRANS  -DNOSTDFUNCT -DNOSTDCLASS -DUSE_SIMPLE_GET' + ;
       ' -DFlagShip5 -DFlagShip6 ' + ;
       _fsiFox + _tmpFile + ' ' + _pipeName
#endif

   /* invoke FlagShip_p _cmdLine
    */
#ifdef FS2LIB 
   wselect(_winOut)     // execute on subwindow
   run(_cmdLine)
   wselect(0)           // back to main window
#else
   run(_cmdLine)
#endif

   if doserror() > 0
      _prep := "Preprocessor Error"
      return "ERROR"
   endif
   
#ifdef FS_WIN32
   _dfPipe := FOpen(_pipeName)
   if _dfPipe <= 0
      _prep := "Could not open debugger temp i/o file " + _pipeName
      return "ERROR" 
   endif
#endif
   
   /* and read it back
    */       
#ifdef HAS_EGREP
   _cmd := FreadStr(_dfPipe, 4000)
   _tmp2 := FreadStr(_dfPipe, 4000)
   while !empty(_tmp2)
      _cmd += _tmp2
      _tmp2 := FreadStr(_dfPipe, 4000)
   enddo   
   _cmd := strtran(@_cmd, NEWLINE, " ;")
#else
   _cmd  := ""
   _tmp2 := ltrim(FreadTxt(_dfPipe))
   while Ferror() == 0
      if !( empty(_tmp2)               .or. ;
            left(_tmp2,5) == "#line"   .or. ;
            left(_tmp2,1) == "*"       .or. ;
            left(_tmp2,2) == "//"      .or. ;
            left(_tmp2,7) == "_PROTO " .or. ;
            left(_tmp2,6) == "_EXPO "  .or. ;
            left(_tmp2,8) == "_PROTCT " )

         _cmd += if(len(_cmd) == 0, "", " ;") + _tmp2
      endif
      _tmp2 := ltrim(FreadTxt(_dfPipe))
   enddo
#endif
#ifdef FS_WIN32
   Fclose(_dfPipe)
#endif
   
   while right(_cmd,1) == ";"
      _cmd := trim(left(_cmd, len(_cmd) -1))
   enddo
   _prep := _cmd                             // pass back preprocessed string to display
   _transl := .F.
     
   /* do some of the job which usually FlagShip compiler do
    */ 
   _translKeywd(@_cmd, @_transl)             // translate MEMVAR, PRIVATE, DO [with]
   if empty(_cmd)
      return ""
   endif
   if ! _testIfCmd(@_cmd, @_transl)          // test/prepare IF..[ELSE]..ENDIF
      return _cmd
   endif
   if ! _testForCmd(@_cmd, @_transl)         // test/prepare FOR...NEXT
      return _cmd
   endif
   
   /* fix syntax if required
    */
   _cmd := strtran(@_cmd, ";  ;", ";")
   _cmd := strtran(@_cmd, "; ;",  ";")
   _cmd := strtran(@_cmd, ";;",   ";")
   if at(";", _cmd) > 0
      _cmd := "(" + strtran(_cmd, ";", ",") + ")"
      _cmd := strtran(@_cmd, "( ,", "(")
      _cmd := strtran(@_cmd, "(,",  "(")
      _cmd := strtran(@_cmd, ", )", ")")
      _cmd := strtran(@_cmd, ",)",  ")")
   endif
   _cEval := _cmd

   /* display the preprocessed command
    */
   if valtype(_prep) == "C" .and. _fsiTra    // INTON set & preprocessor transl diff?
      _tmp1 := strtran(_cmd,  " ", "")
      _tmp2 := strtran(_prep, " ", "")
      if !(upper(_tmp1) == upper(_tmp2))
         ? "  prepr:" ; _formout(_prep)      // display preprocessor transl.
         _prep = NIL                         // if differs from user entry
         if _transl
            ? "   eval:" ; _formout(_cmd)    // display evaluated string
         endif                               // if differs from preprocessor
      endif
   endif
   
   /* macro-evaluate (or "macro-compile") the command 
    * using the standard FlagShip run-time system
    */
#ifdef FS2LIB 
   wselect(_winOut)     // execute on subwindow
   __tmpEval := &(_cEval)   
   wselect(0)           // back to main window
#else
   __tmpEval := &(_cEval)   
#endif
   return __tmpEval

/************************************************************************** 
 * remove and handle some FS preprocessor keywords
 * - this us usually done (far more effective) in the FS compiler...
 */
FUNCTION _translKeywd(_cmd, _transl as logic)
   local _cmd1, _cmd2, _cmd3, _cDelim, _name  as char
   local _dummy   as usual
   local _ii      as intvar
   private _var
   memvar  _var

   /* user or std.fh declared PRIVATE and PUBLIC vars
    * are created as PUBLIC, since must propagate above
    */
   _ii := at("_PRIV ", _cmd)    
   _ii := if(_ii == 0, at("_PUB ", _cmd), _ii)
   while _ii == 1 .or. (_ii > 1 .and. substr(_cmd,_ii-1,1) == ";")
      _cmd1 := if(_ii > 0, left(_cmd, _ii -1), "")  
      _cmd2 := ltrim(substr(_cmd, _ii + 5))   
      _var := _token(@_cmd2, " ,;", @_cDelim) 
      while !empty(_var) 
         PUBLIC &_var
         if left(_cmd2,2) == ":="         // PRIV or PUBL assignment
            _cmd2 := ltrim(substr(_cmd2,3))
            _cmd3 := _token(@_cmd2, " ,;", @_cDelim)
            _var += " := " + _cmd3
            _dummy := &_var
         endif
         if _cDelim == ";"
            exit
         endif
         _var := _token(@_cmd2, " ,;", @_cDelim) 
      enddo
      _cmd  := _cmd1 + ;
               if(!empty(_cmd1) .and. !empty(_cmd2), _cDelim, "") + _cmd2
      _ii := at("_PRIV ", _cmd) 
      _ii := if(_ii == 0, at("_PUB ", _cmd), _ii)
   enddo   

   /* MEMVAR's are simply ignored
    */
   _ii := at("_MEMV ", _cmd)              // MEMVAR declaration
   while _ii == 1 .or. (_ii > 1 .and. substr(_cmd,_ii-1,1) == ";")
      _cmd1 := if(_ii > 0, left(_cmd, _ii -1), "")
      _cmd2 := ltrim(substr(_cmd, _ii + 6)) 
      _ii   := at(";", _cmd2)
      _cmd2 := if(_ii > 0, substr(_cmd2, _ii +1), "")
      _cmd  := _cmd1 + _cmd2
      while len(_cmd) > 0 .and. right(_cmd,1) == ";"
         _cmd := trim(left(_cmd, len(_cmd -1)))
      enddo
      _ii   := at("_MEMV ", _cmd)
   enddo

   /* DO <proc> [with p1,p2,..] ist translated to the 
    * equivalent proc([p1,p2,..])
    */
   _ii := at("_DO ", _cmd) 
   while _ii == 1 .or. (_ii > 1 .and. substr(_cmd,_ii-1,1) == ";")
      _transl := .T.
      _cmd1 := if(_ii > 0, left(_cmd, _ii -1), "")  
      _cmd2 := ltrim(substr(_cmd, _ii + 4))  
      _cmd3 := _token(@_cmd2, " ,;", @_cDelim)
      _name := _cmd3 + "("
      if !(_cDelim == ";") .and. left(_cmd2, 5) == "WITH "       
         _cmd2 := ltrim(substr(_cmd2,5)) 
         _cmd3 := _token(@_cmd2, " ,;", @_cDelim)
         while !empty(_cmd3) .and. !(_cDelim == ";")
            _name += _cmd3 + _cDelim 
            _cmd3 := _token(@_cmd2, " ,;", @_cDelim)
         enddo
      endif
      _name += ")"
      _cmd  := _cmd1 + _name + _cDelim + _cmd2
      _ii := at("_DO ", _cmd)        
   enddo

   _cmd := alltrim(_cmd) 
   return ""   

/************************************************************************** 
 * This is a rough FOR ... NEXT loop checking and "evaluating" by hand,
 * since the FlagShip compiler cannot do the complex job here...
 */
FUNCTION _testForCmd(_cmd as char, _transl as logic) as logical
   local _iForBeg, _iForEnd   as intvar
   local _cmd1, _cmd2, _body  as char
   local _tmp1, _tmp2, _tmp3  as char
   local _forFrom, _forTo, _forStep, _tmp as char
   private _forVar
   memvar  _forVar
   
   if left(_cmd,5) == "_FOR " 
      _iForBeg := 1
   else
      _iForBeg := at(";_FOR ", _cmd)
   endif
   if _iForBeg == 0
      return .T.
   endif
   
   _cmd1 := if(_iForBeg > 1, left(_cmd,_iForBeg -1), "") // part before FOR...
   _body := alltrim(substr(_cmd, _iForBeg +5)) + " "
   
   _iForEnd := at(";next ", lower(_body))
   if _iForEnd == 0
      _cmd := "*** FOR ... structure: missing NEXT"
      return .F.
   endif

   _cmd2 := alltrim(substr(_body, _iForEnd + 5))      // rest after ENDIF
   _body := left(_body, _iForEnd -1)                  // body of FOR..NEXT
   
   _forVar := _token(@_body, " ")                     // FOR variable
   if empty(_forVar) 
      _cmd := "*** FOR ... structure: missing loop variable"
      return .F.
   endif
   _body := ltrim(_body)
   if !(left(_body, 2) == ":=")
      if !(left(_body, 1) == "=")
         _cmd := "*** FOR ... structure: missing assignment"
         return .F.
      else
         _body := ltrim(substr(_body,2))
      endif
   else
      _body := ltrim(substr(_body,3))   
   endif
   
   _forFrom := _token(@_body, " ;")                   // start value
   if empty(_forFrom) 
      _cmd := "*** FOR ... structure: missing start value"
      return .F.
   endif
   _tmp   := _token(@_body, " ;")                     // TO
   if !(upper(alltrim(_tmp)) == "TO")
      _cmd := "*** FOR ... structure: missing TO clause"
      return .F.
   endif
   
   _forTo := _token(@_body, " ;")                     // TO value
   if empty(_forTo)
      _cmd := "*** FOR ... structure: missing TO value"
      return .F.
   endif
   
   _body := ltrim(_body)
   if upper(left(_body, 4)) == "STEP"
      _body := ltrim(substr(_body,5))
      _forStep := _token(@_body, " ;")                // STEP value
   else   
      _forStep := "1"
   endif
   if empty(_forStep)
      _cmd := "*** FOR ... structure: wrong STEP value"
      return .F.
   endif
   
   _body := alltrim(_body)
   while left(_body,1) == ";"
      _body := ltrim(substr(_body,2))                 // final body
   enddo
   if ";" $ _body
      _body := strtran(_body, ";", ",")               // body -> list
      _body := "(" + _body + ")"                      // if necessary
   endif
   
   if type(_forVar) == "U"                            // create FOR variable
      public &_forVar                                 // if necessary
   endif
   
   _tmp1 := if(!empty(_cmd1), _cmd1 + ";", "")
   _tmp2 := "__tmpEval := _for_eval('" + _forVar + "','" + _forFrom + "','" + ;
            _forTo + "','" + _forStep + "','" + _body + "') "
   _tmp3 := if(!empty(_cmd2), ";" + _cmd2, "")
   _cmd  := _tmp1 + _tmp2 + _tmp3
   _transl := .T.                                     // command modified
   return .T.
   
/************************************************************************** 
 * This is a rough IF .. END[IF] checking and "evaluating" by hand,
 * since the FlagShip compiler cannot do the complex job here...
 */
FUNCTION _testIfCmd(_cmd as char, _transl as logic) as logical
   local _iIfBeg, _iIfElse, _iIfEnd, _ii as intvar
   local _ifCond, _body as char
   local _cmd1, _cmd2   as char
   local _texp, _fexp   as char
   
   if left(_cmd,4) == "_IF " 
      _iIfBeg := 1
   else
      _iIfBeg := at(";_IF ", _cmd)
   endif
   if _iIfBeg == 0
      return .T.
   endif

   _cmd1 := if(_iIfBeg > 1, left(_cmd, _iIfBeg -1), "")   // part before IF...
   _body := alltrim(substr(_cmd, _iIfBeg +4)) + " "
   
   _iIfEnd := at(";endif ", lower(_body))
   if _iIfEnd == 0
      _iIfEnd := at(";end ", lower(_body))
      if _iIfEnd == 0
         _cmd := "*** IF ... structure: missing END/ENDIF"
         return .F.
      endif
   endif

   _cmd2 := substr(_body, _iIfEnd)
   _ii   := at(" ", _cmd2)
   _cmd2 := if(_ii > 0, substr(_cmd2, _ii), "")       // rest after ENDIF
   _body := left(_body, _iIfEnd -1)                   // body of IF..END

   _ii := at(";", _body)
   if _ii < 2
      _cmd := "*** IF ... structure: no condition"
      return .F.
   endif
   _ifCond := alltrim(left(_body, _ii -1))            // condition
   _body   := alltrim(substr(_body, _ii +1 )) + " "   // body w/o condition
   
   _iIfElse := at(";else ", lower(_body))
   if _iIfElse > 0
      _texp := if(_iIfElse > 1, left(_body, _iIfElse -1), "")
      _fexp := substr(_body, _iIfElse +5)
   else
      _texp := _body
      _fexp := ""
   endif 
   
   _texp := alltrim(_texp)
   _fexp := alltrim(_fexp)
   _texp := if(!empty(_texp) .and. left(_texp,1) == ";", substr(_texp,2), _texp)
   _fexp := if(!empty(_fexp) .and. left(_fexp,1) == ";", substr(_fexp,2), _fexp)
   _texp := if(!empty(_texp) .and. right(_texp,1)== ";", left(_texp,len(_texp)-1), _texp)
   _fexp := if(!empty(_fexp) .and. right(_fexp,1)== ";", left(_fexp,len(_fexp)-1), _fexp)
    
   _texp := if(empty (_texp), "NIL", alltrim(_texp))  // true part
   _fexp := if(empty (_fexp), "NIL", alltrim(_fexp))  // false part

   if ";" $ _texp
      _tExp := strtran(_texp, ";", ",")               // true part -> list
      _tExp := "(" + _texp + ")"
   endif
   
   if ";" $ _fexp   
      _fExp := strtran(_fexp, ";", ",")               // false part -> list
      _fExp := "(" + _fexp + ")"
   endif
   
   _cmd1 := alltrim(_cmd1)
   if right(_cmd1,1) == ";"
      _cmd1 := trim(left(_cmd1, len(_cmd1) -1))
   endif
   _cmd2 := alltrim(_cmd2)
   if left(_cmd2,1) == ";"
      _cmd2 := ltrim(substr(_cmd2, 2))
   endif

   _cmd := if(!empty(_cmd1), _cmd1 + " ; ", "")       // rebuild command
   _cmd += "iif(" + _ifCond + ", " + _texp + ", " + _fexp + ")"
   _cmd += if(!empty(_cmd2), " ; " + _cmd2, "")       
   
   _transl := .T.                                     // command changed
   return .T.

/************************************************************************** 
 * helper for FOR...NEXT structure
 * since the FlagShip compiler cannot do the complex job here...
 */ 
FUNCTION _FOR_EVAL(_cVal, _cFrom, _cTo, _cStep, _cCmd)
   local _tmp as usual
   local _nFrom, _nTo, _nStep, _cc as usual
   private _nVal 
   memvar  _nval
   
   _nFrom := &(_cFrom)
   if !((_cc := valtype(_nFrom)) $ "NI")
      return "*** FOR ... structure: start value not numeric but '" + _cc + "'"
   endif
   _nTo   := &(_cTo)
   if !((_cc := valtype(_nTo)) $ "NI")
      return "*** FOR ... structure: TO value not numeric but '" + _cc + "'"
   endif
   _nStep := &(_cStep)
   if !((_cc := valtype(_nStep)) $ "NI")
      return "*** FOR ... structure: STEP value not numeric but '" + _cc + "'"
   endif

   for _nVal := _nFrom TO _nTo STEP _nStep
      _tmp := &(_cVal + " := _nVal")
      _tmp := &(_cCmd)
   next
   return _tmp

/************************************************************************** 
 * helper: extract a token from the leftmost command part
 */ 
FUNCTION _token(_cmd, _delimMask, _delimFound)
   local _ii, _jj, _iPos := 9999 as intvar
   local _retVal, _delim as char
   
   _delimFound := ""
   if len(_cmd) == 0
      return ""
   endif
   if valtype(_delimMask) != "C" .or. empty(_delimMask) 
      _delimMask := " "
   endif
   for _ii := 1 to len(_delimMask)
      _delim := substr(_delimMask,_ii,1)
      _jj    := at(_delim, _cmd)
      if _jj > 0 .and. _jj < _iPos
         _iPos := _jj
         _delimFound := _delim
      endif
   next
   if _iPos == 9999 .or. _iPos == 0                // not found,
      _retVal := _cmd                              // token = command
      _cmd := ""                                   // no command rest remain
      _delimFound := ""
   else
      _retVal := alltrim(left(_cmd,_iPos -1))      // return trimmed token
      _cmd    := ltrim(substr(_cmd, _iPos +1))     // remove token from command
   endif
   return _retVal   
  
/************************************************************************** 
 * beautify the output for "prep:" and "eval:"
 */
FUNCTION _FormOut(_txt)    
   local _ii, _iMax  as intvar
   local _iLine := 0 as intvar
   
   if ! _fsiWrp
      ?? " " + _txt
      return NIL
   endif
   
   _iMax := maxcol() -11
   while !empty(_txt)
      _ii := max( rat(";", left(_txt, _iMax)), rat(",", left(_txt, _iMax)) )
      if _iLine == 0
         ?? " "
      else
         ? space(9)
      endif
      if _ii == 0 .or. len(_txt) < _iMax
         ?? _txt
         _txt := ""
      else
         ?? left(_txt, _ii) + " ;"
         _txt := substr(_txt, _ii +1)
      endif
      _iLine ++
   enddo
   return NIL

/************************************************************************** 
 * display variables set, called from HELP VAR
 * internal variables, i.e. ones prefaced by underscore, are suppressed
 * The list is available only, if the compiler switch -nd was NOT given
 */
FUNCTION _DisplVars()
   local _aTmp, _arr as array
   local _ii, _iSize, _iPos, _iLast as intvar
   local _type  as char
   local _value as usual
   
   _aTmp  := __displpriv()
   _arr   := {}
   _iSize := 0
      
   for _ii := 1 to len(_aTmp)
      if !(left(_aTmp[_ii,2], 1) == "_") .and. ;
         !("FS2YY_LAST" $ _aTmp[_ii,2])
         aadd(_arr, _aTmp[_ii]) 
         _iSize := max(_iSize, len(_aTmp[_ii,2]))     // length of var name
         _iLast := len(_arr)
         if valtype(_arr[_iLast,1]) == "C"            // char = public
            _arr[_iLast,1] := "(PUBL) "
         else                                         // numeric = private on level
            _arr[_iLast,1] := "(PRIV) "
         endif
      endif
   next

    if used()                                          // consider fields
      for _ii := 1 to Fcount()
         _iPos := ascan(_arr, {|x| padr(FieldName(_ii),10) == padr(x[2],10) ;
                                   .and. x[1] == "(PRIV) "})
         if _iPos == 0 .or. _arr[_iPos,3] != NIL      // field hides private & publ
            aadd(_arr, {NIL, trim(FieldName(_ii)), NIL})
            _iSize := max(_iSize, len(trim(FieldName(_ii)))) 
            _iPos := len(_arr)
         endif                  
         _arr[_iPos,1] := "(FIELD)"
         _arr[_iPos,3] := fieldget(_ii)               // field value
      next
   endif
   
   if len(_arr) > 0
      _arr := asort (_arr, , , {|x,y| x[2] < y[2] .or. ;
                                     (x[2] == y[2] .and. x[1] <= y[1]) } )
      for _ii := 1 to len(_arr)
         _value := _arr[_ii,3]
         _type  := valtype(_value)
         if _arr[_ii,1] == "(FIELD)"
            _type := FieldType(_arr[_ii,2])
         endif
         ? "       >", _arr[_ii,1], padr(_arr[_ii,2], _iSize), _type + " "
         do case
         case _type == "C" .or. _type == "M"
            _iPos := maxcol() - col() -2
            ?? '"' + if(len(_value) > _iPos, left(_value,_iPos-3) + "...", _value) + '"'  
         case _type == "N"
            ?? ltrim(str(_value)) 
         case _type == "A"
            ?? "len() =", ltrim(len(_value)) 
         otherwise
            ?? _value
         endcase       
      next
   endif
   return ""

/************************************************************************** 
 * display command history, comes from HELP HIS
 */
FUNCTION _DisplHist(_subCmd)
   local _ii, _iLast       as intVar
   local _tmp, _cDelim     as char
   local _error            as logical
   local _aRet := {}       as array
   
   if valtype(_subCmd) == "C"
      _tmp   := _token(@_subCmd, " ")              // rest of HIST subcommands
      if empty(_subCmd)                            // default = all
         _subCmd := "1-" + ltrim(len(_aHistory))
      elseif !("-" $ _subCmd .or. "*" $ _subCmd .or. "," $ _subCmd)
         _subCmd += "-" + ltrim(len(_aHistory))
      endif
      _aRet := _RecursHist(_subCmd, len(_aHistory))
   endif
   if len(_aRet) == 0
      return ""
   endif
   _iLast := len(_aHistory)
   for _ii := 1 to len(_aRet)
      ? str(_aRet[_ii],7) + "# " + _aHistory[_aRet[_ii]]
      if _aRet[_ii] == _iLast
         _iLast := 0
      endif
   next
   if _iLast != 0                                  // always display last item#
      ? str(_iLast,7) + "# " + _aHistory[_iLast]
   endif
   return ""
 
/************************************************************************** 
 * Process !SAVE or !REST commands, or
 * execute command from history, returns commands (as index of history) 
 * in an array. History recursions are resolved here. Empty array
 * signals error.
 */
FUNCTION _ExecuteHist(_subCmd, _iLast)
   local _retArr, _tmpArr, _tmpArr2, _tmpArr3         as array
   local _ii, _i2, _i3                                as intVar
   local _iFrom, _iTo, _iFd, _iHist                   as intVar
   local _cFrom, _cTo, _cFile, _cDelim, _cStr, _cHist as char
   local _error, _SaveCmd, _RestCmd, _lCent           as logical
   local _cDform                                      as usual

   _subCmd := alltrim(_subCmd)         // remove spaces after ! command
   
   BEGIN SEQUENCE
      /* 
       *  [!] REST [file]
       */
      if upper(left(_subCmd,4)) == "REST" 
         _RestCmd := .T.
         _cStr := _token(@_subCmd, " ")      // remove REST keyword     
         if upper(left(_subCmd,4)) == "FROM" 
            _cStr := _token(@_subCmd, " ")   // remove FROM keyword if any
         endif
         _cFile := _token(@_subCmd, " ")     // file name or ""     
         if empty(_cFile)
            _cFile := _getDefaFile()
         endif
         _iFd := Fopen(_cFile)
         if _iFd == 0
             break "cannot open file to read"
         endif
         _aHistory := {}
         while Ferror() == 0
            _cStr := FreadTxt(_iFd)
            _cStr := strtran(@_cStr, chr(13), "")
            _cStr := strtran(@_cStr, chr(26), "")
            if Ferror() == 0 .and. !empty(_cStr) .and. !(left(ltrim(_cStr),1) == "#")
               aadd(_aHistory, _cStr )
            endif
         enddo
         aadd(_aHistory, "* current history discarded, new history read from " + _cFile)
         Fclose(_iFd)
         ? "       > HISTORY CHANGED, read from " + _cFile
         ? "       > If you want to (re-)execute the history, enter '! 1-" + ;
                     ltrim(len(_aHistory) -1) + "'"
         return {}
      endif
     
      /* [!] SAVE numPositive - numPositive [, numPositive, ...]  [TO file]
       * first part
       */
      if upper(left(_subCmd,4)) == "SAVE" 
         _SaveCmd := .T.
         _cStr  := _token(@_subCmd, " ")              // remove SAVE keyword
         _ii := at(" TO ", upper(_subCmd))
         if _ii > 0
            _cFile  := alltrim(substr(_subCmd, _ii +4))
            _subCmd := trim(left(_subCmd, _ii))
         endif
         if empty(_subCmd)
            break "enter item number(s) to be saved"
         endif
         if left(ltrim(_subCmd),1) == "-"  
            break "only positive number(s) allowed"
         endif
         if empty(_cFile)
            _cFile := _getDefaFile()
         endif
         _iFd := Fcreate(_cFile)
         if _iFd < 1
            break "cannot open/create file" 
         endif
      endif
      
   RECOVER USING _cStr
      ? "       > ERROR, " + _cStr
      return {}
   ENDSEQUENCE
   
   /* retrieve the required line numbers of history,
    * and process the [!] ... items
    */
   _tmpArr := _RecursHist(_subCmd, _iLast)
   if len(_tmpArr) == 0
      return {}
   endif
   
   /* re-process, resolve recursions
    */
   BEGIN SEQUENCE
   for _ii := 1 to len(_tmpArr)
      if !(left(_aHistory[_tmpArr[_ii]],1) == "!")
         aadd(_retArr, _tmpArr[_ii])
      else
         _cHist := ltrim(substr(_aHistory[_tmpArr[_ii]], 2))
         if upper(left(_cHist,5)) == "SAVE " .or. upper(left(_cHist,5)) == "REST "
            aadd(_retArr, _tmpArr[_ii])
         else
            _tmpArr2 := _RecursHist(_cHist, _tmpArr[_ii])
            _iHist := if(len(_tmpArr2) > 0, _tmpArr2[1], 0)
            _cHist := if(_iHist > 0, _aHistory[_iHist], "")
            while len(_tmpArr2) == 1 .and. left(_cHist,1) == "!"
               _tmpArr2 := _RecursHist(substr(_cHist, 2), _iHist)
               _iHist := if(len(_tmpArr2) > 0, _tmpArr2[1], 0)
               _cHist := if(_iHist > 0, _aHistory[_iHist], "")
            enddo   
            for _i2 := 1 to len(_tmpArr2)
               if !(left(_aHistory[_tmpArr2[_i2]],1) == "!") 
                  aadd( _retArr, _tmpArr2[_i2] )
               else  
                  _tmpArr3 := _RecursHist(substr(_aHistory[_tmpArr2[_i2]], 2), _tmpArr2[_i2])
                  _iHist := if(len(_tmpArr3) > 0, _tmpArr3[1], 0)
                  _cHist := if(_iHist > 0, _aHistory[_iHist], "")
                  while len(_tmpArr3) == 1 .and. left(_cHist,1) == "!"
                     _tmpArr3 := _RecursHist(substr(_cHist, 2), _iHist)
                     _iHist := if(len(_tmpArr3) > 0, _tmpArr3[1], 0)
                     _cHist := if(_iHist > 0, _aHistory[_iHist], "")
                  enddo   
                  for _i3 := 1 to len(_tmpArr3)
                     if left(_aHistory[_tmpArr3[_i3]],1) == "!"
                        break "recursions too deep"
                     endif
                     aadd( _retArr, _tmpArr3[_i3] )
                  next
               endif
            next
         endif
      endif
   next
   
   /* write the !SAVE file 
    */
   if _SaveCmd
      _lCent := set(_SET_CENTURY)
      _cDform:= set(_SET_DATEFORMAT)
      set century on
      set date format "dd-mm-yyyy"
       
      Fwrite(_iFd, "# 'fsi' (FlagShip Interpreter) history," + NEWLINE )
      Fwrite(_iFd, "# saved on " + dtoc(date()) + " " + time() + ;
                   " by " + netname() + NEWLINE )
      Fwrite(_iFd, "#" + NEWLINE )
      Fwrite(_iFd, "# You may modify this file by any ascii editor," + NEWLINE )
      Fwrite(_iFd, "# but consider the 'fsi' syntax." + NEWLINE )
      Fwrite(_iFd, "# Empty lines and lines starting with '#' are ignored." + ;
                    NEWLINE + NEWLINE )
      set(_SET_DATEFORMAT, _cDform)
      set(_SET_CENTURY, _lCent)
      for _ii := 1 to len(_retArr) -1
         if left(_aHistory[_retArr[_ii]],1) == "!"
            Fwrite( _iFd, "* " + _aHistory[_retArr[_ii]] + NEWLINE ) 
         else
            Fwrite( _iFd, _aHistory[_retArr[_ii]] + NEWLINE ) 
         endif
         if Ferror() != 0
            _error := .T.
            exit
         endif
      next
      Fwrite(_iFd, NEWLINE + "# eof" + NEWLINE )
      Fclose(_iFd)
      ? "       > " + if(_error, "ERROR on writting history file ", ;
                                 "History written to ") + _cFile
      _retArr := {}
   endif
   
   RECOVER USING _cStr
      ? "       > ERROR, " + _cStr + if(_SaveCmd, ", data NOT saved", "")
      return {}
   END SEQUENCE
   
   return _retArr    
   
/************************************************************************** 
 * Step through command history, helper for _ExecuteHist()
 * <_subCmd> The first ! command exclam.mark is already removed, but may
 *           contain !... again = either !! or recursion. Accepted are:
 *       [!] numPositive
 *       [!] numPositive - numPositive
 *       [!] numPositive , numPositive
 *       [!] numPositive , numPositive , numPositive [ , numPositive ...]
 *       [!] numNegative 
 *       [!] numNegative * numPositiveSteps
 * Space delimiters are not relevant
 * <iLast> last valid history item to be processed
 */
FUNCTION _RecursHist(_subCmd, _iLast as intvar)
   local _ii, _iFrom, _iTo, _sign := 1 as intVar
   local _cFrom, _cTo, _cDelim, _cErr  as char
   local _retArr := {}                 as array
   local _error, _removed, _continue   as logical

   _iFrom := 1
   _iTo   := _iLast
   if !valtype(_subCmd) == "C"
      return {}
   endif    
   
   BEGIN SEQUENCE
      if left(_subCmd,1) == "-"
         _sign := -1
         _subCmd := ltrim(substr(_subCmd,2))
      endif
      _cFrom := _token(@_subCmd, ",-*", @_cDelim)      // start#
      if " " $ _cFrom                                 // [!] <num> <num> is invalid
         break "invalid item# '" + _cFrom + "' (missing , - * delimiter?)"
      endif

      _iFrom := val(_cFrom) * _sign
      if _iFrom == 0 
         break "invalid item# " + _cFrom
      endif
      
      /*  [!] numNegative [ * 1 ]
       *  [!] numNegative * numPositiveSteps
       *  [!] numPositive * numPositiveSteps
       */ 
      if _iFrom < 0                                   // go back <_iFrom> items
         if !empty(_subCmd) .and. !(_cDelim == "*" )  // accepted:
            break "Syntax = numNegative * steps"      // [!] numNeg [ * numPosit]
         endif   
         _iFrom := max(_iLast + _iFrom, 1)            // starting item#
      endif
      if _cDelim == "*"
         if _iFrom > _iLast
            break "invalid item#, range is 1..." + ltrim(_iFrom)
         endif
         _cTo   := "1"                                // number of steps
         if !empty(_subCmd)
            _cTo := _token(@_subCmd, " ,*-", @_cDelim) // #steps  
         endif
         _iTo := val(_cTo)                          
         if _iTo <= 0 .or. " " $ _cTo
            break "invalid steps value '" + _cTo + "'"
         endif
         if !empty(_subCmd)
            break "Syntax = num * steps (no subsequent items)"
         endif
         _iTo := min(_iFrom + _iTo -1, _iLast)
         for _ii := _iFrom to _iTo
            aadd( _retArr, _ii )
         next
         return _retArr
      endif

      /* following commands can be repeated
       */
      _continue := .T.
      while .T.
         _removed := .F.
         if _iFrom > _iLast .or. _iFrom < 1
            break "invalid item#, range is 1..." + ltrim(_iLast)
         endif
         
         /*  [!] numPositive
          */
         if empty(_subCmd)
            aadd( _retArr, _iFrom )
            exit
         endif
          
         /*  [!] numPositive - numPositive
          */
         if _cDelim == "-" 
            _cTo := _token(@_subCmd, ",*-", @_cDelim)              
            _iTo := val(_cTo)                         // end item#
            if " " $ _cTo
               break "invalid item# '" + _cTo + "' (missing comma?)"
            endif
            if _iTo < _iFrom 
               break "invalid item#, " + _cTo + " < " + ltrim(_iFrom) 
            endif
            _iTo := min(_iTo, _iLast)                    // corrected end item#
            for _ii := _iFrom to _iTo
               aadd( _retArr, _ii )
            next
            _removed := .T.
         endif

         /*  [!] numPositive , numPositive
          *  [!] numPositive , numPositive , numPositive [ , numPos - numPos ...]
          */
         if !_removed
            aadd( _retArr, _iFrom )
         endif
         
         if empty(_subCmd)
            exit
         endif
         if !(_cDelim == ",")
            break "invalid '" + _cDelim + "' delimiter"
         endif
         _cFrom := _token(@_subCmd, ",*-", @_cDelim)
         _iFrom := val(_cFrom)
      enddo
   RECOVER USING _cErr
      ? "       > ERROR, " + _cErr
      return {}
   ENDSEQUENCE    
   return _retArr 

/************************************************************************** 
 * default save/restore file, system dependant
 */
FUNCTION _getDefaFile()
   local _file as char
   
   _file := getenv("HOME")
   if empty(_file)
      _file := getenv("TMP")
      if empty(_file)
         _file := getenv("TEMP")
      endif
   endif
#ifdef MS_WINDOWS
   _file := if(empty(_file), curdir(), _file) + "\_fsi.set"
#else
   _file := if(empty(_file), "./", _file + "/") + ".fsi"
#endif
   return _file
             
/************************************************************************** 
 * manage fsi internal settings (one or more separated by space)
 * via FSISET opt1 [opt2] [opt3] [...]
 */
FUNCTION _setInternal (_cmdIn)
   local _cmd, _iPos as usual
   
   _cmdIn := alltrim(_cmdIn)
   
   while !empty(_cmdIn)
      _iPos := at(" ", _cmdIn)
      if _iPos > 0
         _cmd   := left(_cmdIn, _iPos -1)
         _cmdIn := alltrim(substr(_cmdIn, _iPos +1))
      else
         _cmd   := _cmdIn
         _cmdIn := ""
      endif
      
      do case
      case _cmd == "ERRON"
         if !_fsiErr
            ErrorBlock(_fsiDefErr)
            _fsiErr := .T.
         endif   
      case _cmd == "ERROFF"
         if _fsiErr
            ErrorBlock(_fsiNoErr)
            _fsiErr := .F.
         endif   
      case _cmd == "LOWON"
         fs_set("lower", .T.) ; fs_set("pathlower", .T.)
         _fsiLow := .T.
         _fsiUpp := .F.
      case _cmd == "LOWOFF"
         fs_set("lower", .F.) ; fs_set("pathlower", .F.)
         _fsiLow := .F.
         _fsiUpp := .F.
      case _cmd == "UPPON"
         fs_set("upper", .T.) ; fs_set("pathupper", .T.)
         _fsiLow := .F.
         _fsiUpp := .T.
      case _cmd == "UPPOFF"
         fs_set("upper", .F.) ; fs_set("pathupper", .F.)
         _fsiLow := .F.
         _fsiUpp := .F.
      case _cmd == "FOXON"
         _fsiFox := "-DFOXPRO "
      case _cmd == "FOXOFF"
         _fsiFox := ""
      case _cmd == "TRAON"
         _fsiTra := .T.
      case _cmd == "TRAOFF"
         _fsiTra := .F.
      case _cmd == "WRPON"
         _fsiWrp := .T.
      case _cmd == "WRPOFF"
         _fsiWrp := .F.
      case _cmd == "CMDON"
         if !_fsiCmdOpen
            SET EXTRA TO fsi.cmd
            _fsiCmdOpen := .T.
         endif
         SET EXTRA on
         set console off
         ? "* fsi command mirroring started on", date(), ;
           "at " + time() + " by " + netname()
         ? "* at history item#" + ltrim(len(_aHistory))
         set console on
         _fsiCmd := .T.
      case _cmd == "CMDOFF"
         set console off
         ? "* fsi command mirroring stopped on", date(), ;
           "at " + time() 
         ? "* at history item#" + ltrim(len(_aHistory)) ; ?
         set console on
         SET EXTRA off
         _fsiCmd := .F.
      endcase
   enddo
   
   ? "       > curr.settings: ERR" + if(_fsiErr, "ON","OFF") + ;
                            " TRA" + if(_fsiTra, "ON","OFF") + ;
                            " CMD" + if(_fsiCmd, "ON","OFF") + ;
                            " WRP" + if(_fsiWrp, "ON","OFF") + ;
                            " LOW" + if(_fsiLow, "ON","OFF") + ;
                            " UPP" + if(_fsiUpp, "ON","OFF") + ;
                            " FOX" + if(empty(_fsiFox), "OFF", "ON")
   return NIL               

/************************************************************************** 
 * on-line help
 */
FUNCTION _Help (_cSubHelp)

   if upper(left(_cSubHelp,3)) == "VAR"
      _DisplVars()
      return "" 
   elseif upper(left(_cSubHelp,3)) == "HIS"
      _DisplHist(_cSubHelp)
      return ""
   endif
   
   do case
   case upper(left(_cSubHelp,3)) == "SYN"
      ? repli("-", 30) + "SYNTAX HELP" + repli("-",35)
         *---------------------------------------------------------------------------*
      ? "The syntax of 'fsi' is nearly equivalent to the usual source programs. But"
      ? "since the statements are macro-evaluated (some says macro-compiled) here,"
      ? "few restrictions apply in this interpreter.  Of course, also the speed of"
      ? "these interpreted commands is not comparable to native compiled executable."
      ? "You can enter any"
      ? "  <command>    [; <command> ; <expression> ; <list> ; ...]" 
      ? "  <expression> [; <expression> ; <command> ; <list> ; ...]"
      ? "  <list>       [; <command> ; <expression> ; <list> ; ...]" 
      ? "  IF <condition> ; <expression>|<command>|<list> ; ENDIF"
      ? "  IF <condition> ; <expression>|<command>|<list> ; ELSE ; <..dito..> ; ENDIF"
      ? "  FOR <start> TO <end> [STEP <step>] ; <expression>|<command>|<list> ; NEXT"
      ? "where"
      ? "  <command>    is a usual command like USE ..., SET INDEX ... etc." 
      ? "  <expression> is e.g. an assignment, variable handling, function call etc."
      ? "  <condition>  is a logical and/or mathematical comparison"
      ? "  <list>       contain 2 or more expressions separated by comma & altogether"
      ? "               parenthesed, e.g. '( a := 2, b := a -3 , ++a , c := date() )'"
      ? "               Note: don't use commands in a list, since not translated there"
      if maxrow() < 27
         wait
      endif
      ? "Each entry may contain up to 4000 characters. Split long entry to next lines"
      ? "by using a semicolon (;) at the line end, same as in the prg source.  This"
      ? "also means: whereever you use 'newline' in the source structures, enter the" 
      ? "semicolon here.  The final result is displayed behind the '>' sign, inter-" 
      ? "mediate results will be displayed when FSISET INTON is set."
      ? "See also: HELP, HELP COM, HELP FUN, HELP DATA, HELP SET (FlagShip manual LNG)"

   case upper(left(_cSubHelp,3)) == "COM"
      ? repli("-", 30) + "COMMANDS HELP" + repli("-",33)
      ? "Enter any by FlagShip supported commands or expression (see also HELP SYN)."
      ? "It will be translated by the FS Preprocessor to corresponding function(s)"
      ? "and executed. Supported control structures are IF...ELSE...ENDIF and FOR..."
      ? "...NEXT (but w/o LOOP and EXIT), see HELP SYN for syntax. Complex structures"
      ? "like IF...ELSEIF...ELSE...ENDIF,  DO WHILE...ENDDO,  BEGIN SEQUEN...BREAK..."
      ? "..ENDSEQUENCE are not supported in this fsi interpreter, same as declarators"
      ? "like LOCAL, STATIC, PROCEDURE, FUNCTION, CLASS etc., since meaningless here."
      ? "Code blocks, macros and comments like '* <comment>' or '... //<comment>' are"
      ? "supported.  Examples:" 
      ? '  SET PRINTER ON                                           //enable printer'
      ? '  USE mydatabase SHARED NEW ; used()                       //open a database'
      ? '  Aeval(DbStruct(),{|a| qout(padr(a[1],10),a[2],a[3],a[4]) } )   //structure'
      ? '  SET PRINTER OFF ; RUN ("lpr -Plaser " + fs_set("print")) //print on spooler'
      ? '  iVar := 25 ; cVar := "hello"                             //set variables'
      ? '  cVar                                                     //display vars'
      ? '  FOR i := 10 TO 5 STEP -1 ; (--iVar , cVar := cVar + ltrim(iVar) ) ; NEXT'
      ? '  a := 5 ; IF a < 10 .or. iVar > 3 ; cVar := "ok" ; ELSE ; "hello" ; ENDIF'
      ? '  iif(iVar < 10, qout("iVar is smaller than 10"), qout("iVar >= 10") )'
      ? '  ? "iVar is " + iif(iVar < 10, "<", ">=") + " 10"'
      ? "See also: HELP, HELP FUN, HELP DATA, HELP SYN, HELP SET (FS manual LNG,CMD)"
      
   case upper(left(_cSubHelp,3)) == "FUN"
      ? repli("-", 30) + "FUNCTIONS HELP" + repli("-",32)
      ? "You may use all standard FlagShip functions, also in combination with "
      ? "commands and expression (see HELP COM). If you additionally need to use your"
      ? "own functions (UDFs), you will need to re-compile fsi and add these objects."
      ? "Examples:"
      ? '  DbCreate("mydatabase", {{"name","c",20,0},{"date","d",8,0}, ;'
      ? '     {"num","n",10,3} } ) ; USE mydatabase'
      ? '  dDate := date()'
      ? '  version()'
      ? '  arr := array(5)' 
      ? '  FOR ii := 1 to len(arr) ; arr[ii] := ii ; NEXT'
      ? '  Aeval(arr, {|x| qout(x) } )'
      ? '  iElem := Ascan(arr, {|x| x == 3})'
      ? '  file("mydatabase.dbf")'
      ? '  iElem := Ascan(Directory("*.dbf"), {|x| "mydata" $ x[1] })' 
      ? '  iif(iElem > 0, qout( Directory("*.dbf")[iElem,1] ), NIL)'
      ? "See also: HELP, HELP COM, HELP DATA, HELP SYN, HELP SET (FS manual FUN)"

   case upper(left(_cSubHelp,3)) == "DAT" .or. ;
        upper(left(_cSubHelp,3)) == "DBF"
      ? repli("-", 20) + "Common Database Commands & Functions" + repli("-",20)
      ? "This is only a short overview of commonly used database commands and"
      ? "functions, see the FlagShip manual section LNG, CMD and FUN for a full"
      ? "reference."
      ? '  DbCreate("mydatabase", {{"name","c",25,0}, {"date","d",8,0}, ;'
      ? '   {"num","n",12,3}, {"logic","L",1,0}, {"Memo","m",10,0} } )  //create dbf'
      ? '  USE mydatabase EXCL NEW ALIAS DBF1; used()          //open mydatabase.dbf'
      ? '  Aeval(DbStruct(),{|a| qout(padr(a[1],10),a[2],a[3],a[4]) } )  //structure'
      ? '  INDEX ON name TO myindex1 ; INDEX ON date TO myindex2  //create 2 indexes'
      ? '  SET INDEX TO myindex1, myindex2                        //assign index(es)'
      ? '  FOR ii := 1 to 10 ; APPEND BLANK ; REPLACE name with "foo", ;'
      ? '     date with date(), num with ii, logic with .F. ; ii ; NEXT //add 10 rec'
      ? '  LastRec()                                              //dbf size in rec.'
      ? '  GO TOP ; RecNo()                                       //move to 1st rec.'
      ? '  DbEdit() ; CLS                                         //browse open dbf'
      ? '  Browse() ; CLS                                   //browse/modify/add/edit'
      ? '  GO BOTTOM ; REPLACE num with 99, name with "Smith"     //move to last rec'
      ? '  SET ORDER TO 2 ; IndexOrd()                            //set active index'
      ? '  today := date() ; SEEK (today) ; found()               //seek index value'
      ? '  LOCATE FOR upper(left(name,5)) == "SMITH" ; .not.eof() //sequent. search'
      ? "See also: HELP, HELP COM, HELP FUN, HELP SYN, HELP SET, FlagShip manual"

   case upper(left(_cSubHelp,3)) == "SET"
      ? repli("-", 30) + "fsi Internal Settings" + repli("-",25)
      ? "Show current fsi settings                      : FSISET"
      ? "Enable or disable error handling               : FSISET ERRON | ERROFF "
      ? "Enable or disable display translation/preproc. : FSISET TRAON | TRAOFF "
      ? "Enable or disable command mirroring to fsi.cmd : FSISET CMDON | CMDOFF "
      ? "Enable or disable line wrap on interm. display : FSISET WRPON | WRPOFF "
      ? "Enable or disable auto lower-case translation  : FSISET LOWON | LOWOFF "
      ? "Enable or disable auto upper-case translation  : FSISET UPPON | UPPOFF "
      ? "Enable or disable FoxPro enhancements          : FSISET FOXON | FOXOFF "
      ? '  Note, you also may set it by fs_set("low",.T.|.F.), fs_set("upp",T.|.F.)'
      ? '  fs_set("pathlow",.T.|.F.), fs_set("pathupp",.T.|.F.) etc.'
      ? "See also: HELP, HELP COM, HELP FUN, HELP DATA, HELP SYN, FlagShip manual"

   case left(_cSubHelp,1) == "!"
      ? repli("-", 30) + "Execute from History" + repli("-",26)
      if _useAccept
       ? "Instead of typing commands anew (you cannot use Cursor Up here, but can use"
       ? "cut-and-paste if supported by the OS), you can repeat the already given (or"
       ? "restored) commands by using the '!' command.  Note: for the xBase command !"
       ? "(execute an application or system command), use the equivalent RUN command."
      else
       ? "Instead of typing commands anew, you can either use Cursor up/down, PgUp/Dn,"
       ? "cut-and-paste (if supported by the OS), or simply repeat the already given"
       ? "entries by using the '!' command.  Note: for the xBase command ! (execute"
       ? "an application or system command), use the equivalent RUN command."
      endif
      ? "  HELP HIS [nFrom [- nTo]   Display commands history [starting with <nFrom>]"
      ? "  HELP HIS -nBack [* nStep] Display commands history, last <nBack> commands"
      ? "  !?                        Same as HELP HIS -10*10 = display last 10 commds"
      ? "  !? 1                      Shortcut for HELP HIS, here same as 'HELP HIS 1'"
      ? "  !!                        Execute last command from history, same as !-1"
      ? "  ! -num [* steps]          Execute item# <last> - <num> [and add. <steps>]"
      ? "                            i.e. '! -1*1' is equivalent to the '!!' command"
      ? "  ! num [* steps]           Execute item# <num> [and additional <steps>]"
      ? "  ! num [- nTo]             Execute history item #<num> or #<num> to #<nTo>"
      ? "  ! num1, num2 [,num3...]   Execute history items# <num1>, <num2>, ..."
      ? "  !SAVE num1-num2 [TO file] Save history items# <num1> to <num2> into <file>"
      ? "  !SAVE num1,num2,num3-num4,num5,num6-num7...[TO file] Possible combinations"
      ? "  !REST [[FROM] file]       Restore (replace) history from given <file>, "
#ifdef MS_WINDOWS
      ? "                            using %HOME|TMP%\_fsi.set if <file> is omitted."
#else
      ? "                            using ~/.fsi if <file> is omitted."
#endif
      ? "See also: HELP, HELP COM, HELP FUN, HELP DATA, HELP SYN"
      
   otherwise
      ? repli("-", 30) + "HELP" + repli("-",42)
      ? "Requires installed FlagShip rel. 4.48 or later, found via PATH environment."
      ? "Use: enter a command or statement and press RETURN. The statement can"
      ? "include several commands separated by semicolon (;) same as in the program"
      ? "source file. The statement or command can be continued at the next line,"
      ? "if the last character is a semicolon - which itself will not count as a"
      ? "command delimiter. Additional help is available by entering"
      ? "  HELP COM[mand]    help for commands use, examples"
      ? "  HELP FUN[ction]   help for functions use, examples"
      ? "  HELP DAT[abase]   help for database commands and functions, examples"
      ? "  HELP DBF          same as HELP DAT"
      ? "  HELP SYN[tax]     help with general syntax explanation"
      ? "  HELP VAR[iables]  display used variables (w/o internal vars with _*)"
      ? "  HELP SET[tinng]   help for & display internal fsi settings"
      ? "  HELP !            help for executing commands from history, save/restore"
      ? "  HELP HIS[tory] [nFrom [- nTo]] show commands history [starting with nFrom]"
      ? "  HELP HIS[tory] -nRest [* nCnt] show commands history, last nRest commands"
      ? "  !? ...            shortcut for 'HELP HISTORY', see HELP !"
      ? "To terminate the fsi interpreter, enter 'q' or 'quit' and press RETURN"
   endcase  
   ? repli("-", 76)
   return NIL

/************************************************************************** 
 * switch between execution and processing window by F2 key
 */
function _winSwitch
   static _execWin := .F.
#ifdef FS2LIB 
   if !_execWin 
      _execWin := .T.
      StatusMessage("F2 to switch back to process window", "R/N", _color2)
      wselect(_winOut)
   else      
      _execWin := .F.
      StatusMessage("F2 to switch to execution window", "R+/N", _color1)
      wselect(0)
   endif
#endif
   return NIL
   
/************************************************************************** 
 * enable externals/std.functions for macro evaluation
 */
function _addExternals()   
   extern dbedit
   extern fox_sys    // enable FoxPro API (called here by macro)
#ifdef FS2LIB  
   extern addmonth, asciisum, Center, crc16, clearbit, crypt, csetkey
   extern dirchange, currentget, charpack, com_init 
   extern csetwrap, ClearEol, dosparam, fact, like, numcol, numrow
   extern screensize, token, waitperiod, Week, wordrepl, wopen, wselect, xtoc
#endif
   return NIL

* eof *
