/* fsmake: utility to create Makefile and build projects for VFS
 *
 * For FlagShip 6.x
 * Copyright (c) 2007..08 by multisoft Datentechnik, Munich, Germany
 *
 * This is free add-on tool for FlagShip. You may modify this source
 * according to your needs and distribute the original or modified source
 * freely with your application, but may not remove this copyright notice.
 * You also may distribute the executable, but with the source file only.
 * Author: Jan V.Balek, multisoft. Report bugs to <bugs@flagship.de>
 * Release $Date: 2007/11/02 13:59:00 $
 *
 * Compile : FlagShip fsmake.prg -o fsmake 
 *      or : FlagShip fsmake.prg -io=g -o fsmake
 * Run     : fsmake [projectName]
 * -------------------------------------------------------------------------
 */

#include "fileio.fh"
#include "dialog.fh"
#include "mouse.fh"
#include "button.fh"

parameters cProject   // optional cmd-parameter: project name

#define IS_EXE          (nCreate == 1)
#define IS_EXELIB       (nCreate == 2)
#define IS_LIB          (nCreate == 3)
#define IS_WITHEXE      (nCreate <= 2)
#define IS_WITHLIB      (nCreate >= 2)

#define IS_RELEASE      (nRelease == 1)
#define IS_DEBUG_P      (nRelease == 2)
#define IS_DEBUG_C      (nRelease == 3)
#define IS_DEBUG_PC     (nRelease == 4)
#define IS_DEBUG_WITHC  (nRelease >= 3)

#define IS_LINUX        (nSystem == 1)
#define IS_WINDOWS      (nSystem >= 2)
#define IS_WIN_BCC      (nSystem == 2)
#define IS_WIN_MSC      (nSystem == 3)

#define IS_IOMODE_A     (nIoMode == 1)
#define IS_IOMODE_G     (nIoMode == 2)
#define IS_IOMODE_T     (nIoMode == 3)
#define IS_IOMODE_B     (nIoMode == 4)

#define APROJ_TYPE      1  /* order of aProjectFiles[n] sub-array */
#define APROJ_FILE      2
#define APROJ_PATH      3
#define APROJ_SWITCH    4
#define APROJ_FSMARK    5
#define APROJ_DEPEND    6
#define APROJ_FSMARKNUM 7  /* not in Tbrowse() */
#define APROJ_DEPENDIDX 8  /* not in Tbrowse() */


PUBLIC nCreate := 0, nRelease := 0, nSystem := 0, nIoMode := 0
PUBLIC cExeName, cLibName, cMain, cMainSwitch, cAddSwitch, cAddObjLib, cSearchPath, cOutputDir
PUBLIC cPathDelim := ":", cSTDFH := "", cSEARCH := ""
PUBLIC swnew_prg := .T., swnew_fmt := .T., swnew_c := .F.
local cTmp, cMake := "make"

#ifdef FS_WIN32
   cPathDelim := ";"
   cMake := "make/nmake"
#endif

SET FONT "courier",10    // decrease size if required

if AppIoMode() == "G"
   iRows := int(Pixel2row(m->oApplic:DesktopHeight - 120))
   m->oApplic:Resize(iRows,81,.F.,.T.,.T.)
   m->oApplic:Move(10,100, .T.)
   SET GUITRANSL TEXT  ON    // allow semi-grapgic chars in text
   SET GUITRANSL LINES ON    // draw lines by semi-grapgic chars
endif

set color to "W+/B,R+/GR,,,B/W"  // for terminal i/o only
cls

if empty(cProject)
   ?
   ? "This is a build utility for development with FlagShip."
   ?
   ? "FSmake manages your project files and creates Makefile for"
   ? "standard " + cMake + " utility which speed-up your development."
   ?
   ? "You may invoke it also by '" + execname() + " your_project_name'"
   ?
endif

if AppIoMode() == "G"
   cTmp := "Decrease SET FONT size;and recompile fsmake.prg"
else
   cTmp := "Increase the terminal-window size (property);and restart " + execname()
endif

if MaxRow() < 35
   alert("Sorry, screen with 35 or more lines required;" + cTmp)
   quit
endif
if MaxCol() < 79
   alert("Sorry, screen with 80 or more columns required;" + cTmp)
   quit
endif

cDbf := OpenProject()     // open available or create new project file
MenuProject(cDbf)         // main menu
use                       // close database

// setpos(20,0)
// wait
quit

/**************************************************************************
 * Open or create project file
 *    cProject is passed from/to main
 *    returns file name w/o extension
 */
FUNCTION OpenProject()
local cFile, ii, ok
local lNew := .T., lReuse := .T.
private nRow

local aStru := {{"File",  "C",30,0}, ; // file name (index upper)
                {"Path",  "M",10,0}, ; // path or empty for current dir
                {"Switch","M",10,0}, ; // special file switches
                {"FSsysNum","N",3,0},; // > 0 if FlagShip system file
                {"Depend","M",10,0}, ; // #include dependencies for this file
                {"Type",  "C",1, 0} }  // "F"=file, "L"=lib

/* Note: File == " .1" = header, Path contains default setings
 *                               Switch contains default switches
 *       File == " .2" = header, Path contains addit. objects and libs,
 *                               Switch contains addit. switches
 *       File == " .3" = header, Path contains search path for dependencies
 *       File == " .4" = header, Path contains output directory
 */

nRow := row() +2
if empty(cProject)
   cProject := space(20)
else
   if  file("fsmake_" + cProject + ".dbf") .and. file("fsmake_" + cProject + ".dbt")
      lNew := .F.
   else
      @ nRow,0 say "Project file fsmake_" + cProject + ".db* not available. Create new (T/Y or f/n) ? " ;
               get lNew
      read clear
      if !lNew .or. lastkey() == K_ESC
         quit
      endif
   endif
endif
if lNew
   cProject := padr(cProject, 20)
   while .T.
      set key K_F2 to ListProjects
      @ nRow,0 say "Project name or F2 for list or ESC" get cProject valid !empty(cProject)
      read
      if lastkey() == K_ESC
         quit
      endif
      set key K_F2 to
      @ nRow +1,0 clear
      if file("fsmake_" + trim(cProject) + ".dbf")
         lNew := .F.
      endif
      ok := .T.
      for ii := 1 to len(trim(cProject))   // check validity
         cc := substr(cProject,ii,1)
         if cc != "_" .and. !isalpha(cc) .and. !isdigit(cc)
            @ nRow +4,0 say "Project name may contain only charcters, digits and underscore"
            ok   := .F.
            lNew := .T.
            exit // exit the for loop
         endif
      next
      if ok
         exit    // exit the retry loop
      endif
   enddo
endif

cProject := trim(cProject)
cFile := "fsmake_" + cProject
if lNew
   if dbCreate(cFile, aStru)
      use (cFile) EXCLUSIVE
   endif
   if used()
      append blank
#ifdef FS_WIN32
      replace File with " .1", Path with " BCC32 " Switch with " -delc "
#else
      replace File with " .1", Path with " UNIX "  Switch with " -delc "
#endif
      append blank
      replace File with " .2"
      append blank
      replace File with " .3"
      append blank
      replace File with " .4"
      use
   endif
endif

use (cFile) EXCLUSIVE ALIAS Project
if !used()
   alert("Sorry, cannot open " + cFile + ".db*;probably used by others;or insufficient access rights")
   quit
endif
INDEX ON upper(File) TO (cFile)
cls
return cFile

/**************************************************************************
 * Main menu
 */
FUNCTION MenuProject(cDbf)
local iRow := 2, iCol := 0, iLen := 1, iMenu := 1
local bBlock := {|p1,p2obj,p3| iCol := p2obj:Column(), iLen := len(p2obj:Caption) }

while .T.
   cls
   @ iRow,0 say "Project: "
   @ iRow,col() say cProject COLOR "GR+/B" GUICOLOR "R+"
   @ iRow,col() +2 PROMPT " Settings "        SELECT bBlock STYLE BOX_RAISED GUICOLOR "B/W,W/B"
   @ iRow,col() +2 PROMPT " Source files "    SELECT bBlock STYLE BOX_RAISED GUICOLOR "B/W,W/B"
   @ iRow,col() +2 PROMPT " Create Makefile " SELECT bBlock STYLE BOX_RAISED GUICOLOR "B/W,W/B"
   @ iRow,col() +2 PROMPT " Quit "            SELECT bBlock STYLE BOX_RAISED GUICOLOR "B/W,W/B"
   MENU TO iMenu

   do case
   case iMenu == 0 .or. iMenu == 4
      return
   case iMenu == 1
      DrawTab(iRow,iCol,iLen)
      EditSettings(iRow,iCol,iLen)
   case iMenu == 2
      DrawTab(iRow,iCol,iLen)
      EditFiles(iRow,iCol,iLen)
   case iMenu == 3
      DrawTab(iRow,iCol,iLen)
      CreateMake(iRow,iCol,iLen)
   endcase
   iMenu += 1
enddo
return

/**************************************************************************
 * Submmenu: Settings
 */
FUNCTION EditSettings(iRow, iCol, iLen)
local cTemp, iPos, iRecord, cc, r1, r2
local lSave := .F., lAbort := .F.
local sw_m := .T., sw_c := .T.
private sw_na, sw_am, sw_q, sw_v, sw_I, sw_w1, sw_w2, sw_w3, sw_w4, sw_fox, sw_del, sw_mdi
MEMVAR nCreate, nRelease, nSystem, nIoMode
MEMVAR cExeName, cLibName, cMain, cAddSwitch, cAddObjLib, cSearchPath, cOutputDir

nCreate := nRelease := 1
nSystem := 0

sw_del := .T.

ReadSettins()

if nSystem == 0
# ifdef FS_WIN32
   nSystem := 2            // default WIN_BCC on Windows
# else
   nSystem := 1            // default LINUX if not Windows
# endif
endif

if empty(cExeName)
   cExeName := trim(cProject) + if(IS_WINDOWS, ".exe", "")
endif
cExeName := padr(cExeName, 25)

if empty(cLibName)
   if IS_LINUX
      cLibName := "lib" + upper(left(cProject,1)) + substr(cProject,2) + ".a"
   else
      cLibName := trim(cProject) + ".lib"
   endif
endif
cLibName := padr(cLibName, 25)

cMain := padr(cMain, 25)
cAddSwitch  += space(100)
cAddObjLib  += space(200)
cOutputDir  += space(100)

GetSwitches(nSystem)

if empty(cSearchPath)
   cSearchPath := GetIncludePaths()
endif
cSearchPath += space(100)

// @ iRow+2,0 say repli("0123456789",8)
iRow += 3
#ifdef FS_WIN32
@ iRow, 2, iRow+4,18 GET nCreate  RADIOGROUP {"Exe file","Exe+Library","Library"}  ;
                                  TOOLTIP "Select the kind of produced output"
#else
@ iRow, 2, iRow+4,18 GET nCreate  RADIOGROUP {"Executable","Exe+Library","Library"}  ;
                                  TOOLTIP "Select the kind of produced output"
#endif
@ iRow,20, iRow+5,38 GET nRelease RADIOGROUP {"Release","Debug .prg","Debug .c","Debug .prg .c"} ;
                                  TOOLTIP "Build without or with Debugger information"
@ iRow,40, iRow+4,58 GET nSystem  RADIOGROUP {"Linux","Windows BCC32","Windows MSVC6"} ;
                                  VALID GetSwitches(nSystem) ;
                                  TOOLTIP "Select the target system" + chr(10) + ;
                                          "for compiling your application"
@ iRow,60, iRow+5,79 GET nIoMode  RADIOGROUP {"-io=a Hybrid","-io=g GUI","-io=t Term","-io=b Basic"} ;
                                  TOOLTIP "Compile for all modes," + chr(10) + ;
                                          "or only for GUI, Terminal," + chr(10) + ;
                                          "or Basic i/o mode," + chr(10) + ;
                                          "see fsman LNG.1.2 and FSC.1.3"
iRow += 7
#ifdef FS_WIN32
 @ iRow,2 say "Name of the .EXE file "
#else
 @ iRow,2 say "Name of the executable "
#endif
cc := col()
r1 := row()
@ r1,cc GET cExeName VALID !empty(cExeName) WHEN IS_WITHEXE ;
                     TOOLTIP "Name of the produced executable," + chr(10) + ;
                             "apply only if Exe or Exe+Library is selected"
iRow += 0.5
@ ++iRow,2 say "Name of main module"
r2 := row()
@ r2,cc GET cMain VALID !empty(cMain) WHEN IS_WITHEXE ;
                  TOOLTIP "Name of the PROCEDURE or FUNCTION" + chr(10) + ;
                          "where your application should start," + chr(10) + ;
                          "e.g. 'main' or 'start' or '" + cProject + "' etc." + chr(10) + ;
                          "Required when Exe or Exe+Library is selected"
@ ++iRow,2 say "= start PROCEDURE, FUNCTION or .prg file"
iRow += 0.5
@ ++iRow,2 say "Name of the library "
@ row(),cc GET cLibName WHEN IS_WITHLIB VALID (IS_WITHLIB .and. !empty(cLibName)) ;
                        TOOLTIP "Name of the created library," + chr(10) + ;
                                "apply only if Exe+Library or Library is selected"

iRow ++
@ ++iRow,2 say "Default compiler switches:"
++iRow
@ ++iRow,2 GET sw_m  CHECKBOX CAPTION "-m : Perform modular compilation (required for make)" ;
                     TOOLTIP "see switches in fsman FSC.1.3"
Atail(GetList):GetEnabled := .f.  // disable
@ ++iRow,2 GET sw_c  CHECKBOX CAPTION "-c : compile each .prg to .obj   (required for make)" ;
                     TOOLTIP "see switches in fsman FSC.1.3"
Atail(GetList):GetEnabled := .f.  // disable
@ ++iRow,2 GET sw_v  CHECKBOX CAPTION "-v : Compile in verbose mode, display FlagShip and C settings" ;
                     TOOLTIP "see switches in fsman FSC.1.3"
@ ++iRow,2 GET sw_na CHECKBOX CAPTION "-na: Suppress generation of automatic procedure (same as filename), or"  ;
                     TOOLTIP "see switches in fsman FSC.1.3" + chr(10) + ;
                             "If set, apply for all source files here." + chr(10) + ;
                             "When only some sources should use this" + chr(10) + ; 
                             "switch, set it in 'Source files' tab" 
cc := Atail(GetList):CapCol
iRow += 0.8
@   iRow,cc say "     when using filewide vars. May be changed for specified files."
iRow += 1.2
@   iRow,2 GET sw_q  CHECKBOX CAPTION "-q : Compile in quiet mode, suppress display line numbers" ;
                     TOOLTIP "see switches in fsman FSC.1.3"
@ ++iRow,2 GET sw_w1 CHECKBOX CAPTION "-w1: Warnings for ambiguous variables"  ;
                     TOOLTIP "see switches in fsman FSC.1.3"
@ ++iRow,2 GET sw_w2 CHECKBOX CAPTION "-w2: Warnings for untyped LOCAL and STATICS" ;
                     TOOLTIP "see switches in fsman FSC.1.3"
// @ ++iRow,2 GET sw_w3 CHECKBOX CAPTION "-w3: Warnings for unknown parameter and return types" ;
//                      TOOLTIP "see switches in fsman FSC.1.3"
// @ ++iRow,2 GET sw_w4 CHECKBOX CAPTION "-w4: Warnings for unknown classes" ;
//                      TOOLTIP "see switches in fsman FSC.1.3"
@ ++iRow,2 GET sw_fox CHECKBOX CAPTION "-fox : compile FoxPro or Foxbase sources" ;
                      TOOLTIP "see switches in fsman FSC.1.3"
@ ++iRow,2 GET sw_mdi CHECKBOX CAPTION "-mdi : compile for using MDI subwindows" ;
                      TOOLTIP "see switches in fsman FSC.1.3" + chr(10) + ;
                              "Simulates Wopen() from FS2 Toolbox"
@ ++iRow,2 GET sw_del CHECKBOX CAPTION "-delc: remove .c files" ;
                      WHEN (if(IS_DEBUG_WITHC, sw_del := .F. , NIL), !IS_DEBUG_WITHC) ;
                      TOOLTIP "see switches in fsman FSC.1.3" + chr(10) + ;
                              "Effective only for build without C Debugger"
++iRow
@ ++iRow,2 say "Additional switches         " GET cAddSwitch PICTURE "@S45" ;
                     TOOLTIP "see switches in fsman FSC.1.3" + ;
                             chr(10) + "Add here additional switches if required, e.g." + ;
                             chr(10) + "-DFS2 -DMYSWITCH -I/my/includes -I/other/include -L/my/libs etc."
?? ">"
@ ++iRow,2 say "Additional objects and libs " GET cAddObjLib PICTURE "@S45" ;
                     TOOLTIP "Add objects and/or libraries not included in" + chr(10) + ;
                             "source in this project to be linked with, e.g." + chr(10) + ;
                             if(IS_WINDOWS, "myfile1.obj  \mypath\myfile2.obj  \other\path\mylib.lib", ;
                                            "myfile1.o  /mypath/myfile2.o  ../other/path/libMylib.a")
?? ">"
@ ++iRow,2 say "Search path for dependency  " GET cSearchPath PICTURE "@S45" ;
                     VALID (if(empty(cSearchPath), cSearchPath := GetIncludePaths(), NIL), .T.) ;
                     TOOLTIP "Apply when you select" + ;
                             chr(10) + "(in the section 'Source Files')" + ;
                             chr(10) + "to check also #include dependencies." + ;
                             chr(10) + "To re-read defaults, clear this" + ;
                             chr(10) + "field and press ENTER" 
?? ">"
@ ++iRow,2 say "Output directory for objects" GET cOutputDir PICTURE "@S45" ;
                     VALID CheckDir(cOutputDir) ;
                     TOOLTIP "Leave empty to create " + ;
                                 if(IS_WINDOWS, ".obj",".o") + " objects" + ;
                             chr(10) + "(and .c files) in current directory"
// ??## "min required row=",ltrim(iRow)                   

@ r1, 57 GET lSave  PUSHBUTTON CAPTION " Save (or PgDn) "  ;
                    TOOLTIP "Click to save chandes or default settings"
@ r2, 57 GET lAbort PUSHBUTTON CAPTION " Abort (or Esc) " WHEN !lSave ;
                    TOOLTIP "Click to exit w/o save changes"
read
if lastkey() == K_ESC .or. lAbort
   InfoBox("Aborted")
   return
endif

/* store default settings
 */
cTemp := " EXENAME=" + alltrim(cExeName)
if "." $ cMain
   cMain := left(cMain, at(".", cMain) -1)
endif
cTemp += " MAIN=" + alltrim(cMain)
if !empty(cLibName)
   cTemp += " LIBNAME=" + alltrim(cLibName)
endif

do case
case IS_EXE             // nCreate == 1
   cTemp += " EXEONLY"
case IS_EXELIB          // nCreate == 2
   cTemp += " EXELIB"
case IS_LIB             // nCreate == 3
   cTemp += " LIBONLY"
endcase

do case
case IS_WIN_BCC         // nSystem == 2
   cTemp += " BCC32"
case IS_WIN_MSC         // nSystem == 3
   cTemp += " MSVC6"
otherwise
   cTemp += " UNIX"
endcase

do case
case IS_DEBUG_PC        // nRelease == 4
   cTemp += " DEBUG_PRG DEBUG_C"
case IS_DEBUG_P         // nRelease == 2
   cTemp += " DEBUG_PRG"
case IS_DEBUG_C         // nRelease == 3
   cTemp += " DEBUG_C"
otherwise
   cTemp += " DEBUG_NO"
endcase

do case
case IS_IOMODE_A        // nIoMode == 1
   cTemp += " IOMODE_A"
case IS_IOMODE_G        // nIoMode == 2
   cTemp += " IOMODE_G"
case IS_IOMODE_T        // nIoMode == 3
   cTemp += " IOMODE_T"
case IS_IOMODE_B        // nIoMode == 4
   cTemp += " IOMODE_B"
endcase

cTemp += " "
replace Path with cTemp

/* store switches
 */
if IS_DEBUG_WITHC       // nRelease >= 3
   sw_del := .F.
endif
cTemp := " -m"
if sw_fox
   cTemp += " -fox"
endif
if sw_del
   cTemp += " -delc"
endif
if sw_v
   cTemp += " -v"
endif
if sw_na
   cTemp += " -na"
endif
if sw_q
   cTemp += " -q"
endif
if sw_w1
   cTemp += " -w1"
endif
if sw_w2
   cTemp += " -w2"
endif
if sw_w3
   cTemp += " -w3"
endif
if sw_w4
   cTemp += " -w4"
endif
if sw_mdi
   cTemp += " -mdi"
endif

cTemp += " "
replace Switch with cTemp

iRecord := recno()
seek " .2"
if found()
   if !empty(cAddObjLib) .or. !empty(Project->Path)
      replace Project->Path with trim(cAddObjLib)
   endif
   if !empty(cAddSwitch) .or. !empty(Project->Switch)
      replace Project->Switch with trim(cAddSwitch)
   endif
endif

seek " .3"
if found()
   replace Project->Path with trim(cSearchPath)
endif

seek " .4"
if found()
   replace Project->Path with trim(cOutputDir)
endif

goto (iRecord)
commit
return

/**************************************************************************
 * Submmenu: Edit Files
 */
FUNCTION EditFiles(iRow, iCol, iLen)
local _oPrompt, cType, iSelect := 1
local aHeader := {"Use","Name","Path","Switch","FS","Dep"}
local cName, cOldName,cExt, cName1, cName2, isDepend, lExit := .F.
local iRec, iFileLen, iPathLen, iSwitchLen, iDepLen, lWarn, bRedir
local iFiles := 0, iLibs := 0, iEmpty := 0
PRIVATE lSaved := .F., lNeedSave := .F., aProjectFiles := {}, aDepend := {}
PRIVATE oBrProj, oSaveButt, nButtRow, nButtCol
MEMVAR nCreate, nRelease, nSystem, nIoMode

bRedir := Onkey(0)
ON ANY KEY

@ ++iRow, iCol say chr(179)
@ ++iRow, iCol say chr(179)
iRow += 0.4
@ iRow, iCol say chr(192)
@ iRow, iCol+1 say chr(196)

while !lExit
   @ iRow +1,0 clear
   aProjectFiles := {}

   if valtype(oSaveButt) == "O"
      oSaveButt:Axit()
   endif
   oSaveButt := NIL


   go top
   while !eof() .and. left(Project->File,1) == " "
       skip     // skip header records
   enddo
   iRec := recno()
   iFileLen := iPathLen := iSwitchLen := iDepLen := 0
   while !eof()
      iFileLen   := max(iFileLen, len(trim(Project->File)) )
      iPathLen   := max(iPathLen, len(trim(Project->Path)) )
      iSwitchLen := max(iSwitchLen, len(trim(Project->Switch)) )
      iDepLen    := max(iDepLen,  len(trim(Project->Depend)) )
      skip
   enddo
   goto (iRec)
   iFileLen   := min(max(15, iFileLen +3), len(Project->File))
   iPathLen   := max(30, iPathLen +5)
   iSwitchLen := max(20, iSwitchLen +5)
   iDepLen    := max(50, iDepLen + 20)
   cOldName   := ""
   lWarn := .F.

   while !eof()
      isDepend := .F.
      if !empty(Project->Depend)
         isDepend := .T.
         aadd(aDepend, alltrim(Project->Depend))         // dependency string
      endif
      aadd(aProjectFiles, ;
             { Project->Type, ;                           // APROJ_TYPE
               padr(trim(Project->File), iFileLen), ;     // APROJ_FILE
               padr(trim(Project->Path), iPathLen), ;     // APROJ_PATH
               padr(trim(Project->Switch), iSwitchLen), ; // APROJ_SWITCH
               if(Project->FSsysNum == 0, " ", "Y"), ;    // APROJ_FSMARK
               if(isDepend, "Y", " "), ;                  // APROJ_DEPEND
               Project->FSsysNum, ;                       // APROJ_FSMARKNUM
               if(isDepend, len(aDepend), 0) } )          // APROJ_DEPENDIDX
      do case
      case Project->Type == "F"
         iFiles ++
      case Project->Type == "L"
         iLibs ++
      case empty(Project->Type)
         iEmpty ++
      endcase

      /* check doublettes
       */
      if left(upper(Project->Type),1) $ "FL"
         cName := trim(Project->Path) + "/" + trim(Project->File)
         if IS_WINDOWS
            cName := upper(cName)
         endif
         if !empty(cOldName) .and. cOldName == cName
            aProjectFiles[len(aProjectFiles),1] := "-"
            lWarn := .T.
         endif
         cOldName := cName
      endif
      skip
   enddo

   if lWarn
      Alert("Doublet(s) found, ea. second marked to delete")
   endif

   if empty(aProjectFiles)
      @ iRow +4, 10 say "None files available yet, press Add New"
      iSelect := 2
// else
//    /* display preview (non-edit)
//     */
//    @ iRow+2,0 clear
//    @ iRow+2,0 say 'Preview only, select "Edit current files" to view/edit all' COLOR "GR+/B"
//    if AppIoMode() == "T"
//       @ iRow+3,0 TO iRow+3,maxcol()
//       oBrProj := TbrowseArr(iRow+4,0, maxrow() -1, maxcol(), .F., .T.)
//       oBrProj:ColSep  := chr(179)      // or "|"
//       oBrProj:HeadSep := chr(197,196)  // or "+-"
//       oBrProj:ColSepEof := .F.
//    else
//       oBrProj := TbrowseArr(iRow+3,0, maxrow() -1, maxcol(), .F., .T.)
// //      oBrProj:RowCache := len(aProjectFiles)  // optional: to scroll by mouse
//    endif
//    oBrProj:UserArray := aProjectFiles
//    for ii := 1 to len(aHeader)
//       oBrProj:AddColumn( TbColumnNew(aHeader[ii], .T.))
//    next
//    oBrProj:CanAppend := .F.
//    oBrProj:ReadOnly  := .T.
//    oBrProj:ForceStable()
//    oBrProj:Show()
   endif

   @ iRow,iCol  +2 PROMPT "Edit current files " STYLE BOX_RAISED GUICOLOR "B+/W,W/B+"
   @ iRow,col() +1 PROMPT "Add new files "      STYLE BOX_RAISED GUICOLOR "B+/W,W/B+"
   @ iRow,col() +1 PROMPT "Exit "               STYLE BOX_RAISED GUICOLOR "B+/W,W/B+"
   MENU TO iSelect

   do case
   case iSelect == 0 .or. iSelect == 3
      @ iRow,0 clear
      Onkey(0, bRedir)
      return

   case iSelect == 1 .and. empty(aProjectFiles)
         @ iRow +4, 10 say "None files available yet, press Add New"

   case iSelect == 1
      /* try to automatically assign "Use" flag as file or lib
       */
      if iEmpty > 0
         ReadSettins()
         for ii := 1 to len(aProjectFiles)
            if empty(aProjectFiles[ii,APROJ_TYPE])
               do case
               case IS_EXE             // exe nCreate == 1
                  aProjectFiles[ii,APROJ_TYPE] := "F"
                  iFiles ++
               case IS_LIB             // libnCreate == 3
                  aProjectFiles[ii,APROJ_TYPE] := "L"
                  iLibs ++
               case  IS_EXELIB         // exe + lib nCreate == 2
                  if iFiles > 0
                     aProjectFiles[ii,APROJ_TYPE] := "L"
                     iLibs ++
                  else
                     cName1 := alltrim(aProjectFiles[ii,APROJ_FILE])
                     cName2 := cMain + ".prg"
                     if IS_WINDOWS     // nSystem >= 2
                        cName1 := upper(cName1)
                        cName2 := upper(cName2)
                     endif
                     if cName1 == cName2
                        aProjectFiles[ii,APROJ_TYPE] := "F"
                        iFiles ++
                     else
                        aProjectFiles[ii,APROJ_TYPE] := "L"
                        iLibs ++
                     endif
                  endif
               endcase
            endif
         next
      endif

      /* display/edit
       */
      @ iRow+2,0 clear
//                   0....5...10....5...20....5...30....5...40....5...50....5...60....5...70....5...9
      @ iRow+2,1 say "[PgDn][PgUp]skip   [C]clear all [E]all=file [F]file [D]dependency     [S]save!" COLOR "GR+/B" GUICOLOR "B"
      @ iRow+3,1 say "[Enter]edit [A]add [-]remove    [B]all=lib  [L]lib  [V]view depend  [Esc]abort" COLOR "GR+/B" GUICOLOR "B"
      nButtRow := iRow+2
      nButtCol := 71

      if AppIoMode() == "T"
         @ iRow+4,0 TO iRow+4,maxcol()
         oBrProj := TbrowseArr(iRow+5,0, maxrow() -1, maxcol(), .F., .T.)
         oBrProj:ColSep  := chr(179)      // or "|"
         oBrProj:HeadSep := chr(197,196)  // or "+-"
         oBrProj:ColSepEof := .F.
      else
         oBrProj := TbrowseArr(iRow+4,0, maxrow()-1, maxcol(), .F., .T.)
         // disable to browse via PgDn/PgUp:
         oBrProj:RowCache := max(oBrProj:RowCount, len(aProjectFiles))  // optional: to scroll by mouse
      endif
      oBrProj:UserArray := aProjectFiles  // only first 6 columns
      for ii := 1 to len(aHeader)
         oBrProj:AddColumn( TbColumnNew(aHeader[ii], .T.) )
      next
      oBrProj:OnUpdate := {|obj,mode| SetSaveButt(obj,mode) }

      /* Note: we need to handle the above special keys.
       * We either can copy the .../system/tbrowsehand.prg here and extend it
       * for these keys,  or redirect keys to sub-handler and precess only
       * these special keys. We use the last, since cheaper :-)
       */
      // oBrProj:Handler := {|| MyTbHandler() }
      inkey()                    // clear possible ENTER
      SET CONFIRM ON             // edit with ENTER
      ON ANY KEY DO TbrHandEdit  // redirect any key to TbrHandEdit()
      oBrProj:Exec()             // process Tbrowse()
      ON ANY KEY                 // disable key redirection
      SET CONFIRM OFF

      if lSaved
         delete all for left(Project->File,1) != " "  // remove file entries
         pack
         for ii := 1 to len(aProjectFiles)            // add newly edited files
            if aProjectFiles[ii,APROJ_TYPE] != "-"
               append blank
               replace File      with aProjectFiles[ii,APROJ_FILE]
               replace Type      with alltrim(aProjectFiles[ii,APROJ_TYPE])
               replace Path      with alltrim(aProjectFiles[ii,APROJ_PATH])
               replace Switch    with alltrim(aProjectFiles[ii,APROJ_SWITCH])
               if empty(aProjectFiles[ii,APROJ_FSMARK])
                  replace FSsysNum  with 0
               else
                  replace FSsysNum  with aProjectFiles[ii,APROJ_FSMARKNUM]
               endif
               if empty(aProjectFiles[ii,APROJ_DEPEND]) .or. aProjectFiles[ii,APROJ_DEPENDIDX] == 0
                  replace Depend with ""
               else
                  replace Depend with aDepend[aProjectFiles[ii,APROJ_DEPENDIDX]]
               endif
            else
               /* remove deleted object and .c file, if exist
                */
               cName := aProjectFiles[ii,APROJ_FILE]
               cExt  := upper(substr(cName, rat(".", cName) -1))
               cName2:= left(cName, rat(".", cName) -1)
               cName := cName2 + if(IS_WINDOWS, ".obj", ".o")
               if cExt $ ".PRG,.FMT,.C" .and. file(cName)
                  delete file (cName)
               endif
               cName := cName2 + ".c"
               if cExt $ ".PRG,.FMT" .and. file(cName)
                  delete file (cName)
               endif
            endif
         next
         commit
         lExit := .T.
      else
         InfoBox("Aborted")
      endif

   case iSelect == 2
      NewFiles(iRow+2)
      iSelect := 1
   endcase

enddo

Onkey(0, bRedir)           // restore previous redirection
return

/**************************************************************************
 * Submmenu: Create Makefile
 */
FUNCTION CreateMake(iRow, iCol, iLen)
local ok := .T., cMainFile := "", iMainRec := 0
local LF, ii, fd, cTmp, cType, cName, cName1, cName2, cPath, cExt, cObjExt
local iMainPos, iFsType, iNumLibs, iLenLibs, iPartLib, iPos
local cMakeFile, cMakeName, cMakeInvoc, cSystem, cTmpSize
local isDepend, cCompSwitch := "", cIoModeSwitch := "", cCompSystem := ""
local cObjDir := "", cObjOutSwitch := "-o ", cExeOutSwitch := "-o "
local aProjectFiles := {}, aDepend := {}, aObj := {}, aLib := {}, aFSfile := {}
MEMVAR nCreate, nRelease, nSystem, nIoMode
MEMVAR cExeName, cLibName, cMain, cAddSwitch, cAddObjLib, cMainSwitch, cSearchPath, cOutputDir

nCreate := nRelease := nSystem := 0

@ iRow +2,0 clear
ReadSettings()

do case
case nCreate == 0
   ?
   ? 'Please go to "Settings" and enter the data there'
   ok := .F.
case IS_WITHEXE .and. (empty(cExeName) .or. empty(cMain))   // nCreate <= 2
   ?
   ? 'Please go to "Settings" and enter there the executable name and main module'
   ok := .F.
case IS_WITHLIB .and. empty(cLibName)     // nCreate >= 2
   ?
   ? 'Please go to "Settings" and enter there the library name'
   ok := .F.
endcase
if !ok .or. nSystem == 0
   wait
   return
endif

cObjDir := alltrim(cOutputDir)
if !CheckDir(cObjDir)
   return
endif
if !empty(cObjDir)
   if !(right(cObjDir,1) $ "/\")
      cObjDir += if(IS_WINDOWS, "\", "/")
   endif 
endif
cObjDir := translSlash(cObjDir)

go top
while !eof() .and. left(Project->File,1) == " "
    skip     // skip header records
enddo
ok := .F.
cName1 := if(IS_WINDOWS, upper(cMain), cMain) + ".prg"
while !eof()
   cName := trim(Project->Path)
   if !empty(cName) .and. !(right(cName,1) $ "/\:")
      cName += if(IS_LINUX, "/", "\")
      cName := translSlash(cName)
   endif
   isDepend := .F.
   if !empty(Project->Depend)
      isDepend := .T.
      aadd(aDepend, alltrim(Project->Depend))   // dependency string
   endif
   
   aadd(aProjectFiles, ;
          { Project->Type, ;                    // APROJ_TYPE
            Project->File, ;                    // APROJ_FILE
            cName, ;                            // APROJ_PATH
            trim(Project->Switch), ;            // APROJ_SWITCH
            NIL, ;                              // APROJ_FSMARK
            NIL, ;                              // APROJ_DEPEND
            Project->FSsysNum, ;                // APROJ_FSMARKNUM
            if(isDepend, len(aDepend), 0) } )   // APROJ_DEPENDIDX

   if Project->Type == "F"
      ok := .T.
   endif
   skip
enddo

if len(aProjectFiles) == 0
   ?
   ? 'Please go back to "Source Files" and add project files'
   ?
   wait
   return
endif

cTmpSize := ""
for ii := 1 to len(aProjectFiles)
   cType  := left(upper(aProjectFiles[ii,APROJ_TYPE]),1)
   cName1 := aProjectFiles[ii,APROJ_FILE]
   cName1 := left(cName1, rat(".", cName1) -1)
   if cType $ "FL" .and. aProjectFiles[ii,APROJ_FSMARKNUM] > 0 // FS files
      iFsType := aProjectFiles[ii,APROJ_FSMARKNUM]
      if binAnd(iFsType, 1) > 0
         aadd(aFSfile, {cName1, ii, " "})
      endif
      if binAnd(iFsType, 16) > 0
         aadd(aFSfile, {cName1 + "_a", ii, "A"})
      endif
      if binAnd(iFsType, 2) > 0
         aadd(aFSfile, {cName1 + "_g", ii, "G"})
      endif
      if binAnd(iFsType, 4) > 0
         aadd(aFSfile, {cName1 + "_t", ii, "T"})
      endif
      if binAnd(iFsType, 8) > 0
         aadd(aFSfile, {cName1 + "_b", ii, "B"})
      endif
   elseif cType == "F"
      aadd(aObj, cName1)         // obj for files w/o extension
      cTmpSize += cName1 + " "
   elseif cType == "L"
      aadd(aLib, cName1)         // obj for lib w/o extension
   endif
next

ok := .T.
if IS_WITHEXE .and. len(aObj) == 0           // nCreate <= 2
   ?
   ? 'Please go back to "Source Files" and select at least one file (F)'
   ? 'It should be preferably the main source file.'
   ok := .F.
endif
if ok .and. IS_WITHLIB .and. len(aLib) == 0  // nCreate >= 2
   ?
   ? 'Please go back to "Source Files" and select at least one library file (L)'
   ok := .F.
endif
if !ok
   ?
   wait
   return
endif

cTmpSize += cAddObjLib
if IS_WIN_BCC .and. len(cTmpSize) > 3000 .and. IS_WITHEXE    // nCreate <= 2
   ?
   ? "Warning: the number of files will most probably oversize"
   ? "         BCC32 command line argument buffer."
   if IS_EXE                                                   // nCreate == 1
      ? "         Suggestion: select Exe+Library in 'Settings'"
   else
      ? "         Suggestion: add some files into Library in 'Settings'"
   endif
   ?
   wait "Do you wish to continue to create Makefile (N/y) ? "
   if !(chr(lastkey()) $ "yY")
      return
   endif
   ?
elseif IS_WIN_MSC .and. len(cTmpSize) > 3000 .and. IS_WITHEXE // nCreate <= 2
   ?
   ? "Warning: the number of files may probably oversize"
   ? "         MS-Windows command line argument buffer."
   if IS_EXE                                                   // nCreate == 1
      ? "         If you get linker error saying 'buffer size overflow',"
      ? "         or similarly, select Exe+Library in 'Settings'"
   else
      ? "         Suggestion: add some files into Library in 'Settings'"
   endif
   ?
   wait "Do you wish to continue to create Makefile (n/Y) ? "
   if (chr(lastkey()) $ "nN")
      return
   endif
   ?
endif

if IS_LINUX
   LF := chr(10)
else
   LF := chr(13,10)
endif
#define TAB chr(9)

cCompSwitch := ""
if IS_DEBUG_P .or. IS_DEBUG_PC      // nRelease == 2 .or. nRelease == 4
   cCompSwitch += "-d "
endif
if IS_DEBUG_WITHC                   // nRelease >= 3
   cCompSwitch += "-g "
endif
cCompSwitch += alltrim(cMainSwitch)
cCompSystem := cCompSwitch + " "
if "-fox " $ cCompSystem
   cCompSystem := strtran(cCompSystem, "-fox ", "")
endif
if "-foxpro " $ cCompSystem
   cCompSystem := strtran(cCompSystem, "-foxpro ", "")
endif
if "-fxp " $ cCompSystem
   cCompSystem := strtran(cCompSystem, "-fxp ", "")
endif
if !IS_DEBUG_WITHC .and. !("-delc" $ cCompSystem)
   cCompSystem += " -delc "
endif
cCompSystem := alltrim(strtran(cCompSystem, "  ", " "))

do case
case IS_IOMODE_A
   cIoModeSwitch := " -io=a"
case IS_IOMODE_G
   cIoModeSwitch := " -io=g"
case IS_IOMODE_T
   cIoModeSwitch := " -io=t"
case IS_IOMODE_B
   cIoModeSwitch := " -io=b"
endcase

do case
case IS_LINUX                       // nSystem == 1
   cMakeName := "make"
   cSystem   := "for Linux"
case IS_WIN_BCC                     // nSystem == 2
   cMakeName := "make"
   cSystem   := "with BCC32 compiler"
case IS_WIN_MSC                     // nSystem == 3
   cMakeName := "nmake"
   cSystem   := "with MSVC6 compiler"
endcase

// if IS_WIN_MSC    // not required, done by FS compiler
// cObjOutSwitch := "-Wc,/Fo"
// cExeOutSwitch := "-Wc,/Fe"
// else
   cObjOutSwitch := "-o "
   cExeOutSwitch := "-o "
// endif

cMakeFile := "Makefile"
cMakeInvoc := cMakeName
if file(cMakeFile)
   if !checkMakefile(cMakeFile)
      cMakeFile  := cProject + ".make"
      cMakeInvoc := cMakeName + " -f " + cMakeFile
      ii := 1
      while !checkMakefile(cMakeFile)
         cMakeFile := cProject + ".make" + ltrim(ii)
         if checkMakefile(cMakeFile)
            cMakeInvoc := cMakeName + " -f " + cMakeFile
            exit
         endif
         ii++
      enddo
   endif
endif
if cMakeFile != "Makefile"
   ? "Note: cannot create 'Makefile' since it already exist for another project"
   wait "      will create file named '" + cMakeFile + "' - continue (Y/n) ? "
   if chr(lastkey()) $ "Nn"
      return
   endif
   ?
endif

if file(cMakefile)
   delete file (cMakefile)
endif
fd := fopen(cMakefile, FO_WRITE, FO_CREATE)
if fd <= 0
   ? "Cannot create file '" + cMakefile + "'"
   ? "Check for permission/access rights to directory " + curdir()
#ifndef FS_WIN32
   ? 'by "ls -lad .", should be at least "rwx" for user/group/all'
   if file(cMakefile)
      ? 'and the file permission by "ls -la ' + cMakefile + '"' 
      ? 'which you may change by "sudo chmod 666 ' + cMakefile + '"'
   endif
#endif
   wait
   return
endif

fwrite(fd, "# Makefile created by FSmake on " + dtoc(date()) + " " + time() + LF)
fwrite(fd, "# using FlagShip " + cSystem + LF)
fwrite(fd, "# for FSmake Project: " + cProject + LF)
fwrite(fd, "#" + LF)
do case
case IS_RELEASE                                 // nRelease == 1
   fwrite(fd, "# = Project-Release version" + LF)
case IS_DEBUG_P                                 // nRelease == 2
   fwrite(fd, "# = Project-Debug version" + LF)
case IS_DEBUG_C                                 // nRelease == 3
   fwrite(fd, "# = Project-Debug version for .c" + LF)
case IS_DEBUG_PC                                // nRelease == 4
   fwrite(fd, "# = Project-Debug version for .prg and .c" + LF)
endcase
fwrite(fd, "#" + LF)
fwrite(fd, "# invoke: " + cMakeInvoc + "          to rebuild new/changed sources"+ LF)
fwrite(fd, "# invoke: " + cMakeInvoc + " clean    to remove objects and old executable/lib"+ LF)
if IS_EXELIB .and. !empty(cLibName) .and. !empty(aLib)      // nCreate == 2
   fwrite(fd, "# invoke: " + cMakeInvoc + " " + cLibName + "   to rebuild library only"+ LF)
endif
fwrite(fd, "#" + LF)
fwrite(fd, "# Note: you may freely modify this file, but it may be overwritten" + LF)
fwrite(fd, "#       by the next invocation of 'fsmake' for the same project." + LF)
fwrite(fd, "#       If you rename it to 'anything.else', invoke it by" + LF)
fwrite(fd, "#       " + cMakeName + " -f anything.else [clean]" + LF)
fwrite(fd, LF)
fwrite(fd, "# comment-out next statement to display make statements" + LF)
fwrite(fd, ".SILENT:" + LF)
fwrite(fd, LF)
fwrite(fd, ".SUFFIXES:" + LF)
if IS_LINUX         // nSystem == 1
   fwrite(fd, ".SUFFIXES: .o .prg .PRG .fmt .FMT" + LF)
else
   fwrite(fd, ".SUFFIXES: .obj .prg .fmt" + LF)
endif
fwrite(fd, LF)
fwrite(fd, "EXENAME  = " + cExeName + LF)
fwrite(fd, "LIBNAME  = " + cLibName + LF)
if IS_WIN_MSC                             // nSystem == 3
   fwrite(fd, 'LIBPATH  = /LIBPATH:"$(PROGRAMFILES)\Microsoft Visual Studio\VC98\lib"' + LF)
endif
fwrite(fd, "OBJ      = " + if(IS_LINUX, ".o", ".obj") + LF)
fwrite(fd, "FS       = FlagShip" + LF)
fwrite(fd, "FSCOMP   = -c " + cCompSwitch + LF)
fwrite(fd, "FSCOMPSYS= -c " + cCompSystem + LF)
fwrite(fd, "FSLINK   = " + cCompSwitch + cIoModeSwitch + LF)
fwrite(fd, "FSADDOPT = " + cAddSwitch + LF)
fwrite(fd, "MAINPROC = -M" + lower(cMain) + LF)
fwrite(fd, LF)
fwrite(fd, "DEL      = " + if(IS_LINUX, "rm -f", "del") + LF)
fwrite(fd, "ECHO     = echo" + LF)
do case
case IS_LINUX                             // nSystem == 1
   fwrite(fd, "LIB      = ar qvS" + LF)
case IS_WIN_BCC                           // nSystem == 2
   fwrite(fd, "LIB      = TLIB /C /P512" + LF)
case IS_WIN_MSC                           // nSystem == 3
   fwrite(fd, "LIB      = LIB /NOLOGO /SUBSYSTEM:windows" + LF)
endcase
fwrite(fd, LF)
fwrite(fd, "ADDOBJ  = " + cAddObjLib + LF)
fwrite(fd, LF)
fwrite(fd, "OBJECTS  = " )
cObjExt := if(IS_WIN_MSC, ".obj ", "$(OBJ) ")
for ii := 1 to len(aObj)
   fwrite(fd, "\" + LF + TAB + cObjDir + aObj[ii] + cObjExt)
next
fwrite(fd, LF)
fwrite(fd, LF)

iNumLibs := 0
iLenLibs := 0
iPartLib := 0
for ii := 1 to len(aLib)
   if ii == 1 .or. ;
       (ii < len(aLib) .and. ;
         (iLenLibs > 8000 .or. iPartLib >= 500) )   // max 500 files or 8kb ea
      iNumLibs++
      iLenLibs := 0
      iPartLib := 0
      if ii > 1
         fwrite(fd, LF)
         fwrite(fd, LF)
      endif
      fwrite(fd, "OBJLIB" + ltrim(iNumLibs) + "  = " )
   endif
   fwrite(fd, "\" + LF + TAB + cObjDir + trim(aLib[ii]) + cObjExt)
   iLenLibs += len(cObjDir + trim(aLib[ii]) + cObjExt) + 2
   iPartLib++
next
fwrite(fd, LF)
fwrite(fd, LF)

fwrite(fd, "OBJSYS   = " )
for ii := 1 to len(aFSfile)
   fwrite(fd, "\" + LF + TAB + cObjDir + aFSfile[ii,1] + cObjExt)
next
fwrite(fd, LF)
fwrite(fd, LF)

/* -------- all:
 */
fwrite(fd, LF)
fwrite(fd, "all: " )
if !empty(cLibName) .and. !empty(aLib)
   fwrite(fd, "$(LIBNAME) ")
endif
if !empty(cExeName)
   fwrite(fd, "$(EXENAME)")
endif
fwrite(fd, LF)
fwrite(fd, LF)

/* -------- EXE
 */
if !empty(cExeName) .and. !empty(aObj)
   fwrite(fd, "$(EXENAME): $(OBJECTS) $(OBJSYS)" + LF)
   fwrite(fd, TAB + "$(FS) $(FSLINK) $(MAINPROC) $(FSADDOPT) $(OBJECTS) $(OBJSYS) ")
   if IS_EXELIB .and. !empty(cLibName) .and. !empty(aLib)   // nCreate == 2
      fwrite(fd, "$(LIBNAME) ")
   endif
   fwrite(fd, "$(ADDOBJ) " + cExeOutSwitch + "$(EXENAME)" + LF)
   if IS_WINDOWS
      fwrite(fd, TAB + "$(ECHO)."  + LF)
      cTmp := TAB + "$(ECHO) ** File '$(EXENAME)' created, invoke: "
      do case
      case IS_IOMODE_A
         cTmp += "$(EXENAME)  -or-  $(EXENAME) -io=t"
      otherwise
         cTmp += "$(EXENAME)"
      endcase
      fwrite(fd, cTmp + LF)
   else
      fwrite(fd, TAB + "$(ECHO)" + LF)
      cTmp := TAB + '$(ECHO) "** Executable ' + "'$(EXENAME)' created, "
      do case
      case IS_IOMODE_A
         fwrite(fd, cTmp + '"' + LF)
         cTmp := TAB + '$(ECHO) "** invoke: ./$(EXENAME)  -or-  newfswin ./$(EXENAME)'
      case IS_IOMODE_T
         cTmp += "invoke: newfswin ./$(EXENAME)"
      otherwise
         cTmp += "invoke: ./$(EXENAME)"
      endcase
      fwrite(fd, cTmp + '"' + LF)
   endif
   fwrite(fd, LF)
endif

/* -------- LIB
 */
if !empty(cLibName) .and. !empty(aLib)
   cTmp := ""
   for ii := 1 to iNumLibs
      cTmp += " $(OBJLIB" + ltrim(ii) + ")"
   next
   fwrite(fd, "$(LIBNAME):" + cTmp + LF)
   if !empty(cExeName)
      fwrite(fd, TAB + "$(DEL) $(EXENAME)" + LF)
   endif
   do case
   case IS_LINUX                    // nSystem == 1
      fwrite(fd, TAB + "$(DEL) $(LIBNAME)" + LF)
      fwrite(fd, TAB + "$(LIB) $(LIBNAME) $(OBJLIB)" + LF)
      fwrite(fd, TAB + "ranlib $(LIBNAME)" + LF)
   case IS_WIN_BCC                  // nSystem == 2
      cName := left(cLibName, rat(".",cLibName)-1)
      fwrite(fd, TAB + "$(DEL) $(LIBNAME)" + LF)
      fwrite(fd, TAB + "$(LIB) $(LIBNAME) @&&| " + LF)
      for ii := 1 to len(aLib)
         if ii == len(aLib)
            fwrite(fd, "+" + cObjDir + aLib[ii] + "$(OBJ)" + LF)
            fwrite(fd, "| , " + cName + ".map" + LF)
            fwrite(fd, TAB + "-del " + cName + ".bak" + LF)
         elseif ii % 200 == 0
            fwrite(fd, "+" + cObjDir + aLib[ii] + "$(OBJ)" + LF)
            fwrite(fd, "| " + LF)
            fwrite(fd, TAB + "$(LIB) $(LIBNAME) @&&| " + LF)
         else
            fwrite(fd, "+" + cObjDir + aLib[ii] + "$(OBJ) \" + LF)
         endif
      next
   case IS_WIN_MSC                  // nSystem == 3
      fwrite(fd, TAB + "$(LIB) $(LIBPATH) /OUT:$(LIBNAME) @<<" + LF)
      fwrite(fd, TAB + "$(OBJLIB)" + LF)
      fwrite(fd, "<<" + LF)
   endcase
   fwrite(fd, LF)
   fwrite(fd, TAB + "$(ECHO)" + if(IS_WINDOWS, ".", "") + LF)
   fwrite(fd, TAB + "$(ECHO) $(LIBNAME) created" + LF)
   fwrite(fd, LF)
endif


/* -------- clean:
 */
fwrite(fd, LF)
fwrite(fd, "clean:" + LF)
if !empty(cExeName)
   fwrite(fd, TAB + "$(DEL) $(EXENAME)" + LF)
endif
if !empty(cLibName) .and. IS_WITHLIB               // nCreate >= 2
   fwrite(fd, TAB + "$(DEL) $(LIBNAME)" + LF)
endif
if IS_LINUX                         // nSystem == 1
   if len(aObj) > 0
      fwrite(fd, TAB + "$(DEL) $(OBJECTS)" + LF)
   endif
   if len(aLib) > 0
      fwrite(fd, TAB + "$(DEL) $(OBJLIB)" + LF)
   endif
   if len(aFSfile) > 0
      fwrite(fd, TAB + "$(DEL) $(OBJSYS)" + LF)
   endif
else
   cObjExt := if(IS_WIN_MSC, ".obj ", "$(OBJ) ")
   cName1 := ""
   for ii := 1 to len(aObj)
      cName1 += cObjDir + aObj[ii] + cObjExt
      if ii == len(aObj)
         fwrite(fd, TAB + "$(DEL) " + cName1 + LF)
      elseif ii > 0 .and. ii % 20 == 0
         fwrite(fd, TAB + "$(DEL) " + cName1 + LF)
         cName1 := ""
      endif
   next
   cName1 := ""
   for ii := 1 to len(aLib)
      cName1 += cObjDir + aLib[ii] + cObjExt
      if ii == len(aLib)
         fwrite(fd, TAB + "$(DEL) " + cName1 + LF)
      elseif ii > 0 .and. ii % 20 == 0
         fwrite(fd, TAB + "$(DEL) " + cName1 + LF)
         cName1 := ""
      endif
   next
   cName1 := ""
   for ii := 1 to len(aFSfile)
      cName1 += cObjDir + aFSfile[ii,1] + cObjExt
      if ii == len(aFSfile)
         fwrite(fd, TAB + "$(DEL) " + cName1 + LF)
      elseif ii > 0 .and. ii % 20 == 0
         fwrite(fd, TAB + "$(DEL) " + cName1 + LF)
         cName1 := ""
      endif
   next
endif
fwrite(fd, LF)

/* -------- dependency rule: compile files (with dependencies)
 */
cExt := if(IS_WIN_MSC, ".obj ", "$(OBJ) ")
for ii := 1 to len(aProjectFiles)
   if aProjectFiles[ii,APROJ_FSMARKNUM] > 0  // all except FlagShip system files
      loop
   endif
   cName1 := translSlash(aProjectFiles[ii,APROJ_FILE])
   cName1 := left(cName1, rat(".", cName1) -1)
   fwrite(fd, cObjDir + cName1 + cExt + ": " + ;
              alltrim(translSlash(aProjectFiles[ii,APROJ_PATH])) + ;
              alltrim(translSlash(aProjectFiles[ii,APROJ_FILE])) )
   if aProjectFiles[ii,APROJ_DEPENDIDX] > 0        // dependencies
      cTmp := alltrim(aDepend[aProjectFiles[ii,APROJ_DEPENDIDX]])
      cTmp := strtran(cTmp, "+", " ")
      cTmp := strtran(cTmp, "  ", " ")
      cTmp := translSlash(cTmp)
      fwrite(fd, " " + alltrim(cTmp))
   endif
   fwrite(fd, LF)
   cTmp := alltrim(aProjectFiles[ii,APROJ_SWITCH]) // addit. file switches
   if !("-io=" $ cTmp) .and. !empty(cIoModeSwitch)
      cTmp += " " + alltrim(cIoModeSwitch) 
   endif
   if !empty(cTmp)
      cTmp += " "
   endif
   fwrite(fd, TAB + "$(FS) $(FSCOMP) $(FSADDOPT) " + cTmp + ;
              cObjOutSwitch + cObjDir + cName1 + cExt + " " + ;
              trim(translSlash(aProjectFiles[ii,APROJ_PATH])) + ;
              trim(translSlash(aProjectFiles[ii,APROJ_FILE])) + LF)
   fwrite(fd, LF)
next
fwrite(fd, LF)

/* FlagShip system files
 */
for ii := 1 to len(aFSfile)
   iPos    := aFSfile[ii,2]
   iFsType := aProjectFiles[iPos,APROJ_FSMARKNUM]

   fwrite(fd, cObjDir + translSlash(aFSfile[ii,1]) + cExt + ": " + ;
              alltrim(translSlash(aProjectFiles[iPos,APROJ_PATH])) + ;
              alltrim(translSlash(aProjectFiles[iPos,APROJ_FILE])) )
   if aProjectFiles[iPos,APROJ_DEPENDIDX] > 0        // dependencies
      cTmp := alltrim(aDepend[aProjectFiles[iPos,APROJ_DEPENDIDX]])
      cTmp := strtran(cTmp, "+", " ")
      cTmp := strtran(cTmp, "  ", " ")
      cTmp := translSlash(cTmp)
      fwrite(fd, " " + alltrim(cTmp))
   endif
   fwrite(fd, LF)

   cTmp := alltrim(aProjectFiles[iPos,APROJ_SWITCH]) // addit. file switches
   if aFSfile[ii,3] == "A"
      cTmp += " -io=a "
   elseif aFSfile[ii,3] == "G"
      cTmp += " -io=g "
   elseif aFSfile[ii,3] == "T"
      cTmp += " -io=t "
   elseif aFSfile[ii,3] == "B"
      cTmp += " -io=b "
   endif

   if !("-io=" $ cTmp)
      cTmp += " " + cIoModeSwitch + " "
   endif
   cTmp := alltrim(strtran(cTmp, "  ", " "))

   fwrite(fd, TAB + "$(FS) $(FSCOMPSYS) $(FSADDOPT) " + cTmp + " " + ;
              cObjOutSwitch + cObjDir + aFSfile[ii,1] + cExt + " " + ;
              trim(translSlash(aProjectFiles[iPos,APROJ_PATH])) + ;
              trim(translSlash(aProjectFiles[iPos,APROJ_FILE])) + LF)
   fwrite(fd, LF)
next

fwrite(fd, LF)
fwrite(fd, "#eof" + LF)
fclose(fd)

?
? "Makefile named '" + cMakefile + "' created"
?
? "Invoke '" + cMakeInvoc + "' to (re)compile and build "
if IS_EXE                     // nCreate == 1
   ?? "executable"
elseif IS_EXELIB              // nCreate == 2
   ?? "executable and lib"
elseif IS_LIB                 // nCreate == 3
   ?? "library"
endif
? "Invoke '" + cMakeInvoc + " clean' to remove old build"
?
wait

return

/**************************************************************************
 * Helper: read settings
 */
FUNCTION ReadSettings()
local cTemp, iPos, iRecord
MEMVAR nCreate, nRelease, nSystem, nIoMode
MEMVAR cExeName, cLibName, cMain, cAddSwitch, cAddObjLib, cSearchPath, cOutputDir

nCreate := nRelease := nSystem := 0
cExeName := cLibName := cMain := cAddSwitch := cAddObjLib := cSearchPath := cOutputDir := ""

seek " .1"
if !found()
   alert("Sorry, wrong project database")
   quit
endif

if " EXEONLY " $ Project->Path
   nCreate := 1                     // check by IS_EXE    or IS_WITHEXE
elseif " EXELIB " $ Project->Path
   nCreate := 2                     // check by IS_EXELIB or IS_WITHEXE
elseif " LIBONLY " $ Project->Path
   nCreate := 3                     // check by IS_LIB    or IS_WITHLIB
endif

do case
case " DEBUG_PRG " $ Project->Path .and. " DEBUG_C " $ Project->Path
   nRelease := 4                    // check by IS_DEBUG_PC or IS_DEBUG_WITHC
case " DEBUG_PRG " $ Project->Path
   nRelease := 2                    // check by IS_DEBUG_P
case " DEBUG_C " $ Project->Path
   nRelease := 3                    // check by IS_DEBUG_C or IS_DEBUG_WITHC
case " DEBUG_NO "  $ Project->Path
   nRelease := 1                    // check by IS_RELEASE
endcase

do case
case " BCC32 " $ Project->Path
   nSystem := 2                     // check by IS_WIN_BCC or IS_WINDOWS
case " MSVC6 " $ Project->Path
   nSystem := 3                     // check by IS_WIN_MSC or IS_WINDOWS
case " UNIX " $ Project->Path
   nSystem := 1                     // check by IS_LINUX
endcase

do case
case " IOMODE_A " $ Project->Path
   nIoMode := 1                     // check by IS_IOMODE_A
case " IOMODE_G " $ Project->Path
   nIoMode := 2                     // check by IS_IOMODE_G
case " IOMODE_T " $ Project->Path
   nIoMode := 3                     // check by IS_IOMODE_T
case " IOMODE_B " $ Project->Path
   nIoMode := 4                     // check by IS_IOMODE_B
endcase

iPos := at(" EXENAME=", Project->Path)
if iPos > 0
   cTemp := substr(Project->Path, iPos+9)
   cExeName := left(cTemp,at(" ", cTemp)-1)
endif

iPos := at(" LIBNAME=", Project->Path)
if iPos > 0
   cTemp := substr(Project->Path, iPos+9)
   cLibName := left(cTemp,at(" ", cTemp)-1)
endif

iPos := at(" MAIN=", Project->Path)
if iPos > 0
   cTemp := substr(Project->Path, iPos+6)
   cMain := left(cTemp,at(" ", cTemp)-1)
endif

cMainSwitch := trim(Project->Switch)

iRecord := recno()
seek " .2"
if !found()
   alert("Sorry, wrong project database")
   quit
endif
if !empty(Project->Path)
   cAddObjLib := trim(Project->Path)
endif
if !empty(Project->Switch)
   cAddSwitch := trim(Project->Switch)
endif

seek " .3"
if !found()
   alert("Sorry, wrong project database")
   quit
endif
if !empty(Project->Path)
   cSearchPath := trim(Project->Path)
endif

seek " .4"
if !found()
   alert("Sorry, wrong project database")
   quit
endif
if !empty(Project->Path)
   cOutputDir := trim(Project->Path)
endif

goto (iRecord)

return

/**************************************************************************
 * Helper for EditSettings()
 */
FUNCTION GetSwitches(nSystem)
MEMVAR sw_na, sw_am, sw_q, sw_v, sw_I, sw_w1, sw_w2, sw_w3, sw_w4, sw_fox, sw_del, sw_mdi

sw_na := " -na " $ Project->Switch
sw_am := " -am " $ Project->Switch
sw_q  := " -q "  $ Project->Switch
sw_v  := " -v "  $ Project->Switch
sw_I  := " -I="  $ Project->Switch
sw_w1 := " -w1 " $ Project->Switch
sw_w2 := " -w2 " $ Project->Switch
sw_w3 := " -w3 " $ Project->Switch
sw_w4 := " -w4 " $ Project->Switch
sw_fox := " -fox "  $ Project->Switch
sw_del := " -delc " $ Project->Switch
sw_mdi := " -mdi "  $ Project->Switch
return .T.

/**************************************************************************
 * Helper: list project files
 */
FUNCTION ListProjects()
local cTxt, ii, aList := {}
local aDir := Directory("fsmake_*.dbf")
MEMVAR nRow, cProject

if empty(aDir)
   InfoBox("Sorry, no project databases available here")
   return
endif
for ii := 1 to len(aDir)
   cTxt := substr(aDir[ii,1],8)         // remove "fsmake_"
   cTxt := left(cTxt, at(".",cTxt) -1)  // remove ".dbf"
   aadd(aList, cTxt)
next
ii := achoice(nRow +3,col(),,, aList)         // selection
if ii > 0
   cProject := padr(aList[ii], 20)
endif
@ nRow +3,0 clear
return

/**************************************************************************
 * Helper: draw tab
 */
FUNCTION DrawTab(iRow, iCol, iLen)
local cColor, aColorLine := {190,190,190}
#define USEGUIDRAW   // comment-out to draw by graph chars also in GUI mode

@ iRow+1,0 clear

#ifdef USEGUIDRAW
   // apply for GUI mode only
   @ iRow+1.3,0 DRAW TO iRow+1.3,iCol-1 COLOR (aColorLine) WIDTH 4
   @ DRAW TO iRow-1, iCol-1          COLOR (aColorLine) WIDTH 4
   @ DRAW TO iRow-1, iCol+iLen+1     COLOR (aColorLine) WIDTH 4
   @ DRAW TO iRow+1.3, iCol+iLen+1   COLOR (aColorLine) WIDTH 4
   @ DRAW TO iRow+1.3, maxcol()      COLOR (aColorLine) WIDTH 4
   if AppIoMode() == "G"
      return
   endif
#endif

// apply for Terminal and optional for GUI mode
SET GUICOLOR ON
cColor := if(AppIoMode() == "G", "GR+", "GR+/B")
@ iRow+1,0 TO iRow+1,iCol-2       DOUBLE COLOR (cColor)
@ iRow+1,iCol-1 say chr(188)             COLOR (cColor)
@ iRow  ,iCol-1 say chr(186)             COLOR (cColor)
@ iRow-1,iCol-1 say chr(201)             COLOR (cColor)
@ iRow-1,iCol TO iRow-1,iCol+iLen DOUBLE COLOR (cColor)
@ iRow-1,iCol+iLen+1 say chr(187)        COLOR (cColor)
@ iRow  ,iCol+iLen+1 say chr(186)        COLOR (cColor)
@ iRow+1,iCol+iLen+1 say chr(200)        COLOR (cColor)
@ iRow+1,iCol+iLen+2 TO iRow+1,maxcol() DOUBLE COLOR (cColor)
SET GUICOLOR OFF
return

/**************************************************************************
 * NewFiles() = Helper for EditFiles()
 */
FUNCTION NewFiles(iRow)
local ii, iPos, cc, y, yy, cPath := space(100), cUsedPath, cFile
local aKnownFSfile, bRedir, iOldRow := iRow
local lNeedSave := .F.
local aDir := {}, aDir2 := {}, aPrg := {}, aFmt := {}, aDirC := {}
local aHeader := {"Use","FileName","Size","Date","Time","Attr","Comment"}
private oBrNew, lSaved := .F., swnew_current := .T.
private aNewFiles := {}
MEMVAR aProjectFiles
MEMVAR swnew_prg, swnew_fmt, swnew_c


bRedir := Onkey(0)
ON ANY KEY            // disable redirection

swnew_prg := .T.

@ iRow,0 clear
@ iRow,1 get swnew_prg CHECKBOX CAPTION ".prg source files" ;
                       TOOLTIP ".prg are the default source" + chr(10) + ;
                               "files of your application" + ;
                               if(IS_WINDOWS, "", chr(10) + "(considered are *.prg and *.PRG)")
                               
@ ++iRow,1 get swnew_fmt CHECKBOX CAPTION ".fmt optional SAY/GETformat files" ;
                       TOOLTIP ".fmt are optional source files" + chr(10) + ;
                               "contaning @..SAY or @..GET formatting," + chr(10) + ;
                               "used mainly in dBASE applications" + ;
                               if(IS_WINDOWS, "", chr(10) + "(considered are *.fmt and *.FMT)")
@ ++iRow,1 get swnew_c CHECKBOX CAPTION ".c source files (except auto created)" ;
                       TOOLTIP "the by FlagShip automatically created .c files" + chr(10) + ;
                               "are excluded from the search, i.e. when both" + chr(10) + ;
                               "name.prg and name.c exist, name.c is excluded" + ;
                               if(IS_WINDOWS, "", chr(10) + "(considered are *.c and *.C)")
                       
// @ ++iRow,1 get swnew_current CHECKBOX CAPTION "in current dir" ; 
//                        TOOLTIP "get files in current directory" + chr(10) + ;
//                                curdir()
// cc := Atail(GetList):CapCol + strlen2col(Atail(GetList):Caption) +1                               
// @ iRow, cc say "or in directory" GET cPath PICTURE "@S50" ;
//                        WHEN (!swnew_current) ;
//                        TOOLTIP "enter absolute or relative path" + chr(10) + ;
//                                "of the directory containing" + chr(10) + ;
//                                "above selected source files." + chr(10) + ;
//                                "Apply only when 'current dir'" + chr(10) + ;
//                                "is unchecked."
@ ++iRow, 1 say "Source files directory (or empty)" GET cPath PICTURE "@S40" ;
                       TOOLTIP "enter absolute or relative path" + chr(10) + ;
                               "of the directory containing" + chr(10) + ;
                               "above selected source files." + chr(10) + ;
                               "Empty = current dir = '" + curdir() + "'" + chr(10) + ;
                               "Default: empty field = current dir" 

@ iRow+2,1 say "Select by Mouse click or Space key, finish by ENTER or PgDn" 
read

iRow := iOldRow
@ iRow,0 clear

// if swnew_current
//    cPath := ""
// endif

if lastkey() == K_ESC
   Onkey(0, bRedir)
   return
endif

if empty(cPath)
   cUsedPath := curdir()
else
   cUsedPath := trim(cPath)
endif
@ iRow,0 say "Path for source files: " + cUsedPath
@ iRow +3, 20 say "Please wait, working..."

cPath := alltrim(cPath)
if cPath == "./" .or. cPath == ".\" .or. cPath == "."
   cPath := ""
endif
if !empty(cPath)
   if !(right(cPath,1) $ "\/:")
#ifdef FS_WIN32
      cPath += "\"
#else
      cPath += "/"
#endif
   endif
endif

aPrg := directory(cPath + "*.prg")
#ifndef FS_WIN32
   aDir2 := directory(cPath + "*.PRG")
   for ii := 1 to len(aDir2)
      aadd(aPrg, aDir2[ii])
   next
#endif

aFmt := directory(cPath + "*.fmt")
#ifndef FS_WIN32
   aDir2 := directory(cPath + "*.FMT")
   for ii := 1 to len(aDir2)
      aadd(aFmt, aDir2[ii])
   next
#endif

if swnew_prg
   for ii := 1 to len(aPrg)
      aadd(aDir, aPrg[ii])
   next
endif
for ii := 1 to len(aFmt)
   aadd(aPrg, aFmt[ii])
   if swnew_fmt
      aadd(aDir, aFmt[ii])
   endif
next

if swnew_c
   aDirC := directory(cPath + "*.c")
#ifndef FS_WIN32
   aDir2 := directory(cPath + "*.C")
   for ii := 1 to len(aDir2)
      aadd(aDirC, aDir2[ii])
   next
#endif
   /* check .c against .prg and .fmt
    */
   for ii := 1 to len(aDirC)
      y := aDirC[ii,1]
      yy := left(if(IS_WINDOWS, upper(y), y), at(".",y) -1)
      iPos := ascan(aPrg, {|x, xx| ;
                  xx := left(if(IS_WINDOWS, upper(x[1]), x[1]), at(".",x[1]) -1), ;
                  xx == yy } )
      if iPos == 0
         iPos := ascan(aProjectFiles, {|x, xx| ;
                  xx := left(if(IS_WINDOWS, upper(x[APROJ_FILE]), x[APROJ_FILE]), at(".",x[APROJ_FILE]) -1), ;
                  xx == yy } )
      endif
      if iPos == 0
         aadd(aDir, aDirC[ii])
      endif
   next
endif

@ iRow +3, 0 clear
if len(aDir) == 0
   setpos(iRow+1,0)
   ? "Sorry, no source files available in this directory"
   wait
   @ iRow,0 clear
   Onkey(0, bRedir)
   return
endif

for ii := 1 to len(aDir)
   aadd(aNewFiles, {"+", aDir[ii,1],aDir[ii,2],aDir[ii,3],aDir[ii,4],aDir[ii,5],""} )
next
aNewFiles := asort(aNewFiles, , , {|a,b| upper(a[2]) <= upper(b[2]) } )

/* check
 */
for ii := 1 to len(aNewFiles)
   cFile := aNewFiles[ii,2]
   if !IsAlpha(left(cFile,1)) .and. !IsDigit(left(cFile,1)) .and. left(cFile,1) != "_"
      aNewFiles[ii,1] := "-"
      aNewFiles[ii,7] := "file name must begin with character, digit or underscore"
   else
#ifdef FS_WIN32
      iPos := ascan(aProjectFiles, {|x| upper(trim(cFile)) == upper(trim(x[APROJ_FILE])) } )
#else
      iPos := ascan(aProjectFiles, {|x| trim(cFile) == trim(x[APROJ_FILE]) } )
#endif
      if iPos > 0
         aNewFiles[ii,1] := "-"
         aNewFiles[ii,7] := "file already in project"
         if !(trim(cPath) == trim(aProjectFiles[iPos,APROJ_PATH]))
            aNewFiles[ii,7] += " (in different directories)"
         endif
      endif
   endif
next

/* display/edit
 */
@ iRow+2,0 say "[PgDn][PgUp]skip  [C]clear all  [+]select  [-]deselect  [S]save  [Esc]exit" COLOR "GR+/B"

if AppIoMode() == "T"
   @ iRow+4,0 TO iRow+4,maxcol()
   oBrNew := TbrowseArr(iRow+5,0, maxrow() -1, maxcol(), .F., .T.)
   oBrNew:ColSep  := chr(179)      // or "|"
   oBrNew:HeadSep := chr(197,196)  // or "+-"
   oBrNew:ColSepEof := .F.
else
   oBrNew := TbrowseArr(iRow+4,0, maxrow() -1, maxcol(), .F., .T.)
   // disable to browse via PgDn/PgUp:
   oBrNew:RowCache := max(oBrNew:RowCount, len(aNewFiles))  // optional: to scroll by mouse
endif
oBrNew:UserArray := aNewFiles
for ii := 1 to len(aNewFiles[1])
   oBrNew:AddColumn( TbColumnNew(aHeader[ii], .T.))
next

/* Note: we need to handle the above special keys.
 * We either can copy the .../system/tbrowsehand.prg here and extend it
 * for these keys, or redirect keys to sub-handler and precess only
 * these special keys. We use the last, since cheaper :-)
 */
// oBrNew:Handler := {|| MyTbHandler() }
ON ANY KEY DO TbrHandNew   // redirect any key to TbrHandNew()
oBrNew:Exec()
ON ANY KEY

if lSaved
   for ii := 1 to len(aNewFiles)
      if aNewFiles[ii,1] == "+"
         aKnownFSfile := checkKnownFn(aNewFiles[ii,2], cPath)
         append blank
         replace File with aNewFiles[ii,2], Path with alltrim(cPath)
         if valtype(aKnownFSfile) == "A"
            replace Switch   with aKnownFSfile[4]
            replace FSsysNum with aKnownFSfile[3]
            replace Type     with "F"
            replace Depend   with ReadDependency(aKnownFSfile[1], alltrim(cPath), .T.)
         endif
         lNeedSave := .T.
      endif
   next
else
   InfoBox("Aborted")
endif

@ iRow,0 clear
// wait
Onkey(0, bRedir)         // restore redirection

if lNeedSave
   commit
   lSaved := .F.
endif
return

/**************************************************************************
 * Sub-Handler for NewFiles(), triggered by ON ANY KEY
 */
FUNCTION TbrHandNew()
MEMVAR aNewFiles, oBrNew, lSaved
local ok, bRedir

bRedir := Onkey(0)
ON ANY KEY           // disable (recursive) redirection
ok := TbrSubHand(aNewFiles, oBrNew, "Cc+-Ss")
if ok
   lSaved := .T.
endif
Onkey(0, bRedir)     // restore redirection
return

/**************************************************************************
 * Sub-Handler for EditFiles(), triggered by ON ANY KEY
 */
FUNCTION TbrHandEdit()
PRIVATE lNeedSave := .F.
MEMVAR aProjectFiles, oBrProj, lSaved, oSaveButt, nButtRow, nButtCol
local ok, bRedir

bRedir := Onkey(0)
ON ANY KEY           // disable (recursive) redirection
ok := TbrSubHand(aProjectFiles, oBrProj, "AaCc-EeBbLlFfDdVvSs" + chr(K_ENTER))
if ok
   lSaved := .T.
elseif lNeedSave .or. oBrProj:Updated
   SetSaveButt(oBrProj,0)
endif
Onkey(0, bRedir)     // restore redirection
return


/**************************************************************************
 * Helper, triggered by Tbrowse() or TbrHandEdit()
 */
FUNCTION SetSaveButt(oBrProj, nMode)
local nCol := col(), nRow := row()
MEMVAR oSaveButt, nButtRow, nButtCol

if AppIoMode() == "G"
   if valtype(oSaveButt) != "O"
      oSaveButt := PushButton{nButtRow,nButtCol,"[S]save!"}
      oSaveButt:GuiColor := "W+/B"
//    oSaveButt:SetStyle(BS_FLAT)
      oSaveButt:Height(1.0,.F.)
      oSaveButt:Width(len("[S]save!"),.F.)
      oSaveButt:ToolTip := "Data changed, select save or abort"
      oSaveButt:Notify := {|obj| __keyboard("S") }    // simulate "S" key
      oSaveButt:Show()

      @ nButtRow+1,nButtCol-2 say "[Esc]abort" GUICOLOR "B"
      ProcessEvents()         // display button now
   endif
else
   @ nButtRow,nButtCol say "[S]save!" COLOR "R+/B"
   setpos(nRow, nCol)
endif
return

/**************************************************************************
 * General Tbrowse sub-handler, called from TbrHandNew() and TbrHandEdit()
 * returns .T. on 'Ss' (save) key and .F. otherwise
 */
FUNCTION TbrSubHand(aFiles, oTbr, cSuppKeys)
local ii, iPos, cKey, oTbCol, cDepend, ok, key := lastkey()
MEMVAR lNeedSave

if key == K_ENTER .and. (chr(key) $ cSuppKeys)
   lNeedSave := .T.
   return .F.
endif

if key < 32 .or. key > 255 .or. !(chr(key) $ cSuppKeys)
   return .F.
endif

cKey := upper(chr(key))
do case
case cKey == "S"        // save
   keyboard chr(K_ESC)  // exit Tbrowse()
   return .T.
case cKey $ "+-LF"      // replace Use column with key
   iPos := oTbr:UserArrayPos
   if iPos > 0 .and. iPos <= len(aFiles)
      oTbCol := oTbr:GetColumn(APROJ_TYPE)
      Eval(oTbCol:Block, cKey)
      oTbr:RefreshCurrent()
      lNeedSave := .T.
   endif
case cKey == "A"        // add new record
   keyboard chr(K_CTRL_PGDN) + chr(K_DOWN) + chr(K_ENTER)
case cKey $ "CEB"       // replace all Tbrowse data with -, F or L
   ok := .F.
   if cKey == "C"
      ok := alert("Are you sure to clear all files?", {"No", "Yes"}) == 2
   elseif cKey == "E"
      ok := alert("Are you sure to mark;all files as [F] = file?", {"No", "Yes"}) == 2
   elseif cKey == "B"
      ok := alert("Are you sure to mark;all files as [L]=library?", {"No", "Yes"}) == 2
   endif
   lastkey(.T.)         // clear all lastkey()
   if ok
      oTbr:GoTop()
      while !oTbr:HitBottom
         oTbCol := oTbr:GetColumn(1)
         if cKey == "C"
            Eval(oTbCol:Block, "-")
         elseif cKey == "E"
            Eval(oTbCol:Block, "F")
         elseif cKey == "B"
            Eval(oTbCol:Block, "L")
         endif
         oTbr:Down()
      enddo
      oTbr:GoTop()
      oTbr:RefreshAll()
      lNeedSave := .T.
   endif
case cKey == "D"        // read #include dependencies
   iPos := oTbr:UserArrayPos
   if iPos > 0 .and. iPos <= len(aFiles)
      cDepend := ReadDependency(aFiles[iPos,APROJ_FILE], ;
                                aFiles[iPos,APROJ_PATH], ;
                                if(aProjectFiles[iPos,APROJ_FSMARKNUM] > 0, .T.,.F.))
      oTbCol  := oTbr:GetColumn(APROJ_DEPEND)
      if !empty(cDepend)
         if aFiles[iPos,APROJ_DEPENDIDX] > 0 .and. ;
               aFiles[iPos,APROJ_DEPENDIDX] <= len(aDepend)
            aDepend[aFiles[iPos,APROJ_DEPENDIDX]] := cDepend
         else
            aadd(aDepend,cDepend)
            aFiles[iPos,APROJ_DEPENDIDX] := len(aDepend)
         endif
         Eval(oTbCol:Block, "Y")
      else
         ii := 2
         if aFiles[iPos,APROJ_DEPENDIDX] > 0
            ii := alert("No dependency found.;Clear current ?", {"No", "Yes"}, ;
                        "GR+/G,W+/R", B_SINGLE + " ", MBOX_QUEST)
         endif
         if ii == 2
            aFiles[iPos,APROJ_DEPENDIDX] := 0
            Eval(oTbCol:Block, " ")
         endif
      endif
      oTbr:RefreshCurrent()
      lNeedSave := .T.
   endif
case cKey == "V"        // view dependencies
   iPos := oTbr:UserArrayPos
   if iPos > 0 .and. iPos <= len(aFiles)
      ViewDepend(iPos, aFiles, oTbr)
   endif
endcase
return .F.

/**************************************************************************
 * Helper for submenu Create: check available Makefile
 * returns .T. if file n/a or if for the same project
 */
FUNCTION checkMakefile(cMakefile)
local fd, cLine, iCount := 0, ok := .F.

if !file(cMakefile)
   return .T.
endif
fd := fopen(cMakefile, FO_READ)
if fd <= 0
   return .T.
endif

/* file available, check if for the same project
 */
while Ferror() == 0 .and. iCount < 100
   cLine := FreadTxt(fd)
   if "# for FSmake Project: " $ cLine
      if upper("# for FSmake Project: " + cProject) $ upper(cLine)
         ok := .T.
      endif
      exit
   endif
   iCount++
enddo
fclose(fd)
return ok

/**************************************************************************
 * Helper for check dependencies: get include paths
 * executed by setup in EditSettings() when cSearchPath is empty
 */
FUNCTION GetIncludePaths()
local fd, cLine, iPos, cTrans, cSlash := "/"
local cTmp := "", cFSDIR := "", cFSPATH := "", cConfig := "", cPath := "", aTmp := {}
MEMVAR cPathDelim, cSTDFH, cSEARCH

cTmp := getenv("FSCONFIG")
if empty(cTmp)
   cTmp := getenv("FLAGSHIP_DIR")
endif
if empty(cTmp)
#ifdef FS_WIN32
   cTmp := Which("FlagShip.exe")
#else
   cTmp := Which("FlagShip")
#endif
endif

if empty(cTmp)
  wait "no path to FS6config"
  return ""
endif
cSlash := if(IS_WINDOWS, "\", "/")

cConfig := cTmp + "/FS6config"
if !file(cConfig)
   cConfig := cTmp + "/etc/FS6config"
endif
if !file(cConfig)
   cConfig := cTmp + "/../etc/FS6config"
endif
if !file(cConfig)
  wait "no FS6config"
  return ""
endif

? "cConfig:",cConfig

/* read FS6config
 */
cConfig := alltrim(translSlash(cConfig))
fd := fopen(cConfig, FO_READ)
if fd <= 0
   wait "cannot open FSconfig in '" + cConfig + "'"
   return ""
endif
while Ferror() == 0
   cLine := FreadTxt(fd) + " "
   if left(cTmp,6) == "FSPATH"
      cFSPATH := alltrim(substr(cLine, at(":", cLine) +1))
   endif
   if left(cTmp,5) == "FSDIR"
      cFSDIR := alltrim(substr(cLine, at(":", cLine) +1))
   endif
   if left(cLine,9) == "FSOPTIONS"
      iPos := at("-nI=",cLine)
      if iPos > 0
         cTmp := substr(cLine, iPos +4)
         cSTDFH := left(cTmp, at(" ", cTmp) -1)
      endif
      iPos := at("-I",cLine)
      while iPos > 0
         cTmp := substr(cLine, iPos+2)
         if left(cTmp,1) == '"'
            aadd(aTmp, left(cTmp, at('"',cTmp, 2)))
         else
            aadd(aTmp, left(cTmp, at(' ',cTmp)-1))
         endif
         iPos := at("-I",cLine, iPos +2)
      enddo
   endif
enddo
fclose(fh)


cTrans := cFSDIR   
if empty(cTrans) 
   if !empty(cFSPATH)
      cTrans := cFSPATH 
   elseif !empty(getenv("FLAGSHIP_DIR"))
      cTrans := getenv("FLAGSHIP_DIR")
   elseif !empty(getenv("FSDEBUG_INCLUDE"))
      cTrans := getenv("FSDEBUG_INCLUDE")
      cTrans := left(cTrans, rat(cSlash,cTrans) -1)      // w/o /include
   else
      cTrans := which("FlagShip" + if(IS_WINDOWS, ".exe", ""))
      if !empty(cTrans)
         cTrans := left(cTrans, rat(cSlash,cTrans) -1)   // w/o /bin
      endif
   endif
endif

for ii := 1 to len(aTmp)
   cTmp := aTmp[ii]
   if "$(FSDIR)" $ cTmp
      cTmp := strtran(cTmp, "$(FSDIR)", cTrans)
   endif
   cTmp := alltrim(cTmp)
   if left(cTmp,1) $ ["'] .and. right(cTmp,1) $ ["']
      cTmp := substr(cTmp,2, len(cTmp) -2)
   endif
   cSEARCH += if(ii > 1, cPathDelim, "") + cTmp
   if empty(cSTDFH)
      if file(cTmp + "/std.fh")
         cSTDFH := cTmp + "/std.fh"
      endif
   endif
next
cTmp := ""
if !empty(cSTDFH)
   cTmp += "STDFH=" + cSTDFH + " "
endif
cTmp += "INCLUDES=."
if !empty(cSEARCH)
   cTmp += cPathDelim + cSEARCH + " "
endif
if "$(FSDIR)" $ cTmp
   cTmp := strtran(cTmp, "$(FSDIR)", cTrans)
endif
cTmp := translSlash(cTmp)
return cTmp


FUNCTION Which(cExe)
local iPos, ii, cTmp, aPath, lAdd := .F.

#ifdef FS_WIN32
   lAdd := .T.       // consider current dir (default in Windows)
#endif
aPath := SplitPath(getenv("PATH"), lAdd)
for ii := 1 to len(aPath)
   cTmp := alltrim(aPath[ii])
   if !(right(cTmp,1) $ "/\:")
      cTmp += "/"
   endif
   if file(cTmp + cExe)
      if len(cTmp) > 1 .and. right(cTmp,1) $ "/\"
         cTmp := left(cTmp, len(cTmp) -1)
      endif
      return cTmp    // returns path (w/o "/") if exe found in PATH environment
   endif
next
return ""            // not found


FUNCTION SplitPath(cPath, lAddCurrent)
local iPos, cTmp, aPath := {}
MEMVAR cPathDelim    // PATH delimiter is ':' in Unix, ';' in Windows

if lAddCurrent
   aadd(aPath, curdir())
endif
while !empty(cPath)
   iPos := at(cPathDelim, cPath)
   if iPos > 0
      cTmp  := left(cPath, iPos -1)
      cPath := substr(cPath, iPos +1)
   else
      cTmp  := cPath
      cPath := ""
   endif
   if !empty(cTmp)
      aadd(aPath, cTmp)
   endif
enddo
return aPath         // path splitted into array

/**************************************************************************
 * Helper for Edit Files: Dependencies collection, triggered by "D"
 * Reads current .prg file and collect #include files, including std.fh and
 * recursion in each #include
 * The returned string concatenates these files by "+" sign, the string
 * is passed via database to CreateMake()
 */
FUNCTION ReadDependency(cName, cPath, lFSdepend)
local fd, cLine, cFile, cTmp, ii
local cDepend := ""
PRIVATE aTmpDep := {}, aInclPaths := {}
MEMVAR  nSystem

if empty(cName)
   return ""
endif

@ 0,0 say "Please wait, processing..." COLOR "R+/B" GUICOLOR "R+"

if nSystem == 0
   ReadSettings()
endif

cPath := alltrim(cPath)
if !empty(cPath) .and. !(right(cPath,1) $ ":/\")
   if right(cPath,1) != ":"
      cPath += "/"
   endif
endif
cFile := cPath + alltrim(cName)

cFile := alltrim(translSlash(cFile))
fd := fopen(cFile, FO_READ)
if fd <= 0
   @ 0,0 clear to 0,50
   InfoBox("Cannot read file;" + cFile + ";check path and access rights")
   return ""
endif

if "INCLUDES=" $ cSearchPath  // from Settings
   cTmp := substr(cSearchPath, at("INCLUDES=",cSearchPath) + 9) + " "
   cTmp := left(cTmp, at(" ",cTmp) -1)
   aInclPaths := aclone( SplitPath(cTmp, .F.))
   for ii := 1 to len(aInclPaths)
      cTmp := aInclPaths[ii]
      if !empty(cTmp) .and. !(right(cTmp,1) $ ":/\")
         if right(cTmp,1) != ":"
            cTmp += "/"
         endif
      endif
      cTmp := translSlash(cTmp)
      aInclPaths[ii] := cTmp
   next
endif

RecursRead(fd, cFile)
if ascan(aTmpDep, "std.fh") == 0
   if "STDFH=" $ cSearchPath
      cFile := substr(cSearchPath, at("STDFH=",cSearchPath) + 6) + " "
      cFile := left(cFile, at(" ",cFile) -1)
   else
      cFile := "std.fh"
   endif
   if lFSdepend   // FlagShip system file
      ii := 2
   else
      ii := alert("Process also 'std.fh' file ?", {"No", "Yes"}, ;
                  "GR/G,R/W", B_SINGLE + " ", MBOX_QUEST)
   endif
   if ii == 2
      RecursRead(0, cFile)
   endif
endif
for ii := 1 to len(aTmpDep)
   cDepend += if(ii > 1, " + ", "") + aTmpDep[ii]
next
@ 0,0 clear to 0,50
return cDepend


FUNCTION RecursRead(fd, cFile)
local ii, cLine, cTmp, cc, cPath := "", cAddMsg := "", lAddCurrent := .F.
MEMVAR aTmpDep, aInclPaths

if fd == 0                          // fd > 0 signals source .prg file
   lAddCurrent := .T.
   if left(cFile,1) $ ["'] .and. right(cFile,1) == left(cFile,1)
      cFile := substr(cFile,2)
      cFile := left(cFile,len(cFile)-1)
   endif   
   if "\" $ cFile .or. "/" $ cFile
      fd := fopen(alltrim(translSlash(cFile)), FO_READ)   // known path
   else
      cAddMsg := "in known paths (see settings) :;"
      for ii := 1 to len(aInclPaths)
         cPath := aInclPaths[ii]
         fd := fopen(alltrim(cPath + cFile), FO_READ)  // add path
         if fd > 0
            exit
         endif
         cAddMsg += aInclPaths[ii] + ";"
      next
   endif
endif
if fd <= 0
   InfoBox("Cannot read file;" + cFile + ";" + cAddMsg + "check path and access rights")
   return
endif

if lAddCurrent
   if len(aTmpDep) > 0
      if ascan(aTmpDep, cPath + cFile) > 0
         fclose(fd)
         return                     // already processed
      endif
   endif
   aadd(aTmpDep, cPath + cFile)     // add current file
endif

while Ferror() == 0
   cLine := alltrim(FreadTxt(fd))
   cTmp  := strtran(cLine, " ", "")
   if lower(left(cTmp,8)) == "#include"
      iPos := at("include", lower(cLine))
      if iPos > 0
         cTmp := ltrim(substr(cLine, iPos+7))
         cc := left(cTmp,1)               // first " or '
         iPos := at(cc, cTmp, 2)          // last  " or '
         cTmp := substr(cTmp,2,iPos -2)   // include name
         if len(aTmpDep) > 0
            if ascan(aTmpDep, cTmp) > 0   // already available
               loop                       // continue read
            endif
         endif
         RecursRead(0, cTmp)              // recursion: process current
      endif
   endif
enddo
fclose(fd)
return

/**************************************************************************
 * Helper for Edit Files: View/edit dependencies, triggered by "V"
 */
FUNCTION ViewDepend(iPos, aFiles, oTbr)
local bRedir, cColor, cDepend := ""
MEMVAR lNeedSave

#define CRLF chr(13,10)
#define SOFTCR chr(141,10)

cColor := SetColor()
SET COLOR TO "GR+/G,R+/G"

bRedir := Onkey(0)
ON ANY KEY    // disable redirection

cDepend := "# Include files added to Makefile dependency," + CRLF + ;
           "#   separated by ' + '.  Files must be fully"  + CRLF + ;
           "#   qualified, the default FlagShip search" + CRLF + ;
           "#   path is specified in 'Settings' tab."   + CRLF + ;
           "# Save changes by Ctrl-W, Exit with ESC"    + CRLF + ;
           "# --- Do not remove or change this header ---" + CRLF + CRLF
if aFiles[iPos,APROJ_DEPENDIDX] > 0
   cDepend += aDepend[aFiles[iPos,APROJ_DEPENDIDX]]
endif
cDepend += space(50)
if AppIoMode() == "T"
   cSaveScr := SaveScreen(9,16,30,65)
   @ 9,17,26,63 BOX B_SINGLE + " "
endif

cDepend := MemoEdit(cDepend, 10,17,29,64,,,,,8,,,,,,.T., ; // clear at finish
                             "#000000/#89DDFF")            // black/light blue
//                           "#000000/#7FFF7F,,#7FFF7F")   // black/light green with frame
@ 0,0 clear to 0,maxcol()

if lastkey() == K_ESC
   KEYBOARD " "
else
   lNeedSave := .T.
   cDepend := substr(cDepend, at("header ---",cDepend) + 10)
   cDepend := strtran(cDepend, CRLF,    "")
   cDepend := strtran(cDepend, chr(10), "")
   cDepend := strtran(cDepend, chr(13), "")
   cDepend := strtran(cDepend, chr(141), "")
   cDepend := alltrim(strtran(cDepend, "  ", " "))
   if empty(cDepend)                               // delete
      aDepend[aFiles[iPos,APROJ_DEPENDIDX]] := 0
      aFiles[iPos,APROJ_DEPENDIDX] := ""
   elseif aFiles[iPos,APROJ_DEPENDIDX] == 0        // add new
      aadd(aDepend, cDepend)
      aDepend[aFiles[iPos,APROJ_DEPENDIDX]] := len(aDepend)
   else                                            // replace
      aDepend[aFiles[iPos,APROJ_DEPENDIDX]] := cDepend
   endif
endif
if AppIoMode() == "T"
   RestScreen(cSaveScr, 9,16,30,65)
endif

SetColor(cColor)
Onkey(0, bRedir)           // restore redirection
return

/**************************************************************************
 * helper: checks for given directory
 */
FUNCTION CheckDir(cDir)
local iPos, aDir

cDir := alltrim(cDir)
if empty(cDir)
   return .T.
endif
if (right(cDir,1) $ "/\")
   cObjDir := left(cDir, len(cDir) -1)
endif 
if empty(cDir) .or. cDir == "."
   return .T.
endif
   
while !empty(cDir)
   aDir := directory(cDir, "D")
   if IS_WINDOWS
      iPos := ascan(aDir, {|x| upper(x[1]) == upper(cDir) .and. "D" $ x[5] })
   else
      iPos := ascan(aDir, {|x| x[1] == cDir .and. left(x[5],1) == "d" })
   endif
   if iPos == 0
      iPos := alert("Directory '" + cDir + "';not available. Create?", {"Yes", "No"})
      if iPos != 1
         return .F.
      else
         run ("mkdir " + cDir)
      endif
   else
      exit
   endif
enddo
return .T.

/**************************************************************************
 * helper: translate slash
 */
FUNCTION translSlash(cTmp)
if IS_LINUX
   cTmp := strtran(cTmp, "\", "/")
else
   cTmp := strtran(cTmp, "/", "\")
endif 
return cTmp  

/**************************************************************************
 * Helper for submenu Edit: check known programs
 * returns subarray if FlagShip system file or NIL otherwise
 */
FUNCTION checkKnownFn(cFileName, cPath)
local iPos, iAnsw, cFullName

/* aKnownPrg[n,3] =  1: FlagShip system file, compile as is
                   | 2: compile with -io=g GUI
                   | 4: compile with -io=t Terminal
                   | 8: compile with -io=b Basic
                   |16: compile with -io=a All, hybrid
                   |32: not -fox
*/
static aKnownPrg := {{"achoicehand.prg",   .T.,       1, "-na -io=a"}, ;
                     {"autolock.prg",      .T.,       1, "-na -io=a"}, ;
                     {"boxfunct.prg",      .T.,       1, "-na -io=a"}, ;
                     {"browse.prg",        .T.,   2+4+8, "-na"},       ;
                     {"browsedb.prg",      .T.,       1, "-na -io=a"}, ;
                     {"checkboxhand.prg",  .T.,       1, "-na -io=a"}, ;
                     {"dbedit.prg",        .T.,   2+4+8, "-na"},       ;
                     {"extern_b.prg",      .T.,       1, "-na -io=b"}, ;
                     {"extern_g.prg",      .T.,       1, "-na -io=g"}, ;
                     {"extern_t.prg",      .T.,       1, "-na -io=t"}, ;
                     {"foxpro_api.prg",    .T.,    1+32, "-na -io=a"}, ;
                     {"fsi.prg",           .F.,       1, "-na -io=a"}, ;
                     {"fs2_com.prg",       .T.,       1, "-na -io=a"}, ;
                     {"fs2_udfs.prg",      .T.,       1, "-na -io=a"}, ;
                     {"FSerror.prg",       .T.,       1, "-na -io=a -FSintern"}, ;
                     {"getsys.prg",        .T.,       1, "-na -io=a"}, ;
                     {"getsys4.prg",       .F.,       1, "-na -io=a"}, ;
                     {"initio.prg",        .T.,2+4+8+16, "-na"},       ;
                     {"initiomenu.prg",    .T.,  2+4+16, "-na"},       ;
                     {"labedit.prg",       .T.,       1, "-na -io=a"}, ;
                     {"listboxhand.prg",   .T.,       1, "-na -io=a"}, ;
                     {"memoedithand.prg",  .T.,       1, "-na -io=a"}, ;
                     {"oprinter_b.prg",    .T.,       1, "-na -io=b"}, ;
                     {"oprinter_g.prg",    .T.,       1, "-na -io=g"}, ;
                     {"oprinter_t.prg",    .T.,       1, "-na -io=t"}, ;
                     {"o_printer.prg",     .T.,       1, "-na -io=a"}, ;
                     {"pushbutthand.prg",  .T.,       1, "-na -io=a"}, ;
                     {"radbutthand.prg ",  .T.,       1, "-na -io=a"}, ;
                     {"radgrouphand.prg",  .T.,       1, "-na -io=a"}, ;
                     {"rddsys.prg",        .T.,       1, "-na -io=a"}, ;
                     {"repoedit.prg",      .T.,       1, "-na -io=a"}, ;
                     {"scor_mem.prg",      .T.,       1, "-na -io=a"}, ;
                     {"scrnmode.prg",      .T.,       1, "-na -io=a"}, ;
                     {"setsortab.prg",     .T.,       1, "-na -io=a"}, ;
                     {"tbrowsearr.prg",    .T.,   2+4+8, "-na"},       ;
                     {"tbrowsedb.prg",     .T.,   2+4+8, "-na"},       ;
                     {"tbrowsedbhand.prg", .T.,       1, "-na -io=a"}, ;
                     {"tbrowsehand.prg",   .T.,       1, "-na -io=a"} }

cPath := alltrim(cPath)
if !empty(cPath) .and. !(right(cPath,1) $ ":\/")
   cPath += "/"
endif
cFullName := cPath + cFileName
cFileName := alltrim(cFileName)
if IS_WINDOWS
   cFileName := lower(cFileName)
endif
cFullName := translSlash(cFullName)
iPos := ascan(aKnownPrg, {|x| x[1] == cFileName } )
if iPos > 0
   iAnsw := Alert("File: " + cFullName + ;
                  ";;Is this (modified) FlagShip system file;" + ;
                  "with known special switches,;or your own source ?", ;
                  {"FlagShip","My own"}, ;
                  "GR+/G,W+/R", B_SINGLE + " ", MBOX_QUEST)
   if iAnsw == 1 .and. aKnownPrg[iPos,2]
      return aKnownPrg[iPos]
   endif
endif
return NIL


// eof

