/* A small utility to determine index key and other header info from
 * specified index file. It also displays the required INDEX ON... string.
 *
 * For FlagShip 5.x or 6.x (c) 2004 by multisoft GmbH, Munich, Germany
 * You may modify this source according to your needs and distribute the
 * object code freely with your executable. You may distribute also the 
 * modified source code when your application is distributed in source, 
 * but may not remove the copyright notice.
 * Release $Date: 2004/02/16 07:56:49 $
 *
 * Compile : FlagShip indexdump.prg [-io=b] [-stat] -o indexdump
 * Optional: ln -s `pwd`/indexdump /usr/local/bin/.
 * Execute : indexdump [-l] myfile.ext [../myPath/OtherFile.ext ...]
 */

PARAM p1, p2, p3, p4, p4, p6, p7, p8, p9, p10 // will be autom. extended

LOCAL ii, iPar, iRead, cFile, cExt, fh
LOCAL lShortOutput := .T.
LOCAL buffer := space(2048)
LOCAL aTmp, nTmp, cKey, lUniq, lDesc, cFor, cDbf, cIdxFile, cIdxOpt
LOCAL aFiles := {}

#ifdef FlagShip5
   set font "courier", 10
#endif

? "This ", execname(), " program determies the index key of given file(s)."
? "On-line help is available via '" + execname() + " -h'  or  '" + execname() + " /?'"
if pcount() == 0 .or. lower(p1) == "-h" .or. p1 == "/?"
   ?
   ? "The kind of the index is determined from the file extension."
   ? "Supported are .idx, .ntx, .ndx and .cdx indices (regardless case)."
   ? "The controlling database is not required/used here."
   ?
   ? "Syntax: " + execname() + " [-l] [--long] myfile.<ext> [../myPath/OtherFile.<ext> ...]"
   ? "  e.g.: " + execname() + " --long myFile.idx /server/Clipper/otherFile.ntx"
   ?
   ? "where the file (and path) names are case sensitive (on Unix)."
   ? "The '-l' or '--long' option creates full cut & paste source output."
   ? "Hint: you may redirect the output to a file to easy edit, e.g.:"
   ? "  " + execname() + " -io=b --long file*.ntx file2.idx file3.cdx >out.txt"
   ?
   wait
   quit
endif

/* get cmd-line parameters and resolve wildcards (this is required for
 * MS-Windows only, Unix resolves wildcards at command-line level)
 */
for iPar := 1 to pCount()
   cFile := param(iPar)      // get n-th command-line parameter's value 
   if lower(cFile) == "-l" .or. cFile == "--long"
      lShortOutput := .F.
      loop
   endif
   if "*" $ cFile .or. "?" $ cFile
      aTmp := directory(cFile)
      for ii := 1 to len(aTmp)
         aadd(aFiles, aTmp[ii,1])
      next
   else
      aadd(aFiles, cFile)
   endif
next

if len(aFiles) == 0
   ? "** sorry, no matching files found"
   wait
   quit
endif
  
/* just to avoid problems, ensure that
 * at least the buffer is \0 terminated
 */
buffer += chr(0)

/* walk thru the parameter list 
 * incl. resolved wildcards
 */
for iPar := 1 to len(aFiles)
   if iPar > 1
      wait
      if lastkey() = K_ESC
         ? "aborted"
         quit
      endif
   endif
   
   cFile := aFiles[iPar]  
   ?   
   ? "-- File   : " + cFile
   if !file(cFile)
      ? "** sorry, file not available"
      loop
   endif
   fh := fopen(cFile, 0)
   if fh > 0
      iRead := fread(fh, @buffer, 2048)
      fclose(fh)
   endif
   if fh <= 0 .or. iRead < 512  // at least first header required
      ? "** sorry, unsufficient permission to open file in read-only mode,"
      ? "   or this is not an index file, or the index is exlusively locked" 
      loop
   endif
   
   cKey  := ""
   cFor  := ""
   cDbf  := ""
   lUniq := .F.
   lDesc := .F.
   cExt := lower(substr(cFile, rat(".",cFile) +1))
   
   do case
   case cExt == "idx"
      if strpeek(buffer,1) == 82       // FlagShip 5.x and 6.x index
         ? "   Index  : FlagShip 5.x or 6.x"
         ? "   Key    :", cKey := getBufStr(@buffer, 83, 430)
         ? "   Type   :", chr(strpeek(buffer,5))
         ? "   Len,Dec:", ltrim(strpeek(buffer,3)) + ;
                          if( strpeek(buffer,21) != 78, "", ","+ltrim(strpeek(buffer,4)))
         ? "   #Keys  :", ltrim(bin2L(substr(buffer,13,4)))
         ? "   Unique :", lUniq := binAnd(strpeek(buffer,2), 1) > 0
         ? "   Descend:", lDesc := binAnd(strpeek(buffer,2), 2) > 0
         ? "   ForCond:", cFor  := getBufStr(@buffer, 539, 230)
         ? "   .dbf   :", cDbf  := getBufStr(@buffer, 519, 20)
         ? "   idxChk#:", ltrim(strpeek(buffer,8))
         ? "   sortFil:", getBufStr(@buffer, 739, 20)
         ? "   sortOrd:", getBufSort(@buffer, 769)
      elseif strpeek(buffer,1) == 81 // FlagShip 3.x and 4.x index
         ? "   Index  : FlagShip 4.x"
         ? "   Key    :", cKey := getBufStr(@buffer, 33, 420)
         ? "   Type   :", chr(strpeek(buffer,21))
         ? "   Len,Dec:", if( strpeek(buffer,21) != 78, "", ltrim(strpeek(buffer,3)) + ",") + ; 
                          ltrim(strpeek(buffer,4))
         ? "   #Keys  :", ltrim(bin2L(substr(buffer,5,4)))
         ? "   Unique :", lUniq := binAnd(strpeek(buffer,2), 1) > 0
         ? "   Descend:", lDesc := binAnd(strpeek(buffer,2), 2) > 0
         ? "   ForCond:", cFor  := getBufStr(@buffer, 1051, 200)
         ? "   .dbf   :", cDbf  := getBufStr(@buffer, 1031, 20)
         ? "   idxChk#:", ltrim(strpeek(buffer,1026))
         ? "   sortOrd:", getBufSort(@buffer, 1281)
      else
         ? "** NOT ** a FlagShip index, trying FoxBase/FoxPro's .idx (may cause garbage)"
         ? "   Index  : FoxBase ???"
         ? "   Key    :", cKey := getBufStr(@buffer, 17, 220)   
         ? "   Type   : n/a"
         ? "   Len,Dec:", ltrim(bin2i(substr(buffer,13,2))) + ",n/a"
         ? "   #Keys  : n/a"
         nTmp := strpeek(buffer,15)               
         ? "   Unique :", lUniq := binAnd(nTmp, 1) > 0
         ? "   Descend: n/a" 
         ? "   ForCond:", cFor  := getBufStr(@buffer, 237, 220) 
         ? "   .dbf   : n/a"
         ? "   Compoun:", binAnd(nTmp, 64) > 0
         ? "   Compact:", binAnd(nTmp, 32) > 0
      endif
      
   case cExt == "ntx"      // Clipper index
      if strpeek(buffer,1) == 3
         ? "   Index  : Clipper 87"
      elseif strpeek(buffer,1) == 6
         ? "   Index  : Clipper 5.x"
      else
         ? "** sorry, not a Clipper index"
         loop
      endif
      ? "   Key    :", cKey := getBufStr(@buffer, 23, 256)
      ? "   Type   : n/a"
      ? "   Len,Dec:", ltrim(bin2i(substr(buffer,15,2))) + "," + ;
                       ltrim(bin2i(substr(buffer,17,2)))
      ? "   #Keys  : n/a"
      ? "   Unique :", lUniq := strpeek(buffer,279) == 1
      ? "   Descend: n/a"  
      ? "   ForCond: n/a"
      ? "   .dbf   : n/a"

   case cExt == "ndx"      // dBaseIII index
      ? "   Index  : dBaseIII+"
      ? "   Key    :", cKey := getBufStr(@buffer, 25, 488)
      ? "   Type   :", if(strpeek(buffer,17) == 1, "N", "C")
      ? "   Len,Dec:", ltrim(bin2i(substr(buffer,13,2))) + ;
                       if(strpeek(buffer,17) == 1, ",n/a", "")
      ? "   #Keys  : n/a"
      ? "   Unique :", lUniq := strpeek(buffer,24) > 0
      ? "   Descend: n/a"  
      ? "   ForCond: n/a"
      ? "   .dbf   : n/a"
      
   case cExt == "cdx"      // FoxPro index
      ? "   Index  : FoxPro"
      ? "   Key    :", cKey := getBufStr(@buffer, 513, 512) // ... FOR will continue
      nTmp := strpeek(buffer,15)                            // byte#14 is or-ed status
      if binAnd(nTmp, 8) > 0                                // has FOR condition?
         cFor := getBufStr(@buffer, -1, 512)                // FOR expression after key
      endif
      ? "   Type   : n/a"
      ? "   Len,Dec:", ltrim(bin2i(substr(buffer,13,2))) + ",n/a"
      ? "   #Keys  : n/a"
      ? "   Unique :", lUniq := binAnd(nTmp, 1) > 0
      ? "   Descend:", lDesc := strpeek(buffer,503) > 0  
      ? "   ForCond: " + cFor
      ? "   .dbf   : n/a"
      ? "   Compoun:", binAnd(nTmp, 64) > 0
      ? "   Compact:", binAnd(nTmp, 32) > 0

   otherwise
      ? "** sorry, unsupported file extension, invoke help via " + execname() + " -h"
      loop
   endcase  
    
   if !empty(cKey) .and. left(cKey,9) != "n/a = NOT"
      cIdxFile := left(cFile, len(cFile) -4) + ".idx"
      cIdxOpt  := if(lDesc, " DESCEND", "") + ;
                  if(lUniq, " UNIQUE", "") + ;
                  if(!empty(cFor), " FOR " + cFor, "")

      if lShortOutput
         ? "   How to"
         ? "   create : INDEX ON " + cKey + " TO " + cIdxFile + cIdxOpt
      else
         ? "   How to"
         ? "   create : ************ snip ************"
         ? '            // #include "fspreset.fh"  /* auto lower-case transl. */'
         ? '            cIdxFile := "' + cIdxFile + '"'
         ? '            cDbfFile := "' + ;
                              if(empty(cDbf), " ...specify... ", cDbf) + '"'
         ? "            IF ! FILE(cIdxFile)"
         ? "               USE (cDbfFile) EXCLUSIVE NEW"
         ? "               IF USED()"
         ? "                  INDEX ON " + cKey + " TO (cIdxFile)"  + cIdxOpt
         ? "                  USE"
         ? "               ELSE"
         ? '                  ALERT("Cannot open " + cDbfFile + " EXCLUSIVE")'
         ? "               ENDIF"
         ? "            ENDIF"
         ? "            ************ snip ************"
      endif
   endif
   
next

wait "done... "
quit

/* ----------------------------------------------------------------
 * This helper reads a \0 terminated string at specified 
 * position and returns this string, but checks the string 
 * boundary (since we can work on unknown/uspecified file).
 * It of course could be written in .prg too, we want to
 * demonstrates the use of Open-C API; you can roughly 
 * see it as a kind of substr()
 */
function getBufStr(cBuffer, nStart, nMaxLng)
local cRet := "n/a = NOT an index file !"
if nStart != 0                      // 0 is invalid value
   #Cinline
   {
   #include "FSopenc.h"
      static int iNextBuff = 0;     // for .cdx only
      static int iRemain   = 0;     // for .cdx only
      
      char *ptrBuffer = VAR_CHR(VAR_NAME_LOCPAR(cbuffer));
      char *ptrCurr;
      int   iStart    = VAR_INTNUM(VAR_NAME_LOCPAR(nstart));   
      int   iMaxLng   = VAR_INTNUM(VAR_NAME_LOCPAR(nmaxlng));
      int   ii;
      
      /* The .cdx FOR condition is stored just after the key, both  
       * teminated by \0, so we continue. This is triggered by passing 
       * nStart < 0 by using the same cBuffer and nMaxLng.
       */
      if(iStart < 0 && iNextBuff > 0 && iNextBuff <= iMaxLng) {
         ptrCurr = ptrBuffer + iNextBuff;    // continue (.cdx) FOR...
         iMaxLng = iRemain;
      } else if(iStart > 0) {                // the usual case
         ptrCurr = ptrBuffer + iStart -1;    // -1 since nStart is 1..n 
      } else
         iMaxLng = 0;                        // don't process
          
      iNextBuff = 0; 
      iRemain   = 0;
      
      /* The string is \0 terminated, so we could directly use it.
       * But to avoid garbage on reading invalid/corrupted index file, 
       * we will check if \0 is really within the nMaxLng boundary.
       * We also cannot simply use SET_VAR_CHRNUM(), since this would
       * include chr(0) into the string what we don't want here.
       */
      for(ii=0; ii < iMaxLng ; ii++) {
         if(! *(ptrCurr + ii)) {                          // ok, \0 found,
            SET_VAR_CHR(VAR_NAME_LOCAL(cret), ptrCurr );  // strcpy() -> cRet
            iNextBuff = ptrCurr + ii +1 - ptrBuffer;      // save next ptr
            iRemain   = iMaxLng - ii -1;                  // and rest size
            break;
         }
      } 
   }
   #endCinline
endif
return cRet

/* ----------------------------------------------------------------
 * This helper checks and display sort order table, if any.
 */
function getBufSort(cBuffer, nStart)
local cRet := ""
local ii, iCurr, iFirst , iLast

if strpeek(@cBuffer, nStart +1) == 0 .and. strpeek(@cBuffer, nStart +2) == 0
   cRet := "natural 0-255"
else
   iFirst := iLast := strpeek(@cBuffer, nStart)
   cRet   := ltrim(iFirst)
   for ii := 1 to 255
      iCurr := strpeek(@cBuffer, nStart + ii)
      if iCurr == iLast +1 
         iLast := iCurr
         if ii == 255
            cRet += "-" + ltrim(iCurr)
         endif
      else
         if iLast > iFirst + 1
            cRet += "-" + ltrim(iLast) + "," + ltrim(iCurr)
         else
            cRet += "," + ltrim(iCurr) 
         endif
         iFirst := iLast := iCurr
      endif
   next
endif
return cRet

** eof **
