#ifdef COMMENT_ONLY
/* webmail.c   (see also webutil.prg, freadstd.c, bbsleep.c) 
 *
 * Some commonly used functions for Web based FlagShip applications
 * (c) 1999..2007 by multisoft, Munich
 * This source may be freely used or modified by the FlagShip Licensee
 * It is also available in the std. FlagShip library 
 *
 * Release $Date: 2007/10/31 08:10:18 $
 *
 * compile: FlagShip mywebapplic.prg [webutil.prg] webmail.c ....
 *      or: FlagShip -c webmail.c
 *     and: FlagShip myapp*.prg webmail.o (or webmail.obj)
 */
#endif

static char *fileID="@@# $Name:  $ $Revision: 1.12 $ $Date: 2007/10/31 08:10:18 $";
 
#include "FSopenc.h"
#include <stdio.h>

#ifdef FS_WIN32
# define POPEN(p1,p2)             _popen(p1,p2)
# define PCLOSE(p1)               _pclose(p1)
#else
# define POPEN(p1,p2)             popen(p1,p2)
# define PCLOSE(p1)               pclose(p1)
#endif

/* #define TESTING */

/* ------------------------------------------------------------------------
 * WebSendMail (caFrom, cSubject, cText, cTo, [cCc], [cBcc], [cExe])  --> nError
 *
 * Invokes sendmail on the server to submit an e-mail
 * This is a low level function, only marginal checks are performed - 
 * so check the parameters and valid addresses before invoking!
 *
 * (1) caFrom  : own e-mail address (from) as string,
 *               or an array {"own_email_address", ;   required
 *                            "organization", ;        may be empty
 *                            "send_reply_to" ;        may be empty 
 *                            "return-path" }          may be empty 
 * (2) cSubject: subject ("" is accepted but not recommended)
 * (3) cText   : e-mail body (cannot be empty, at least " " is required)
 *               sent "as is"
 * (4) cTo     : adress to, e.g. "name@any.domain.com", see note below
 * (5) cCc     : optional CC adress[es] delimited by comma. If not required,
 *               an empty value ("" or NIL) is ok.
 * (6) cBcc    : optional BCC adress[es] delimited by comma. If not required,
 *               an empty value ("" or NIL) is ok.
 * (7) cExe    : optional path and name of sendmail executable
 *
 * nError      : 0 = ok
 *               1..6 = wrong argument n
 *               7 = not enought arguments
 *              10 = command-line is too long (check the cTo size) 
 *              11 = could not open sendmail process (correct path ?)
 *
 * Note: general form for specifying e-mail adresses (par 1,3,4):
 *       "any_name@my.domain.com"
 *   or: '"My Name" <any_name@my.domain.com>'
 *   or: '<any_name@my.domain.co>, "My Name" <any_name@my.domain.com>'
 * ----------------------------------------------------------------------- */
UDF_DECL(websendmai)
{
  
#define CMDBUFF 1024                           /* replace with 250 for DOS */
#define SENDMAIL_WPATH "/usr/lib/sendmail"     /* check on YOUR server !!  */

// #define OLDER_SENDMAIL
#ifdef OLDER_SENDMAIL
# define SENDMAILFLAGS "-t -n"   /* for older sendmail */
#else
# define SENDMAILFLAGS "-t"      /* for current sendmail versions */
#endif

/* Explanation for above SENDMAILFLAGS flags:
 *
 * The -t flag tells sendmail to look for To:, Cc:, and Bcc: lines
 * in the mail message to determine the addresses to send to, the
 * Bcc line will be removed. Note: It is not supported anymore in newer 
 * sendmail versions (e.g. Linux kernel 2.6 etc) when recipiens are
 * already specified
 *
 * The -n flag tells sendmail not to use the alias list for the UNIX server.
 * Usually we do not want outsiders using the alias list of the webserver.
 * Note: It is not supported anymore in newer sendmail versions (e.g. Linux 
 * kernel 2.6 etc or MS-Windows) and is hence disabled per default, see
 * above #define OLDER_SENDMAIL
 *
 * Note for MS-Windows: you will probably need the sendmail.exe 
 * available from http://glob.com.au/sendmail or via
 * http://www.fship.com/tools/sendmail.zip or via
 * ftp://fship.com/pub/multisoft/FlagShip/tools/sendmail.zip
 */

   FILE *mail;
   char command[CMDBUFF], *cSendmail;
   unsigned char *cFrom, *cSubject, *cTo, *cText;
   unsigned char  *cCc = 0, *cBcc = 0, *cOrgan = 0, *cReply = 0, *cRetPath = 0;
   int nError = 0;
   FSvar *var, *elem;

   if(argc < 4) {
      nError = 7;
      goto done;
   }

   nError = 1;
   var = argv[0];
   if( IS_VAR_CHR(var) ) {                                       /* par 1 */
      if (! IS_VAR_EMPTY(var)) {
         nError = 0;
         cFrom = VAR_CHR(var);
      }
   }else if( IS_VAR_ARR(var) && (VAR_ISDIM(var) > 0 ) ) {
      elem = VAR_ARRELEM (var, 1);
      if( IS_VAR_CHR(elem) && (! IS_VAR_EMPTY(elem) ) ) {
         nError = 0;
         cFrom = VAR_CHR(elem);
      }
      if (VAR_ISDIM(var) > 1) {
         elem = VAR_ARRELEM (var, 2);
         if( IS_VAR_CHR(elem) && (! IS_VAR_EMPTY(elem) ) )
            cOrgan = VAR_CHR(elem);
      }
      if (VAR_ISDIM(var) > 2) {
         elem = VAR_ARRELEM (var, 3);
         if( IS_VAR_CHR(elem) && (! IS_VAR_EMPTY(elem) ) ) 
            cReply = VAR_CHR(elem);
      }
      if (VAR_ISDIM(var) > 3) {
         elem = VAR_ARRELEM (var, 4);        /* fixed 28-Mar-2000 */
         if( IS_VAR_CHR(elem) && (! IS_VAR_EMPTY(elem) ) ) 
            cRetPath = VAR_CHR(elem);
      }
   }
   if(nError)
      goto done;
       
   if(! IS_VAR_CHR(argv[1])) {                           /* par 2, subject */
      nError = 2;
      goto done;
   }
   cSubject = VAR_CHR(argv[1]);
       
   if(! IS_VAR_CHR(argv[2]) || strlen(VAR_CHR(argv[2])) == 0) {  /* par 3, text */
      nError = 3;
      goto done;
   }
   cText = VAR_CHR(argv[2]);

   if(! IS_VAR_CHR(argv[3]) || IS_VAR_EMPTY(argv[3])) {      /* par 4, to */
      nError = 4;
      goto done;
   }
   cTo = VAR_CHR(argv[3]);
       
   if( argc > 4 && IS_VAR_CHR(argv[4]) && ! IS_VAR_EMPTY(argv[4]) )      /* par 5, cc */
      cCc = VAR_CHR(argv[4]);

   if( argc > 5 && IS_VAR_CHR(argv[5]) && ! IS_VAR_EMPTY(argv[5]) )      /* par 6, bcc */
      cBcc = VAR_CHR(argv[5]);

   if( argc > 6 && IS_VAR_CHR(argv[6]) && ! IS_VAR_EMPTY(argv[6]) )      /* par 7, exe */
      cSendmail = VAR_CHR(argv[6]);
   else
      cSendmail = SENDMAIL_WPATH;
      
   /* form the command-line: /path/sendmail switches to_address
    * see idetails in "man sendmail"
    */
   if( (strlen(cTo) + strlen(SENDMAILFLAGS) + 
        strlen(cSendmail)) > CMDBUFF -5) {
      nError = 10;
      goto done;
   }
   strcpy(command, cSendmail);
   strcat(command, " ");
   strcat(command, SENDMAILFLAGS);
#if OLDER_SENDMAIL
   strcat(command, " '");
   strcat(command, cTo);
   strcat(command, "'");
#endif

#ifdef TESTING
      printf("\n-----------sending mail------------\n");
      printf("via (command) %s\n",command);
      if(cRetPath && *cRetPath)
         printf("Return-Path: %s\n",cRetPath);
      else
         printf("Return-Path: %s\n",cFrom);
      printf("From: %s\n",cFrom);
      if(cOrgan)
         printf("Organization: %s\n",cOrgan);
      if(cReply)
         printf("Reply-to: %s\n",cReply);
      printf("To: %s\n",cTo);
      if(cCc)
         printf("Cc: %s\n",cCc);
      if(cBcc)
         printf("Bcc: %s\n",cBcc);
      printf("Subject: %s\n",cSubject);
#endif

   /* invoke the sendmail program/daemon ...
    */
   mail = POPEN(command,"w");     /* open subprocess and pipe */
   if (mail == NULL) { 
      nError = 11;
      goto done;
   }

   /* ... and pass the e-mail header + text 
    * via stdin/pipe
    */
   if(cRetPath && *cRetPath)
      fprintf(mail,"Return-Path: %s\n",cRetPath);
   else
      fprintf(mail,"Return-Path: %s\n",cFrom);
   fprintf(mail,"To: %s\n",cTo);
   fprintf(mail,"From: %s\n",cFrom);
   if(cOrgan && *cOrgan)
      fprintf(mail,"Organization: %s\n",cOrgan);
   if(cReply && *cReply)
      fprintf(mail,"Reply-to: %s\n",cReply);
   if(cCc && *cCc)
      fprintf(mail,"Cc: %s\n",cCc);
   if(cBcc && *cBcc)
      fprintf(mail,"Bcc: %s\n",cBcc);
   fprintf(mail,"Subject: %s\n",cSubject);
   fprintf(mail,"\n%s\n",cText);
   PCLOSE(mail);                   /* close subprocess */

done:
   return set_Iv(nError);
}

/* eof */ 
